(function(){const l=document.createElement("link").relList;if(l&&l.supports&&l.supports("modulepreload"))return;for(const c of document.querySelectorAll('link[rel="modulepreload"]'))o(c);new MutationObserver(c=>{for(const f of c)if(f.type==="childList")for(const d of f.addedNodes)d.tagName==="LINK"&&d.rel==="modulepreload"&&o(d)}).observe(document,{childList:!0,subtree:!0});function i(c){const f={};return c.integrity&&(f.integrity=c.integrity),c.referrerPolicy&&(f.referrerPolicy=c.referrerPolicy),c.crossOrigin==="use-credentials"?f.credentials="include":c.crossOrigin==="anonymous"?f.credentials="omit":f.credentials="same-origin",f}function o(c){if(c.ep)return;c.ep=!0;const f=i(c);fetch(c.href,f)}})();function Sv(n){return n&&n.__esModule&&Object.prototype.hasOwnProperty.call(n,"default")?n.default:n}var Js={exports:{}},Kr={};/**
 * @license React
 * react-jsx-runtime.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var y0;function Xb(){if(y0)return Kr;y0=1;var n=Symbol.for("react.transitional.element"),l=Symbol.for("react.fragment");function i(o,c,f){var d=null;if(f!==void 0&&(d=""+f),c.key!==void 0&&(d=""+c.key),"key"in c){f={};for(var h in c)h!=="key"&&(f[h]=c[h])}else f=c;return c=f.ref,{$$typeof:n,type:o,key:d,ref:c!==void 0?c:null,props:f}}return Kr.Fragment=l,Kr.jsx=i,Kr.jsxs=i,Kr}var v0;function Gb(){return v0||(v0=1,Js.exports=Xb()),Js.exports}var g=Gb(),Ws={exports:{}},gt={};/**
 * @license React
 * react.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var p0;function Vb(){if(p0)return gt;p0=1;var n=Symbol.for("react.transitional.element"),l=Symbol.for("react.portal"),i=Symbol.for("react.fragment"),o=Symbol.for("react.strict_mode"),c=Symbol.for("react.profiler"),f=Symbol.for("react.consumer"),d=Symbol.for("react.context"),h=Symbol.for("react.forward_ref"),y=Symbol.for("react.suspense"),m=Symbol.for("react.memo"),v=Symbol.for("react.lazy"),S=Symbol.iterator;function _(T){return T===null||typeof T!="object"?null:(T=S&&T[S]||T["@@iterator"],typeof T=="function"?T:null)}var x={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},w=Object.assign,N={};function B(T,X,I){this.props=T,this.context=X,this.refs=N,this.updater=I||x}B.prototype.isReactComponent={},B.prototype.setState=function(T,X){if(typeof T!="object"&&typeof T!="function"&&T!=null)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,T,X,"setState")},B.prototype.forceUpdate=function(T){this.updater.enqueueForceUpdate(this,T,"forceUpdate")};function O(){}O.prototype=B.prototype;function j(T,X,I){this.props=T,this.context=X,this.refs=N,this.updater=I||x}var z=j.prototype=new O;z.constructor=j,w(z,B.prototype),z.isPureReactComponent=!0;var L=Array.isArray,$={H:null,A:null,T:null,S:null,V:null},V=Object.prototype.hasOwnProperty;function Z(T,X,I,P,nt,ut){return I=ut.ref,{$$typeof:n,type:T,key:X,ref:I!==void 0?I:null,props:ut}}function W(T,X){return Z(T.type,X,void 0,void 0,void 0,T.props)}function it(T){return typeof T=="object"&&T!==null&&T.$$typeof===n}function ht(T){var X={"=":"=0",":":"=2"};return"$"+T.replace(/[=:]/g,function(I){return X[I]})}var pt=/\/+/g;function at(T,X){return typeof T=="object"&&T!==null&&T.key!=null?ht(""+T.key):X.toString(36)}function lt(){}function R(T){switch(T.status){case"fulfilled":return T.value;case"rejected":throw T.reason;default:switch(typeof T.status=="string"?T.then(lt,lt):(T.status="pending",T.then(function(X){T.status==="pending"&&(T.status="fulfilled",T.value=X)},function(X){T.status==="pending"&&(T.status="rejected",T.reason=X)})),T.status){case"fulfilled":return T.value;case"rejected":throw T.reason}}throw T}function Q(T,X,I,P,nt){var ut=typeof T;(ut==="undefined"||ut==="boolean")&&(T=null);var rt=!1;if(T===null)rt=!0;else switch(ut){case"bigint":case"string":case"number":rt=!0;break;case"object":switch(T.$$typeof){case n:case l:rt=!0;break;case v:return rt=T._init,Q(rt(T._payload),X,I,P,nt)}}if(rt)return nt=nt(T),rt=P===""?"."+at(T,0):P,L(nt)?(I="",rt!=null&&(I=rt.replace(pt,"$&/")+"/"),Q(nt,X,I,"",function(Ut){return Ut})):nt!=null&&(it(nt)&&(nt=W(nt,I+(nt.key==null||T&&T.key===nt.key?"":(""+nt.key).replace(pt,"$&/")+"/")+rt)),X.push(nt)),1;rt=0;var bt=P===""?".":P+":";if(L(T))for(var vt=0;vt<T.length;vt++)P=T[vt],ut=bt+at(P,vt),rt+=Q(P,X,I,ut,nt);else if(vt=_(T),typeof vt=="function")for(T=vt.call(T),vt=0;!(P=T.next()).done;)P=P.value,ut=bt+at(P,vt++),rt+=Q(P,X,I,ut,nt);else if(ut==="object"){if(typeof T.then=="function")return Q(R(T),X,I,P,nt);throw X=String(T),Error("Objects are not valid as a React child (found: "+(X==="[object Object]"?"object with keys {"+Object.keys(T).join(", ")+"}":X)+"). If you meant to render a collection of children, use an array instead.")}return rt}function E(T,X,I){if(T==null)return T;var P=[],nt=0;return Q(T,P,"","",function(ut){return X.call(I,ut,nt++)}),P}function Y(T){if(T._status===-1){var X=T._result;X=X(),X.then(function(I){(T._status===0||T._status===-1)&&(T._status=1,T._result=I)},function(I){(T._status===0||T._status===-1)&&(T._status=2,T._result=I)}),T._status===-1&&(T._status=0,T._result=X)}if(T._status===1)return T._result.default;throw T._result}var k=typeof reportError=="function"?reportError:function(T){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var X=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof T=="object"&&T!==null&&typeof T.message=="string"?String(T.message):String(T),error:T});if(!window.dispatchEvent(X))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",T);return}console.error(T)};function et(){}return gt.Children={map:E,forEach:function(T,X,I){E(T,function(){X.apply(this,arguments)},I)},count:function(T){var X=0;return E(T,function(){X++}),X},toArray:function(T){return E(T,function(X){return X})||[]},only:function(T){if(!it(T))throw Error("React.Children.only expected to receive a single React element child.");return T}},gt.Component=B,gt.Fragment=i,gt.Profiler=c,gt.PureComponent=j,gt.StrictMode=o,gt.Suspense=y,gt.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=$,gt.__COMPILER_RUNTIME={__proto__:null,c:function(T){return $.H.useMemoCache(T)}},gt.cache=function(T){return function(){return T.apply(null,arguments)}},gt.cloneElement=function(T,X,I){if(T==null)throw Error("The argument must be a React element, but you passed "+T+".");var P=w({},T.props),nt=T.key,ut=void 0;if(X!=null)for(rt in X.ref!==void 0&&(ut=void 0),X.key!==void 0&&(nt=""+X.key),X)!V.call(X,rt)||rt==="key"||rt==="__self"||rt==="__source"||rt==="ref"&&X.ref===void 0||(P[rt]=X[rt]);var rt=arguments.length-2;if(rt===1)P.children=I;else if(1<rt){for(var bt=Array(rt),vt=0;vt<rt;vt++)bt[vt]=arguments[vt+2];P.children=bt}return Z(T.type,nt,void 0,void 0,ut,P)},gt.createContext=function(T){return T={$$typeof:d,_currentValue:T,_currentValue2:T,_threadCount:0,Provider:null,Consumer:null},T.Provider=T,T.Consumer={$$typeof:f,_context:T},T},gt.createElement=function(T,X,I){var P,nt={},ut=null;if(X!=null)for(P in X.key!==void 0&&(ut=""+X.key),X)V.call(X,P)&&P!=="key"&&P!=="__self"&&P!=="__source"&&(nt[P]=X[P]);var rt=arguments.length-2;if(rt===1)nt.children=I;else if(1<rt){for(var bt=Array(rt),vt=0;vt<rt;vt++)bt[vt]=arguments[vt+2];nt.children=bt}if(T&&T.defaultProps)for(P in rt=T.defaultProps,rt)nt[P]===void 0&&(nt[P]=rt[P]);return Z(T,ut,void 0,void 0,null,nt)},gt.createRef=function(){return{current:null}},gt.forwardRef=function(T){return{$$typeof:h,render:T}},gt.isValidElement=it,gt.lazy=function(T){return{$$typeof:v,_payload:{_status:-1,_result:T},_init:Y}},gt.memo=function(T,X){return{$$typeof:m,type:T,compare:X===void 0?null:X}},gt.startTransition=function(T){var X=$.T,I={};$.T=I;try{var P=T(),nt=$.S;nt!==null&&nt(I,P),typeof P=="object"&&P!==null&&typeof P.then=="function"&&P.then(et,k)}catch(ut){k(ut)}finally{$.T=X}},gt.unstable_useCacheRefresh=function(){return $.H.useCacheRefresh()},gt.use=function(T){return $.H.use(T)},gt.useActionState=function(T,X,I){return $.H.useActionState(T,X,I)},gt.useCallback=function(T,X){return $.H.useCallback(T,X)},gt.useContext=function(T){return $.H.useContext(T)},gt.useDebugValue=function(){},gt.useDeferredValue=function(T,X){return $.H.useDeferredValue(T,X)},gt.useEffect=function(T,X,I){var P=$.H;if(typeof I=="function")throw Error("useEffect CRUD overload is not enabled in this build of React.");return P.useEffect(T,X)},gt.useId=function(){return $.H.useId()},gt.useImperativeHandle=function(T,X,I){return $.H.useImperativeHandle(T,X,I)},gt.useInsertionEffect=function(T,X){return $.H.useInsertionEffect(T,X)},gt.useLayoutEffect=function(T,X){return $.H.useLayoutEffect(T,X)},gt.useMemo=function(T,X){return $.H.useMemo(T,X)},gt.useOptimistic=function(T,X){return $.H.useOptimistic(T,X)},gt.useReducer=function(T,X,I){return $.H.useReducer(T,X,I)},gt.useRef=function(T){return $.H.useRef(T)},gt.useState=function(T){return $.H.useState(T)},gt.useSyncExternalStore=function(T,X,I){return $.H.useSyncExternalStore(T,X,I)},gt.useTransition=function(){return $.H.useTransition()},gt.version="19.1.0",gt}var g0;function Lf(){return g0||(g0=1,Ws.exports=Vb()),Ws.exports}var C=Lf();const Yf=Sv(C);var Ps={exports:{}},Jr={},Fs={exports:{}},Is={};/**
 * @license React
 * scheduler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var b0;function kb(){return b0||(b0=1,function(n){function l(E,Y){var k=E.length;E.push(Y);t:for(;0<k;){var et=k-1>>>1,T=E[et];if(0<c(T,Y))E[et]=Y,E[k]=T,k=et;else break t}}function i(E){return E.length===0?null:E[0]}function o(E){if(E.length===0)return null;var Y=E[0],k=E.pop();if(k!==Y){E[0]=k;t:for(var et=0,T=E.length,X=T>>>1;et<X;){var I=2*(et+1)-1,P=E[I],nt=I+1,ut=E[nt];if(0>c(P,k))nt<T&&0>c(ut,P)?(E[et]=ut,E[nt]=k,et=nt):(E[et]=P,E[I]=k,et=I);else if(nt<T&&0>c(ut,k))E[et]=ut,E[nt]=k,et=nt;else break t}}return Y}function c(E,Y){var k=E.sortIndex-Y.sortIndex;return k!==0?k:E.id-Y.id}if(n.unstable_now=void 0,typeof performance=="object"&&typeof performance.now=="function"){var f=performance;n.unstable_now=function(){return f.now()}}else{var d=Date,h=d.now();n.unstable_now=function(){return d.now()-h}}var y=[],m=[],v=1,S=null,_=3,x=!1,w=!1,N=!1,B=!1,O=typeof setTimeout=="function"?setTimeout:null,j=typeof clearTimeout=="function"?clearTimeout:null,z=typeof setImmediate<"u"?setImmediate:null;function L(E){for(var Y=i(m);Y!==null;){if(Y.callback===null)o(m);else if(Y.startTime<=E)o(m),Y.sortIndex=Y.expirationTime,l(y,Y);else break;Y=i(m)}}function $(E){if(N=!1,L(E),!w)if(i(y)!==null)w=!0,V||(V=!0,at());else{var Y=i(m);Y!==null&&Q($,Y.startTime-E)}}var V=!1,Z=-1,W=5,it=-1;function ht(){return B?!0:!(n.unstable_now()-it<W)}function pt(){if(B=!1,V){var E=n.unstable_now();it=E;var Y=!0;try{t:{w=!1,N&&(N=!1,j(Z),Z=-1),x=!0;var k=_;try{e:{for(L(E),S=i(y);S!==null&&!(S.expirationTime>E&&ht());){var et=S.callback;if(typeof et=="function"){S.callback=null,_=S.priorityLevel;var T=et(S.expirationTime<=E);if(E=n.unstable_now(),typeof T=="function"){S.callback=T,L(E),Y=!0;break e}S===i(y)&&o(y),L(E)}else o(y);S=i(y)}if(S!==null)Y=!0;else{var X=i(m);X!==null&&Q($,X.startTime-E),Y=!1}}break t}finally{S=null,_=k,x=!1}Y=void 0}}finally{Y?at():V=!1}}}var at;if(typeof z=="function")at=function(){z(pt)};else if(typeof MessageChannel<"u"){var lt=new MessageChannel,R=lt.port2;lt.port1.onmessage=pt,at=function(){R.postMessage(null)}}else at=function(){O(pt,0)};function Q(E,Y){Z=O(function(){E(n.unstable_now())},Y)}n.unstable_IdlePriority=5,n.unstable_ImmediatePriority=1,n.unstable_LowPriority=4,n.unstable_NormalPriority=3,n.unstable_Profiling=null,n.unstable_UserBlockingPriority=2,n.unstable_cancelCallback=function(E){E.callback=null},n.unstable_forceFrameRate=function(E){0>E||125<E?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):W=0<E?Math.floor(1e3/E):5},n.unstable_getCurrentPriorityLevel=function(){return _},n.unstable_next=function(E){switch(_){case 1:case 2:case 3:var Y=3;break;default:Y=_}var k=_;_=Y;try{return E()}finally{_=k}},n.unstable_requestPaint=function(){B=!0},n.unstable_runWithPriority=function(E,Y){switch(E){case 1:case 2:case 3:case 4:case 5:break;default:E=3}var k=_;_=E;try{return Y()}finally{_=k}},n.unstable_scheduleCallback=function(E,Y,k){var et=n.unstable_now();switch(typeof k=="object"&&k!==null?(k=k.delay,k=typeof k=="number"&&0<k?et+k:et):k=et,E){case 1:var T=-1;break;case 2:T=250;break;case 5:T=1073741823;break;case 4:T=1e4;break;default:T=5e3}return T=k+T,E={id:v++,callback:Y,priorityLevel:E,startTime:k,expirationTime:T,sortIndex:-1},k>et?(E.sortIndex=k,l(m,E),i(y)===null&&E===i(m)&&(N?(j(Z),Z=-1):N=!0,Q($,k-et))):(E.sortIndex=T,l(y,E),w||x||(w=!0,V||(V=!0,at()))),E},n.unstable_shouldYield=ht,n.unstable_wrapCallback=function(E){var Y=_;return function(){var k=_;_=Y;try{return E.apply(this,arguments)}finally{_=k}}}}(Is)),Is}var S0;function Zb(){return S0||(S0=1,Fs.exports=kb()),Fs.exports}var tf={exports:{}},de={};/**
 * @license React
 * react-dom.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var _0;function Qb(){if(_0)return de;_0=1;var n=Lf();function l(y){var m="https://react.dev/errors/"+y;if(1<arguments.length){m+="?args[]="+encodeURIComponent(arguments[1]);for(var v=2;v<arguments.length;v++)m+="&args[]="+encodeURIComponent(arguments[v])}return"Minified React error #"+y+"; visit "+m+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function i(){}var o={d:{f:i,r:function(){throw Error(l(522))},D:i,C:i,L:i,m:i,X:i,S:i,M:i},p:0,findDOMNode:null},c=Symbol.for("react.portal");function f(y,m,v){var S=3<arguments.length&&arguments[3]!==void 0?arguments[3]:null;return{$$typeof:c,key:S==null?null:""+S,children:y,containerInfo:m,implementation:v}}var d=n.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function h(y,m){if(y==="font")return"";if(typeof m=="string")return m==="use-credentials"?m:""}return de.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=o,de.createPortal=function(y,m){var v=2<arguments.length&&arguments[2]!==void 0?arguments[2]:null;if(!m||m.nodeType!==1&&m.nodeType!==9&&m.nodeType!==11)throw Error(l(299));return f(y,m,null,v)},de.flushSync=function(y){var m=d.T,v=o.p;try{if(d.T=null,o.p=2,y)return y()}finally{d.T=m,o.p=v,o.d.f()}},de.preconnect=function(y,m){typeof y=="string"&&(m?(m=m.crossOrigin,m=typeof m=="string"?m==="use-credentials"?m:"":void 0):m=null,o.d.C(y,m))},de.prefetchDNS=function(y){typeof y=="string"&&o.d.D(y)},de.preinit=function(y,m){if(typeof y=="string"&&m&&typeof m.as=="string"){var v=m.as,S=h(v,m.crossOrigin),_=typeof m.integrity=="string"?m.integrity:void 0,x=typeof m.fetchPriority=="string"?m.fetchPriority:void 0;v==="style"?o.d.S(y,typeof m.precedence=="string"?m.precedence:void 0,{crossOrigin:S,integrity:_,fetchPriority:x}):v==="script"&&o.d.X(y,{crossOrigin:S,integrity:_,fetchPriority:x,nonce:typeof m.nonce=="string"?m.nonce:void 0})}},de.preinitModule=function(y,m){if(typeof y=="string")if(typeof m=="object"&&m!==null){if(m.as==null||m.as==="script"){var v=h(m.as,m.crossOrigin);o.d.M(y,{crossOrigin:v,integrity:typeof m.integrity=="string"?m.integrity:void 0,nonce:typeof m.nonce=="string"?m.nonce:void 0})}}else m==null&&o.d.M(y)},de.preload=function(y,m){if(typeof y=="string"&&typeof m=="object"&&m!==null&&typeof m.as=="string"){var v=m.as,S=h(v,m.crossOrigin);o.d.L(y,v,{crossOrigin:S,integrity:typeof m.integrity=="string"?m.integrity:void 0,nonce:typeof m.nonce=="string"?m.nonce:void 0,type:typeof m.type=="string"?m.type:void 0,fetchPriority:typeof m.fetchPriority=="string"?m.fetchPriority:void 0,referrerPolicy:typeof m.referrerPolicy=="string"?m.referrerPolicy:void 0,imageSrcSet:typeof m.imageSrcSet=="string"?m.imageSrcSet:void 0,imageSizes:typeof m.imageSizes=="string"?m.imageSizes:void 0,media:typeof m.media=="string"?m.media:void 0})}},de.preloadModule=function(y,m){if(typeof y=="string")if(m){var v=h(m.as,m.crossOrigin);o.d.m(y,{as:typeof m.as=="string"&&m.as!=="script"?m.as:void 0,crossOrigin:v,integrity:typeof m.integrity=="string"?m.integrity:void 0})}else o.d.m(y)},de.requestFormReset=function(y){o.d.r(y)},de.unstable_batchedUpdates=function(y,m){return y(m)},de.useFormState=function(y,m,v){return d.H.useFormState(y,m,v)},de.useFormStatus=function(){return d.H.useHostTransitionStatus()},de.version="19.1.0",de}var x0;function _v(){if(x0)return tf.exports;x0=1;function n(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(n)}catch(l){console.error(l)}}return n(),tf.exports=Qb(),tf.exports}/**
 * @license React
 * react-dom-client.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var T0;function Kb(){if(T0)return Jr;T0=1;var n=Zb(),l=Lf(),i=_v();function o(t){var e="https://react.dev/errors/"+t;if(1<arguments.length){e+="?args[]="+encodeURIComponent(arguments[1]);for(var a=2;a<arguments.length;a++)e+="&args[]="+encodeURIComponent(arguments[a])}return"Minified React error #"+t+"; visit "+e+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function c(t){return!(!t||t.nodeType!==1&&t.nodeType!==9&&t.nodeType!==11)}function f(t){var e=t,a=t;if(t.alternate)for(;e.return;)e=e.return;else{t=e;do e=t,(e.flags&4098)!==0&&(a=e.return),t=e.return;while(t)}return e.tag===3?a:null}function d(t){if(t.tag===13){var e=t.memoizedState;if(e===null&&(t=t.alternate,t!==null&&(e=t.memoizedState)),e!==null)return e.dehydrated}return null}function h(t){if(f(t)!==t)throw Error(o(188))}function y(t){var e=t.alternate;if(!e){if(e=f(t),e===null)throw Error(o(188));return e!==t?null:t}for(var a=t,r=e;;){var u=a.return;if(u===null)break;var s=u.alternate;if(s===null){if(r=u.return,r!==null){a=r;continue}break}if(u.child===s.child){for(s=u.child;s;){if(s===a)return h(u),t;if(s===r)return h(u),e;s=s.sibling}throw Error(o(188))}if(a.return!==r.return)a=u,r=s;else{for(var p=!1,b=u.child;b;){if(b===a){p=!0,a=u,r=s;break}if(b===r){p=!0,r=u,a=s;break}b=b.sibling}if(!p){for(b=s.child;b;){if(b===a){p=!0,a=s,r=u;break}if(b===r){p=!0,r=s,a=u;break}b=b.sibling}if(!p)throw Error(o(189))}}if(a.alternate!==r)throw Error(o(190))}if(a.tag!==3)throw Error(o(188));return a.stateNode.current===a?t:e}function m(t){var e=t.tag;if(e===5||e===26||e===27||e===6)return t;for(t=t.child;t!==null;){if(e=m(t),e!==null)return e;t=t.sibling}return null}var v=Object.assign,S=Symbol.for("react.element"),_=Symbol.for("react.transitional.element"),x=Symbol.for("react.portal"),w=Symbol.for("react.fragment"),N=Symbol.for("react.strict_mode"),B=Symbol.for("react.profiler"),O=Symbol.for("react.provider"),j=Symbol.for("react.consumer"),z=Symbol.for("react.context"),L=Symbol.for("react.forward_ref"),$=Symbol.for("react.suspense"),V=Symbol.for("react.suspense_list"),Z=Symbol.for("react.memo"),W=Symbol.for("react.lazy"),it=Symbol.for("react.activity"),ht=Symbol.for("react.memo_cache_sentinel"),pt=Symbol.iterator;function at(t){return t===null||typeof t!="object"?null:(t=pt&&t[pt]||t["@@iterator"],typeof t=="function"?t:null)}var lt=Symbol.for("react.client.reference");function R(t){if(t==null)return null;if(typeof t=="function")return t.$$typeof===lt?null:t.displayName||t.name||null;if(typeof t=="string")return t;switch(t){case w:return"Fragment";case B:return"Profiler";case N:return"StrictMode";case $:return"Suspense";case V:return"SuspenseList";case it:return"Activity"}if(typeof t=="object")switch(t.$$typeof){case x:return"Portal";case z:return(t.displayName||"Context")+".Provider";case j:return(t._context.displayName||"Context")+".Consumer";case L:var e=t.render;return t=t.displayName,t||(t=e.displayName||e.name||"",t=t!==""?"ForwardRef("+t+")":"ForwardRef"),t;case Z:return e=t.displayName||null,e!==null?e:R(t.type)||"Memo";case W:e=t._payload,t=t._init;try{return R(t(e))}catch{}}return null}var Q=Array.isArray,E=l.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,Y=i.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,k={pending:!1,data:null,method:null,action:null},et=[],T=-1;function X(t){return{current:t}}function I(t){0>T||(t.current=et[T],et[T]=null,T--)}function P(t,e){T++,et[T]=t.current,t.current=e}var nt=X(null),ut=X(null),rt=X(null),bt=X(null);function vt(t,e){switch(P(rt,e),P(ut,t),P(nt,null),e.nodeType){case 9:case 11:t=(t=e.documentElement)&&(t=t.namespaceURI)?Gy(t):0;break;default:if(t=e.tagName,e=e.namespaceURI)e=Gy(e),t=Vy(e,t);else switch(t){case"svg":t=1;break;case"math":t=2;break;default:t=0}}I(nt),P(nt,t)}function Ut(){I(nt),I(ut),I(rt)}function ae(t){t.memoizedState!==null&&P(bt,t);var e=nt.current,a=Vy(e,t.type);e!==a&&(P(ut,t),P(nt,a))}function pe(t){ut.current===t&&(I(nt),I(ut)),bt.current===t&&(I(bt),Gr._currentValue=k)}var $e=Object.prototype.hasOwnProperty,ln=n.unstable_scheduleCallback,Sn=n.unstable_cancelCallback,pa=n.unstable_shouldYield,ce=n.unstable_requestPaint,Ht=n.unstable_now,Wl=n.unstable_getCurrentPriorityLevel,Ei=n.unstable_ImmediatePriority,Hu=n.unstable_UserBlockingPriority,Ci=n.unstable_NormalPriority,Tg=n.unstable_LowPriority,Ed=n.unstable_IdlePriority,wg=n.log,Eg=n.unstable_setDisableYieldValue,Pl=null,Ae=null;function Vn(t){if(typeof wg=="function"&&Eg(t),Ae&&typeof Ae.setStrictMode=="function")try{Ae.setStrictMode(Pl,t)}catch{}}var je=Math.clz32?Math.clz32:jg,Cg=Math.log,Ag=Math.LN2;function jg(t){return t>>>=0,t===0?32:31-(Cg(t)/Ag|0)|0}var Ai=256,ji=4194304;function ga(t){var e=t&42;if(e!==0)return e;switch(t&-t){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:return 128;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return t&4194048;case 4194304:case 8388608:case 16777216:case 33554432:return t&62914560;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return t}}function Ni(t,e,a){var r=t.pendingLanes;if(r===0)return 0;var u=0,s=t.suspendedLanes,p=t.pingedLanes;t=t.warmLanes;var b=r&134217727;return b!==0?(r=b&~s,r!==0?u=ga(r):(p&=b,p!==0?u=ga(p):a||(a=b&~t,a!==0&&(u=ga(a))))):(b=r&~s,b!==0?u=ga(b):p!==0?u=ga(p):a||(a=r&~t,a!==0&&(u=ga(a)))),u===0?0:e!==0&&e!==u&&(e&s)===0&&(s=u&-u,a=e&-e,s>=a||s===32&&(a&4194048)!==0)?e:u}function Fl(t,e){return(t.pendingLanes&~(t.suspendedLanes&~t.pingedLanes)&e)===0}function Ng(t,e){switch(t){case 1:case 2:case 4:case 8:case 64:return e+250;case 16:case 32:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return e+5e3;case 4194304:case 8388608:case 16777216:case 33554432:return-1;case 67108864:case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function Cd(){var t=Ai;return Ai<<=1,(Ai&4194048)===0&&(Ai=256),t}function Ad(){var t=ji;return ji<<=1,(ji&62914560)===0&&(ji=4194304),t}function $u(t){for(var e=[],a=0;31>a;a++)e.push(t);return e}function Il(t,e){t.pendingLanes|=e,e!==268435456&&(t.suspendedLanes=0,t.pingedLanes=0,t.warmLanes=0)}function Og(t,e,a,r,u,s){var p=t.pendingLanes;t.pendingLanes=a,t.suspendedLanes=0,t.pingedLanes=0,t.warmLanes=0,t.expiredLanes&=a,t.entangledLanes&=a,t.errorRecoveryDisabledLanes&=a,t.shellSuspendCounter=0;var b=t.entanglements,A=t.expirationTimes,H=t.hiddenUpdates;for(a=p&~a;0<a;){var K=31-je(a),F=1<<K;b[K]=0,A[K]=-1;var q=H[K];if(q!==null)for(H[K]=null,K=0;K<q.length;K++){var G=q[K];G!==null&&(G.lane&=-536870913)}a&=~F}r!==0&&jd(t,r,0),s!==0&&u===0&&t.tag!==0&&(t.suspendedLanes|=s&~(p&~e))}function jd(t,e,a){t.pendingLanes|=e,t.suspendedLanes&=~e;var r=31-je(e);t.entangledLanes|=e,t.entanglements[r]=t.entanglements[r]|1073741824|a&4194090}function Nd(t,e){var a=t.entangledLanes|=e;for(t=t.entanglements;a;){var r=31-je(a),u=1<<r;u&e|t[r]&e&&(t[r]|=e),a&=~u}}function Lu(t){switch(t){case 2:t=1;break;case 8:t=4;break;case 32:t=16;break;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:t=128;break;case 268435456:t=134217728;break;default:t=0}return t}function Yu(t){return t&=-t,2<t?8<t?(t&134217727)!==0?32:268435456:8:2}function Od(){var t=Y.p;return t!==0?t:(t=window.event,t===void 0?32:c0(t.type))}function Mg(t,e){var a=Y.p;try{return Y.p=t,e()}finally{Y.p=a}}var kn=Math.random().toString(36).slice(2),se="__reactFiber$"+kn,ge="__reactProps$"+kn,Qa="__reactContainer$"+kn,qu="__reactEvents$"+kn,Rg="__reactListeners$"+kn,zg="__reactHandles$"+kn,Md="__reactResources$"+kn,tr="__reactMarker$"+kn;function Xu(t){delete t[se],delete t[ge],delete t[qu],delete t[Rg],delete t[zg]}function Ka(t){var e=t[se];if(e)return e;for(var a=t.parentNode;a;){if(e=a[Qa]||a[se]){if(a=e.alternate,e.child!==null||a!==null&&a.child!==null)for(t=Ky(t);t!==null;){if(a=t[se])return a;t=Ky(t)}return e}t=a,a=t.parentNode}return null}function Ja(t){if(t=t[se]||t[Qa]){var e=t.tag;if(e===5||e===6||e===13||e===26||e===27||e===3)return t}return null}function er(t){var e=t.tag;if(e===5||e===26||e===27||e===6)return t.stateNode;throw Error(o(33))}function Wa(t){var e=t[Md];return e||(e=t[Md]={hoistableStyles:new Map,hoistableScripts:new Map}),e}function Ft(t){t[tr]=!0}var Rd=new Set,zd={};function ba(t,e){Pa(t,e),Pa(t+"Capture",e)}function Pa(t,e){for(zd[t]=e,t=0;t<e.length;t++)Rd.add(e[t])}var Dg=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),Dd={},Bd={};function Bg(t){return $e.call(Bd,t)?!0:$e.call(Dd,t)?!1:Dg.test(t)?Bd[t]=!0:(Dd[t]=!0,!1)}function Oi(t,e,a){if(Bg(e))if(a===null)t.removeAttribute(e);else{switch(typeof a){case"undefined":case"function":case"symbol":t.removeAttribute(e);return;case"boolean":var r=e.toLowerCase().slice(0,5);if(r!=="data-"&&r!=="aria-"){t.removeAttribute(e);return}}t.setAttribute(e,""+a)}}function Mi(t,e,a){if(a===null)t.removeAttribute(e);else{switch(typeof a){case"undefined":case"function":case"symbol":case"boolean":t.removeAttribute(e);return}t.setAttribute(e,""+a)}}function _n(t,e,a,r){if(r===null)t.removeAttribute(a);else{switch(typeof r){case"undefined":case"function":case"symbol":case"boolean":t.removeAttribute(a);return}t.setAttributeNS(e,a,""+r)}}var Gu,Ud;function Fa(t){if(Gu===void 0)try{throw Error()}catch(a){var e=a.stack.trim().match(/\n( *(at )?)/);Gu=e&&e[1]||"",Ud=-1<a.stack.indexOf(`
    at`)?" (<anonymous>)":-1<a.stack.indexOf("@")?"@unknown:0:0":""}return`
`+Gu+t+Ud}var Vu=!1;function ku(t,e){if(!t||Vu)return"";Vu=!0;var a=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var r={DetermineComponentFrameRoot:function(){try{if(e){var F=function(){throw Error()};if(Object.defineProperty(F.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(F,[])}catch(G){var q=G}Reflect.construct(t,[],F)}else{try{F.call()}catch(G){q=G}t.call(F.prototype)}}else{try{throw Error()}catch(G){q=G}(F=t())&&typeof F.catch=="function"&&F.catch(function(){})}}catch(G){if(G&&q&&typeof G.stack=="string")return[G.stack,q.stack]}return[null,null]}};r.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var u=Object.getOwnPropertyDescriptor(r.DetermineComponentFrameRoot,"name");u&&u.configurable&&Object.defineProperty(r.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var s=r.DetermineComponentFrameRoot(),p=s[0],b=s[1];if(p&&b){var A=p.split(`
`),H=b.split(`
`);for(u=r=0;r<A.length&&!A[r].includes("DetermineComponentFrameRoot");)r++;for(;u<H.length&&!H[u].includes("DetermineComponentFrameRoot");)u++;if(r===A.length||u===H.length)for(r=A.length-1,u=H.length-1;1<=r&&0<=u&&A[r]!==H[u];)u--;for(;1<=r&&0<=u;r--,u--)if(A[r]!==H[u]){if(r!==1||u!==1)do if(r--,u--,0>u||A[r]!==H[u]){var K=`
`+A[r].replace(" at new "," at ");return t.displayName&&K.includes("<anonymous>")&&(K=K.replace("<anonymous>",t.displayName)),K}while(1<=r&&0<=u);break}}}finally{Vu=!1,Error.prepareStackTrace=a}return(a=t?t.displayName||t.name:"")?Fa(a):""}function Ug(t){switch(t.tag){case 26:case 27:case 5:return Fa(t.type);case 16:return Fa("Lazy");case 13:return Fa("Suspense");case 19:return Fa("SuspenseList");case 0:case 15:return ku(t.type,!1);case 11:return ku(t.type.render,!1);case 1:return ku(t.type,!0);case 31:return Fa("Activity");default:return""}}function Hd(t){try{var e="";do e+=Ug(t),t=t.return;while(t);return e}catch(a){return`
Error generating stack: `+a.message+`
`+a.stack}}function Le(t){switch(typeof t){case"bigint":case"boolean":case"number":case"string":case"undefined":return t;case"object":return t;default:return""}}function $d(t){var e=t.type;return(t=t.nodeName)&&t.toLowerCase()==="input"&&(e==="checkbox"||e==="radio")}function Hg(t){var e=$d(t)?"checked":"value",a=Object.getOwnPropertyDescriptor(t.constructor.prototype,e),r=""+t[e];if(!t.hasOwnProperty(e)&&typeof a<"u"&&typeof a.get=="function"&&typeof a.set=="function"){var u=a.get,s=a.set;return Object.defineProperty(t,e,{configurable:!0,get:function(){return u.call(this)},set:function(p){r=""+p,s.call(this,p)}}),Object.defineProperty(t,e,{enumerable:a.enumerable}),{getValue:function(){return r},setValue:function(p){r=""+p},stopTracking:function(){t._valueTracker=null,delete t[e]}}}}function Ri(t){t._valueTracker||(t._valueTracker=Hg(t))}function Ld(t){if(!t)return!1;var e=t._valueTracker;if(!e)return!0;var a=e.getValue(),r="";return t&&(r=$d(t)?t.checked?"true":"false":t.value),t=r,t!==a?(e.setValue(t),!0):!1}function zi(t){if(t=t||(typeof document<"u"?document:void 0),typeof t>"u")return null;try{return t.activeElement||t.body}catch{return t.body}}var $g=/[\n"\\]/g;function Ye(t){return t.replace($g,function(e){return"\\"+e.charCodeAt(0).toString(16)+" "})}function Zu(t,e,a,r,u,s,p,b){t.name="",p!=null&&typeof p!="function"&&typeof p!="symbol"&&typeof p!="boolean"?t.type=p:t.removeAttribute("type"),e!=null?p==="number"?(e===0&&t.value===""||t.value!=e)&&(t.value=""+Le(e)):t.value!==""+Le(e)&&(t.value=""+Le(e)):p!=="submit"&&p!=="reset"||t.removeAttribute("value"),e!=null?Qu(t,p,Le(e)):a!=null?Qu(t,p,Le(a)):r!=null&&t.removeAttribute("value"),u==null&&s!=null&&(t.defaultChecked=!!s),u!=null&&(t.checked=u&&typeof u!="function"&&typeof u!="symbol"),b!=null&&typeof b!="function"&&typeof b!="symbol"&&typeof b!="boolean"?t.name=""+Le(b):t.removeAttribute("name")}function Yd(t,e,a,r,u,s,p,b){if(s!=null&&typeof s!="function"&&typeof s!="symbol"&&typeof s!="boolean"&&(t.type=s),e!=null||a!=null){if(!(s!=="submit"&&s!=="reset"||e!=null))return;a=a!=null?""+Le(a):"",e=e!=null?""+Le(e):a,b||e===t.value||(t.value=e),t.defaultValue=e}r=r??u,r=typeof r!="function"&&typeof r!="symbol"&&!!r,t.checked=b?t.checked:!!r,t.defaultChecked=!!r,p!=null&&typeof p!="function"&&typeof p!="symbol"&&typeof p!="boolean"&&(t.name=p)}function Qu(t,e,a){e==="number"&&zi(t.ownerDocument)===t||t.defaultValue===""+a||(t.defaultValue=""+a)}function Ia(t,e,a,r){if(t=t.options,e){e={};for(var u=0;u<a.length;u++)e["$"+a[u]]=!0;for(a=0;a<t.length;a++)u=e.hasOwnProperty("$"+t[a].value),t[a].selected!==u&&(t[a].selected=u),u&&r&&(t[a].defaultSelected=!0)}else{for(a=""+Le(a),e=null,u=0;u<t.length;u++){if(t[u].value===a){t[u].selected=!0,r&&(t[u].defaultSelected=!0);return}e!==null||t[u].disabled||(e=t[u])}e!==null&&(e.selected=!0)}}function qd(t,e,a){if(e!=null&&(e=""+Le(e),e!==t.value&&(t.value=e),a==null)){t.defaultValue!==e&&(t.defaultValue=e);return}t.defaultValue=a!=null?""+Le(a):""}function Xd(t,e,a,r){if(e==null){if(r!=null){if(a!=null)throw Error(o(92));if(Q(r)){if(1<r.length)throw Error(o(93));r=r[0]}a=r}a==null&&(a=""),e=a}a=Le(e),t.defaultValue=a,r=t.textContent,r===a&&r!==""&&r!==null&&(t.value=r)}function tl(t,e){if(e){var a=t.firstChild;if(a&&a===t.lastChild&&a.nodeType===3){a.nodeValue=e;return}}t.textContent=e}var Lg=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function Gd(t,e,a){var r=e.indexOf("--")===0;a==null||typeof a=="boolean"||a===""?r?t.setProperty(e,""):e==="float"?t.cssFloat="":t[e]="":r?t.setProperty(e,a):typeof a!="number"||a===0||Lg.has(e)?e==="float"?t.cssFloat=a:t[e]=(""+a).trim():t[e]=a+"px"}function Vd(t,e,a){if(e!=null&&typeof e!="object")throw Error(o(62));if(t=t.style,a!=null){for(var r in a)!a.hasOwnProperty(r)||e!=null&&e.hasOwnProperty(r)||(r.indexOf("--")===0?t.setProperty(r,""):r==="float"?t.cssFloat="":t[r]="");for(var u in e)r=e[u],e.hasOwnProperty(u)&&a[u]!==r&&Gd(t,u,r)}else for(var s in e)e.hasOwnProperty(s)&&Gd(t,s,e[s])}function Ku(t){if(t.indexOf("-")===-1)return!1;switch(t){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var Yg=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),qg=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function Di(t){return qg.test(""+t)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":t}var Ju=null;function Wu(t){return t=t.target||t.srcElement||window,t.correspondingUseElement&&(t=t.correspondingUseElement),t.nodeType===3?t.parentNode:t}var el=null,nl=null;function kd(t){var e=Ja(t);if(e&&(t=e.stateNode)){var a=t[ge]||null;t:switch(t=e.stateNode,e.type){case"input":if(Zu(t,a.value,a.defaultValue,a.defaultValue,a.checked,a.defaultChecked,a.type,a.name),e=a.name,a.type==="radio"&&e!=null){for(a=t;a.parentNode;)a=a.parentNode;for(a=a.querySelectorAll('input[name="'+Ye(""+e)+'"][type="radio"]'),e=0;e<a.length;e++){var r=a[e];if(r!==t&&r.form===t.form){var u=r[ge]||null;if(!u)throw Error(o(90));Zu(r,u.value,u.defaultValue,u.defaultValue,u.checked,u.defaultChecked,u.type,u.name)}}for(e=0;e<a.length;e++)r=a[e],r.form===t.form&&Ld(r)}break t;case"textarea":qd(t,a.value,a.defaultValue);break t;case"select":e=a.value,e!=null&&Ia(t,!!a.multiple,e,!1)}}}var Pu=!1;function Zd(t,e,a){if(Pu)return t(e,a);Pu=!0;try{var r=t(e);return r}finally{if(Pu=!1,(el!==null||nl!==null)&&(So(),el&&(e=el,t=nl,nl=el=null,kd(e),t)))for(e=0;e<t.length;e++)kd(t[e])}}function nr(t,e){var a=t.stateNode;if(a===null)return null;var r=a[ge]||null;if(r===null)return null;a=r[e];t:switch(e){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(r=!r.disabled)||(t=t.type,r=!(t==="button"||t==="input"||t==="select"||t==="textarea")),t=!r;break t;default:t=!1}if(t)return null;if(a&&typeof a!="function")throw Error(o(231,e,typeof a));return a}var xn=!(typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"),Fu=!1;if(xn)try{var ar={};Object.defineProperty(ar,"passive",{get:function(){Fu=!0}}),window.addEventListener("test",ar,ar),window.removeEventListener("test",ar,ar)}catch{Fu=!1}var Zn=null,Iu=null,Bi=null;function Qd(){if(Bi)return Bi;var t,e=Iu,a=e.length,r,u="value"in Zn?Zn.value:Zn.textContent,s=u.length;for(t=0;t<a&&e[t]===u[t];t++);var p=a-t;for(r=1;r<=p&&e[a-r]===u[s-r];r++);return Bi=u.slice(t,1<r?1-r:void 0)}function Ui(t){var e=t.keyCode;return"charCode"in t?(t=t.charCode,t===0&&e===13&&(t=13)):t=e,t===10&&(t=13),32<=t||t===13?t:0}function Hi(){return!0}function Kd(){return!1}function be(t){function e(a,r,u,s,p){this._reactName=a,this._targetInst=u,this.type=r,this.nativeEvent=s,this.target=p,this.currentTarget=null;for(var b in t)t.hasOwnProperty(b)&&(a=t[b],this[b]=a?a(s):s[b]);return this.isDefaultPrevented=(s.defaultPrevented!=null?s.defaultPrevented:s.returnValue===!1)?Hi:Kd,this.isPropagationStopped=Kd,this}return v(e.prototype,{preventDefault:function(){this.defaultPrevented=!0;var a=this.nativeEvent;a&&(a.preventDefault?a.preventDefault():typeof a.returnValue!="unknown"&&(a.returnValue=!1),this.isDefaultPrevented=Hi)},stopPropagation:function(){var a=this.nativeEvent;a&&(a.stopPropagation?a.stopPropagation():typeof a.cancelBubble!="unknown"&&(a.cancelBubble=!0),this.isPropagationStopped=Hi)},persist:function(){},isPersistent:Hi}),e}var Sa={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(t){return t.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},$i=be(Sa),lr=v({},Sa,{view:0,detail:0}),Xg=be(lr),tc,ec,rr,Li=v({},lr,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:ac,button:0,buttons:0,relatedTarget:function(t){return t.relatedTarget===void 0?t.fromElement===t.srcElement?t.toElement:t.fromElement:t.relatedTarget},movementX:function(t){return"movementX"in t?t.movementX:(t!==rr&&(rr&&t.type==="mousemove"?(tc=t.screenX-rr.screenX,ec=t.screenY-rr.screenY):ec=tc=0,rr=t),tc)},movementY:function(t){return"movementY"in t?t.movementY:ec}}),Jd=be(Li),Gg=v({},Li,{dataTransfer:0}),Vg=be(Gg),kg=v({},lr,{relatedTarget:0}),nc=be(kg),Zg=v({},Sa,{animationName:0,elapsedTime:0,pseudoElement:0}),Qg=be(Zg),Kg=v({},Sa,{clipboardData:function(t){return"clipboardData"in t?t.clipboardData:window.clipboardData}}),Jg=be(Kg),Wg=v({},Sa,{data:0}),Wd=be(Wg),Pg={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},Fg={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},Ig={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function t1(t){var e=this.nativeEvent;return e.getModifierState?e.getModifierState(t):(t=Ig[t])?!!e[t]:!1}function ac(){return t1}var e1=v({},lr,{key:function(t){if(t.key){var e=Pg[t.key]||t.key;if(e!=="Unidentified")return e}return t.type==="keypress"?(t=Ui(t),t===13?"Enter":String.fromCharCode(t)):t.type==="keydown"||t.type==="keyup"?Fg[t.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:ac,charCode:function(t){return t.type==="keypress"?Ui(t):0},keyCode:function(t){return t.type==="keydown"||t.type==="keyup"?t.keyCode:0},which:function(t){return t.type==="keypress"?Ui(t):t.type==="keydown"||t.type==="keyup"?t.keyCode:0}}),n1=be(e1),a1=v({},Li,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0}),Pd=be(a1),l1=v({},lr,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:ac}),r1=be(l1),i1=v({},Sa,{propertyName:0,elapsedTime:0,pseudoElement:0}),o1=be(i1),u1=v({},Li,{deltaX:function(t){return"deltaX"in t?t.deltaX:"wheelDeltaX"in t?-t.wheelDeltaX:0},deltaY:function(t){return"deltaY"in t?t.deltaY:"wheelDeltaY"in t?-t.wheelDeltaY:"wheelDelta"in t?-t.wheelDelta:0},deltaZ:0,deltaMode:0}),c1=be(u1),s1=v({},Sa,{newState:0,oldState:0}),f1=be(s1),d1=[9,13,27,32],lc=xn&&"CompositionEvent"in window,ir=null;xn&&"documentMode"in document&&(ir=document.documentMode);var h1=xn&&"TextEvent"in window&&!ir,Fd=xn&&(!lc||ir&&8<ir&&11>=ir),Id=" ",th=!1;function eh(t,e){switch(t){case"keyup":return d1.indexOf(e.keyCode)!==-1;case"keydown":return e.keyCode!==229;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function nh(t){return t=t.detail,typeof t=="object"&&"data"in t?t.data:null}var al=!1;function m1(t,e){switch(t){case"compositionend":return nh(e);case"keypress":return e.which!==32?null:(th=!0,Id);case"textInput":return t=e.data,t===Id&&th?null:t;default:return null}}function y1(t,e){if(al)return t==="compositionend"||!lc&&eh(t,e)?(t=Qd(),Bi=Iu=Zn=null,al=!1,t):null;switch(t){case"paste":return null;case"keypress":if(!(e.ctrlKey||e.altKey||e.metaKey)||e.ctrlKey&&e.altKey){if(e.char&&1<e.char.length)return e.char;if(e.which)return String.fromCharCode(e.which)}return null;case"compositionend":return Fd&&e.locale!=="ko"?null:e.data;default:return null}}var v1={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function ah(t){var e=t&&t.nodeName&&t.nodeName.toLowerCase();return e==="input"?!!v1[t.type]:e==="textarea"}function lh(t,e,a,r){el?nl?nl.push(r):nl=[r]:el=r,e=Co(e,"onChange"),0<e.length&&(a=new $i("onChange","change",null,a,r),t.push({event:a,listeners:e}))}var or=null,ur=null;function p1(t){$y(t,0)}function Yi(t){var e=er(t);if(Ld(e))return t}function rh(t,e){if(t==="change")return e}var ih=!1;if(xn){var rc;if(xn){var ic="oninput"in document;if(!ic){var oh=document.createElement("div");oh.setAttribute("oninput","return;"),ic=typeof oh.oninput=="function"}rc=ic}else rc=!1;ih=rc&&(!document.documentMode||9<document.documentMode)}function uh(){or&&(or.detachEvent("onpropertychange",ch),ur=or=null)}function ch(t){if(t.propertyName==="value"&&Yi(ur)){var e=[];lh(e,ur,t,Wu(t)),Zd(p1,e)}}function g1(t,e,a){t==="focusin"?(uh(),or=e,ur=a,or.attachEvent("onpropertychange",ch)):t==="focusout"&&uh()}function b1(t){if(t==="selectionchange"||t==="keyup"||t==="keydown")return Yi(ur)}function S1(t,e){if(t==="click")return Yi(e)}function _1(t,e){if(t==="input"||t==="change")return Yi(e)}function x1(t,e){return t===e&&(t!==0||1/t===1/e)||t!==t&&e!==e}var Ne=typeof Object.is=="function"?Object.is:x1;function cr(t,e){if(Ne(t,e))return!0;if(typeof t!="object"||t===null||typeof e!="object"||e===null)return!1;var a=Object.keys(t),r=Object.keys(e);if(a.length!==r.length)return!1;for(r=0;r<a.length;r++){var u=a[r];if(!$e.call(e,u)||!Ne(t[u],e[u]))return!1}return!0}function sh(t){for(;t&&t.firstChild;)t=t.firstChild;return t}function fh(t,e){var a=sh(t);t=0;for(var r;a;){if(a.nodeType===3){if(r=t+a.textContent.length,t<=e&&r>=e)return{node:a,offset:e-t};t=r}t:{for(;a;){if(a.nextSibling){a=a.nextSibling;break t}a=a.parentNode}a=void 0}a=sh(a)}}function dh(t,e){return t&&e?t===e?!0:t&&t.nodeType===3?!1:e&&e.nodeType===3?dh(t,e.parentNode):"contains"in t?t.contains(e):t.compareDocumentPosition?!!(t.compareDocumentPosition(e)&16):!1:!1}function hh(t){t=t!=null&&t.ownerDocument!=null&&t.ownerDocument.defaultView!=null?t.ownerDocument.defaultView:window;for(var e=zi(t.document);e instanceof t.HTMLIFrameElement;){try{var a=typeof e.contentWindow.location.href=="string"}catch{a=!1}if(a)t=e.contentWindow;else break;e=zi(t.document)}return e}function oc(t){var e=t&&t.nodeName&&t.nodeName.toLowerCase();return e&&(e==="input"&&(t.type==="text"||t.type==="search"||t.type==="tel"||t.type==="url"||t.type==="password")||e==="textarea"||t.contentEditable==="true")}var T1=xn&&"documentMode"in document&&11>=document.documentMode,ll=null,uc=null,sr=null,cc=!1;function mh(t,e,a){var r=a.window===a?a.document:a.nodeType===9?a:a.ownerDocument;cc||ll==null||ll!==zi(r)||(r=ll,"selectionStart"in r&&oc(r)?r={start:r.selectionStart,end:r.selectionEnd}:(r=(r.ownerDocument&&r.ownerDocument.defaultView||window).getSelection(),r={anchorNode:r.anchorNode,anchorOffset:r.anchorOffset,focusNode:r.focusNode,focusOffset:r.focusOffset}),sr&&cr(sr,r)||(sr=r,r=Co(uc,"onSelect"),0<r.length&&(e=new $i("onSelect","select",null,e,a),t.push({event:e,listeners:r}),e.target=ll)))}function _a(t,e){var a={};return a[t.toLowerCase()]=e.toLowerCase(),a["Webkit"+t]="webkit"+e,a["Moz"+t]="moz"+e,a}var rl={animationend:_a("Animation","AnimationEnd"),animationiteration:_a("Animation","AnimationIteration"),animationstart:_a("Animation","AnimationStart"),transitionrun:_a("Transition","TransitionRun"),transitionstart:_a("Transition","TransitionStart"),transitioncancel:_a("Transition","TransitionCancel"),transitionend:_a("Transition","TransitionEnd")},sc={},yh={};xn&&(yh=document.createElement("div").style,"AnimationEvent"in window||(delete rl.animationend.animation,delete rl.animationiteration.animation,delete rl.animationstart.animation),"TransitionEvent"in window||delete rl.transitionend.transition);function xa(t){if(sc[t])return sc[t];if(!rl[t])return t;var e=rl[t],a;for(a in e)if(e.hasOwnProperty(a)&&a in yh)return sc[t]=e[a];return t}var vh=xa("animationend"),ph=xa("animationiteration"),gh=xa("animationstart"),w1=xa("transitionrun"),E1=xa("transitionstart"),C1=xa("transitioncancel"),bh=xa("transitionend"),Sh=new Map,fc="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll toggle touchMove waiting wheel".split(" ");fc.push("scrollEnd");function Fe(t,e){Sh.set(t,e),ba(e,[t])}var _h=new WeakMap;function qe(t,e){if(typeof t=="object"&&t!==null){var a=_h.get(t);return a!==void 0?a:(e={value:t,source:e,stack:Hd(e)},_h.set(t,e),e)}return{value:t,source:e,stack:Hd(e)}}var Xe=[],il=0,dc=0;function qi(){for(var t=il,e=dc=il=0;e<t;){var a=Xe[e];Xe[e++]=null;var r=Xe[e];Xe[e++]=null;var u=Xe[e];Xe[e++]=null;var s=Xe[e];if(Xe[e++]=null,r!==null&&u!==null){var p=r.pending;p===null?u.next=u:(u.next=p.next,p.next=u),r.pending=u}s!==0&&xh(a,u,s)}}function Xi(t,e,a,r){Xe[il++]=t,Xe[il++]=e,Xe[il++]=a,Xe[il++]=r,dc|=r,t.lanes|=r,t=t.alternate,t!==null&&(t.lanes|=r)}function hc(t,e,a,r){return Xi(t,e,a,r),Gi(t)}function ol(t,e){return Xi(t,null,null,e),Gi(t)}function xh(t,e,a){t.lanes|=a;var r=t.alternate;r!==null&&(r.lanes|=a);for(var u=!1,s=t.return;s!==null;)s.childLanes|=a,r=s.alternate,r!==null&&(r.childLanes|=a),s.tag===22&&(t=s.stateNode,t===null||t._visibility&1||(u=!0)),t=s,s=s.return;return t.tag===3?(s=t.stateNode,u&&e!==null&&(u=31-je(a),t=s.hiddenUpdates,r=t[u],r===null?t[u]=[e]:r.push(e),e.lane=a|536870912),s):null}function Gi(t){if(50<Br)throw Br=0,bs=null,Error(o(185));for(var e=t.return;e!==null;)t=e,e=t.return;return t.tag===3?t.stateNode:null}var ul={};function A1(t,e,a,r){this.tag=t,this.key=a,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=e,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=r,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function Oe(t,e,a,r){return new A1(t,e,a,r)}function mc(t){return t=t.prototype,!(!t||!t.isReactComponent)}function Tn(t,e){var a=t.alternate;return a===null?(a=Oe(t.tag,e,t.key,t.mode),a.elementType=t.elementType,a.type=t.type,a.stateNode=t.stateNode,a.alternate=t,t.alternate=a):(a.pendingProps=e,a.type=t.type,a.flags=0,a.subtreeFlags=0,a.deletions=null),a.flags=t.flags&65011712,a.childLanes=t.childLanes,a.lanes=t.lanes,a.child=t.child,a.memoizedProps=t.memoizedProps,a.memoizedState=t.memoizedState,a.updateQueue=t.updateQueue,e=t.dependencies,a.dependencies=e===null?null:{lanes:e.lanes,firstContext:e.firstContext},a.sibling=t.sibling,a.index=t.index,a.ref=t.ref,a.refCleanup=t.refCleanup,a}function Th(t,e){t.flags&=65011714;var a=t.alternate;return a===null?(t.childLanes=0,t.lanes=e,t.child=null,t.subtreeFlags=0,t.memoizedProps=null,t.memoizedState=null,t.updateQueue=null,t.dependencies=null,t.stateNode=null):(t.childLanes=a.childLanes,t.lanes=a.lanes,t.child=a.child,t.subtreeFlags=0,t.deletions=null,t.memoizedProps=a.memoizedProps,t.memoizedState=a.memoizedState,t.updateQueue=a.updateQueue,t.type=a.type,e=a.dependencies,t.dependencies=e===null?null:{lanes:e.lanes,firstContext:e.firstContext}),t}function Vi(t,e,a,r,u,s){var p=0;if(r=t,typeof t=="function")mc(t)&&(p=1);else if(typeof t=="string")p=Nb(t,a,nt.current)?26:t==="html"||t==="head"||t==="body"?27:5;else t:switch(t){case it:return t=Oe(31,a,e,u),t.elementType=it,t.lanes=s,t;case w:return Ta(a.children,u,s,e);case N:p=8,u|=24;break;case B:return t=Oe(12,a,e,u|2),t.elementType=B,t.lanes=s,t;case $:return t=Oe(13,a,e,u),t.elementType=$,t.lanes=s,t;case V:return t=Oe(19,a,e,u),t.elementType=V,t.lanes=s,t;default:if(typeof t=="object"&&t!==null)switch(t.$$typeof){case O:case z:p=10;break t;case j:p=9;break t;case L:p=11;break t;case Z:p=14;break t;case W:p=16,r=null;break t}p=29,a=Error(o(130,t===null?"null":typeof t,"")),r=null}return e=Oe(p,a,e,u),e.elementType=t,e.type=r,e.lanes=s,e}function Ta(t,e,a,r){return t=Oe(7,t,r,e),t.lanes=a,t}function yc(t,e,a){return t=Oe(6,t,null,e),t.lanes=a,t}function vc(t,e,a){return e=Oe(4,t.children!==null?t.children:[],t.key,e),e.lanes=a,e.stateNode={containerInfo:t.containerInfo,pendingChildren:null,implementation:t.implementation},e}var cl=[],sl=0,ki=null,Zi=0,Ge=[],Ve=0,wa=null,wn=1,En="";function Ea(t,e){cl[sl++]=Zi,cl[sl++]=ki,ki=t,Zi=e}function wh(t,e,a){Ge[Ve++]=wn,Ge[Ve++]=En,Ge[Ve++]=wa,wa=t;var r=wn;t=En;var u=32-je(r)-1;r&=~(1<<u),a+=1;var s=32-je(e)+u;if(30<s){var p=u-u%5;s=(r&(1<<p)-1).toString(32),r>>=p,u-=p,wn=1<<32-je(e)+u|a<<u|r,En=s+t}else wn=1<<s|a<<u|r,En=t}function pc(t){t.return!==null&&(Ea(t,1),wh(t,1,0))}function gc(t){for(;t===ki;)ki=cl[--sl],cl[sl]=null,Zi=cl[--sl],cl[sl]=null;for(;t===wa;)wa=Ge[--Ve],Ge[Ve]=null,En=Ge[--Ve],Ge[Ve]=null,wn=Ge[--Ve],Ge[Ve]=null}var me=null,Yt=null,At=!1,Ca=null,rn=!1,bc=Error(o(519));function Aa(t){var e=Error(o(418,""));throw hr(qe(e,t)),bc}function Eh(t){var e=t.stateNode,a=t.type,r=t.memoizedProps;switch(e[se]=t,e[ge]=r,a){case"dialog":wt("cancel",e),wt("close",e);break;case"iframe":case"object":case"embed":wt("load",e);break;case"video":case"audio":for(a=0;a<Hr.length;a++)wt(Hr[a],e);break;case"source":wt("error",e);break;case"img":case"image":case"link":wt("error",e),wt("load",e);break;case"details":wt("toggle",e);break;case"input":wt("invalid",e),Yd(e,r.value,r.defaultValue,r.checked,r.defaultChecked,r.type,r.name,!0),Ri(e);break;case"select":wt("invalid",e);break;case"textarea":wt("invalid",e),Xd(e,r.value,r.defaultValue,r.children),Ri(e)}a=r.children,typeof a!="string"&&typeof a!="number"&&typeof a!="bigint"||e.textContent===""+a||r.suppressHydrationWarning===!0||Xy(e.textContent,a)?(r.popover!=null&&(wt("beforetoggle",e),wt("toggle",e)),r.onScroll!=null&&wt("scroll",e),r.onScrollEnd!=null&&wt("scrollend",e),r.onClick!=null&&(e.onclick=Ao),e=!0):e=!1,e||Aa(t)}function Ch(t){for(me=t.return;me;)switch(me.tag){case 5:case 13:rn=!1;return;case 27:case 3:rn=!0;return;default:me=me.return}}function fr(t){if(t!==me)return!1;if(!At)return Ch(t),At=!0,!1;var e=t.tag,a;if((a=e!==3&&e!==27)&&((a=e===5)&&(a=t.type,a=!(a!=="form"&&a!=="button")||Bs(t.type,t.memoizedProps)),a=!a),a&&Yt&&Aa(t),Ch(t),e===13){if(t=t.memoizedState,t=t!==null?t.dehydrated:null,!t)throw Error(o(317));t:{for(t=t.nextSibling,e=0;t;){if(t.nodeType===8)if(a=t.data,a==="/$"){if(e===0){Yt=tn(t.nextSibling);break t}e--}else a!=="$"&&a!=="$!"&&a!=="$?"||e++;t=t.nextSibling}Yt=null}}else e===27?(e=Yt,ua(t.type)?(t=Ls,Ls=null,Yt=t):Yt=e):Yt=me?tn(t.stateNode.nextSibling):null;return!0}function dr(){Yt=me=null,At=!1}function Ah(){var t=Ca;return t!==null&&(xe===null?xe=t:xe.push.apply(xe,t),Ca=null),t}function hr(t){Ca===null?Ca=[t]:Ca.push(t)}var Sc=X(null),ja=null,Cn=null;function Qn(t,e,a){P(Sc,e._currentValue),e._currentValue=a}function An(t){t._currentValue=Sc.current,I(Sc)}function _c(t,e,a){for(;t!==null;){var r=t.alternate;if((t.childLanes&e)!==e?(t.childLanes|=e,r!==null&&(r.childLanes|=e)):r!==null&&(r.childLanes&e)!==e&&(r.childLanes|=e),t===a)break;t=t.return}}function xc(t,e,a,r){var u=t.child;for(u!==null&&(u.return=t);u!==null;){var s=u.dependencies;if(s!==null){var p=u.child;s=s.firstContext;t:for(;s!==null;){var b=s;s=u;for(var A=0;A<e.length;A++)if(b.context===e[A]){s.lanes|=a,b=s.alternate,b!==null&&(b.lanes|=a),_c(s.return,a,t),r||(p=null);break t}s=b.next}}else if(u.tag===18){if(p=u.return,p===null)throw Error(o(341));p.lanes|=a,s=p.alternate,s!==null&&(s.lanes|=a),_c(p,a,t),p=null}else p=u.child;if(p!==null)p.return=u;else for(p=u;p!==null;){if(p===t){p=null;break}if(u=p.sibling,u!==null){u.return=p.return,p=u;break}p=p.return}u=p}}function mr(t,e,a,r){t=null;for(var u=e,s=!1;u!==null;){if(!s){if((u.flags&524288)!==0)s=!0;else if((u.flags&262144)!==0)break}if(u.tag===10){var p=u.alternate;if(p===null)throw Error(o(387));if(p=p.memoizedProps,p!==null){var b=u.type;Ne(u.pendingProps.value,p.value)||(t!==null?t.push(b):t=[b])}}else if(u===bt.current){if(p=u.alternate,p===null)throw Error(o(387));p.memoizedState.memoizedState!==u.memoizedState.memoizedState&&(t!==null?t.push(Gr):t=[Gr])}u=u.return}t!==null&&xc(e,t,a,r),e.flags|=262144}function Qi(t){for(t=t.firstContext;t!==null;){if(!Ne(t.context._currentValue,t.memoizedValue))return!0;t=t.next}return!1}function Na(t){ja=t,Cn=null,t=t.dependencies,t!==null&&(t.firstContext=null)}function fe(t){return jh(ja,t)}function Ki(t,e){return ja===null&&Na(t),jh(t,e)}function jh(t,e){var a=e._currentValue;if(e={context:e,memoizedValue:a,next:null},Cn===null){if(t===null)throw Error(o(308));Cn=e,t.dependencies={lanes:0,firstContext:e},t.flags|=524288}else Cn=Cn.next=e;return a}var j1=typeof AbortController<"u"?AbortController:function(){var t=[],e=this.signal={aborted:!1,addEventListener:function(a,r){t.push(r)}};this.abort=function(){e.aborted=!0,t.forEach(function(a){return a()})}},N1=n.unstable_scheduleCallback,O1=n.unstable_NormalPriority,Kt={$$typeof:z,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function Tc(){return{controller:new j1,data:new Map,refCount:0}}function yr(t){t.refCount--,t.refCount===0&&N1(O1,function(){t.controller.abort()})}var vr=null,wc=0,fl=0,dl=null;function M1(t,e){if(vr===null){var a=vr=[];wc=0,fl=Cs(),dl={status:"pending",value:void 0,then:function(r){a.push(r)}}}return wc++,e.then(Nh,Nh),e}function Nh(){if(--wc===0&&vr!==null){dl!==null&&(dl.status="fulfilled");var t=vr;vr=null,fl=0,dl=null;for(var e=0;e<t.length;e++)(0,t[e])()}}function R1(t,e){var a=[],r={status:"pending",value:null,reason:null,then:function(u){a.push(u)}};return t.then(function(){r.status="fulfilled",r.value=e;for(var u=0;u<a.length;u++)(0,a[u])(e)},function(u){for(r.status="rejected",r.reason=u,u=0;u<a.length;u++)(0,a[u])(void 0)}),r}var Oh=E.S;E.S=function(t,e){typeof e=="object"&&e!==null&&typeof e.then=="function"&&M1(t,e),Oh!==null&&Oh(t,e)};var Oa=X(null);function Ec(){var t=Oa.current;return t!==null?t:Bt.pooledCache}function Ji(t,e){e===null?P(Oa,Oa.current):P(Oa,e.pool)}function Mh(){var t=Ec();return t===null?null:{parent:Kt._currentValue,pool:t}}var pr=Error(o(460)),Rh=Error(o(474)),Wi=Error(o(542)),Cc={then:function(){}};function zh(t){return t=t.status,t==="fulfilled"||t==="rejected"}function Pi(){}function Dh(t,e,a){switch(a=t[a],a===void 0?t.push(e):a!==e&&(e.then(Pi,Pi),e=a),e.status){case"fulfilled":return e.value;case"rejected":throw t=e.reason,Uh(t),t;default:if(typeof e.status=="string")e.then(Pi,Pi);else{if(t=Bt,t!==null&&100<t.shellSuspendCounter)throw Error(o(482));t=e,t.status="pending",t.then(function(r){if(e.status==="pending"){var u=e;u.status="fulfilled",u.value=r}},function(r){if(e.status==="pending"){var u=e;u.status="rejected",u.reason=r}})}switch(e.status){case"fulfilled":return e.value;case"rejected":throw t=e.reason,Uh(t),t}throw gr=e,pr}}var gr=null;function Bh(){if(gr===null)throw Error(o(459));var t=gr;return gr=null,t}function Uh(t){if(t===pr||t===Wi)throw Error(o(483))}var Kn=!1;function Ac(t){t.updateQueue={baseState:t.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function jc(t,e){t=t.updateQueue,e.updateQueue===t&&(e.updateQueue={baseState:t.baseState,firstBaseUpdate:t.firstBaseUpdate,lastBaseUpdate:t.lastBaseUpdate,shared:t.shared,callbacks:null})}function Jn(t){return{lane:t,tag:0,payload:null,callback:null,next:null}}function Wn(t,e,a){var r=t.updateQueue;if(r===null)return null;if(r=r.shared,(Nt&2)!==0){var u=r.pending;return u===null?e.next=e:(e.next=u.next,u.next=e),r.pending=e,e=Gi(t),xh(t,null,a),e}return Xi(t,r,e,a),Gi(t)}function br(t,e,a){if(e=e.updateQueue,e!==null&&(e=e.shared,(a&4194048)!==0)){var r=e.lanes;r&=t.pendingLanes,a|=r,e.lanes=a,Nd(t,a)}}function Nc(t,e){var a=t.updateQueue,r=t.alternate;if(r!==null&&(r=r.updateQueue,a===r)){var u=null,s=null;if(a=a.firstBaseUpdate,a!==null){do{var p={lane:a.lane,tag:a.tag,payload:a.payload,callback:null,next:null};s===null?u=s=p:s=s.next=p,a=a.next}while(a!==null);s===null?u=s=e:s=s.next=e}else u=s=e;a={baseState:r.baseState,firstBaseUpdate:u,lastBaseUpdate:s,shared:r.shared,callbacks:r.callbacks},t.updateQueue=a;return}t=a.lastBaseUpdate,t===null?a.firstBaseUpdate=e:t.next=e,a.lastBaseUpdate=e}var Oc=!1;function Sr(){if(Oc){var t=dl;if(t!==null)throw t}}function _r(t,e,a,r){Oc=!1;var u=t.updateQueue;Kn=!1;var s=u.firstBaseUpdate,p=u.lastBaseUpdate,b=u.shared.pending;if(b!==null){u.shared.pending=null;var A=b,H=A.next;A.next=null,p===null?s=H:p.next=H,p=A;var K=t.alternate;K!==null&&(K=K.updateQueue,b=K.lastBaseUpdate,b!==p&&(b===null?K.firstBaseUpdate=H:b.next=H,K.lastBaseUpdate=A))}if(s!==null){var F=u.baseState;p=0,K=H=A=null,b=s;do{var q=b.lane&-536870913,G=q!==b.lane;if(G?(Et&q)===q:(r&q)===q){q!==0&&q===fl&&(Oc=!0),K!==null&&(K=K.next={lane:0,tag:b.tag,payload:b.payload,callback:null,next:null});t:{var yt=t,dt=b;q=e;var zt=a;switch(dt.tag){case 1:if(yt=dt.payload,typeof yt=="function"){F=yt.call(zt,F,q);break t}F=yt;break t;case 3:yt.flags=yt.flags&-65537|128;case 0:if(yt=dt.payload,q=typeof yt=="function"?yt.call(zt,F,q):yt,q==null)break t;F=v({},F,q);break t;case 2:Kn=!0}}q=b.callback,q!==null&&(t.flags|=64,G&&(t.flags|=8192),G=u.callbacks,G===null?u.callbacks=[q]:G.push(q))}else G={lane:q,tag:b.tag,payload:b.payload,callback:b.callback,next:null},K===null?(H=K=G,A=F):K=K.next=G,p|=q;if(b=b.next,b===null){if(b=u.shared.pending,b===null)break;G=b,b=G.next,G.next=null,u.lastBaseUpdate=G,u.shared.pending=null}}while(!0);K===null&&(A=F),u.baseState=A,u.firstBaseUpdate=H,u.lastBaseUpdate=K,s===null&&(u.shared.lanes=0),la|=p,t.lanes=p,t.memoizedState=F}}function Hh(t,e){if(typeof t!="function")throw Error(o(191,t));t.call(e)}function $h(t,e){var a=t.callbacks;if(a!==null)for(t.callbacks=null,t=0;t<a.length;t++)Hh(a[t],e)}var hl=X(null),Fi=X(0);function Lh(t,e){t=Dn,P(Fi,t),P(hl,e),Dn=t|e.baseLanes}function Mc(){P(Fi,Dn),P(hl,hl.current)}function Rc(){Dn=Fi.current,I(hl),I(Fi)}var Pn=0,St=null,Mt=null,kt=null,Ii=!1,ml=!1,Ma=!1,to=0,xr=0,yl=null,z1=0;function Xt(){throw Error(o(321))}function zc(t,e){if(e===null)return!1;for(var a=0;a<e.length&&a<t.length;a++)if(!Ne(t[a],e[a]))return!1;return!0}function Dc(t,e,a,r,u,s){return Pn=s,St=e,e.memoizedState=null,e.updateQueue=null,e.lanes=0,E.H=t===null||t.memoizedState===null?xm:Tm,Ma=!1,s=a(r,u),Ma=!1,ml&&(s=qh(e,a,r,u)),Yh(t),s}function Yh(t){E.H=io;var e=Mt!==null&&Mt.next!==null;if(Pn=0,kt=Mt=St=null,Ii=!1,xr=0,yl=null,e)throw Error(o(300));t===null||It||(t=t.dependencies,t!==null&&Qi(t)&&(It=!0))}function qh(t,e,a,r){St=t;var u=0;do{if(ml&&(yl=null),xr=0,ml=!1,25<=u)throw Error(o(301));if(u+=1,kt=Mt=null,t.updateQueue!=null){var s=t.updateQueue;s.lastEffect=null,s.events=null,s.stores=null,s.memoCache!=null&&(s.memoCache.index=0)}E.H=Y1,s=e(a,r)}while(ml);return s}function D1(){var t=E.H,e=t.useState()[0];return e=typeof e.then=="function"?Tr(e):e,t=t.useState()[0],(Mt!==null?Mt.memoizedState:null)!==t&&(St.flags|=1024),e}function Bc(){var t=to!==0;return to=0,t}function Uc(t,e,a){e.updateQueue=t.updateQueue,e.flags&=-2053,t.lanes&=~a}function Hc(t){if(Ii){for(t=t.memoizedState;t!==null;){var e=t.queue;e!==null&&(e.pending=null),t=t.next}Ii=!1}Pn=0,kt=Mt=St=null,ml=!1,xr=to=0,yl=null}function Se(){var t={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return kt===null?St.memoizedState=kt=t:kt=kt.next=t,kt}function Zt(){if(Mt===null){var t=St.alternate;t=t!==null?t.memoizedState:null}else t=Mt.next;var e=kt===null?St.memoizedState:kt.next;if(e!==null)kt=e,Mt=t;else{if(t===null)throw St.alternate===null?Error(o(467)):Error(o(310));Mt=t,t={memoizedState:Mt.memoizedState,baseState:Mt.baseState,baseQueue:Mt.baseQueue,queue:Mt.queue,next:null},kt===null?St.memoizedState=kt=t:kt=kt.next=t}return kt}function $c(){return{lastEffect:null,events:null,stores:null,memoCache:null}}function Tr(t){var e=xr;return xr+=1,yl===null&&(yl=[]),t=Dh(yl,t,e),e=St,(kt===null?e.memoizedState:kt.next)===null&&(e=e.alternate,E.H=e===null||e.memoizedState===null?xm:Tm),t}function eo(t){if(t!==null&&typeof t=="object"){if(typeof t.then=="function")return Tr(t);if(t.$$typeof===z)return fe(t)}throw Error(o(438,String(t)))}function Lc(t){var e=null,a=St.updateQueue;if(a!==null&&(e=a.memoCache),e==null){var r=St.alternate;r!==null&&(r=r.updateQueue,r!==null&&(r=r.memoCache,r!=null&&(e={data:r.data.map(function(u){return u.slice()}),index:0})))}if(e==null&&(e={data:[],index:0}),a===null&&(a=$c(),St.updateQueue=a),a.memoCache=e,a=e.data[e.index],a===void 0)for(a=e.data[e.index]=Array(t),r=0;r<t;r++)a[r]=ht;return e.index++,a}function jn(t,e){return typeof e=="function"?e(t):e}function no(t){var e=Zt();return Yc(e,Mt,t)}function Yc(t,e,a){var r=t.queue;if(r===null)throw Error(o(311));r.lastRenderedReducer=a;var u=t.baseQueue,s=r.pending;if(s!==null){if(u!==null){var p=u.next;u.next=s.next,s.next=p}e.baseQueue=u=s,r.pending=null}if(s=t.baseState,u===null)t.memoizedState=s;else{e=u.next;var b=p=null,A=null,H=e,K=!1;do{var F=H.lane&-536870913;if(F!==H.lane?(Et&F)===F:(Pn&F)===F){var q=H.revertLane;if(q===0)A!==null&&(A=A.next={lane:0,revertLane:0,action:H.action,hasEagerState:H.hasEagerState,eagerState:H.eagerState,next:null}),F===fl&&(K=!0);else if((Pn&q)===q){H=H.next,q===fl&&(K=!0);continue}else F={lane:0,revertLane:H.revertLane,action:H.action,hasEagerState:H.hasEagerState,eagerState:H.eagerState,next:null},A===null?(b=A=F,p=s):A=A.next=F,St.lanes|=q,la|=q;F=H.action,Ma&&a(s,F),s=H.hasEagerState?H.eagerState:a(s,F)}else q={lane:F,revertLane:H.revertLane,action:H.action,hasEagerState:H.hasEagerState,eagerState:H.eagerState,next:null},A===null?(b=A=q,p=s):A=A.next=q,St.lanes|=F,la|=F;H=H.next}while(H!==null&&H!==e);if(A===null?p=s:A.next=b,!Ne(s,t.memoizedState)&&(It=!0,K&&(a=dl,a!==null)))throw a;t.memoizedState=s,t.baseState=p,t.baseQueue=A,r.lastRenderedState=s}return u===null&&(r.lanes=0),[t.memoizedState,r.dispatch]}function qc(t){var e=Zt(),a=e.queue;if(a===null)throw Error(o(311));a.lastRenderedReducer=t;var r=a.dispatch,u=a.pending,s=e.memoizedState;if(u!==null){a.pending=null;var p=u=u.next;do s=t(s,p.action),p=p.next;while(p!==u);Ne(s,e.memoizedState)||(It=!0),e.memoizedState=s,e.baseQueue===null&&(e.baseState=s),a.lastRenderedState=s}return[s,r]}function Xh(t,e,a){var r=St,u=Zt(),s=At;if(s){if(a===void 0)throw Error(o(407));a=a()}else a=e();var p=!Ne((Mt||u).memoizedState,a);p&&(u.memoizedState=a,It=!0),u=u.queue;var b=kh.bind(null,r,u,t);if(wr(2048,8,b,[t]),u.getSnapshot!==e||p||kt!==null&&kt.memoizedState.tag&1){if(r.flags|=2048,vl(9,ao(),Vh.bind(null,r,u,a,e),null),Bt===null)throw Error(o(349));s||(Pn&124)!==0||Gh(r,e,a)}return a}function Gh(t,e,a){t.flags|=16384,t={getSnapshot:e,value:a},e=St.updateQueue,e===null?(e=$c(),St.updateQueue=e,e.stores=[t]):(a=e.stores,a===null?e.stores=[t]:a.push(t))}function Vh(t,e,a,r){e.value=a,e.getSnapshot=r,Zh(e)&&Qh(t)}function kh(t,e,a){return a(function(){Zh(e)&&Qh(t)})}function Zh(t){var e=t.getSnapshot;t=t.value;try{var a=e();return!Ne(t,a)}catch{return!0}}function Qh(t){var e=ol(t,2);e!==null&&Be(e,t,2)}function Xc(t){var e=Se();if(typeof t=="function"){var a=t;if(t=a(),Ma){Vn(!0);try{a()}finally{Vn(!1)}}}return e.memoizedState=e.baseState=t,e.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:jn,lastRenderedState:t},e}function Kh(t,e,a,r){return t.baseState=a,Yc(t,Mt,typeof r=="function"?r:jn)}function B1(t,e,a,r,u){if(ro(t))throw Error(o(485));if(t=e.action,t!==null){var s={payload:u,action:t,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(p){s.listeners.push(p)}};E.T!==null?a(!0):s.isTransition=!1,r(s),a=e.pending,a===null?(s.next=e.pending=s,Jh(e,s)):(s.next=a.next,e.pending=a.next=s)}}function Jh(t,e){var a=e.action,r=e.payload,u=t.state;if(e.isTransition){var s=E.T,p={};E.T=p;try{var b=a(u,r),A=E.S;A!==null&&A(p,b),Wh(t,e,b)}catch(H){Gc(t,e,H)}finally{E.T=s}}else try{s=a(u,r),Wh(t,e,s)}catch(H){Gc(t,e,H)}}function Wh(t,e,a){a!==null&&typeof a=="object"&&typeof a.then=="function"?a.then(function(r){Ph(t,e,r)},function(r){return Gc(t,e,r)}):Ph(t,e,a)}function Ph(t,e,a){e.status="fulfilled",e.value=a,Fh(e),t.state=a,e=t.pending,e!==null&&(a=e.next,a===e?t.pending=null:(a=a.next,e.next=a,Jh(t,a)))}function Gc(t,e,a){var r=t.pending;if(t.pending=null,r!==null){r=r.next;do e.status="rejected",e.reason=a,Fh(e),e=e.next;while(e!==r)}t.action=null}function Fh(t){t=t.listeners;for(var e=0;e<t.length;e++)(0,t[e])()}function Ih(t,e){return e}function tm(t,e){if(At){var a=Bt.formState;if(a!==null){t:{var r=St;if(At){if(Yt){e:{for(var u=Yt,s=rn;u.nodeType!==8;){if(!s){u=null;break e}if(u=tn(u.nextSibling),u===null){u=null;break e}}s=u.data,u=s==="F!"||s==="F"?u:null}if(u){Yt=tn(u.nextSibling),r=u.data==="F!";break t}}Aa(r)}r=!1}r&&(e=a[0])}}return a=Se(),a.memoizedState=a.baseState=e,r={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Ih,lastRenderedState:e},a.queue=r,a=bm.bind(null,St,r),r.dispatch=a,r=Xc(!1),s=Kc.bind(null,St,!1,r.queue),r=Se(),u={state:e,dispatch:null,action:t,pending:null},r.queue=u,a=B1.bind(null,St,u,s,a),u.dispatch=a,r.memoizedState=t,[e,a,!1]}function em(t){var e=Zt();return nm(e,Mt,t)}function nm(t,e,a){if(e=Yc(t,e,Ih)[0],t=no(jn)[0],typeof e=="object"&&e!==null&&typeof e.then=="function")try{var r=Tr(e)}catch(p){throw p===pr?Wi:p}else r=e;e=Zt();var u=e.queue,s=u.dispatch;return a!==e.memoizedState&&(St.flags|=2048,vl(9,ao(),U1.bind(null,u,a),null)),[r,s,t]}function U1(t,e){t.action=e}function am(t){var e=Zt(),a=Mt;if(a!==null)return nm(e,a,t);Zt(),e=e.memoizedState,a=Zt();var r=a.queue.dispatch;return a.memoizedState=t,[e,r,!1]}function vl(t,e,a,r){return t={tag:t,create:a,deps:r,inst:e,next:null},e=St.updateQueue,e===null&&(e=$c(),St.updateQueue=e),a=e.lastEffect,a===null?e.lastEffect=t.next=t:(r=a.next,a.next=t,t.next=r,e.lastEffect=t),t}function ao(){return{destroy:void 0,resource:void 0}}function lm(){return Zt().memoizedState}function lo(t,e,a,r){var u=Se();r=r===void 0?null:r,St.flags|=t,u.memoizedState=vl(1|e,ao(),a,r)}function wr(t,e,a,r){var u=Zt();r=r===void 0?null:r;var s=u.memoizedState.inst;Mt!==null&&r!==null&&zc(r,Mt.memoizedState.deps)?u.memoizedState=vl(e,s,a,r):(St.flags|=t,u.memoizedState=vl(1|e,s,a,r))}function rm(t,e){lo(8390656,8,t,e)}function im(t,e){wr(2048,8,t,e)}function om(t,e){return wr(4,2,t,e)}function um(t,e){return wr(4,4,t,e)}function cm(t,e){if(typeof e=="function"){t=t();var a=e(t);return function(){typeof a=="function"?a():e(null)}}if(e!=null)return t=t(),e.current=t,function(){e.current=null}}function sm(t,e,a){a=a!=null?a.concat([t]):null,wr(4,4,cm.bind(null,e,t),a)}function Vc(){}function fm(t,e){var a=Zt();e=e===void 0?null:e;var r=a.memoizedState;return e!==null&&zc(e,r[1])?r[0]:(a.memoizedState=[t,e],t)}function dm(t,e){var a=Zt();e=e===void 0?null:e;var r=a.memoizedState;if(e!==null&&zc(e,r[1]))return r[0];if(r=t(),Ma){Vn(!0);try{t()}finally{Vn(!1)}}return a.memoizedState=[r,e],r}function kc(t,e,a){return a===void 0||(Pn&1073741824)!==0?t.memoizedState=e:(t.memoizedState=a,t=yy(),St.lanes|=t,la|=t,a)}function hm(t,e,a,r){return Ne(a,e)?a:hl.current!==null?(t=kc(t,a,r),Ne(t,e)||(It=!0),t):(Pn&42)===0?(It=!0,t.memoizedState=a):(t=yy(),St.lanes|=t,la|=t,e)}function mm(t,e,a,r,u){var s=Y.p;Y.p=s!==0&&8>s?s:8;var p=E.T,b={};E.T=b,Kc(t,!1,e,a);try{var A=u(),H=E.S;if(H!==null&&H(b,A),A!==null&&typeof A=="object"&&typeof A.then=="function"){var K=R1(A,r);Er(t,e,K,De(t))}else Er(t,e,r,De(t))}catch(F){Er(t,e,{then:function(){},status:"rejected",reason:F},De())}finally{Y.p=s,E.T=p}}function H1(){}function Zc(t,e,a,r){if(t.tag!==5)throw Error(o(476));var u=ym(t).queue;mm(t,u,e,k,a===null?H1:function(){return vm(t),a(r)})}function ym(t){var e=t.memoizedState;if(e!==null)return e;e={memoizedState:k,baseState:k,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:jn,lastRenderedState:k},next:null};var a={};return e.next={memoizedState:a,baseState:a,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:jn,lastRenderedState:a},next:null},t.memoizedState=e,t=t.alternate,t!==null&&(t.memoizedState=e),e}function vm(t){var e=ym(t).next.queue;Er(t,e,{},De())}function Qc(){return fe(Gr)}function pm(){return Zt().memoizedState}function gm(){return Zt().memoizedState}function $1(t){for(var e=t.return;e!==null;){switch(e.tag){case 24:case 3:var a=De();t=Jn(a);var r=Wn(e,t,a);r!==null&&(Be(r,e,a),br(r,e,a)),e={cache:Tc()},t.payload=e;return}e=e.return}}function L1(t,e,a){var r=De();a={lane:r,revertLane:0,action:a,hasEagerState:!1,eagerState:null,next:null},ro(t)?Sm(e,a):(a=hc(t,e,a,r),a!==null&&(Be(a,t,r),_m(a,e,r)))}function bm(t,e,a){var r=De();Er(t,e,a,r)}function Er(t,e,a,r){var u={lane:r,revertLane:0,action:a,hasEagerState:!1,eagerState:null,next:null};if(ro(t))Sm(e,u);else{var s=t.alternate;if(t.lanes===0&&(s===null||s.lanes===0)&&(s=e.lastRenderedReducer,s!==null))try{var p=e.lastRenderedState,b=s(p,a);if(u.hasEagerState=!0,u.eagerState=b,Ne(b,p))return Xi(t,e,u,0),Bt===null&&qi(),!1}catch{}finally{}if(a=hc(t,e,u,r),a!==null)return Be(a,t,r),_m(a,e,r),!0}return!1}function Kc(t,e,a,r){if(r={lane:2,revertLane:Cs(),action:r,hasEagerState:!1,eagerState:null,next:null},ro(t)){if(e)throw Error(o(479))}else e=hc(t,a,r,2),e!==null&&Be(e,t,2)}function ro(t){var e=t.alternate;return t===St||e!==null&&e===St}function Sm(t,e){ml=Ii=!0;var a=t.pending;a===null?e.next=e:(e.next=a.next,a.next=e),t.pending=e}function _m(t,e,a){if((a&4194048)!==0){var r=e.lanes;r&=t.pendingLanes,a|=r,e.lanes=a,Nd(t,a)}}var io={readContext:fe,use:eo,useCallback:Xt,useContext:Xt,useEffect:Xt,useImperativeHandle:Xt,useLayoutEffect:Xt,useInsertionEffect:Xt,useMemo:Xt,useReducer:Xt,useRef:Xt,useState:Xt,useDebugValue:Xt,useDeferredValue:Xt,useTransition:Xt,useSyncExternalStore:Xt,useId:Xt,useHostTransitionStatus:Xt,useFormState:Xt,useActionState:Xt,useOptimistic:Xt,useMemoCache:Xt,useCacheRefresh:Xt},xm={readContext:fe,use:eo,useCallback:function(t,e){return Se().memoizedState=[t,e===void 0?null:e],t},useContext:fe,useEffect:rm,useImperativeHandle:function(t,e,a){a=a!=null?a.concat([t]):null,lo(4194308,4,cm.bind(null,e,t),a)},useLayoutEffect:function(t,e){return lo(4194308,4,t,e)},useInsertionEffect:function(t,e){lo(4,2,t,e)},useMemo:function(t,e){var a=Se();e=e===void 0?null:e;var r=t();if(Ma){Vn(!0);try{t()}finally{Vn(!1)}}return a.memoizedState=[r,e],r},useReducer:function(t,e,a){var r=Se();if(a!==void 0){var u=a(e);if(Ma){Vn(!0);try{a(e)}finally{Vn(!1)}}}else u=e;return r.memoizedState=r.baseState=u,t={pending:null,lanes:0,dispatch:null,lastRenderedReducer:t,lastRenderedState:u},r.queue=t,t=t.dispatch=L1.bind(null,St,t),[r.memoizedState,t]},useRef:function(t){var e=Se();return t={current:t},e.memoizedState=t},useState:function(t){t=Xc(t);var e=t.queue,a=bm.bind(null,St,e);return e.dispatch=a,[t.memoizedState,a]},useDebugValue:Vc,useDeferredValue:function(t,e){var a=Se();return kc(a,t,e)},useTransition:function(){var t=Xc(!1);return t=mm.bind(null,St,t.queue,!0,!1),Se().memoizedState=t,[!1,t]},useSyncExternalStore:function(t,e,a){var r=St,u=Se();if(At){if(a===void 0)throw Error(o(407));a=a()}else{if(a=e(),Bt===null)throw Error(o(349));(Et&124)!==0||Gh(r,e,a)}u.memoizedState=a;var s={value:a,getSnapshot:e};return u.queue=s,rm(kh.bind(null,r,s,t),[t]),r.flags|=2048,vl(9,ao(),Vh.bind(null,r,s,a,e),null),a},useId:function(){var t=Se(),e=Bt.identifierPrefix;if(At){var a=En,r=wn;a=(r&~(1<<32-je(r)-1)).toString(32)+a,e="«"+e+"R"+a,a=to++,0<a&&(e+="H"+a.toString(32)),e+="»"}else a=z1++,e="«"+e+"r"+a.toString(32)+"»";return t.memoizedState=e},useHostTransitionStatus:Qc,useFormState:tm,useActionState:tm,useOptimistic:function(t){var e=Se();e.memoizedState=e.baseState=t;var a={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return e.queue=a,e=Kc.bind(null,St,!0,a),a.dispatch=e,[t,e]},useMemoCache:Lc,useCacheRefresh:function(){return Se().memoizedState=$1.bind(null,St)}},Tm={readContext:fe,use:eo,useCallback:fm,useContext:fe,useEffect:im,useImperativeHandle:sm,useInsertionEffect:om,useLayoutEffect:um,useMemo:dm,useReducer:no,useRef:lm,useState:function(){return no(jn)},useDebugValue:Vc,useDeferredValue:function(t,e){var a=Zt();return hm(a,Mt.memoizedState,t,e)},useTransition:function(){var t=no(jn)[0],e=Zt().memoizedState;return[typeof t=="boolean"?t:Tr(t),e]},useSyncExternalStore:Xh,useId:pm,useHostTransitionStatus:Qc,useFormState:em,useActionState:em,useOptimistic:function(t,e){var a=Zt();return Kh(a,Mt,t,e)},useMemoCache:Lc,useCacheRefresh:gm},Y1={readContext:fe,use:eo,useCallback:fm,useContext:fe,useEffect:im,useImperativeHandle:sm,useInsertionEffect:om,useLayoutEffect:um,useMemo:dm,useReducer:qc,useRef:lm,useState:function(){return qc(jn)},useDebugValue:Vc,useDeferredValue:function(t,e){var a=Zt();return Mt===null?kc(a,t,e):hm(a,Mt.memoizedState,t,e)},useTransition:function(){var t=qc(jn)[0],e=Zt().memoizedState;return[typeof t=="boolean"?t:Tr(t),e]},useSyncExternalStore:Xh,useId:pm,useHostTransitionStatus:Qc,useFormState:am,useActionState:am,useOptimistic:function(t,e){var a=Zt();return Mt!==null?Kh(a,Mt,t,e):(a.baseState=t,[t,a.queue.dispatch])},useMemoCache:Lc,useCacheRefresh:gm},pl=null,Cr=0;function oo(t){var e=Cr;return Cr+=1,pl===null&&(pl=[]),Dh(pl,t,e)}function Ar(t,e){e=e.props.ref,t.ref=e!==void 0?e:null}function uo(t,e){throw e.$$typeof===S?Error(o(525)):(t=Object.prototype.toString.call(e),Error(o(31,t==="[object Object]"?"object with keys {"+Object.keys(e).join(", ")+"}":t)))}function wm(t){var e=t._init;return e(t._payload)}function Em(t){function e(D,M){if(t){var U=D.deletions;U===null?(D.deletions=[M],D.flags|=16):U.push(M)}}function a(D,M){if(!t)return null;for(;M!==null;)e(D,M),M=M.sibling;return null}function r(D){for(var M=new Map;D!==null;)D.key!==null?M.set(D.key,D):M.set(D.index,D),D=D.sibling;return M}function u(D,M){return D=Tn(D,M),D.index=0,D.sibling=null,D}function s(D,M,U){return D.index=U,t?(U=D.alternate,U!==null?(U=U.index,U<M?(D.flags|=67108866,M):U):(D.flags|=67108866,M)):(D.flags|=1048576,M)}function p(D){return t&&D.alternate===null&&(D.flags|=67108866),D}function b(D,M,U,J){return M===null||M.tag!==6?(M=yc(U,D.mode,J),M.return=D,M):(M=u(M,U),M.return=D,M)}function A(D,M,U,J){var ot=U.type;return ot===w?K(D,M,U.props.children,J,U.key):M!==null&&(M.elementType===ot||typeof ot=="object"&&ot!==null&&ot.$$typeof===W&&wm(ot)===M.type)?(M=u(M,U.props),Ar(M,U),M.return=D,M):(M=Vi(U.type,U.key,U.props,null,D.mode,J),Ar(M,U),M.return=D,M)}function H(D,M,U,J){return M===null||M.tag!==4||M.stateNode.containerInfo!==U.containerInfo||M.stateNode.implementation!==U.implementation?(M=vc(U,D.mode,J),M.return=D,M):(M=u(M,U.children||[]),M.return=D,M)}function K(D,M,U,J,ot){return M===null||M.tag!==7?(M=Ta(U,D.mode,J,ot),M.return=D,M):(M=u(M,U),M.return=D,M)}function F(D,M,U){if(typeof M=="string"&&M!==""||typeof M=="number"||typeof M=="bigint")return M=yc(""+M,D.mode,U),M.return=D,M;if(typeof M=="object"&&M!==null){switch(M.$$typeof){case _:return U=Vi(M.type,M.key,M.props,null,D.mode,U),Ar(U,M),U.return=D,U;case x:return M=vc(M,D.mode,U),M.return=D,M;case W:var J=M._init;return M=J(M._payload),F(D,M,U)}if(Q(M)||at(M))return M=Ta(M,D.mode,U,null),M.return=D,M;if(typeof M.then=="function")return F(D,oo(M),U);if(M.$$typeof===z)return F(D,Ki(D,M),U);uo(D,M)}return null}function q(D,M,U,J){var ot=M!==null?M.key:null;if(typeof U=="string"&&U!==""||typeof U=="number"||typeof U=="bigint")return ot!==null?null:b(D,M,""+U,J);if(typeof U=="object"&&U!==null){switch(U.$$typeof){case _:return U.key===ot?A(D,M,U,J):null;case x:return U.key===ot?H(D,M,U,J):null;case W:return ot=U._init,U=ot(U._payload),q(D,M,U,J)}if(Q(U)||at(U))return ot!==null?null:K(D,M,U,J,null);if(typeof U.then=="function")return q(D,M,oo(U),J);if(U.$$typeof===z)return q(D,M,Ki(D,U),J);uo(D,U)}return null}function G(D,M,U,J,ot){if(typeof J=="string"&&J!==""||typeof J=="number"||typeof J=="bigint")return D=D.get(U)||null,b(M,D,""+J,ot);if(typeof J=="object"&&J!==null){switch(J.$$typeof){case _:return D=D.get(J.key===null?U:J.key)||null,A(M,D,J,ot);case x:return D=D.get(J.key===null?U:J.key)||null,H(M,D,J,ot);case W:var xt=J._init;return J=xt(J._payload),G(D,M,U,J,ot)}if(Q(J)||at(J))return D=D.get(U)||null,K(M,D,J,ot,null);if(typeof J.then=="function")return G(D,M,U,oo(J),ot);if(J.$$typeof===z)return G(D,M,U,Ki(M,J),ot);uo(M,J)}return null}function yt(D,M,U,J){for(var ot=null,xt=null,ct=M,mt=M=0,ee=null;ct!==null&&mt<U.length;mt++){ct.index>mt?(ee=ct,ct=null):ee=ct.sibling;var Ct=q(D,ct,U[mt],J);if(Ct===null){ct===null&&(ct=ee);break}t&&ct&&Ct.alternate===null&&e(D,ct),M=s(Ct,M,mt),xt===null?ot=Ct:xt.sibling=Ct,xt=Ct,ct=ee}if(mt===U.length)return a(D,ct),At&&Ea(D,mt),ot;if(ct===null){for(;mt<U.length;mt++)ct=F(D,U[mt],J),ct!==null&&(M=s(ct,M,mt),xt===null?ot=ct:xt.sibling=ct,xt=ct);return At&&Ea(D,mt),ot}for(ct=r(ct);mt<U.length;mt++)ee=G(ct,D,mt,U[mt],J),ee!==null&&(t&&ee.alternate!==null&&ct.delete(ee.key===null?mt:ee.key),M=s(ee,M,mt),xt===null?ot=ee:xt.sibling=ee,xt=ee);return t&&ct.forEach(function(ha){return e(D,ha)}),At&&Ea(D,mt),ot}function dt(D,M,U,J){if(U==null)throw Error(o(151));for(var ot=null,xt=null,ct=M,mt=M=0,ee=null,Ct=U.next();ct!==null&&!Ct.done;mt++,Ct=U.next()){ct.index>mt?(ee=ct,ct=null):ee=ct.sibling;var ha=q(D,ct,Ct.value,J);if(ha===null){ct===null&&(ct=ee);break}t&&ct&&ha.alternate===null&&e(D,ct),M=s(ha,M,mt),xt===null?ot=ha:xt.sibling=ha,xt=ha,ct=ee}if(Ct.done)return a(D,ct),At&&Ea(D,mt),ot;if(ct===null){for(;!Ct.done;mt++,Ct=U.next())Ct=F(D,Ct.value,J),Ct!==null&&(M=s(Ct,M,mt),xt===null?ot=Ct:xt.sibling=Ct,xt=Ct);return At&&Ea(D,mt),ot}for(ct=r(ct);!Ct.done;mt++,Ct=U.next())Ct=G(ct,D,mt,Ct.value,J),Ct!==null&&(t&&Ct.alternate!==null&&ct.delete(Ct.key===null?mt:Ct.key),M=s(Ct,M,mt),xt===null?ot=Ct:xt.sibling=Ct,xt=Ct);return t&&ct.forEach(function(qb){return e(D,qb)}),At&&Ea(D,mt),ot}function zt(D,M,U,J){if(typeof U=="object"&&U!==null&&U.type===w&&U.key===null&&(U=U.props.children),typeof U=="object"&&U!==null){switch(U.$$typeof){case _:t:{for(var ot=U.key;M!==null;){if(M.key===ot){if(ot=U.type,ot===w){if(M.tag===7){a(D,M.sibling),J=u(M,U.props.children),J.return=D,D=J;break t}}else if(M.elementType===ot||typeof ot=="object"&&ot!==null&&ot.$$typeof===W&&wm(ot)===M.type){a(D,M.sibling),J=u(M,U.props),Ar(J,U),J.return=D,D=J;break t}a(D,M);break}else e(D,M);M=M.sibling}U.type===w?(J=Ta(U.props.children,D.mode,J,U.key),J.return=D,D=J):(J=Vi(U.type,U.key,U.props,null,D.mode,J),Ar(J,U),J.return=D,D=J)}return p(D);case x:t:{for(ot=U.key;M!==null;){if(M.key===ot)if(M.tag===4&&M.stateNode.containerInfo===U.containerInfo&&M.stateNode.implementation===U.implementation){a(D,M.sibling),J=u(M,U.children||[]),J.return=D,D=J;break t}else{a(D,M);break}else e(D,M);M=M.sibling}J=vc(U,D.mode,J),J.return=D,D=J}return p(D);case W:return ot=U._init,U=ot(U._payload),zt(D,M,U,J)}if(Q(U))return yt(D,M,U,J);if(at(U)){if(ot=at(U),typeof ot!="function")throw Error(o(150));return U=ot.call(U),dt(D,M,U,J)}if(typeof U.then=="function")return zt(D,M,oo(U),J);if(U.$$typeof===z)return zt(D,M,Ki(D,U),J);uo(D,U)}return typeof U=="string"&&U!==""||typeof U=="number"||typeof U=="bigint"?(U=""+U,M!==null&&M.tag===6?(a(D,M.sibling),J=u(M,U),J.return=D,D=J):(a(D,M),J=yc(U,D.mode,J),J.return=D,D=J),p(D)):a(D,M)}return function(D,M,U,J){try{Cr=0;var ot=zt(D,M,U,J);return pl=null,ot}catch(ct){if(ct===pr||ct===Wi)throw ct;var xt=Oe(29,ct,null,D.mode);return xt.lanes=J,xt.return=D,xt}finally{}}}var gl=Em(!0),Cm=Em(!1),ke=X(null),on=null;function Fn(t){var e=t.alternate;P(Jt,Jt.current&1),P(ke,t),on===null&&(e===null||hl.current!==null||e.memoizedState!==null)&&(on=t)}function Am(t){if(t.tag===22){if(P(Jt,Jt.current),P(ke,t),on===null){var e=t.alternate;e!==null&&e.memoizedState!==null&&(on=t)}}else In()}function In(){P(Jt,Jt.current),P(ke,ke.current)}function Nn(t){I(ke),on===t&&(on=null),I(Jt)}var Jt=X(0);function co(t){for(var e=t;e!==null;){if(e.tag===13){var a=e.memoizedState;if(a!==null&&(a=a.dehydrated,a===null||a.data==="$?"||$s(a)))return e}else if(e.tag===19&&e.memoizedProps.revealOrder!==void 0){if((e.flags&128)!==0)return e}else if(e.child!==null){e.child.return=e,e=e.child;continue}if(e===t)break;for(;e.sibling===null;){if(e.return===null||e.return===t)return null;e=e.return}e.sibling.return=e.return,e=e.sibling}return null}function Jc(t,e,a,r){e=t.memoizedState,a=a(r,e),a=a==null?e:v({},e,a),t.memoizedState=a,t.lanes===0&&(t.updateQueue.baseState=a)}var Wc={enqueueSetState:function(t,e,a){t=t._reactInternals;var r=De(),u=Jn(r);u.payload=e,a!=null&&(u.callback=a),e=Wn(t,u,r),e!==null&&(Be(e,t,r),br(e,t,r))},enqueueReplaceState:function(t,e,a){t=t._reactInternals;var r=De(),u=Jn(r);u.tag=1,u.payload=e,a!=null&&(u.callback=a),e=Wn(t,u,r),e!==null&&(Be(e,t,r),br(e,t,r))},enqueueForceUpdate:function(t,e){t=t._reactInternals;var a=De(),r=Jn(a);r.tag=2,e!=null&&(r.callback=e),e=Wn(t,r,a),e!==null&&(Be(e,t,a),br(e,t,a))}};function jm(t,e,a,r,u,s,p){return t=t.stateNode,typeof t.shouldComponentUpdate=="function"?t.shouldComponentUpdate(r,s,p):e.prototype&&e.prototype.isPureReactComponent?!cr(a,r)||!cr(u,s):!0}function Nm(t,e,a,r){t=e.state,typeof e.componentWillReceiveProps=="function"&&e.componentWillReceiveProps(a,r),typeof e.UNSAFE_componentWillReceiveProps=="function"&&e.UNSAFE_componentWillReceiveProps(a,r),e.state!==t&&Wc.enqueueReplaceState(e,e.state,null)}function Ra(t,e){var a=e;if("ref"in e){a={};for(var r in e)r!=="ref"&&(a[r]=e[r])}if(t=t.defaultProps){a===e&&(a=v({},a));for(var u in t)a[u]===void 0&&(a[u]=t[u])}return a}var so=typeof reportError=="function"?reportError:function(t){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var e=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof t=="object"&&t!==null&&typeof t.message=="string"?String(t.message):String(t),error:t});if(!window.dispatchEvent(e))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",t);return}console.error(t)};function Om(t){so(t)}function Mm(t){console.error(t)}function Rm(t){so(t)}function fo(t,e){try{var a=t.onUncaughtError;a(e.value,{componentStack:e.stack})}catch(r){setTimeout(function(){throw r})}}function zm(t,e,a){try{var r=t.onCaughtError;r(a.value,{componentStack:a.stack,errorBoundary:e.tag===1?e.stateNode:null})}catch(u){setTimeout(function(){throw u})}}function Pc(t,e,a){return a=Jn(a),a.tag=3,a.payload={element:null},a.callback=function(){fo(t,e)},a}function Dm(t){return t=Jn(t),t.tag=3,t}function Bm(t,e,a,r){var u=a.type.getDerivedStateFromError;if(typeof u=="function"){var s=r.value;t.payload=function(){return u(s)},t.callback=function(){zm(e,a,r)}}var p=a.stateNode;p!==null&&typeof p.componentDidCatch=="function"&&(t.callback=function(){zm(e,a,r),typeof u!="function"&&(ra===null?ra=new Set([this]):ra.add(this));var b=r.stack;this.componentDidCatch(r.value,{componentStack:b!==null?b:""})})}function q1(t,e,a,r,u){if(a.flags|=32768,r!==null&&typeof r=="object"&&typeof r.then=="function"){if(e=a.alternate,e!==null&&mr(e,a,u,!0),a=ke.current,a!==null){switch(a.tag){case 13:return on===null?_s():a.alternate===null&&qt===0&&(qt=3),a.flags&=-257,a.flags|=65536,a.lanes=u,r===Cc?a.flags|=16384:(e=a.updateQueue,e===null?a.updateQueue=new Set([r]):e.add(r),Ts(t,r,u)),!1;case 22:return a.flags|=65536,r===Cc?a.flags|=16384:(e=a.updateQueue,e===null?(e={transitions:null,markerInstances:null,retryQueue:new Set([r])},a.updateQueue=e):(a=e.retryQueue,a===null?e.retryQueue=new Set([r]):a.add(r)),Ts(t,r,u)),!1}throw Error(o(435,a.tag))}return Ts(t,r,u),_s(),!1}if(At)return e=ke.current,e!==null?((e.flags&65536)===0&&(e.flags|=256),e.flags|=65536,e.lanes=u,r!==bc&&(t=Error(o(422),{cause:r}),hr(qe(t,a)))):(r!==bc&&(e=Error(o(423),{cause:r}),hr(qe(e,a))),t=t.current.alternate,t.flags|=65536,u&=-u,t.lanes|=u,r=qe(r,a),u=Pc(t.stateNode,r,u),Nc(t,u),qt!==4&&(qt=2)),!1;var s=Error(o(520),{cause:r});if(s=qe(s,a),Dr===null?Dr=[s]:Dr.push(s),qt!==4&&(qt=2),e===null)return!0;r=qe(r,a),a=e;do{switch(a.tag){case 3:return a.flags|=65536,t=u&-u,a.lanes|=t,t=Pc(a.stateNode,r,t),Nc(a,t),!1;case 1:if(e=a.type,s=a.stateNode,(a.flags&128)===0&&(typeof e.getDerivedStateFromError=="function"||s!==null&&typeof s.componentDidCatch=="function"&&(ra===null||!ra.has(s))))return a.flags|=65536,u&=-u,a.lanes|=u,u=Dm(u),Bm(u,t,a,r),Nc(a,u),!1}a=a.return}while(a!==null);return!1}var Um=Error(o(461)),It=!1;function le(t,e,a,r){e.child=t===null?Cm(e,null,a,r):gl(e,t.child,a,r)}function Hm(t,e,a,r,u){a=a.render;var s=e.ref;if("ref"in r){var p={};for(var b in r)b!=="ref"&&(p[b]=r[b])}else p=r;return Na(e),r=Dc(t,e,a,p,s,u),b=Bc(),t!==null&&!It?(Uc(t,e,u),On(t,e,u)):(At&&b&&pc(e),e.flags|=1,le(t,e,r,u),e.child)}function $m(t,e,a,r,u){if(t===null){var s=a.type;return typeof s=="function"&&!mc(s)&&s.defaultProps===void 0&&a.compare===null?(e.tag=15,e.type=s,Lm(t,e,s,r,u)):(t=Vi(a.type,null,r,e,e.mode,u),t.ref=e.ref,t.return=e,e.child=t)}if(s=t.child,!rs(t,u)){var p=s.memoizedProps;if(a=a.compare,a=a!==null?a:cr,a(p,r)&&t.ref===e.ref)return On(t,e,u)}return e.flags|=1,t=Tn(s,r),t.ref=e.ref,t.return=e,e.child=t}function Lm(t,e,a,r,u){if(t!==null){var s=t.memoizedProps;if(cr(s,r)&&t.ref===e.ref)if(It=!1,e.pendingProps=r=s,rs(t,u))(t.flags&131072)!==0&&(It=!0);else return e.lanes=t.lanes,On(t,e,u)}return Fc(t,e,a,r,u)}function Ym(t,e,a){var r=e.pendingProps,u=r.children,s=t!==null?t.memoizedState:null;if(r.mode==="hidden"){if((e.flags&128)!==0){if(r=s!==null?s.baseLanes|a:a,t!==null){for(u=e.child=t.child,s=0;u!==null;)s=s|u.lanes|u.childLanes,u=u.sibling;e.childLanes=s&~r}else e.childLanes=0,e.child=null;return qm(t,e,r,a)}if((a&536870912)!==0)e.memoizedState={baseLanes:0,cachePool:null},t!==null&&Ji(e,s!==null?s.cachePool:null),s!==null?Lh(e,s):Mc(),Am(e);else return e.lanes=e.childLanes=536870912,qm(t,e,s!==null?s.baseLanes|a:a,a)}else s!==null?(Ji(e,s.cachePool),Lh(e,s),In(),e.memoizedState=null):(t!==null&&Ji(e,null),Mc(),In());return le(t,e,u,a),e.child}function qm(t,e,a,r){var u=Ec();return u=u===null?null:{parent:Kt._currentValue,pool:u},e.memoizedState={baseLanes:a,cachePool:u},t!==null&&Ji(e,null),Mc(),Am(e),t!==null&&mr(t,e,r,!0),null}function ho(t,e){var a=e.ref;if(a===null)t!==null&&t.ref!==null&&(e.flags|=4194816);else{if(typeof a!="function"&&typeof a!="object")throw Error(o(284));(t===null||t.ref!==a)&&(e.flags|=4194816)}}function Fc(t,e,a,r,u){return Na(e),a=Dc(t,e,a,r,void 0,u),r=Bc(),t!==null&&!It?(Uc(t,e,u),On(t,e,u)):(At&&r&&pc(e),e.flags|=1,le(t,e,a,u),e.child)}function Xm(t,e,a,r,u,s){return Na(e),e.updateQueue=null,a=qh(e,r,a,u),Yh(t),r=Bc(),t!==null&&!It?(Uc(t,e,s),On(t,e,s)):(At&&r&&pc(e),e.flags|=1,le(t,e,a,s),e.child)}function Gm(t,e,a,r,u){if(Na(e),e.stateNode===null){var s=ul,p=a.contextType;typeof p=="object"&&p!==null&&(s=fe(p)),s=new a(r,s),e.memoizedState=s.state!==null&&s.state!==void 0?s.state:null,s.updater=Wc,e.stateNode=s,s._reactInternals=e,s=e.stateNode,s.props=r,s.state=e.memoizedState,s.refs={},Ac(e),p=a.contextType,s.context=typeof p=="object"&&p!==null?fe(p):ul,s.state=e.memoizedState,p=a.getDerivedStateFromProps,typeof p=="function"&&(Jc(e,a,p,r),s.state=e.memoizedState),typeof a.getDerivedStateFromProps=="function"||typeof s.getSnapshotBeforeUpdate=="function"||typeof s.UNSAFE_componentWillMount!="function"&&typeof s.componentWillMount!="function"||(p=s.state,typeof s.componentWillMount=="function"&&s.componentWillMount(),typeof s.UNSAFE_componentWillMount=="function"&&s.UNSAFE_componentWillMount(),p!==s.state&&Wc.enqueueReplaceState(s,s.state,null),_r(e,r,s,u),Sr(),s.state=e.memoizedState),typeof s.componentDidMount=="function"&&(e.flags|=4194308),r=!0}else if(t===null){s=e.stateNode;var b=e.memoizedProps,A=Ra(a,b);s.props=A;var H=s.context,K=a.contextType;p=ul,typeof K=="object"&&K!==null&&(p=fe(K));var F=a.getDerivedStateFromProps;K=typeof F=="function"||typeof s.getSnapshotBeforeUpdate=="function",b=e.pendingProps!==b,K||typeof s.UNSAFE_componentWillReceiveProps!="function"&&typeof s.componentWillReceiveProps!="function"||(b||H!==p)&&Nm(e,s,r,p),Kn=!1;var q=e.memoizedState;s.state=q,_r(e,r,s,u),Sr(),H=e.memoizedState,b||q!==H||Kn?(typeof F=="function"&&(Jc(e,a,F,r),H=e.memoizedState),(A=Kn||jm(e,a,A,r,q,H,p))?(K||typeof s.UNSAFE_componentWillMount!="function"&&typeof s.componentWillMount!="function"||(typeof s.componentWillMount=="function"&&s.componentWillMount(),typeof s.UNSAFE_componentWillMount=="function"&&s.UNSAFE_componentWillMount()),typeof s.componentDidMount=="function"&&(e.flags|=4194308)):(typeof s.componentDidMount=="function"&&(e.flags|=4194308),e.memoizedProps=r,e.memoizedState=H),s.props=r,s.state=H,s.context=p,r=A):(typeof s.componentDidMount=="function"&&(e.flags|=4194308),r=!1)}else{s=e.stateNode,jc(t,e),p=e.memoizedProps,K=Ra(a,p),s.props=K,F=e.pendingProps,q=s.context,H=a.contextType,A=ul,typeof H=="object"&&H!==null&&(A=fe(H)),b=a.getDerivedStateFromProps,(H=typeof b=="function"||typeof s.getSnapshotBeforeUpdate=="function")||typeof s.UNSAFE_componentWillReceiveProps!="function"&&typeof s.componentWillReceiveProps!="function"||(p!==F||q!==A)&&Nm(e,s,r,A),Kn=!1,q=e.memoizedState,s.state=q,_r(e,r,s,u),Sr();var G=e.memoizedState;p!==F||q!==G||Kn||t!==null&&t.dependencies!==null&&Qi(t.dependencies)?(typeof b=="function"&&(Jc(e,a,b,r),G=e.memoizedState),(K=Kn||jm(e,a,K,r,q,G,A)||t!==null&&t.dependencies!==null&&Qi(t.dependencies))?(H||typeof s.UNSAFE_componentWillUpdate!="function"&&typeof s.componentWillUpdate!="function"||(typeof s.componentWillUpdate=="function"&&s.componentWillUpdate(r,G,A),typeof s.UNSAFE_componentWillUpdate=="function"&&s.UNSAFE_componentWillUpdate(r,G,A)),typeof s.componentDidUpdate=="function"&&(e.flags|=4),typeof s.getSnapshotBeforeUpdate=="function"&&(e.flags|=1024)):(typeof s.componentDidUpdate!="function"||p===t.memoizedProps&&q===t.memoizedState||(e.flags|=4),typeof s.getSnapshotBeforeUpdate!="function"||p===t.memoizedProps&&q===t.memoizedState||(e.flags|=1024),e.memoizedProps=r,e.memoizedState=G),s.props=r,s.state=G,s.context=A,r=K):(typeof s.componentDidUpdate!="function"||p===t.memoizedProps&&q===t.memoizedState||(e.flags|=4),typeof s.getSnapshotBeforeUpdate!="function"||p===t.memoizedProps&&q===t.memoizedState||(e.flags|=1024),r=!1)}return s=r,ho(t,e),r=(e.flags&128)!==0,s||r?(s=e.stateNode,a=r&&typeof a.getDerivedStateFromError!="function"?null:s.render(),e.flags|=1,t!==null&&r?(e.child=gl(e,t.child,null,u),e.child=gl(e,null,a,u)):le(t,e,a,u),e.memoizedState=s.state,t=e.child):t=On(t,e,u),t}function Vm(t,e,a,r){return dr(),e.flags|=256,le(t,e,a,r),e.child}var Ic={dehydrated:null,treeContext:null,retryLane:0,hydrationErrors:null};function ts(t){return{baseLanes:t,cachePool:Mh()}}function es(t,e,a){return t=t!==null?t.childLanes&~a:0,e&&(t|=Ze),t}function km(t,e,a){var r=e.pendingProps,u=!1,s=(e.flags&128)!==0,p;if((p=s)||(p=t!==null&&t.memoizedState===null?!1:(Jt.current&2)!==0),p&&(u=!0,e.flags&=-129),p=(e.flags&32)!==0,e.flags&=-33,t===null){if(At){if(u?Fn(e):In(),At){var b=Yt,A;if(A=b){t:{for(A=b,b=rn;A.nodeType!==8;){if(!b){b=null;break t}if(A=tn(A.nextSibling),A===null){b=null;break t}}b=A}b!==null?(e.memoizedState={dehydrated:b,treeContext:wa!==null?{id:wn,overflow:En}:null,retryLane:536870912,hydrationErrors:null},A=Oe(18,null,null,0),A.stateNode=b,A.return=e,e.child=A,me=e,Yt=null,A=!0):A=!1}A||Aa(e)}if(b=e.memoizedState,b!==null&&(b=b.dehydrated,b!==null))return $s(b)?e.lanes=32:e.lanes=536870912,null;Nn(e)}return b=r.children,r=r.fallback,u?(In(),u=e.mode,b=mo({mode:"hidden",children:b},u),r=Ta(r,u,a,null),b.return=e,r.return=e,b.sibling=r,e.child=b,u=e.child,u.memoizedState=ts(a),u.childLanes=es(t,p,a),e.memoizedState=Ic,r):(Fn(e),ns(e,b))}if(A=t.memoizedState,A!==null&&(b=A.dehydrated,b!==null)){if(s)e.flags&256?(Fn(e),e.flags&=-257,e=as(t,e,a)):e.memoizedState!==null?(In(),e.child=t.child,e.flags|=128,e=null):(In(),u=r.fallback,b=e.mode,r=mo({mode:"visible",children:r.children},b),u=Ta(u,b,a,null),u.flags|=2,r.return=e,u.return=e,r.sibling=u,e.child=r,gl(e,t.child,null,a),r=e.child,r.memoizedState=ts(a),r.childLanes=es(t,p,a),e.memoizedState=Ic,e=u);else if(Fn(e),$s(b)){if(p=b.nextSibling&&b.nextSibling.dataset,p)var H=p.dgst;p=H,r=Error(o(419)),r.stack="",r.digest=p,hr({value:r,source:null,stack:null}),e=as(t,e,a)}else if(It||mr(t,e,a,!1),p=(a&t.childLanes)!==0,It||p){if(p=Bt,p!==null&&(r=a&-a,r=(r&42)!==0?1:Lu(r),r=(r&(p.suspendedLanes|a))!==0?0:r,r!==0&&r!==A.retryLane))throw A.retryLane=r,ol(t,r),Be(p,t,r),Um;b.data==="$?"||_s(),e=as(t,e,a)}else b.data==="$?"?(e.flags|=192,e.child=t.child,e=null):(t=A.treeContext,Yt=tn(b.nextSibling),me=e,At=!0,Ca=null,rn=!1,t!==null&&(Ge[Ve++]=wn,Ge[Ve++]=En,Ge[Ve++]=wa,wn=t.id,En=t.overflow,wa=e),e=ns(e,r.children),e.flags|=4096);return e}return u?(In(),u=r.fallback,b=e.mode,A=t.child,H=A.sibling,r=Tn(A,{mode:"hidden",children:r.children}),r.subtreeFlags=A.subtreeFlags&65011712,H!==null?u=Tn(H,u):(u=Ta(u,b,a,null),u.flags|=2),u.return=e,r.return=e,r.sibling=u,e.child=r,r=u,u=e.child,b=t.child.memoizedState,b===null?b=ts(a):(A=b.cachePool,A!==null?(H=Kt._currentValue,A=A.parent!==H?{parent:H,pool:H}:A):A=Mh(),b={baseLanes:b.baseLanes|a,cachePool:A}),u.memoizedState=b,u.childLanes=es(t,p,a),e.memoizedState=Ic,r):(Fn(e),a=t.child,t=a.sibling,a=Tn(a,{mode:"visible",children:r.children}),a.return=e,a.sibling=null,t!==null&&(p=e.deletions,p===null?(e.deletions=[t],e.flags|=16):p.push(t)),e.child=a,e.memoizedState=null,a)}function ns(t,e){return e=mo({mode:"visible",children:e},t.mode),e.return=t,t.child=e}function mo(t,e){return t=Oe(22,t,null,e),t.lanes=0,t.stateNode={_visibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null},t}function as(t,e,a){return gl(e,t.child,null,a),t=ns(e,e.pendingProps.children),t.flags|=2,e.memoizedState=null,t}function Zm(t,e,a){t.lanes|=e;var r=t.alternate;r!==null&&(r.lanes|=e),_c(t.return,e,a)}function ls(t,e,a,r,u){var s=t.memoizedState;s===null?t.memoizedState={isBackwards:e,rendering:null,renderingStartTime:0,last:r,tail:a,tailMode:u}:(s.isBackwards=e,s.rendering=null,s.renderingStartTime=0,s.last=r,s.tail=a,s.tailMode=u)}function Qm(t,e,a){var r=e.pendingProps,u=r.revealOrder,s=r.tail;if(le(t,e,r.children,a),r=Jt.current,(r&2)!==0)r=r&1|2,e.flags|=128;else{if(t!==null&&(t.flags&128)!==0)t:for(t=e.child;t!==null;){if(t.tag===13)t.memoizedState!==null&&Zm(t,a,e);else if(t.tag===19)Zm(t,a,e);else if(t.child!==null){t.child.return=t,t=t.child;continue}if(t===e)break t;for(;t.sibling===null;){if(t.return===null||t.return===e)break t;t=t.return}t.sibling.return=t.return,t=t.sibling}r&=1}switch(P(Jt,r),u){case"forwards":for(a=e.child,u=null;a!==null;)t=a.alternate,t!==null&&co(t)===null&&(u=a),a=a.sibling;a=u,a===null?(u=e.child,e.child=null):(u=a.sibling,a.sibling=null),ls(e,!1,u,a,s);break;case"backwards":for(a=null,u=e.child,e.child=null;u!==null;){if(t=u.alternate,t!==null&&co(t)===null){e.child=u;break}t=u.sibling,u.sibling=a,a=u,u=t}ls(e,!0,a,null,s);break;case"together":ls(e,!1,null,null,void 0);break;default:e.memoizedState=null}return e.child}function On(t,e,a){if(t!==null&&(e.dependencies=t.dependencies),la|=e.lanes,(a&e.childLanes)===0)if(t!==null){if(mr(t,e,a,!1),(a&e.childLanes)===0)return null}else return null;if(t!==null&&e.child!==t.child)throw Error(o(153));if(e.child!==null){for(t=e.child,a=Tn(t,t.pendingProps),e.child=a,a.return=e;t.sibling!==null;)t=t.sibling,a=a.sibling=Tn(t,t.pendingProps),a.return=e;a.sibling=null}return e.child}function rs(t,e){return(t.lanes&e)!==0?!0:(t=t.dependencies,!!(t!==null&&Qi(t)))}function X1(t,e,a){switch(e.tag){case 3:vt(e,e.stateNode.containerInfo),Qn(e,Kt,t.memoizedState.cache),dr();break;case 27:case 5:ae(e);break;case 4:vt(e,e.stateNode.containerInfo);break;case 10:Qn(e,e.type,e.memoizedProps.value);break;case 13:var r=e.memoizedState;if(r!==null)return r.dehydrated!==null?(Fn(e),e.flags|=128,null):(a&e.child.childLanes)!==0?km(t,e,a):(Fn(e),t=On(t,e,a),t!==null?t.sibling:null);Fn(e);break;case 19:var u=(t.flags&128)!==0;if(r=(a&e.childLanes)!==0,r||(mr(t,e,a,!1),r=(a&e.childLanes)!==0),u){if(r)return Qm(t,e,a);e.flags|=128}if(u=e.memoizedState,u!==null&&(u.rendering=null,u.tail=null,u.lastEffect=null),P(Jt,Jt.current),r)break;return null;case 22:case 23:return e.lanes=0,Ym(t,e,a);case 24:Qn(e,Kt,t.memoizedState.cache)}return On(t,e,a)}function Km(t,e,a){if(t!==null)if(t.memoizedProps!==e.pendingProps)It=!0;else{if(!rs(t,a)&&(e.flags&128)===0)return It=!1,X1(t,e,a);It=(t.flags&131072)!==0}else It=!1,At&&(e.flags&1048576)!==0&&wh(e,Zi,e.index);switch(e.lanes=0,e.tag){case 16:t:{t=e.pendingProps;var r=e.elementType,u=r._init;if(r=u(r._payload),e.type=r,typeof r=="function")mc(r)?(t=Ra(r,t),e.tag=1,e=Gm(null,e,r,t,a)):(e.tag=0,e=Fc(null,e,r,t,a));else{if(r!=null){if(u=r.$$typeof,u===L){e.tag=11,e=Hm(null,e,r,t,a);break t}else if(u===Z){e.tag=14,e=$m(null,e,r,t,a);break t}}throw e=R(r)||r,Error(o(306,e,""))}}return e;case 0:return Fc(t,e,e.type,e.pendingProps,a);case 1:return r=e.type,u=Ra(r,e.pendingProps),Gm(t,e,r,u,a);case 3:t:{if(vt(e,e.stateNode.containerInfo),t===null)throw Error(o(387));r=e.pendingProps;var s=e.memoizedState;u=s.element,jc(t,e),_r(e,r,null,a);var p=e.memoizedState;if(r=p.cache,Qn(e,Kt,r),r!==s.cache&&xc(e,[Kt],a,!0),Sr(),r=p.element,s.isDehydrated)if(s={element:r,isDehydrated:!1,cache:p.cache},e.updateQueue.baseState=s,e.memoizedState=s,e.flags&256){e=Vm(t,e,r,a);break t}else if(r!==u){u=qe(Error(o(424)),e),hr(u),e=Vm(t,e,r,a);break t}else{switch(t=e.stateNode.containerInfo,t.nodeType){case 9:t=t.body;break;default:t=t.nodeName==="HTML"?t.ownerDocument.body:t}for(Yt=tn(t.firstChild),me=e,At=!0,Ca=null,rn=!0,a=Cm(e,null,r,a),e.child=a;a;)a.flags=a.flags&-3|4096,a=a.sibling}else{if(dr(),r===u){e=On(t,e,a);break t}le(t,e,r,a)}e=e.child}return e;case 26:return ho(t,e),t===null?(a=Fy(e.type,null,e.pendingProps,null))?e.memoizedState=a:At||(a=e.type,t=e.pendingProps,r=jo(rt.current).createElement(a),r[se]=e,r[ge]=t,ie(r,a,t),Ft(r),e.stateNode=r):e.memoizedState=Fy(e.type,t.memoizedProps,e.pendingProps,t.memoizedState),null;case 27:return ae(e),t===null&&At&&(r=e.stateNode=Jy(e.type,e.pendingProps,rt.current),me=e,rn=!0,u=Yt,ua(e.type)?(Ls=u,Yt=tn(r.firstChild)):Yt=u),le(t,e,e.pendingProps.children,a),ho(t,e),t===null&&(e.flags|=4194304),e.child;case 5:return t===null&&At&&((u=r=Yt)&&(r=vb(r,e.type,e.pendingProps,rn),r!==null?(e.stateNode=r,me=e,Yt=tn(r.firstChild),rn=!1,u=!0):u=!1),u||Aa(e)),ae(e),u=e.type,s=e.pendingProps,p=t!==null?t.memoizedProps:null,r=s.children,Bs(u,s)?r=null:p!==null&&Bs(u,p)&&(e.flags|=32),e.memoizedState!==null&&(u=Dc(t,e,D1,null,null,a),Gr._currentValue=u),ho(t,e),le(t,e,r,a),e.child;case 6:return t===null&&At&&((t=a=Yt)&&(a=pb(a,e.pendingProps,rn),a!==null?(e.stateNode=a,me=e,Yt=null,t=!0):t=!1),t||Aa(e)),null;case 13:return km(t,e,a);case 4:return vt(e,e.stateNode.containerInfo),r=e.pendingProps,t===null?e.child=gl(e,null,r,a):le(t,e,r,a),e.child;case 11:return Hm(t,e,e.type,e.pendingProps,a);case 7:return le(t,e,e.pendingProps,a),e.child;case 8:return le(t,e,e.pendingProps.children,a),e.child;case 12:return le(t,e,e.pendingProps.children,a),e.child;case 10:return r=e.pendingProps,Qn(e,e.type,r.value),le(t,e,r.children,a),e.child;case 9:return u=e.type._context,r=e.pendingProps.children,Na(e),u=fe(u),r=r(u),e.flags|=1,le(t,e,r,a),e.child;case 14:return $m(t,e,e.type,e.pendingProps,a);case 15:return Lm(t,e,e.type,e.pendingProps,a);case 19:return Qm(t,e,a);case 31:return r=e.pendingProps,a=e.mode,r={mode:r.mode,children:r.children},t===null?(a=mo(r,a),a.ref=e.ref,e.child=a,a.return=e,e=a):(a=Tn(t.child,r),a.ref=e.ref,e.child=a,a.return=e,e=a),e;case 22:return Ym(t,e,a);case 24:return Na(e),r=fe(Kt),t===null?(u=Ec(),u===null&&(u=Bt,s=Tc(),u.pooledCache=s,s.refCount++,s!==null&&(u.pooledCacheLanes|=a),u=s),e.memoizedState={parent:r,cache:u},Ac(e),Qn(e,Kt,u)):((t.lanes&a)!==0&&(jc(t,e),_r(e,null,null,a),Sr()),u=t.memoizedState,s=e.memoizedState,u.parent!==r?(u={parent:r,cache:r},e.memoizedState=u,e.lanes===0&&(e.memoizedState=e.updateQueue.baseState=u),Qn(e,Kt,r)):(r=s.cache,Qn(e,Kt,r),r!==u.cache&&xc(e,[Kt],a,!0))),le(t,e,e.pendingProps.children,a),e.child;case 29:throw e.pendingProps}throw Error(o(156,e.tag))}function Mn(t){t.flags|=4}function Jm(t,e){if(e.type!=="stylesheet"||(e.state.loading&4)!==0)t.flags&=-16777217;else if(t.flags|=16777216,!a0(e)){if(e=ke.current,e!==null&&((Et&4194048)===Et?on!==null:(Et&62914560)!==Et&&(Et&536870912)===0||e!==on))throw gr=Cc,Rh;t.flags|=8192}}function yo(t,e){e!==null&&(t.flags|=4),t.flags&16384&&(e=t.tag!==22?Ad():536870912,t.lanes|=e,xl|=e)}function jr(t,e){if(!At)switch(t.tailMode){case"hidden":e=t.tail;for(var a=null;e!==null;)e.alternate!==null&&(a=e),e=e.sibling;a===null?t.tail=null:a.sibling=null;break;case"collapsed":a=t.tail;for(var r=null;a!==null;)a.alternate!==null&&(r=a),a=a.sibling;r===null?e||t.tail===null?t.tail=null:t.tail.sibling=null:r.sibling=null}}function Lt(t){var e=t.alternate!==null&&t.alternate.child===t.child,a=0,r=0;if(e)for(var u=t.child;u!==null;)a|=u.lanes|u.childLanes,r|=u.subtreeFlags&65011712,r|=u.flags&65011712,u.return=t,u=u.sibling;else for(u=t.child;u!==null;)a|=u.lanes|u.childLanes,r|=u.subtreeFlags,r|=u.flags,u.return=t,u=u.sibling;return t.subtreeFlags|=r,t.childLanes=a,e}function G1(t,e,a){var r=e.pendingProps;switch(gc(e),e.tag){case 31:case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return Lt(e),null;case 1:return Lt(e),null;case 3:return a=e.stateNode,r=null,t!==null&&(r=t.memoizedState.cache),e.memoizedState.cache!==r&&(e.flags|=2048),An(Kt),Ut(),a.pendingContext&&(a.context=a.pendingContext,a.pendingContext=null),(t===null||t.child===null)&&(fr(e)?Mn(e):t===null||t.memoizedState.isDehydrated&&(e.flags&256)===0||(e.flags|=1024,Ah())),Lt(e),null;case 26:return a=e.memoizedState,t===null?(Mn(e),a!==null?(Lt(e),Jm(e,a)):(Lt(e),e.flags&=-16777217)):a?a!==t.memoizedState?(Mn(e),Lt(e),Jm(e,a)):(Lt(e),e.flags&=-16777217):(t.memoizedProps!==r&&Mn(e),Lt(e),e.flags&=-16777217),null;case 27:pe(e),a=rt.current;var u=e.type;if(t!==null&&e.stateNode!=null)t.memoizedProps!==r&&Mn(e);else{if(!r){if(e.stateNode===null)throw Error(o(166));return Lt(e),null}t=nt.current,fr(e)?Eh(e):(t=Jy(u,r,a),e.stateNode=t,Mn(e))}return Lt(e),null;case 5:if(pe(e),a=e.type,t!==null&&e.stateNode!=null)t.memoizedProps!==r&&Mn(e);else{if(!r){if(e.stateNode===null)throw Error(o(166));return Lt(e),null}if(t=nt.current,fr(e))Eh(e);else{switch(u=jo(rt.current),t){case 1:t=u.createElementNS("http://www.w3.org/2000/svg",a);break;case 2:t=u.createElementNS("http://www.w3.org/1998/Math/MathML",a);break;default:switch(a){case"svg":t=u.createElementNS("http://www.w3.org/2000/svg",a);break;case"math":t=u.createElementNS("http://www.w3.org/1998/Math/MathML",a);break;case"script":t=u.createElement("div"),t.innerHTML="<script><\/script>",t=t.removeChild(t.firstChild);break;case"select":t=typeof r.is=="string"?u.createElement("select",{is:r.is}):u.createElement("select"),r.multiple?t.multiple=!0:r.size&&(t.size=r.size);break;default:t=typeof r.is=="string"?u.createElement(a,{is:r.is}):u.createElement(a)}}t[se]=e,t[ge]=r;t:for(u=e.child;u!==null;){if(u.tag===5||u.tag===6)t.appendChild(u.stateNode);else if(u.tag!==4&&u.tag!==27&&u.child!==null){u.child.return=u,u=u.child;continue}if(u===e)break t;for(;u.sibling===null;){if(u.return===null||u.return===e)break t;u=u.return}u.sibling.return=u.return,u=u.sibling}e.stateNode=t;t:switch(ie(t,a,r),a){case"button":case"input":case"select":case"textarea":t=!!r.autoFocus;break t;case"img":t=!0;break t;default:t=!1}t&&Mn(e)}}return Lt(e),e.flags&=-16777217,null;case 6:if(t&&e.stateNode!=null)t.memoizedProps!==r&&Mn(e);else{if(typeof r!="string"&&e.stateNode===null)throw Error(o(166));if(t=rt.current,fr(e)){if(t=e.stateNode,a=e.memoizedProps,r=null,u=me,u!==null)switch(u.tag){case 27:case 5:r=u.memoizedProps}t[se]=e,t=!!(t.nodeValue===a||r!==null&&r.suppressHydrationWarning===!0||Xy(t.nodeValue,a)),t||Aa(e)}else t=jo(t).createTextNode(r),t[se]=e,e.stateNode=t}return Lt(e),null;case 13:if(r=e.memoizedState,t===null||t.memoizedState!==null&&t.memoizedState.dehydrated!==null){if(u=fr(e),r!==null&&r.dehydrated!==null){if(t===null){if(!u)throw Error(o(318));if(u=e.memoizedState,u=u!==null?u.dehydrated:null,!u)throw Error(o(317));u[se]=e}else dr(),(e.flags&128)===0&&(e.memoizedState=null),e.flags|=4;Lt(e),u=!1}else u=Ah(),t!==null&&t.memoizedState!==null&&(t.memoizedState.hydrationErrors=u),u=!0;if(!u)return e.flags&256?(Nn(e),e):(Nn(e),null)}if(Nn(e),(e.flags&128)!==0)return e.lanes=a,e;if(a=r!==null,t=t!==null&&t.memoizedState!==null,a){r=e.child,u=null,r.alternate!==null&&r.alternate.memoizedState!==null&&r.alternate.memoizedState.cachePool!==null&&(u=r.alternate.memoizedState.cachePool.pool);var s=null;r.memoizedState!==null&&r.memoizedState.cachePool!==null&&(s=r.memoizedState.cachePool.pool),s!==u&&(r.flags|=2048)}return a!==t&&a&&(e.child.flags|=8192),yo(e,e.updateQueue),Lt(e),null;case 4:return Ut(),t===null&&Os(e.stateNode.containerInfo),Lt(e),null;case 10:return An(e.type),Lt(e),null;case 19:if(I(Jt),u=e.memoizedState,u===null)return Lt(e),null;if(r=(e.flags&128)!==0,s=u.rendering,s===null)if(r)jr(u,!1);else{if(qt!==0||t!==null&&(t.flags&128)!==0)for(t=e.child;t!==null;){if(s=co(t),s!==null){for(e.flags|=128,jr(u,!1),t=s.updateQueue,e.updateQueue=t,yo(e,t),e.subtreeFlags=0,t=a,a=e.child;a!==null;)Th(a,t),a=a.sibling;return P(Jt,Jt.current&1|2),e.child}t=t.sibling}u.tail!==null&&Ht()>go&&(e.flags|=128,r=!0,jr(u,!1),e.lanes=4194304)}else{if(!r)if(t=co(s),t!==null){if(e.flags|=128,r=!0,t=t.updateQueue,e.updateQueue=t,yo(e,t),jr(u,!0),u.tail===null&&u.tailMode==="hidden"&&!s.alternate&&!At)return Lt(e),null}else 2*Ht()-u.renderingStartTime>go&&a!==536870912&&(e.flags|=128,r=!0,jr(u,!1),e.lanes=4194304);u.isBackwards?(s.sibling=e.child,e.child=s):(t=u.last,t!==null?t.sibling=s:e.child=s,u.last=s)}return u.tail!==null?(e=u.tail,u.rendering=e,u.tail=e.sibling,u.renderingStartTime=Ht(),e.sibling=null,t=Jt.current,P(Jt,r?t&1|2:t&1),e):(Lt(e),null);case 22:case 23:return Nn(e),Rc(),r=e.memoizedState!==null,t!==null?t.memoizedState!==null!==r&&(e.flags|=8192):r&&(e.flags|=8192),r?(a&536870912)!==0&&(e.flags&128)===0&&(Lt(e),e.subtreeFlags&6&&(e.flags|=8192)):Lt(e),a=e.updateQueue,a!==null&&yo(e,a.retryQueue),a=null,t!==null&&t.memoizedState!==null&&t.memoizedState.cachePool!==null&&(a=t.memoizedState.cachePool.pool),r=null,e.memoizedState!==null&&e.memoizedState.cachePool!==null&&(r=e.memoizedState.cachePool.pool),r!==a&&(e.flags|=2048),t!==null&&I(Oa),null;case 24:return a=null,t!==null&&(a=t.memoizedState.cache),e.memoizedState.cache!==a&&(e.flags|=2048),An(Kt),Lt(e),null;case 25:return null;case 30:return null}throw Error(o(156,e.tag))}function V1(t,e){switch(gc(e),e.tag){case 1:return t=e.flags,t&65536?(e.flags=t&-65537|128,e):null;case 3:return An(Kt),Ut(),t=e.flags,(t&65536)!==0&&(t&128)===0?(e.flags=t&-65537|128,e):null;case 26:case 27:case 5:return pe(e),null;case 13:if(Nn(e),t=e.memoizedState,t!==null&&t.dehydrated!==null){if(e.alternate===null)throw Error(o(340));dr()}return t=e.flags,t&65536?(e.flags=t&-65537|128,e):null;case 19:return I(Jt),null;case 4:return Ut(),null;case 10:return An(e.type),null;case 22:case 23:return Nn(e),Rc(),t!==null&&I(Oa),t=e.flags,t&65536?(e.flags=t&-65537|128,e):null;case 24:return An(Kt),null;case 25:return null;default:return null}}function Wm(t,e){switch(gc(e),e.tag){case 3:An(Kt),Ut();break;case 26:case 27:case 5:pe(e);break;case 4:Ut();break;case 13:Nn(e);break;case 19:I(Jt);break;case 10:An(e.type);break;case 22:case 23:Nn(e),Rc(),t!==null&&I(Oa);break;case 24:An(Kt)}}function Nr(t,e){try{var a=e.updateQueue,r=a!==null?a.lastEffect:null;if(r!==null){var u=r.next;a=u;do{if((a.tag&t)===t){r=void 0;var s=a.create,p=a.inst;r=s(),p.destroy=r}a=a.next}while(a!==u)}}catch(b){Dt(e,e.return,b)}}function ta(t,e,a){try{var r=e.updateQueue,u=r!==null?r.lastEffect:null;if(u!==null){var s=u.next;r=s;do{if((r.tag&t)===t){var p=r.inst,b=p.destroy;if(b!==void 0){p.destroy=void 0,u=e;var A=a,H=b;try{H()}catch(K){Dt(u,A,K)}}}r=r.next}while(r!==s)}}catch(K){Dt(e,e.return,K)}}function Pm(t){var e=t.updateQueue;if(e!==null){var a=t.stateNode;try{$h(e,a)}catch(r){Dt(t,t.return,r)}}}function Fm(t,e,a){a.props=Ra(t.type,t.memoizedProps),a.state=t.memoizedState;try{a.componentWillUnmount()}catch(r){Dt(t,e,r)}}function Or(t,e){try{var a=t.ref;if(a!==null){switch(t.tag){case 26:case 27:case 5:var r=t.stateNode;break;case 30:r=t.stateNode;break;default:r=t.stateNode}typeof a=="function"?t.refCleanup=a(r):a.current=r}}catch(u){Dt(t,e,u)}}function un(t,e){var a=t.ref,r=t.refCleanup;if(a!==null)if(typeof r=="function")try{r()}catch(u){Dt(t,e,u)}finally{t.refCleanup=null,t=t.alternate,t!=null&&(t.refCleanup=null)}else if(typeof a=="function")try{a(null)}catch(u){Dt(t,e,u)}else a.current=null}function Im(t){var e=t.type,a=t.memoizedProps,r=t.stateNode;try{t:switch(e){case"button":case"input":case"select":case"textarea":a.autoFocus&&r.focus();break t;case"img":a.src?r.src=a.src:a.srcSet&&(r.srcset=a.srcSet)}}catch(u){Dt(t,t.return,u)}}function is(t,e,a){try{var r=t.stateNode;fb(r,t.type,a,e),r[ge]=e}catch(u){Dt(t,t.return,u)}}function ty(t){return t.tag===5||t.tag===3||t.tag===26||t.tag===27&&ua(t.type)||t.tag===4}function os(t){t:for(;;){for(;t.sibling===null;){if(t.return===null||ty(t.return))return null;t=t.return}for(t.sibling.return=t.return,t=t.sibling;t.tag!==5&&t.tag!==6&&t.tag!==18;){if(t.tag===27&&ua(t.type)||t.flags&2||t.child===null||t.tag===4)continue t;t.child.return=t,t=t.child}if(!(t.flags&2))return t.stateNode}}function us(t,e,a){var r=t.tag;if(r===5||r===6)t=t.stateNode,e?(a.nodeType===9?a.body:a.nodeName==="HTML"?a.ownerDocument.body:a).insertBefore(t,e):(e=a.nodeType===9?a.body:a.nodeName==="HTML"?a.ownerDocument.body:a,e.appendChild(t),a=a._reactRootContainer,a!=null||e.onclick!==null||(e.onclick=Ao));else if(r!==4&&(r===27&&ua(t.type)&&(a=t.stateNode,e=null),t=t.child,t!==null))for(us(t,e,a),t=t.sibling;t!==null;)us(t,e,a),t=t.sibling}function vo(t,e,a){var r=t.tag;if(r===5||r===6)t=t.stateNode,e?a.insertBefore(t,e):a.appendChild(t);else if(r!==4&&(r===27&&ua(t.type)&&(a=t.stateNode),t=t.child,t!==null))for(vo(t,e,a),t=t.sibling;t!==null;)vo(t,e,a),t=t.sibling}function ey(t){var e=t.stateNode,a=t.memoizedProps;try{for(var r=t.type,u=e.attributes;u.length;)e.removeAttributeNode(u[0]);ie(e,r,a),e[se]=t,e[ge]=a}catch(s){Dt(t,t.return,s)}}var Rn=!1,Gt=!1,cs=!1,ny=typeof WeakSet=="function"?WeakSet:Set,te=null;function k1(t,e){if(t=t.containerInfo,zs=Do,t=hh(t),oc(t)){if("selectionStart"in t)var a={start:t.selectionStart,end:t.selectionEnd};else t:{a=(a=t.ownerDocument)&&a.defaultView||window;var r=a.getSelection&&a.getSelection();if(r&&r.rangeCount!==0){a=r.anchorNode;var u=r.anchorOffset,s=r.focusNode;r=r.focusOffset;try{a.nodeType,s.nodeType}catch{a=null;break t}var p=0,b=-1,A=-1,H=0,K=0,F=t,q=null;e:for(;;){for(var G;F!==a||u!==0&&F.nodeType!==3||(b=p+u),F!==s||r!==0&&F.nodeType!==3||(A=p+r),F.nodeType===3&&(p+=F.nodeValue.length),(G=F.firstChild)!==null;)q=F,F=G;for(;;){if(F===t)break e;if(q===a&&++H===u&&(b=p),q===s&&++K===r&&(A=p),(G=F.nextSibling)!==null)break;F=q,q=F.parentNode}F=G}a=b===-1||A===-1?null:{start:b,end:A}}else a=null}a=a||{start:0,end:0}}else a=null;for(Ds={focusedElem:t,selectionRange:a},Do=!1,te=e;te!==null;)if(e=te,t=e.child,(e.subtreeFlags&1024)!==0&&t!==null)t.return=e,te=t;else for(;te!==null;){switch(e=te,s=e.alternate,t=e.flags,e.tag){case 0:break;case 11:case 15:break;case 1:if((t&1024)!==0&&s!==null){t=void 0,a=e,u=s.memoizedProps,s=s.memoizedState,r=a.stateNode;try{var yt=Ra(a.type,u,a.elementType===a.type);t=r.getSnapshotBeforeUpdate(yt,s),r.__reactInternalSnapshotBeforeUpdate=t}catch(dt){Dt(a,a.return,dt)}}break;case 3:if((t&1024)!==0){if(t=e.stateNode.containerInfo,a=t.nodeType,a===9)Hs(t);else if(a===1)switch(t.nodeName){case"HEAD":case"HTML":case"BODY":Hs(t);break;default:t.textContent=""}}break;case 5:case 26:case 27:case 6:case 4:case 17:break;default:if((t&1024)!==0)throw Error(o(163))}if(t=e.sibling,t!==null){t.return=e.return,te=t;break}te=e.return}}function ay(t,e,a){var r=a.flags;switch(a.tag){case 0:case 11:case 15:ea(t,a),r&4&&Nr(5,a);break;case 1:if(ea(t,a),r&4)if(t=a.stateNode,e===null)try{t.componentDidMount()}catch(p){Dt(a,a.return,p)}else{var u=Ra(a.type,e.memoizedProps);e=e.memoizedState;try{t.componentDidUpdate(u,e,t.__reactInternalSnapshotBeforeUpdate)}catch(p){Dt(a,a.return,p)}}r&64&&Pm(a),r&512&&Or(a,a.return);break;case 3:if(ea(t,a),r&64&&(t=a.updateQueue,t!==null)){if(e=null,a.child!==null)switch(a.child.tag){case 27:case 5:e=a.child.stateNode;break;case 1:e=a.child.stateNode}try{$h(t,e)}catch(p){Dt(a,a.return,p)}}break;case 27:e===null&&r&4&&ey(a);case 26:case 5:ea(t,a),e===null&&r&4&&Im(a),r&512&&Or(a,a.return);break;case 12:ea(t,a);break;case 13:ea(t,a),r&4&&iy(t,a),r&64&&(t=a.memoizedState,t!==null&&(t=t.dehydrated,t!==null&&(a=tb.bind(null,a),gb(t,a))));break;case 22:if(r=a.memoizedState!==null||Rn,!r){e=e!==null&&e.memoizedState!==null||Gt,u=Rn;var s=Gt;Rn=r,(Gt=e)&&!s?na(t,a,(a.subtreeFlags&8772)!==0):ea(t,a),Rn=u,Gt=s}break;case 30:break;default:ea(t,a)}}function ly(t){var e=t.alternate;e!==null&&(t.alternate=null,ly(e)),t.child=null,t.deletions=null,t.sibling=null,t.tag===5&&(e=t.stateNode,e!==null&&Xu(e)),t.stateNode=null,t.return=null,t.dependencies=null,t.memoizedProps=null,t.memoizedState=null,t.pendingProps=null,t.stateNode=null,t.updateQueue=null}var $t=null,_e=!1;function zn(t,e,a){for(a=a.child;a!==null;)ry(t,e,a),a=a.sibling}function ry(t,e,a){if(Ae&&typeof Ae.onCommitFiberUnmount=="function")try{Ae.onCommitFiberUnmount(Pl,a)}catch{}switch(a.tag){case 26:Gt||un(a,e),zn(t,e,a),a.memoizedState?a.memoizedState.count--:a.stateNode&&(a=a.stateNode,a.parentNode.removeChild(a));break;case 27:Gt||un(a,e);var r=$t,u=_e;ua(a.type)&&($t=a.stateNode,_e=!1),zn(t,e,a),Lr(a.stateNode),$t=r,_e=u;break;case 5:Gt||un(a,e);case 6:if(r=$t,u=_e,$t=null,zn(t,e,a),$t=r,_e=u,$t!==null)if(_e)try{($t.nodeType===9?$t.body:$t.nodeName==="HTML"?$t.ownerDocument.body:$t).removeChild(a.stateNode)}catch(s){Dt(a,e,s)}else try{$t.removeChild(a.stateNode)}catch(s){Dt(a,e,s)}break;case 18:$t!==null&&(_e?(t=$t,Qy(t.nodeType===9?t.body:t.nodeName==="HTML"?t.ownerDocument.body:t,a.stateNode),Qr(t)):Qy($t,a.stateNode));break;case 4:r=$t,u=_e,$t=a.stateNode.containerInfo,_e=!0,zn(t,e,a),$t=r,_e=u;break;case 0:case 11:case 14:case 15:Gt||ta(2,a,e),Gt||ta(4,a,e),zn(t,e,a);break;case 1:Gt||(un(a,e),r=a.stateNode,typeof r.componentWillUnmount=="function"&&Fm(a,e,r)),zn(t,e,a);break;case 21:zn(t,e,a);break;case 22:Gt=(r=Gt)||a.memoizedState!==null,zn(t,e,a),Gt=r;break;default:zn(t,e,a)}}function iy(t,e){if(e.memoizedState===null&&(t=e.alternate,t!==null&&(t=t.memoizedState,t!==null&&(t=t.dehydrated,t!==null))))try{Qr(t)}catch(a){Dt(e,e.return,a)}}function Z1(t){switch(t.tag){case 13:case 19:var e=t.stateNode;return e===null&&(e=t.stateNode=new ny),e;case 22:return t=t.stateNode,e=t._retryCache,e===null&&(e=t._retryCache=new ny),e;default:throw Error(o(435,t.tag))}}function ss(t,e){var a=Z1(t);e.forEach(function(r){var u=eb.bind(null,t,r);a.has(r)||(a.add(r),r.then(u,u))})}function Me(t,e){var a=e.deletions;if(a!==null)for(var r=0;r<a.length;r++){var u=a[r],s=t,p=e,b=p;t:for(;b!==null;){switch(b.tag){case 27:if(ua(b.type)){$t=b.stateNode,_e=!1;break t}break;case 5:$t=b.stateNode,_e=!1;break t;case 3:case 4:$t=b.stateNode.containerInfo,_e=!0;break t}b=b.return}if($t===null)throw Error(o(160));ry(s,p,u),$t=null,_e=!1,s=u.alternate,s!==null&&(s.return=null),u.return=null}if(e.subtreeFlags&13878)for(e=e.child;e!==null;)oy(e,t),e=e.sibling}var Ie=null;function oy(t,e){var a=t.alternate,r=t.flags;switch(t.tag){case 0:case 11:case 14:case 15:Me(e,t),Re(t),r&4&&(ta(3,t,t.return),Nr(3,t),ta(5,t,t.return));break;case 1:Me(e,t),Re(t),r&512&&(Gt||a===null||un(a,a.return)),r&64&&Rn&&(t=t.updateQueue,t!==null&&(r=t.callbacks,r!==null&&(a=t.shared.hiddenCallbacks,t.shared.hiddenCallbacks=a===null?r:a.concat(r))));break;case 26:var u=Ie;if(Me(e,t),Re(t),r&512&&(Gt||a===null||un(a,a.return)),r&4){var s=a!==null?a.memoizedState:null;if(r=t.memoizedState,a===null)if(r===null)if(t.stateNode===null){t:{r=t.type,a=t.memoizedProps,u=u.ownerDocument||u;e:switch(r){case"title":s=u.getElementsByTagName("title")[0],(!s||s[tr]||s[se]||s.namespaceURI==="http://www.w3.org/2000/svg"||s.hasAttribute("itemprop"))&&(s=u.createElement(r),u.head.insertBefore(s,u.querySelector("head > title"))),ie(s,r,a),s[se]=t,Ft(s),r=s;break t;case"link":var p=e0("link","href",u).get(r+(a.href||""));if(p){for(var b=0;b<p.length;b++)if(s=p[b],s.getAttribute("href")===(a.href==null||a.href===""?null:a.href)&&s.getAttribute("rel")===(a.rel==null?null:a.rel)&&s.getAttribute("title")===(a.title==null?null:a.title)&&s.getAttribute("crossorigin")===(a.crossOrigin==null?null:a.crossOrigin)){p.splice(b,1);break e}}s=u.createElement(r),ie(s,r,a),u.head.appendChild(s);break;case"meta":if(p=e0("meta","content",u).get(r+(a.content||""))){for(b=0;b<p.length;b++)if(s=p[b],s.getAttribute("content")===(a.content==null?null:""+a.content)&&s.getAttribute("name")===(a.name==null?null:a.name)&&s.getAttribute("property")===(a.property==null?null:a.property)&&s.getAttribute("http-equiv")===(a.httpEquiv==null?null:a.httpEquiv)&&s.getAttribute("charset")===(a.charSet==null?null:a.charSet)){p.splice(b,1);break e}}s=u.createElement(r),ie(s,r,a),u.head.appendChild(s);break;default:throw Error(o(468,r))}s[se]=t,Ft(s),r=s}t.stateNode=r}else n0(u,t.type,t.stateNode);else t.stateNode=t0(u,r,t.memoizedProps);else s!==r?(s===null?a.stateNode!==null&&(a=a.stateNode,a.parentNode.removeChild(a)):s.count--,r===null?n0(u,t.type,t.stateNode):t0(u,r,t.memoizedProps)):r===null&&t.stateNode!==null&&is(t,t.memoizedProps,a.memoizedProps)}break;case 27:Me(e,t),Re(t),r&512&&(Gt||a===null||un(a,a.return)),a!==null&&r&4&&is(t,t.memoizedProps,a.memoizedProps);break;case 5:if(Me(e,t),Re(t),r&512&&(Gt||a===null||un(a,a.return)),t.flags&32){u=t.stateNode;try{tl(u,"")}catch(G){Dt(t,t.return,G)}}r&4&&t.stateNode!=null&&(u=t.memoizedProps,is(t,u,a!==null?a.memoizedProps:u)),r&1024&&(cs=!0);break;case 6:if(Me(e,t),Re(t),r&4){if(t.stateNode===null)throw Error(o(162));r=t.memoizedProps,a=t.stateNode;try{a.nodeValue=r}catch(G){Dt(t,t.return,G)}}break;case 3:if(Mo=null,u=Ie,Ie=No(e.containerInfo),Me(e,t),Ie=u,Re(t),r&4&&a!==null&&a.memoizedState.isDehydrated)try{Qr(e.containerInfo)}catch(G){Dt(t,t.return,G)}cs&&(cs=!1,uy(t));break;case 4:r=Ie,Ie=No(t.stateNode.containerInfo),Me(e,t),Re(t),Ie=r;break;case 12:Me(e,t),Re(t);break;case 13:Me(e,t),Re(t),t.child.flags&8192&&t.memoizedState!==null!=(a!==null&&a.memoizedState!==null)&&(vs=Ht()),r&4&&(r=t.updateQueue,r!==null&&(t.updateQueue=null,ss(t,r)));break;case 22:u=t.memoizedState!==null;var A=a!==null&&a.memoizedState!==null,H=Rn,K=Gt;if(Rn=H||u,Gt=K||A,Me(e,t),Gt=K,Rn=H,Re(t),r&8192)t:for(e=t.stateNode,e._visibility=u?e._visibility&-2:e._visibility|1,u&&(a===null||A||Rn||Gt||za(t)),a=null,e=t;;){if(e.tag===5||e.tag===26){if(a===null){A=a=e;try{if(s=A.stateNode,u)p=s.style,typeof p.setProperty=="function"?p.setProperty("display","none","important"):p.display="none";else{b=A.stateNode;var F=A.memoizedProps.style,q=F!=null&&F.hasOwnProperty("display")?F.display:null;b.style.display=q==null||typeof q=="boolean"?"":(""+q).trim()}}catch(G){Dt(A,A.return,G)}}}else if(e.tag===6){if(a===null){A=e;try{A.stateNode.nodeValue=u?"":A.memoizedProps}catch(G){Dt(A,A.return,G)}}}else if((e.tag!==22&&e.tag!==23||e.memoizedState===null||e===t)&&e.child!==null){e.child.return=e,e=e.child;continue}if(e===t)break t;for(;e.sibling===null;){if(e.return===null||e.return===t)break t;a===e&&(a=null),e=e.return}a===e&&(a=null),e.sibling.return=e.return,e=e.sibling}r&4&&(r=t.updateQueue,r!==null&&(a=r.retryQueue,a!==null&&(r.retryQueue=null,ss(t,a))));break;case 19:Me(e,t),Re(t),r&4&&(r=t.updateQueue,r!==null&&(t.updateQueue=null,ss(t,r)));break;case 30:break;case 21:break;default:Me(e,t),Re(t)}}function Re(t){var e=t.flags;if(e&2){try{for(var a,r=t.return;r!==null;){if(ty(r)){a=r;break}r=r.return}if(a==null)throw Error(o(160));switch(a.tag){case 27:var u=a.stateNode,s=os(t);vo(t,s,u);break;case 5:var p=a.stateNode;a.flags&32&&(tl(p,""),a.flags&=-33);var b=os(t);vo(t,b,p);break;case 3:case 4:var A=a.stateNode.containerInfo,H=os(t);us(t,H,A);break;default:throw Error(o(161))}}catch(K){Dt(t,t.return,K)}t.flags&=-3}e&4096&&(t.flags&=-4097)}function uy(t){if(t.subtreeFlags&1024)for(t=t.child;t!==null;){var e=t;uy(e),e.tag===5&&e.flags&1024&&e.stateNode.reset(),t=t.sibling}}function ea(t,e){if(e.subtreeFlags&8772)for(e=e.child;e!==null;)ay(t,e.alternate,e),e=e.sibling}function za(t){for(t=t.child;t!==null;){var e=t;switch(e.tag){case 0:case 11:case 14:case 15:ta(4,e,e.return),za(e);break;case 1:un(e,e.return);var a=e.stateNode;typeof a.componentWillUnmount=="function"&&Fm(e,e.return,a),za(e);break;case 27:Lr(e.stateNode);case 26:case 5:un(e,e.return),za(e);break;case 22:e.memoizedState===null&&za(e);break;case 30:za(e);break;default:za(e)}t=t.sibling}}function na(t,e,a){for(a=a&&(e.subtreeFlags&8772)!==0,e=e.child;e!==null;){var r=e.alternate,u=t,s=e,p=s.flags;switch(s.tag){case 0:case 11:case 15:na(u,s,a),Nr(4,s);break;case 1:if(na(u,s,a),r=s,u=r.stateNode,typeof u.componentDidMount=="function")try{u.componentDidMount()}catch(H){Dt(r,r.return,H)}if(r=s,u=r.updateQueue,u!==null){var b=r.stateNode;try{var A=u.shared.hiddenCallbacks;if(A!==null)for(u.shared.hiddenCallbacks=null,u=0;u<A.length;u++)Hh(A[u],b)}catch(H){Dt(r,r.return,H)}}a&&p&64&&Pm(s),Or(s,s.return);break;case 27:ey(s);case 26:case 5:na(u,s,a),a&&r===null&&p&4&&Im(s),Or(s,s.return);break;case 12:na(u,s,a);break;case 13:na(u,s,a),a&&p&4&&iy(u,s);break;case 22:s.memoizedState===null&&na(u,s,a),Or(s,s.return);break;case 30:break;default:na(u,s,a)}e=e.sibling}}function fs(t,e){var a=null;t!==null&&t.memoizedState!==null&&t.memoizedState.cachePool!==null&&(a=t.memoizedState.cachePool.pool),t=null,e.memoizedState!==null&&e.memoizedState.cachePool!==null&&(t=e.memoizedState.cachePool.pool),t!==a&&(t!=null&&t.refCount++,a!=null&&yr(a))}function ds(t,e){t=null,e.alternate!==null&&(t=e.alternate.memoizedState.cache),e=e.memoizedState.cache,e!==t&&(e.refCount++,t!=null&&yr(t))}function cn(t,e,a,r){if(e.subtreeFlags&10256)for(e=e.child;e!==null;)cy(t,e,a,r),e=e.sibling}function cy(t,e,a,r){var u=e.flags;switch(e.tag){case 0:case 11:case 15:cn(t,e,a,r),u&2048&&Nr(9,e);break;case 1:cn(t,e,a,r);break;case 3:cn(t,e,a,r),u&2048&&(t=null,e.alternate!==null&&(t=e.alternate.memoizedState.cache),e=e.memoizedState.cache,e!==t&&(e.refCount++,t!=null&&yr(t)));break;case 12:if(u&2048){cn(t,e,a,r),t=e.stateNode;try{var s=e.memoizedProps,p=s.id,b=s.onPostCommit;typeof b=="function"&&b(p,e.alternate===null?"mount":"update",t.passiveEffectDuration,-0)}catch(A){Dt(e,e.return,A)}}else cn(t,e,a,r);break;case 13:cn(t,e,a,r);break;case 23:break;case 22:s=e.stateNode,p=e.alternate,e.memoizedState!==null?s._visibility&2?cn(t,e,a,r):Mr(t,e):s._visibility&2?cn(t,e,a,r):(s._visibility|=2,bl(t,e,a,r,(e.subtreeFlags&10256)!==0)),u&2048&&fs(p,e);break;case 24:cn(t,e,a,r),u&2048&&ds(e.alternate,e);break;default:cn(t,e,a,r)}}function bl(t,e,a,r,u){for(u=u&&(e.subtreeFlags&10256)!==0,e=e.child;e!==null;){var s=t,p=e,b=a,A=r,H=p.flags;switch(p.tag){case 0:case 11:case 15:bl(s,p,b,A,u),Nr(8,p);break;case 23:break;case 22:var K=p.stateNode;p.memoizedState!==null?K._visibility&2?bl(s,p,b,A,u):Mr(s,p):(K._visibility|=2,bl(s,p,b,A,u)),u&&H&2048&&fs(p.alternate,p);break;case 24:bl(s,p,b,A,u),u&&H&2048&&ds(p.alternate,p);break;default:bl(s,p,b,A,u)}e=e.sibling}}function Mr(t,e){if(e.subtreeFlags&10256)for(e=e.child;e!==null;){var a=t,r=e,u=r.flags;switch(r.tag){case 22:Mr(a,r),u&2048&&fs(r.alternate,r);break;case 24:Mr(a,r),u&2048&&ds(r.alternate,r);break;default:Mr(a,r)}e=e.sibling}}var Rr=8192;function Sl(t){if(t.subtreeFlags&Rr)for(t=t.child;t!==null;)sy(t),t=t.sibling}function sy(t){switch(t.tag){case 26:Sl(t),t.flags&Rr&&t.memoizedState!==null&&Mb(Ie,t.memoizedState,t.memoizedProps);break;case 5:Sl(t);break;case 3:case 4:var e=Ie;Ie=No(t.stateNode.containerInfo),Sl(t),Ie=e;break;case 22:t.memoizedState===null&&(e=t.alternate,e!==null&&e.memoizedState!==null?(e=Rr,Rr=16777216,Sl(t),Rr=e):Sl(t));break;default:Sl(t)}}function fy(t){var e=t.alternate;if(e!==null&&(t=e.child,t!==null)){e.child=null;do e=t.sibling,t.sibling=null,t=e;while(t!==null)}}function zr(t){var e=t.deletions;if((t.flags&16)!==0){if(e!==null)for(var a=0;a<e.length;a++){var r=e[a];te=r,hy(r,t)}fy(t)}if(t.subtreeFlags&10256)for(t=t.child;t!==null;)dy(t),t=t.sibling}function dy(t){switch(t.tag){case 0:case 11:case 15:zr(t),t.flags&2048&&ta(9,t,t.return);break;case 3:zr(t);break;case 12:zr(t);break;case 22:var e=t.stateNode;t.memoizedState!==null&&e._visibility&2&&(t.return===null||t.return.tag!==13)?(e._visibility&=-3,po(t)):zr(t);break;default:zr(t)}}function po(t){var e=t.deletions;if((t.flags&16)!==0){if(e!==null)for(var a=0;a<e.length;a++){var r=e[a];te=r,hy(r,t)}fy(t)}for(t=t.child;t!==null;){switch(e=t,e.tag){case 0:case 11:case 15:ta(8,e,e.return),po(e);break;case 22:a=e.stateNode,a._visibility&2&&(a._visibility&=-3,po(e));break;default:po(e)}t=t.sibling}}function hy(t,e){for(;te!==null;){var a=te;switch(a.tag){case 0:case 11:case 15:ta(8,a,e);break;case 23:case 22:if(a.memoizedState!==null&&a.memoizedState.cachePool!==null){var r=a.memoizedState.cachePool.pool;r!=null&&r.refCount++}break;case 24:yr(a.memoizedState.cache)}if(r=a.child,r!==null)r.return=a,te=r;else t:for(a=t;te!==null;){r=te;var u=r.sibling,s=r.return;if(ly(r),r===a){te=null;break t}if(u!==null){u.return=s,te=u;break t}te=s}}}var Q1={getCacheForType:function(t){var e=fe(Kt),a=e.data.get(t);return a===void 0&&(a=t(),e.data.set(t,a)),a}},K1=typeof WeakMap=="function"?WeakMap:Map,Nt=0,Bt=null,Tt=null,Et=0,Ot=0,ze=null,aa=!1,_l=!1,hs=!1,Dn=0,qt=0,la=0,Da=0,ms=0,Ze=0,xl=0,Dr=null,xe=null,ys=!1,vs=0,go=1/0,bo=null,ra=null,re=0,ia=null,Tl=null,wl=0,ps=0,gs=null,my=null,Br=0,bs=null;function De(){if((Nt&2)!==0&&Et!==0)return Et&-Et;if(E.T!==null){var t=fl;return t!==0?t:Cs()}return Od()}function yy(){Ze===0&&(Ze=(Et&536870912)===0||At?Cd():536870912);var t=ke.current;return t!==null&&(t.flags|=32),Ze}function Be(t,e,a){(t===Bt&&(Ot===2||Ot===9)||t.cancelPendingCommit!==null)&&(El(t,0),oa(t,Et,Ze,!1)),Il(t,a),((Nt&2)===0||t!==Bt)&&(t===Bt&&((Nt&2)===0&&(Da|=a),qt===4&&oa(t,Et,Ze,!1)),sn(t))}function vy(t,e,a){if((Nt&6)!==0)throw Error(o(327));var r=!a&&(e&124)===0&&(e&t.expiredLanes)===0||Fl(t,e),u=r?P1(t,e):xs(t,e,!0),s=r;do{if(u===0){_l&&!r&&oa(t,e,0,!1);break}else{if(a=t.current.alternate,s&&!J1(a)){u=xs(t,e,!1),s=!1;continue}if(u===2){if(s=e,t.errorRecoveryDisabledLanes&s)var p=0;else p=t.pendingLanes&-536870913,p=p!==0?p:p&536870912?536870912:0;if(p!==0){e=p;t:{var b=t;u=Dr;var A=b.current.memoizedState.isDehydrated;if(A&&(El(b,p).flags|=256),p=xs(b,p,!1),p!==2){if(hs&&!A){b.errorRecoveryDisabledLanes|=s,Da|=s,u=4;break t}s=xe,xe=u,s!==null&&(xe===null?xe=s:xe.push.apply(xe,s))}u=p}if(s=!1,u!==2)continue}}if(u===1){El(t,0),oa(t,e,0,!0);break}t:{switch(r=t,s=u,s){case 0:case 1:throw Error(o(345));case 4:if((e&4194048)!==e)break;case 6:oa(r,e,Ze,!aa);break t;case 2:xe=null;break;case 3:case 5:break;default:throw Error(o(329))}if((e&62914560)===e&&(u=vs+300-Ht(),10<u)){if(oa(r,e,Ze,!aa),Ni(r,0,!0)!==0)break t;r.timeoutHandle=ky(py.bind(null,r,a,xe,bo,ys,e,Ze,Da,xl,aa,s,2,-0,0),u);break t}py(r,a,xe,bo,ys,e,Ze,Da,xl,aa,s,0,-0,0)}}break}while(!0);sn(t)}function py(t,e,a,r,u,s,p,b,A,H,K,F,q,G){if(t.timeoutHandle=-1,F=e.subtreeFlags,(F&8192||(F&16785408)===16785408)&&(Xr={stylesheets:null,count:0,unsuspend:Ob},sy(e),F=Rb(),F!==null)){t.cancelPendingCommit=F(wy.bind(null,t,e,s,a,r,u,p,b,A,K,1,q,G)),oa(t,s,p,!H);return}wy(t,e,s,a,r,u,p,b,A)}function J1(t){for(var e=t;;){var a=e.tag;if((a===0||a===11||a===15)&&e.flags&16384&&(a=e.updateQueue,a!==null&&(a=a.stores,a!==null)))for(var r=0;r<a.length;r++){var u=a[r],s=u.getSnapshot;u=u.value;try{if(!Ne(s(),u))return!1}catch{return!1}}if(a=e.child,e.subtreeFlags&16384&&a!==null)a.return=e,e=a;else{if(e===t)break;for(;e.sibling===null;){if(e.return===null||e.return===t)return!0;e=e.return}e.sibling.return=e.return,e=e.sibling}}return!0}function oa(t,e,a,r){e&=~ms,e&=~Da,t.suspendedLanes|=e,t.pingedLanes&=~e,r&&(t.warmLanes|=e),r=t.expirationTimes;for(var u=e;0<u;){var s=31-je(u),p=1<<s;r[s]=-1,u&=~p}a!==0&&jd(t,a,e)}function So(){return(Nt&6)===0?(Ur(0),!1):!0}function Ss(){if(Tt!==null){if(Ot===0)var t=Tt.return;else t=Tt,Cn=ja=null,Hc(t),pl=null,Cr=0,t=Tt;for(;t!==null;)Wm(t.alternate,t),t=t.return;Tt=null}}function El(t,e){var a=t.timeoutHandle;a!==-1&&(t.timeoutHandle=-1,hb(a)),a=t.cancelPendingCommit,a!==null&&(t.cancelPendingCommit=null,a()),Ss(),Bt=t,Tt=a=Tn(t.current,null),Et=e,Ot=0,ze=null,aa=!1,_l=Fl(t,e),hs=!1,xl=Ze=ms=Da=la=qt=0,xe=Dr=null,ys=!1,(e&8)!==0&&(e|=e&32);var r=t.entangledLanes;if(r!==0)for(t=t.entanglements,r&=e;0<r;){var u=31-je(r),s=1<<u;e|=t[u],r&=~s}return Dn=e,qi(),a}function gy(t,e){St=null,E.H=io,e===pr||e===Wi?(e=Bh(),Ot=3):e===Rh?(e=Bh(),Ot=4):Ot=e===Um?8:e!==null&&typeof e=="object"&&typeof e.then=="function"?6:1,ze=e,Tt===null&&(qt=1,fo(t,qe(e,t.current)))}function by(){var t=E.H;return E.H=io,t===null?io:t}function Sy(){var t=E.A;return E.A=Q1,t}function _s(){qt=4,aa||(Et&4194048)!==Et&&ke.current!==null||(_l=!0),(la&134217727)===0&&(Da&134217727)===0||Bt===null||oa(Bt,Et,Ze,!1)}function xs(t,e,a){var r=Nt;Nt|=2;var u=by(),s=Sy();(Bt!==t||Et!==e)&&(bo=null,El(t,e)),e=!1;var p=qt;t:do try{if(Ot!==0&&Tt!==null){var b=Tt,A=ze;switch(Ot){case 8:Ss(),p=6;break t;case 3:case 2:case 9:case 6:ke.current===null&&(e=!0);var H=Ot;if(Ot=0,ze=null,Cl(t,b,A,H),a&&_l){p=0;break t}break;default:H=Ot,Ot=0,ze=null,Cl(t,b,A,H)}}W1(),p=qt;break}catch(K){gy(t,K)}while(!0);return e&&t.shellSuspendCounter++,Cn=ja=null,Nt=r,E.H=u,E.A=s,Tt===null&&(Bt=null,Et=0,qi()),p}function W1(){for(;Tt!==null;)_y(Tt)}function P1(t,e){var a=Nt;Nt|=2;var r=by(),u=Sy();Bt!==t||Et!==e?(bo=null,go=Ht()+500,El(t,e)):_l=Fl(t,e);t:do try{if(Ot!==0&&Tt!==null){e=Tt;var s=ze;e:switch(Ot){case 1:Ot=0,ze=null,Cl(t,e,s,1);break;case 2:case 9:if(zh(s)){Ot=0,ze=null,xy(e);break}e=function(){Ot!==2&&Ot!==9||Bt!==t||(Ot=7),sn(t)},s.then(e,e);break t;case 3:Ot=7;break t;case 4:Ot=5;break t;case 7:zh(s)?(Ot=0,ze=null,xy(e)):(Ot=0,ze=null,Cl(t,e,s,7));break;case 5:var p=null;switch(Tt.tag){case 26:p=Tt.memoizedState;case 5:case 27:var b=Tt;if(!p||a0(p)){Ot=0,ze=null;var A=b.sibling;if(A!==null)Tt=A;else{var H=b.return;H!==null?(Tt=H,_o(H)):Tt=null}break e}}Ot=0,ze=null,Cl(t,e,s,5);break;case 6:Ot=0,ze=null,Cl(t,e,s,6);break;case 8:Ss(),qt=6;break t;default:throw Error(o(462))}}F1();break}catch(K){gy(t,K)}while(!0);return Cn=ja=null,E.H=r,E.A=u,Nt=a,Tt!==null?0:(Bt=null,Et=0,qi(),qt)}function F1(){for(;Tt!==null&&!pa();)_y(Tt)}function _y(t){var e=Km(t.alternate,t,Dn);t.memoizedProps=t.pendingProps,e===null?_o(t):Tt=e}function xy(t){var e=t,a=e.alternate;switch(e.tag){case 15:case 0:e=Xm(a,e,e.pendingProps,e.type,void 0,Et);break;case 11:e=Xm(a,e,e.pendingProps,e.type.render,e.ref,Et);break;case 5:Hc(e);default:Wm(a,e),e=Tt=Th(e,Dn),e=Km(a,e,Dn)}t.memoizedProps=t.pendingProps,e===null?_o(t):Tt=e}function Cl(t,e,a,r){Cn=ja=null,Hc(e),pl=null,Cr=0;var u=e.return;try{if(q1(t,u,e,a,Et)){qt=1,fo(t,qe(a,t.current)),Tt=null;return}}catch(s){if(u!==null)throw Tt=u,s;qt=1,fo(t,qe(a,t.current)),Tt=null;return}e.flags&32768?(At||r===1?t=!0:_l||(Et&536870912)!==0?t=!1:(aa=t=!0,(r===2||r===9||r===3||r===6)&&(r=ke.current,r!==null&&r.tag===13&&(r.flags|=16384))),Ty(e,t)):_o(e)}function _o(t){var e=t;do{if((e.flags&32768)!==0){Ty(e,aa);return}t=e.return;var a=G1(e.alternate,e,Dn);if(a!==null){Tt=a;return}if(e=e.sibling,e!==null){Tt=e;return}Tt=e=t}while(e!==null);qt===0&&(qt=5)}function Ty(t,e){do{var a=V1(t.alternate,t);if(a!==null){a.flags&=32767,Tt=a;return}if(a=t.return,a!==null&&(a.flags|=32768,a.subtreeFlags=0,a.deletions=null),!e&&(t=t.sibling,t!==null)){Tt=t;return}Tt=t=a}while(t!==null);qt=6,Tt=null}function wy(t,e,a,r,u,s,p,b,A){t.cancelPendingCommit=null;do xo();while(re!==0);if((Nt&6)!==0)throw Error(o(327));if(e!==null){if(e===t.current)throw Error(o(177));if(s=e.lanes|e.childLanes,s|=dc,Og(t,a,s,p,b,A),t===Bt&&(Tt=Bt=null,Et=0),Tl=e,ia=t,wl=a,ps=s,gs=u,my=r,(e.subtreeFlags&10256)!==0||(e.flags&10256)!==0?(t.callbackNode=null,t.callbackPriority=0,nb(Ci,function(){return Ny(),null})):(t.callbackNode=null,t.callbackPriority=0),r=(e.flags&13878)!==0,(e.subtreeFlags&13878)!==0||r){r=E.T,E.T=null,u=Y.p,Y.p=2,p=Nt,Nt|=4;try{k1(t,e,a)}finally{Nt=p,Y.p=u,E.T=r}}re=1,Ey(),Cy(),Ay()}}function Ey(){if(re===1){re=0;var t=ia,e=Tl,a=(e.flags&13878)!==0;if((e.subtreeFlags&13878)!==0||a){a=E.T,E.T=null;var r=Y.p;Y.p=2;var u=Nt;Nt|=4;try{oy(e,t);var s=Ds,p=hh(t.containerInfo),b=s.focusedElem,A=s.selectionRange;if(p!==b&&b&&b.ownerDocument&&dh(b.ownerDocument.documentElement,b)){if(A!==null&&oc(b)){var H=A.start,K=A.end;if(K===void 0&&(K=H),"selectionStart"in b)b.selectionStart=H,b.selectionEnd=Math.min(K,b.value.length);else{var F=b.ownerDocument||document,q=F&&F.defaultView||window;if(q.getSelection){var G=q.getSelection(),yt=b.textContent.length,dt=Math.min(A.start,yt),zt=A.end===void 0?dt:Math.min(A.end,yt);!G.extend&&dt>zt&&(p=zt,zt=dt,dt=p);var D=fh(b,dt),M=fh(b,zt);if(D&&M&&(G.rangeCount!==1||G.anchorNode!==D.node||G.anchorOffset!==D.offset||G.focusNode!==M.node||G.focusOffset!==M.offset)){var U=F.createRange();U.setStart(D.node,D.offset),G.removeAllRanges(),dt>zt?(G.addRange(U),G.extend(M.node,M.offset)):(U.setEnd(M.node,M.offset),G.addRange(U))}}}}for(F=[],G=b;G=G.parentNode;)G.nodeType===1&&F.push({element:G,left:G.scrollLeft,top:G.scrollTop});for(typeof b.focus=="function"&&b.focus(),b=0;b<F.length;b++){var J=F[b];J.element.scrollLeft=J.left,J.element.scrollTop=J.top}}Do=!!zs,Ds=zs=null}finally{Nt=u,Y.p=r,E.T=a}}t.current=e,re=2}}function Cy(){if(re===2){re=0;var t=ia,e=Tl,a=(e.flags&8772)!==0;if((e.subtreeFlags&8772)!==0||a){a=E.T,E.T=null;var r=Y.p;Y.p=2;var u=Nt;Nt|=4;try{ay(t,e.alternate,e)}finally{Nt=u,Y.p=r,E.T=a}}re=3}}function Ay(){if(re===4||re===3){re=0,ce();var t=ia,e=Tl,a=wl,r=my;(e.subtreeFlags&10256)!==0||(e.flags&10256)!==0?re=5:(re=0,Tl=ia=null,jy(t,t.pendingLanes));var u=t.pendingLanes;if(u===0&&(ra=null),Yu(a),e=e.stateNode,Ae&&typeof Ae.onCommitFiberRoot=="function")try{Ae.onCommitFiberRoot(Pl,e,void 0,(e.current.flags&128)===128)}catch{}if(r!==null){e=E.T,u=Y.p,Y.p=2,E.T=null;try{for(var s=t.onRecoverableError,p=0;p<r.length;p++){var b=r[p];s(b.value,{componentStack:b.stack})}}finally{E.T=e,Y.p=u}}(wl&3)!==0&&xo(),sn(t),u=t.pendingLanes,(a&4194090)!==0&&(u&42)!==0?t===bs?Br++:(Br=0,bs=t):Br=0,Ur(0)}}function jy(t,e){(t.pooledCacheLanes&=e)===0&&(e=t.pooledCache,e!=null&&(t.pooledCache=null,yr(e)))}function xo(t){return Ey(),Cy(),Ay(),Ny()}function Ny(){if(re!==5)return!1;var t=ia,e=ps;ps=0;var a=Yu(wl),r=E.T,u=Y.p;try{Y.p=32>a?32:a,E.T=null,a=gs,gs=null;var s=ia,p=wl;if(re=0,Tl=ia=null,wl=0,(Nt&6)!==0)throw Error(o(331));var b=Nt;if(Nt|=4,dy(s.current),cy(s,s.current,p,a),Nt=b,Ur(0,!1),Ae&&typeof Ae.onPostCommitFiberRoot=="function")try{Ae.onPostCommitFiberRoot(Pl,s)}catch{}return!0}finally{Y.p=u,E.T=r,jy(t,e)}}function Oy(t,e,a){e=qe(a,e),e=Pc(t.stateNode,e,2),t=Wn(t,e,2),t!==null&&(Il(t,2),sn(t))}function Dt(t,e,a){if(t.tag===3)Oy(t,t,a);else for(;e!==null;){if(e.tag===3){Oy(e,t,a);break}else if(e.tag===1){var r=e.stateNode;if(typeof e.type.getDerivedStateFromError=="function"||typeof r.componentDidCatch=="function"&&(ra===null||!ra.has(r))){t=qe(a,t),a=Dm(2),r=Wn(e,a,2),r!==null&&(Bm(a,r,e,t),Il(r,2),sn(r));break}}e=e.return}}function Ts(t,e,a){var r=t.pingCache;if(r===null){r=t.pingCache=new K1;var u=new Set;r.set(e,u)}else u=r.get(e),u===void 0&&(u=new Set,r.set(e,u));u.has(a)||(hs=!0,u.add(a),t=I1.bind(null,t,e,a),e.then(t,t))}function I1(t,e,a){var r=t.pingCache;r!==null&&r.delete(e),t.pingedLanes|=t.suspendedLanes&a,t.warmLanes&=~a,Bt===t&&(Et&a)===a&&(qt===4||qt===3&&(Et&62914560)===Et&&300>Ht()-vs?(Nt&2)===0&&El(t,0):ms|=a,xl===Et&&(xl=0)),sn(t)}function My(t,e){e===0&&(e=Ad()),t=ol(t,e),t!==null&&(Il(t,e),sn(t))}function tb(t){var e=t.memoizedState,a=0;e!==null&&(a=e.retryLane),My(t,a)}function eb(t,e){var a=0;switch(t.tag){case 13:var r=t.stateNode,u=t.memoizedState;u!==null&&(a=u.retryLane);break;case 19:r=t.stateNode;break;case 22:r=t.stateNode._retryCache;break;default:throw Error(o(314))}r!==null&&r.delete(e),My(t,a)}function nb(t,e){return ln(t,e)}var To=null,Al=null,ws=!1,wo=!1,Es=!1,Ba=0;function sn(t){t!==Al&&t.next===null&&(Al===null?To=Al=t:Al=Al.next=t),wo=!0,ws||(ws=!0,lb())}function Ur(t,e){if(!Es&&wo){Es=!0;do for(var a=!1,r=To;r!==null;){if(t!==0){var u=r.pendingLanes;if(u===0)var s=0;else{var p=r.suspendedLanes,b=r.pingedLanes;s=(1<<31-je(42|t)+1)-1,s&=u&~(p&~b),s=s&201326741?s&201326741|1:s?s|2:0}s!==0&&(a=!0,By(r,s))}else s=Et,s=Ni(r,r===Bt?s:0,r.cancelPendingCommit!==null||r.timeoutHandle!==-1),(s&3)===0||Fl(r,s)||(a=!0,By(r,s));r=r.next}while(a);Es=!1}}function ab(){Ry()}function Ry(){wo=ws=!1;var t=0;Ba!==0&&(db()&&(t=Ba),Ba=0);for(var e=Ht(),a=null,r=To;r!==null;){var u=r.next,s=zy(r,e);s===0?(r.next=null,a===null?To=u:a.next=u,u===null&&(Al=a)):(a=r,(t!==0||(s&3)!==0)&&(wo=!0)),r=u}Ur(t)}function zy(t,e){for(var a=t.suspendedLanes,r=t.pingedLanes,u=t.expirationTimes,s=t.pendingLanes&-62914561;0<s;){var p=31-je(s),b=1<<p,A=u[p];A===-1?((b&a)===0||(b&r)!==0)&&(u[p]=Ng(b,e)):A<=e&&(t.expiredLanes|=b),s&=~b}if(e=Bt,a=Et,a=Ni(t,t===e?a:0,t.cancelPendingCommit!==null||t.timeoutHandle!==-1),r=t.callbackNode,a===0||t===e&&(Ot===2||Ot===9)||t.cancelPendingCommit!==null)return r!==null&&r!==null&&Sn(r),t.callbackNode=null,t.callbackPriority=0;if((a&3)===0||Fl(t,a)){if(e=a&-a,e===t.callbackPriority)return e;switch(r!==null&&Sn(r),Yu(a)){case 2:case 8:a=Hu;break;case 32:a=Ci;break;case 268435456:a=Ed;break;default:a=Ci}return r=Dy.bind(null,t),a=ln(a,r),t.callbackPriority=e,t.callbackNode=a,e}return r!==null&&r!==null&&Sn(r),t.callbackPriority=2,t.callbackNode=null,2}function Dy(t,e){if(re!==0&&re!==5)return t.callbackNode=null,t.callbackPriority=0,null;var a=t.callbackNode;if(xo()&&t.callbackNode!==a)return null;var r=Et;return r=Ni(t,t===Bt?r:0,t.cancelPendingCommit!==null||t.timeoutHandle!==-1),r===0?null:(vy(t,r,e),zy(t,Ht()),t.callbackNode!=null&&t.callbackNode===a?Dy.bind(null,t):null)}function By(t,e){if(xo())return null;vy(t,e,!0)}function lb(){mb(function(){(Nt&6)!==0?ln(Ei,ab):Ry()})}function Cs(){return Ba===0&&(Ba=Cd()),Ba}function Uy(t){return t==null||typeof t=="symbol"||typeof t=="boolean"?null:typeof t=="function"?t:Di(""+t)}function Hy(t,e){var a=e.ownerDocument.createElement("input");return a.name=e.name,a.value=e.value,t.id&&a.setAttribute("form",t.id),e.parentNode.insertBefore(a,e),t=new FormData(t),a.parentNode.removeChild(a),t}function rb(t,e,a,r,u){if(e==="submit"&&a&&a.stateNode===u){var s=Uy((u[ge]||null).action),p=r.submitter;p&&(e=(e=p[ge]||null)?Uy(e.formAction):p.getAttribute("formAction"),e!==null&&(s=e,p=null));var b=new $i("action","action",null,r,u);t.push({event:b,listeners:[{instance:null,listener:function(){if(r.defaultPrevented){if(Ba!==0){var A=p?Hy(u,p):new FormData(u);Zc(a,{pending:!0,data:A,method:u.method,action:s},null,A)}}else typeof s=="function"&&(b.preventDefault(),A=p?Hy(u,p):new FormData(u),Zc(a,{pending:!0,data:A,method:u.method,action:s},s,A))},currentTarget:u}]})}}for(var As=0;As<fc.length;As++){var js=fc[As],ib=js.toLowerCase(),ob=js[0].toUpperCase()+js.slice(1);Fe(ib,"on"+ob)}Fe(vh,"onAnimationEnd"),Fe(ph,"onAnimationIteration"),Fe(gh,"onAnimationStart"),Fe("dblclick","onDoubleClick"),Fe("focusin","onFocus"),Fe("focusout","onBlur"),Fe(w1,"onTransitionRun"),Fe(E1,"onTransitionStart"),Fe(C1,"onTransitionCancel"),Fe(bh,"onTransitionEnd"),Pa("onMouseEnter",["mouseout","mouseover"]),Pa("onMouseLeave",["mouseout","mouseover"]),Pa("onPointerEnter",["pointerout","pointerover"]),Pa("onPointerLeave",["pointerout","pointerover"]),ba("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),ba("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),ba("onBeforeInput",["compositionend","keypress","textInput","paste"]),ba("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),ba("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),ba("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var Hr="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),ub=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(Hr));function $y(t,e){e=(e&4)!==0;for(var a=0;a<t.length;a++){var r=t[a],u=r.event;r=r.listeners;t:{var s=void 0;if(e)for(var p=r.length-1;0<=p;p--){var b=r[p],A=b.instance,H=b.currentTarget;if(b=b.listener,A!==s&&u.isPropagationStopped())break t;s=b,u.currentTarget=H;try{s(u)}catch(K){so(K)}u.currentTarget=null,s=A}else for(p=0;p<r.length;p++){if(b=r[p],A=b.instance,H=b.currentTarget,b=b.listener,A!==s&&u.isPropagationStopped())break t;s=b,u.currentTarget=H;try{s(u)}catch(K){so(K)}u.currentTarget=null,s=A}}}}function wt(t,e){var a=e[qu];a===void 0&&(a=e[qu]=new Set);var r=t+"__bubble";a.has(r)||(Ly(e,t,2,!1),a.add(r))}function Ns(t,e,a){var r=0;e&&(r|=4),Ly(a,t,r,e)}var Eo="_reactListening"+Math.random().toString(36).slice(2);function Os(t){if(!t[Eo]){t[Eo]=!0,Rd.forEach(function(a){a!=="selectionchange"&&(ub.has(a)||Ns(a,!1,t),Ns(a,!0,t))});var e=t.nodeType===9?t:t.ownerDocument;e===null||e[Eo]||(e[Eo]=!0,Ns("selectionchange",!1,e))}}function Ly(t,e,a,r){switch(c0(e)){case 2:var u=Bb;break;case 8:u=Ub;break;default:u=Vs}a=u.bind(null,e,a,t),u=void 0,!Fu||e!=="touchstart"&&e!=="touchmove"&&e!=="wheel"||(u=!0),r?u!==void 0?t.addEventListener(e,a,{capture:!0,passive:u}):t.addEventListener(e,a,!0):u!==void 0?t.addEventListener(e,a,{passive:u}):t.addEventListener(e,a,!1)}function Ms(t,e,a,r,u){var s=r;if((e&1)===0&&(e&2)===0&&r!==null)t:for(;;){if(r===null)return;var p=r.tag;if(p===3||p===4){var b=r.stateNode.containerInfo;if(b===u)break;if(p===4)for(p=r.return;p!==null;){var A=p.tag;if((A===3||A===4)&&p.stateNode.containerInfo===u)return;p=p.return}for(;b!==null;){if(p=Ka(b),p===null)return;if(A=p.tag,A===5||A===6||A===26||A===27){r=s=p;continue t}b=b.parentNode}}r=r.return}Zd(function(){var H=s,K=Wu(a),F=[];t:{var q=Sh.get(t);if(q!==void 0){var G=$i,yt=t;switch(t){case"keypress":if(Ui(a)===0)break t;case"keydown":case"keyup":G=n1;break;case"focusin":yt="focus",G=nc;break;case"focusout":yt="blur",G=nc;break;case"beforeblur":case"afterblur":G=nc;break;case"click":if(a.button===2)break t;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":G=Jd;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":G=Vg;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":G=r1;break;case vh:case ph:case gh:G=Qg;break;case bh:G=o1;break;case"scroll":case"scrollend":G=Xg;break;case"wheel":G=c1;break;case"copy":case"cut":case"paste":G=Jg;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":G=Pd;break;case"toggle":case"beforetoggle":G=f1}var dt=(e&4)!==0,zt=!dt&&(t==="scroll"||t==="scrollend"),D=dt?q!==null?q+"Capture":null:q;dt=[];for(var M=H,U;M!==null;){var J=M;if(U=J.stateNode,J=J.tag,J!==5&&J!==26&&J!==27||U===null||D===null||(J=nr(M,D),J!=null&&dt.push($r(M,J,U))),zt)break;M=M.return}0<dt.length&&(q=new G(q,yt,null,a,K),F.push({event:q,listeners:dt}))}}if((e&7)===0){t:{if(q=t==="mouseover"||t==="pointerover",G=t==="mouseout"||t==="pointerout",q&&a!==Ju&&(yt=a.relatedTarget||a.fromElement)&&(Ka(yt)||yt[Qa]))break t;if((G||q)&&(q=K.window===K?K:(q=K.ownerDocument)?q.defaultView||q.parentWindow:window,G?(yt=a.relatedTarget||a.toElement,G=H,yt=yt?Ka(yt):null,yt!==null&&(zt=f(yt),dt=yt.tag,yt!==zt||dt!==5&&dt!==27&&dt!==6)&&(yt=null)):(G=null,yt=H),G!==yt)){if(dt=Jd,J="onMouseLeave",D="onMouseEnter",M="mouse",(t==="pointerout"||t==="pointerover")&&(dt=Pd,J="onPointerLeave",D="onPointerEnter",M="pointer"),zt=G==null?q:er(G),U=yt==null?q:er(yt),q=new dt(J,M+"leave",G,a,K),q.target=zt,q.relatedTarget=U,J=null,Ka(K)===H&&(dt=new dt(D,M+"enter",yt,a,K),dt.target=U,dt.relatedTarget=zt,J=dt),zt=J,G&&yt)e:{for(dt=G,D=yt,M=0,U=dt;U;U=jl(U))M++;for(U=0,J=D;J;J=jl(J))U++;for(;0<M-U;)dt=jl(dt),M--;for(;0<U-M;)D=jl(D),U--;for(;M--;){if(dt===D||D!==null&&dt===D.alternate)break e;dt=jl(dt),D=jl(D)}dt=null}else dt=null;G!==null&&Yy(F,q,G,dt,!1),yt!==null&&zt!==null&&Yy(F,zt,yt,dt,!0)}}t:{if(q=H?er(H):window,G=q.nodeName&&q.nodeName.toLowerCase(),G==="select"||G==="input"&&q.type==="file")var ot=rh;else if(ah(q))if(ih)ot=_1;else{ot=b1;var xt=g1}else G=q.nodeName,!G||G.toLowerCase()!=="input"||q.type!=="checkbox"&&q.type!=="radio"?H&&Ku(H.elementType)&&(ot=rh):ot=S1;if(ot&&(ot=ot(t,H))){lh(F,ot,a,K);break t}xt&&xt(t,q,H),t==="focusout"&&H&&q.type==="number"&&H.memoizedProps.value!=null&&Qu(q,"number",q.value)}switch(xt=H?er(H):window,t){case"focusin":(ah(xt)||xt.contentEditable==="true")&&(ll=xt,uc=H,sr=null);break;case"focusout":sr=uc=ll=null;break;case"mousedown":cc=!0;break;case"contextmenu":case"mouseup":case"dragend":cc=!1,mh(F,a,K);break;case"selectionchange":if(T1)break;case"keydown":case"keyup":mh(F,a,K)}var ct;if(lc)t:{switch(t){case"compositionstart":var mt="onCompositionStart";break t;case"compositionend":mt="onCompositionEnd";break t;case"compositionupdate":mt="onCompositionUpdate";break t}mt=void 0}else al?eh(t,a)&&(mt="onCompositionEnd"):t==="keydown"&&a.keyCode===229&&(mt="onCompositionStart");mt&&(Fd&&a.locale!=="ko"&&(al||mt!=="onCompositionStart"?mt==="onCompositionEnd"&&al&&(ct=Qd()):(Zn=K,Iu="value"in Zn?Zn.value:Zn.textContent,al=!0)),xt=Co(H,mt),0<xt.length&&(mt=new Wd(mt,t,null,a,K),F.push({event:mt,listeners:xt}),ct?mt.data=ct:(ct=nh(a),ct!==null&&(mt.data=ct)))),(ct=h1?m1(t,a):y1(t,a))&&(mt=Co(H,"onBeforeInput"),0<mt.length&&(xt=new Wd("onBeforeInput","beforeinput",null,a,K),F.push({event:xt,listeners:mt}),xt.data=ct)),rb(F,t,H,a,K)}$y(F,e)})}function $r(t,e,a){return{instance:t,listener:e,currentTarget:a}}function Co(t,e){for(var a=e+"Capture",r=[];t!==null;){var u=t,s=u.stateNode;if(u=u.tag,u!==5&&u!==26&&u!==27||s===null||(u=nr(t,a),u!=null&&r.unshift($r(t,u,s)),u=nr(t,e),u!=null&&r.push($r(t,u,s))),t.tag===3)return r;t=t.return}return[]}function jl(t){if(t===null)return null;do t=t.return;while(t&&t.tag!==5&&t.tag!==27);return t||null}function Yy(t,e,a,r,u){for(var s=e._reactName,p=[];a!==null&&a!==r;){var b=a,A=b.alternate,H=b.stateNode;if(b=b.tag,A!==null&&A===r)break;b!==5&&b!==26&&b!==27||H===null||(A=H,u?(H=nr(a,s),H!=null&&p.unshift($r(a,H,A))):u||(H=nr(a,s),H!=null&&p.push($r(a,H,A)))),a=a.return}p.length!==0&&t.push({event:e,listeners:p})}var cb=/\r\n?/g,sb=/\u0000|\uFFFD/g;function qy(t){return(typeof t=="string"?t:""+t).replace(cb,`
`).replace(sb,"")}function Xy(t,e){return e=qy(e),qy(t)===e}function Ao(){}function Rt(t,e,a,r,u,s){switch(a){case"children":typeof r=="string"?e==="body"||e==="textarea"&&r===""||tl(t,r):(typeof r=="number"||typeof r=="bigint")&&e!=="body"&&tl(t,""+r);break;case"className":Mi(t,"class",r);break;case"tabIndex":Mi(t,"tabindex",r);break;case"dir":case"role":case"viewBox":case"width":case"height":Mi(t,a,r);break;case"style":Vd(t,r,s);break;case"data":if(e!=="object"){Mi(t,"data",r);break}case"src":case"href":if(r===""&&(e!=="a"||a!=="href")){t.removeAttribute(a);break}if(r==null||typeof r=="function"||typeof r=="symbol"||typeof r=="boolean"){t.removeAttribute(a);break}r=Di(""+r),t.setAttribute(a,r);break;case"action":case"formAction":if(typeof r=="function"){t.setAttribute(a,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}else typeof s=="function"&&(a==="formAction"?(e!=="input"&&Rt(t,e,"name",u.name,u,null),Rt(t,e,"formEncType",u.formEncType,u,null),Rt(t,e,"formMethod",u.formMethod,u,null),Rt(t,e,"formTarget",u.formTarget,u,null)):(Rt(t,e,"encType",u.encType,u,null),Rt(t,e,"method",u.method,u,null),Rt(t,e,"target",u.target,u,null)));if(r==null||typeof r=="symbol"||typeof r=="boolean"){t.removeAttribute(a);break}r=Di(""+r),t.setAttribute(a,r);break;case"onClick":r!=null&&(t.onclick=Ao);break;case"onScroll":r!=null&&wt("scroll",t);break;case"onScrollEnd":r!=null&&wt("scrollend",t);break;case"dangerouslySetInnerHTML":if(r!=null){if(typeof r!="object"||!("__html"in r))throw Error(o(61));if(a=r.__html,a!=null){if(u.children!=null)throw Error(o(60));t.innerHTML=a}}break;case"multiple":t.multiple=r&&typeof r!="function"&&typeof r!="symbol";break;case"muted":t.muted=r&&typeof r!="function"&&typeof r!="symbol";break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":break;case"autoFocus":break;case"xlinkHref":if(r==null||typeof r=="function"||typeof r=="boolean"||typeof r=="symbol"){t.removeAttribute("xlink:href");break}a=Di(""+r),t.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",a);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":r!=null&&typeof r!="function"&&typeof r!="symbol"?t.setAttribute(a,""+r):t.removeAttribute(a);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":r&&typeof r!="function"&&typeof r!="symbol"?t.setAttribute(a,""):t.removeAttribute(a);break;case"capture":case"download":r===!0?t.setAttribute(a,""):r!==!1&&r!=null&&typeof r!="function"&&typeof r!="symbol"?t.setAttribute(a,r):t.removeAttribute(a);break;case"cols":case"rows":case"size":case"span":r!=null&&typeof r!="function"&&typeof r!="symbol"&&!isNaN(r)&&1<=r?t.setAttribute(a,r):t.removeAttribute(a);break;case"rowSpan":case"start":r==null||typeof r=="function"||typeof r=="symbol"||isNaN(r)?t.removeAttribute(a):t.setAttribute(a,r);break;case"popover":wt("beforetoggle",t),wt("toggle",t),Oi(t,"popover",r);break;case"xlinkActuate":_n(t,"http://www.w3.org/1999/xlink","xlink:actuate",r);break;case"xlinkArcrole":_n(t,"http://www.w3.org/1999/xlink","xlink:arcrole",r);break;case"xlinkRole":_n(t,"http://www.w3.org/1999/xlink","xlink:role",r);break;case"xlinkShow":_n(t,"http://www.w3.org/1999/xlink","xlink:show",r);break;case"xlinkTitle":_n(t,"http://www.w3.org/1999/xlink","xlink:title",r);break;case"xlinkType":_n(t,"http://www.w3.org/1999/xlink","xlink:type",r);break;case"xmlBase":_n(t,"http://www.w3.org/XML/1998/namespace","xml:base",r);break;case"xmlLang":_n(t,"http://www.w3.org/XML/1998/namespace","xml:lang",r);break;case"xmlSpace":_n(t,"http://www.w3.org/XML/1998/namespace","xml:space",r);break;case"is":Oi(t,"is",r);break;case"innerText":case"textContent":break;default:(!(2<a.length)||a[0]!=="o"&&a[0]!=="O"||a[1]!=="n"&&a[1]!=="N")&&(a=Yg.get(a)||a,Oi(t,a,r))}}function Rs(t,e,a,r,u,s){switch(a){case"style":Vd(t,r,s);break;case"dangerouslySetInnerHTML":if(r!=null){if(typeof r!="object"||!("__html"in r))throw Error(o(61));if(a=r.__html,a!=null){if(u.children!=null)throw Error(o(60));t.innerHTML=a}}break;case"children":typeof r=="string"?tl(t,r):(typeof r=="number"||typeof r=="bigint")&&tl(t,""+r);break;case"onScroll":r!=null&&wt("scroll",t);break;case"onScrollEnd":r!=null&&wt("scrollend",t);break;case"onClick":r!=null&&(t.onclick=Ao);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":break;case"innerText":case"textContent":break;default:if(!zd.hasOwnProperty(a))t:{if(a[0]==="o"&&a[1]==="n"&&(u=a.endsWith("Capture"),e=a.slice(2,u?a.length-7:void 0),s=t[ge]||null,s=s!=null?s[a]:null,typeof s=="function"&&t.removeEventListener(e,s,u),typeof r=="function")){typeof s!="function"&&s!==null&&(a in t?t[a]=null:t.hasAttribute(a)&&t.removeAttribute(a)),t.addEventListener(e,r,u);break t}a in t?t[a]=r:r===!0?t.setAttribute(a,""):Oi(t,a,r)}}}function ie(t,e,a){switch(e){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":wt("error",t),wt("load",t);var r=!1,u=!1,s;for(s in a)if(a.hasOwnProperty(s)){var p=a[s];if(p!=null)switch(s){case"src":r=!0;break;case"srcSet":u=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(o(137,e));default:Rt(t,e,s,p,a,null)}}u&&Rt(t,e,"srcSet",a.srcSet,a,null),r&&Rt(t,e,"src",a.src,a,null);return;case"input":wt("invalid",t);var b=s=p=u=null,A=null,H=null;for(r in a)if(a.hasOwnProperty(r)){var K=a[r];if(K!=null)switch(r){case"name":u=K;break;case"type":p=K;break;case"checked":A=K;break;case"defaultChecked":H=K;break;case"value":s=K;break;case"defaultValue":b=K;break;case"children":case"dangerouslySetInnerHTML":if(K!=null)throw Error(o(137,e));break;default:Rt(t,e,r,K,a,null)}}Yd(t,s,b,A,H,p,u,!1),Ri(t);return;case"select":wt("invalid",t),r=p=s=null;for(u in a)if(a.hasOwnProperty(u)&&(b=a[u],b!=null))switch(u){case"value":s=b;break;case"defaultValue":p=b;break;case"multiple":r=b;default:Rt(t,e,u,b,a,null)}e=s,a=p,t.multiple=!!r,e!=null?Ia(t,!!r,e,!1):a!=null&&Ia(t,!!r,a,!0);return;case"textarea":wt("invalid",t),s=u=r=null;for(p in a)if(a.hasOwnProperty(p)&&(b=a[p],b!=null))switch(p){case"value":r=b;break;case"defaultValue":u=b;break;case"children":s=b;break;case"dangerouslySetInnerHTML":if(b!=null)throw Error(o(91));break;default:Rt(t,e,p,b,a,null)}Xd(t,r,u,s),Ri(t);return;case"option":for(A in a)if(a.hasOwnProperty(A)&&(r=a[A],r!=null))switch(A){case"selected":t.selected=r&&typeof r!="function"&&typeof r!="symbol";break;default:Rt(t,e,A,r,a,null)}return;case"dialog":wt("beforetoggle",t),wt("toggle",t),wt("cancel",t),wt("close",t);break;case"iframe":case"object":wt("load",t);break;case"video":case"audio":for(r=0;r<Hr.length;r++)wt(Hr[r],t);break;case"image":wt("error",t),wt("load",t);break;case"details":wt("toggle",t);break;case"embed":case"source":case"link":wt("error",t),wt("load",t);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(H in a)if(a.hasOwnProperty(H)&&(r=a[H],r!=null))switch(H){case"children":case"dangerouslySetInnerHTML":throw Error(o(137,e));default:Rt(t,e,H,r,a,null)}return;default:if(Ku(e)){for(K in a)a.hasOwnProperty(K)&&(r=a[K],r!==void 0&&Rs(t,e,K,r,a,void 0));return}}for(b in a)a.hasOwnProperty(b)&&(r=a[b],r!=null&&Rt(t,e,b,r,a,null))}function fb(t,e,a,r){switch(e){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var u=null,s=null,p=null,b=null,A=null,H=null,K=null;for(G in a){var F=a[G];if(a.hasOwnProperty(G)&&F!=null)switch(G){case"checked":break;case"value":break;case"defaultValue":A=F;default:r.hasOwnProperty(G)||Rt(t,e,G,null,r,F)}}for(var q in r){var G=r[q];if(F=a[q],r.hasOwnProperty(q)&&(G!=null||F!=null))switch(q){case"type":s=G;break;case"name":u=G;break;case"checked":H=G;break;case"defaultChecked":K=G;break;case"value":p=G;break;case"defaultValue":b=G;break;case"children":case"dangerouslySetInnerHTML":if(G!=null)throw Error(o(137,e));break;default:G!==F&&Rt(t,e,q,G,r,F)}}Zu(t,p,b,A,H,K,s,u);return;case"select":G=p=b=q=null;for(s in a)if(A=a[s],a.hasOwnProperty(s)&&A!=null)switch(s){case"value":break;case"multiple":G=A;default:r.hasOwnProperty(s)||Rt(t,e,s,null,r,A)}for(u in r)if(s=r[u],A=a[u],r.hasOwnProperty(u)&&(s!=null||A!=null))switch(u){case"value":q=s;break;case"defaultValue":b=s;break;case"multiple":p=s;default:s!==A&&Rt(t,e,u,s,r,A)}e=b,a=p,r=G,q!=null?Ia(t,!!a,q,!1):!!r!=!!a&&(e!=null?Ia(t,!!a,e,!0):Ia(t,!!a,a?[]:"",!1));return;case"textarea":G=q=null;for(b in a)if(u=a[b],a.hasOwnProperty(b)&&u!=null&&!r.hasOwnProperty(b))switch(b){case"value":break;case"children":break;default:Rt(t,e,b,null,r,u)}for(p in r)if(u=r[p],s=a[p],r.hasOwnProperty(p)&&(u!=null||s!=null))switch(p){case"value":q=u;break;case"defaultValue":G=u;break;case"children":break;case"dangerouslySetInnerHTML":if(u!=null)throw Error(o(91));break;default:u!==s&&Rt(t,e,p,u,r,s)}qd(t,q,G);return;case"option":for(var yt in a)if(q=a[yt],a.hasOwnProperty(yt)&&q!=null&&!r.hasOwnProperty(yt))switch(yt){case"selected":t.selected=!1;break;default:Rt(t,e,yt,null,r,q)}for(A in r)if(q=r[A],G=a[A],r.hasOwnProperty(A)&&q!==G&&(q!=null||G!=null))switch(A){case"selected":t.selected=q&&typeof q!="function"&&typeof q!="symbol";break;default:Rt(t,e,A,q,r,G)}return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var dt in a)q=a[dt],a.hasOwnProperty(dt)&&q!=null&&!r.hasOwnProperty(dt)&&Rt(t,e,dt,null,r,q);for(H in r)if(q=r[H],G=a[H],r.hasOwnProperty(H)&&q!==G&&(q!=null||G!=null))switch(H){case"children":case"dangerouslySetInnerHTML":if(q!=null)throw Error(o(137,e));break;default:Rt(t,e,H,q,r,G)}return;default:if(Ku(e)){for(var zt in a)q=a[zt],a.hasOwnProperty(zt)&&q!==void 0&&!r.hasOwnProperty(zt)&&Rs(t,e,zt,void 0,r,q);for(K in r)q=r[K],G=a[K],!r.hasOwnProperty(K)||q===G||q===void 0&&G===void 0||Rs(t,e,K,q,r,G);return}}for(var D in a)q=a[D],a.hasOwnProperty(D)&&q!=null&&!r.hasOwnProperty(D)&&Rt(t,e,D,null,r,q);for(F in r)q=r[F],G=a[F],!r.hasOwnProperty(F)||q===G||q==null&&G==null||Rt(t,e,F,q,r,G)}var zs=null,Ds=null;function jo(t){return t.nodeType===9?t:t.ownerDocument}function Gy(t){switch(t){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function Vy(t,e){if(t===0)switch(e){case"svg":return 1;case"math":return 2;default:return 0}return t===1&&e==="foreignObject"?0:t}function Bs(t,e){return t==="textarea"||t==="noscript"||typeof e.children=="string"||typeof e.children=="number"||typeof e.children=="bigint"||typeof e.dangerouslySetInnerHTML=="object"&&e.dangerouslySetInnerHTML!==null&&e.dangerouslySetInnerHTML.__html!=null}var Us=null;function db(){var t=window.event;return t&&t.type==="popstate"?t===Us?!1:(Us=t,!0):(Us=null,!1)}var ky=typeof setTimeout=="function"?setTimeout:void 0,hb=typeof clearTimeout=="function"?clearTimeout:void 0,Zy=typeof Promise=="function"?Promise:void 0,mb=typeof queueMicrotask=="function"?queueMicrotask:typeof Zy<"u"?function(t){return Zy.resolve(null).then(t).catch(yb)}:ky;function yb(t){setTimeout(function(){throw t})}function ua(t){return t==="head"}function Qy(t,e){var a=e,r=0,u=0;do{var s=a.nextSibling;if(t.removeChild(a),s&&s.nodeType===8)if(a=s.data,a==="/$"){if(0<r&&8>r){a=r;var p=t.ownerDocument;if(a&1&&Lr(p.documentElement),a&2&&Lr(p.body),a&4)for(a=p.head,Lr(a),p=a.firstChild;p;){var b=p.nextSibling,A=p.nodeName;p[tr]||A==="SCRIPT"||A==="STYLE"||A==="LINK"&&p.rel.toLowerCase()==="stylesheet"||a.removeChild(p),p=b}}if(u===0){t.removeChild(s),Qr(e);return}u--}else a==="$"||a==="$?"||a==="$!"?u++:r=a.charCodeAt(0)-48;else r=0;a=s}while(a);Qr(e)}function Hs(t){var e=t.firstChild;for(e&&e.nodeType===10&&(e=e.nextSibling);e;){var a=e;switch(e=e.nextSibling,a.nodeName){case"HTML":case"HEAD":case"BODY":Hs(a),Xu(a);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if(a.rel.toLowerCase()==="stylesheet")continue}t.removeChild(a)}}function vb(t,e,a,r){for(;t.nodeType===1;){var u=a;if(t.nodeName.toLowerCase()!==e.toLowerCase()){if(!r&&(t.nodeName!=="INPUT"||t.type!=="hidden"))break}else if(r){if(!t[tr])switch(e){case"meta":if(!t.hasAttribute("itemprop"))break;return t;case"link":if(s=t.getAttribute("rel"),s==="stylesheet"&&t.hasAttribute("data-precedence"))break;if(s!==u.rel||t.getAttribute("href")!==(u.href==null||u.href===""?null:u.href)||t.getAttribute("crossorigin")!==(u.crossOrigin==null?null:u.crossOrigin)||t.getAttribute("title")!==(u.title==null?null:u.title))break;return t;case"style":if(t.hasAttribute("data-precedence"))break;return t;case"script":if(s=t.getAttribute("src"),(s!==(u.src==null?null:u.src)||t.getAttribute("type")!==(u.type==null?null:u.type)||t.getAttribute("crossorigin")!==(u.crossOrigin==null?null:u.crossOrigin))&&s&&t.hasAttribute("async")&&!t.hasAttribute("itemprop"))break;return t;default:return t}}else if(e==="input"&&t.type==="hidden"){var s=u.name==null?null:""+u.name;if(u.type==="hidden"&&t.getAttribute("name")===s)return t}else return t;if(t=tn(t.nextSibling),t===null)break}return null}function pb(t,e,a){if(e==="")return null;for(;t.nodeType!==3;)if((t.nodeType!==1||t.nodeName!=="INPUT"||t.type!=="hidden")&&!a||(t=tn(t.nextSibling),t===null))return null;return t}function $s(t){return t.data==="$!"||t.data==="$?"&&t.ownerDocument.readyState==="complete"}function gb(t,e){var a=t.ownerDocument;if(t.data!=="$?"||a.readyState==="complete")e();else{var r=function(){e(),a.removeEventListener("DOMContentLoaded",r)};a.addEventListener("DOMContentLoaded",r),t._reactRetry=r}}function tn(t){for(;t!=null;t=t.nextSibling){var e=t.nodeType;if(e===1||e===3)break;if(e===8){if(e=t.data,e==="$"||e==="$!"||e==="$?"||e==="F!"||e==="F")break;if(e==="/$")return null}}return t}var Ls=null;function Ky(t){t=t.previousSibling;for(var e=0;t;){if(t.nodeType===8){var a=t.data;if(a==="$"||a==="$!"||a==="$?"){if(e===0)return t;e--}else a==="/$"&&e++}t=t.previousSibling}return null}function Jy(t,e,a){switch(e=jo(a),t){case"html":if(t=e.documentElement,!t)throw Error(o(452));return t;case"head":if(t=e.head,!t)throw Error(o(453));return t;case"body":if(t=e.body,!t)throw Error(o(454));return t;default:throw Error(o(451))}}function Lr(t){for(var e=t.attributes;e.length;)t.removeAttributeNode(e[0]);Xu(t)}var Qe=new Map,Wy=new Set;function No(t){return typeof t.getRootNode=="function"?t.getRootNode():t.nodeType===9?t:t.ownerDocument}var Bn=Y.d;Y.d={f:bb,r:Sb,D:_b,C:xb,L:Tb,m:wb,X:Cb,S:Eb,M:Ab};function bb(){var t=Bn.f(),e=So();return t||e}function Sb(t){var e=Ja(t);e!==null&&e.tag===5&&e.type==="form"?vm(e):Bn.r(t)}var Nl=typeof document>"u"?null:document;function Py(t,e,a){var r=Nl;if(r&&typeof e=="string"&&e){var u=Ye(e);u='link[rel="'+t+'"][href="'+u+'"]',typeof a=="string"&&(u+='[crossorigin="'+a+'"]'),Wy.has(u)||(Wy.add(u),t={rel:t,crossOrigin:a,href:e},r.querySelector(u)===null&&(e=r.createElement("link"),ie(e,"link",t),Ft(e),r.head.appendChild(e)))}}function _b(t){Bn.D(t),Py("dns-prefetch",t,null)}function xb(t,e){Bn.C(t,e),Py("preconnect",t,e)}function Tb(t,e,a){Bn.L(t,e,a);var r=Nl;if(r&&t&&e){var u='link[rel="preload"][as="'+Ye(e)+'"]';e==="image"&&a&&a.imageSrcSet?(u+='[imagesrcset="'+Ye(a.imageSrcSet)+'"]',typeof a.imageSizes=="string"&&(u+='[imagesizes="'+Ye(a.imageSizes)+'"]')):u+='[href="'+Ye(t)+'"]';var s=u;switch(e){case"style":s=Ol(t);break;case"script":s=Ml(t)}Qe.has(s)||(t=v({rel:"preload",href:e==="image"&&a&&a.imageSrcSet?void 0:t,as:e},a),Qe.set(s,t),r.querySelector(u)!==null||e==="style"&&r.querySelector(Yr(s))||e==="script"&&r.querySelector(qr(s))||(e=r.createElement("link"),ie(e,"link",t),Ft(e),r.head.appendChild(e)))}}function wb(t,e){Bn.m(t,e);var a=Nl;if(a&&t){var r=e&&typeof e.as=="string"?e.as:"script",u='link[rel="modulepreload"][as="'+Ye(r)+'"][href="'+Ye(t)+'"]',s=u;switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":s=Ml(t)}if(!Qe.has(s)&&(t=v({rel:"modulepreload",href:t},e),Qe.set(s,t),a.querySelector(u)===null)){switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(a.querySelector(qr(s)))return}r=a.createElement("link"),ie(r,"link",t),Ft(r),a.head.appendChild(r)}}}function Eb(t,e,a){Bn.S(t,e,a);var r=Nl;if(r&&t){var u=Wa(r).hoistableStyles,s=Ol(t);e=e||"default";var p=u.get(s);if(!p){var b={loading:0,preload:null};if(p=r.querySelector(Yr(s)))b.loading=5;else{t=v({rel:"stylesheet",href:t,"data-precedence":e},a),(a=Qe.get(s))&&Ys(t,a);var A=p=r.createElement("link");Ft(A),ie(A,"link",t),A._p=new Promise(function(H,K){A.onload=H,A.onerror=K}),A.addEventListener("load",function(){b.loading|=1}),A.addEventListener("error",function(){b.loading|=2}),b.loading|=4,Oo(p,e,r)}p={type:"stylesheet",instance:p,count:1,state:b},u.set(s,p)}}}function Cb(t,e){Bn.X(t,e);var a=Nl;if(a&&t){var r=Wa(a).hoistableScripts,u=Ml(t),s=r.get(u);s||(s=a.querySelector(qr(u)),s||(t=v({src:t,async:!0},e),(e=Qe.get(u))&&qs(t,e),s=a.createElement("script"),Ft(s),ie(s,"link",t),a.head.appendChild(s)),s={type:"script",instance:s,count:1,state:null},r.set(u,s))}}function Ab(t,e){Bn.M(t,e);var a=Nl;if(a&&t){var r=Wa(a).hoistableScripts,u=Ml(t),s=r.get(u);s||(s=a.querySelector(qr(u)),s||(t=v({src:t,async:!0,type:"module"},e),(e=Qe.get(u))&&qs(t,e),s=a.createElement("script"),Ft(s),ie(s,"link",t),a.head.appendChild(s)),s={type:"script",instance:s,count:1,state:null},r.set(u,s))}}function Fy(t,e,a,r){var u=(u=rt.current)?No(u):null;if(!u)throw Error(o(446));switch(t){case"meta":case"title":return null;case"style":return typeof a.precedence=="string"&&typeof a.href=="string"?(e=Ol(a.href),a=Wa(u).hoistableStyles,r=a.get(e),r||(r={type:"style",instance:null,count:0,state:null},a.set(e,r)),r):{type:"void",instance:null,count:0,state:null};case"link":if(a.rel==="stylesheet"&&typeof a.href=="string"&&typeof a.precedence=="string"){t=Ol(a.href);var s=Wa(u).hoistableStyles,p=s.get(t);if(p||(u=u.ownerDocument||u,p={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},s.set(t,p),(s=u.querySelector(Yr(t)))&&!s._p&&(p.instance=s,p.state.loading=5),Qe.has(t)||(a={rel:"preload",as:"style",href:a.href,crossOrigin:a.crossOrigin,integrity:a.integrity,media:a.media,hrefLang:a.hrefLang,referrerPolicy:a.referrerPolicy},Qe.set(t,a),s||jb(u,t,a,p.state))),e&&r===null)throw Error(o(528,""));return p}if(e&&r!==null)throw Error(o(529,""));return null;case"script":return e=a.async,a=a.src,typeof a=="string"&&e&&typeof e!="function"&&typeof e!="symbol"?(e=Ml(a),a=Wa(u).hoistableScripts,r=a.get(e),r||(r={type:"script",instance:null,count:0,state:null},a.set(e,r)),r):{type:"void",instance:null,count:0,state:null};default:throw Error(o(444,t))}}function Ol(t){return'href="'+Ye(t)+'"'}function Yr(t){return'link[rel="stylesheet"]['+t+"]"}function Iy(t){return v({},t,{"data-precedence":t.precedence,precedence:null})}function jb(t,e,a,r){t.querySelector('link[rel="preload"][as="style"]['+e+"]")?r.loading=1:(e=t.createElement("link"),r.preload=e,e.addEventListener("load",function(){return r.loading|=1}),e.addEventListener("error",function(){return r.loading|=2}),ie(e,"link",a),Ft(e),t.head.appendChild(e))}function Ml(t){return'[src="'+Ye(t)+'"]'}function qr(t){return"script[async]"+t}function t0(t,e,a){if(e.count++,e.instance===null)switch(e.type){case"style":var r=t.querySelector('style[data-href~="'+Ye(a.href)+'"]');if(r)return e.instance=r,Ft(r),r;var u=v({},a,{"data-href":a.href,"data-precedence":a.precedence,href:null,precedence:null});return r=(t.ownerDocument||t).createElement("style"),Ft(r),ie(r,"style",u),Oo(r,a.precedence,t),e.instance=r;case"stylesheet":u=Ol(a.href);var s=t.querySelector(Yr(u));if(s)return e.state.loading|=4,e.instance=s,Ft(s),s;r=Iy(a),(u=Qe.get(u))&&Ys(r,u),s=(t.ownerDocument||t).createElement("link"),Ft(s);var p=s;return p._p=new Promise(function(b,A){p.onload=b,p.onerror=A}),ie(s,"link",r),e.state.loading|=4,Oo(s,a.precedence,t),e.instance=s;case"script":return s=Ml(a.src),(u=t.querySelector(qr(s)))?(e.instance=u,Ft(u),u):(r=a,(u=Qe.get(s))&&(r=v({},a),qs(r,u)),t=t.ownerDocument||t,u=t.createElement("script"),Ft(u),ie(u,"link",r),t.head.appendChild(u),e.instance=u);case"void":return null;default:throw Error(o(443,e.type))}else e.type==="stylesheet"&&(e.state.loading&4)===0&&(r=e.instance,e.state.loading|=4,Oo(r,a.precedence,t));return e.instance}function Oo(t,e,a){for(var r=a.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),u=r.length?r[r.length-1]:null,s=u,p=0;p<r.length;p++){var b=r[p];if(b.dataset.precedence===e)s=b;else if(s!==u)break}s?s.parentNode.insertBefore(t,s.nextSibling):(e=a.nodeType===9?a.head:a,e.insertBefore(t,e.firstChild))}function Ys(t,e){t.crossOrigin==null&&(t.crossOrigin=e.crossOrigin),t.referrerPolicy==null&&(t.referrerPolicy=e.referrerPolicy),t.title==null&&(t.title=e.title)}function qs(t,e){t.crossOrigin==null&&(t.crossOrigin=e.crossOrigin),t.referrerPolicy==null&&(t.referrerPolicy=e.referrerPolicy),t.integrity==null&&(t.integrity=e.integrity)}var Mo=null;function e0(t,e,a){if(Mo===null){var r=new Map,u=Mo=new Map;u.set(a,r)}else u=Mo,r=u.get(a),r||(r=new Map,u.set(a,r));if(r.has(t))return r;for(r.set(t,null),a=a.getElementsByTagName(t),u=0;u<a.length;u++){var s=a[u];if(!(s[tr]||s[se]||t==="link"&&s.getAttribute("rel")==="stylesheet")&&s.namespaceURI!=="http://www.w3.org/2000/svg"){var p=s.getAttribute(e)||"";p=t+p;var b=r.get(p);b?b.push(s):r.set(p,[s])}}return r}function n0(t,e,a){t=t.ownerDocument||t,t.head.insertBefore(a,e==="title"?t.querySelector("head > title"):null)}function Nb(t,e,a){if(a===1||e.itemProp!=null)return!1;switch(t){case"meta":case"title":return!0;case"style":if(typeof e.precedence!="string"||typeof e.href!="string"||e.href==="")break;return!0;case"link":if(typeof e.rel!="string"||typeof e.href!="string"||e.href===""||e.onLoad||e.onError)break;switch(e.rel){case"stylesheet":return t=e.disabled,typeof e.precedence=="string"&&t==null;default:return!0}case"script":if(e.async&&typeof e.async!="function"&&typeof e.async!="symbol"&&!e.onLoad&&!e.onError&&e.src&&typeof e.src=="string")return!0}return!1}function a0(t){return!(t.type==="stylesheet"&&(t.state.loading&3)===0)}var Xr=null;function Ob(){}function Mb(t,e,a){if(Xr===null)throw Error(o(475));var r=Xr;if(e.type==="stylesheet"&&(typeof a.media!="string"||matchMedia(a.media).matches!==!1)&&(e.state.loading&4)===0){if(e.instance===null){var u=Ol(a.href),s=t.querySelector(Yr(u));if(s){t=s._p,t!==null&&typeof t=="object"&&typeof t.then=="function"&&(r.count++,r=Ro.bind(r),t.then(r,r)),e.state.loading|=4,e.instance=s,Ft(s);return}s=t.ownerDocument||t,a=Iy(a),(u=Qe.get(u))&&Ys(a,u),s=s.createElement("link"),Ft(s);var p=s;p._p=new Promise(function(b,A){p.onload=b,p.onerror=A}),ie(s,"link",a),e.instance=s}r.stylesheets===null&&(r.stylesheets=new Map),r.stylesheets.set(e,t),(t=e.state.preload)&&(e.state.loading&3)===0&&(r.count++,e=Ro.bind(r),t.addEventListener("load",e),t.addEventListener("error",e))}}function Rb(){if(Xr===null)throw Error(o(475));var t=Xr;return t.stylesheets&&t.count===0&&Xs(t,t.stylesheets),0<t.count?function(e){var a=setTimeout(function(){if(t.stylesheets&&Xs(t,t.stylesheets),t.unsuspend){var r=t.unsuspend;t.unsuspend=null,r()}},6e4);return t.unsuspend=e,function(){t.unsuspend=null,clearTimeout(a)}}:null}function Ro(){if(this.count--,this.count===0){if(this.stylesheets)Xs(this,this.stylesheets);else if(this.unsuspend){var t=this.unsuspend;this.unsuspend=null,t()}}}var zo=null;function Xs(t,e){t.stylesheets=null,t.unsuspend!==null&&(t.count++,zo=new Map,e.forEach(zb,t),zo=null,Ro.call(t))}function zb(t,e){if(!(e.state.loading&4)){var a=zo.get(t);if(a)var r=a.get(null);else{a=new Map,zo.set(t,a);for(var u=t.querySelectorAll("link[data-precedence],style[data-precedence]"),s=0;s<u.length;s++){var p=u[s];(p.nodeName==="LINK"||p.getAttribute("media")!=="not all")&&(a.set(p.dataset.precedence,p),r=p)}r&&a.set(null,r)}u=e.instance,p=u.getAttribute("data-precedence"),s=a.get(p)||r,s===r&&a.set(null,u),a.set(p,u),this.count++,r=Ro.bind(this),u.addEventListener("load",r),u.addEventListener("error",r),s?s.parentNode.insertBefore(u,s.nextSibling):(t=t.nodeType===9?t.head:t,t.insertBefore(u,t.firstChild)),e.state.loading|=4}}var Gr={$$typeof:z,Provider:null,Consumer:null,_currentValue:k,_currentValue2:k,_threadCount:0};function Db(t,e,a,r,u,s,p,b){this.tag=1,this.containerInfo=t,this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=$u(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=$u(0),this.hiddenUpdates=$u(null),this.identifierPrefix=r,this.onUncaughtError=u,this.onCaughtError=s,this.onRecoverableError=p,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=b,this.incompleteTransitions=new Map}function l0(t,e,a,r,u,s,p,b,A,H,K,F){return t=new Db(t,e,a,p,b,A,H,F),e=1,s===!0&&(e|=24),s=Oe(3,null,null,e),t.current=s,s.stateNode=t,e=Tc(),e.refCount++,t.pooledCache=e,e.refCount++,s.memoizedState={element:r,isDehydrated:a,cache:e},Ac(s),t}function r0(t){return t?(t=ul,t):ul}function i0(t,e,a,r,u,s){u=r0(u),r.context===null?r.context=u:r.pendingContext=u,r=Jn(e),r.payload={element:a},s=s===void 0?null:s,s!==null&&(r.callback=s),a=Wn(t,r,e),a!==null&&(Be(a,t,e),br(a,t,e))}function o0(t,e){if(t=t.memoizedState,t!==null&&t.dehydrated!==null){var a=t.retryLane;t.retryLane=a!==0&&a<e?a:e}}function Gs(t,e){o0(t,e),(t=t.alternate)&&o0(t,e)}function u0(t){if(t.tag===13){var e=ol(t,67108864);e!==null&&Be(e,t,67108864),Gs(t,67108864)}}var Do=!0;function Bb(t,e,a,r){var u=E.T;E.T=null;var s=Y.p;try{Y.p=2,Vs(t,e,a,r)}finally{Y.p=s,E.T=u}}function Ub(t,e,a,r){var u=E.T;E.T=null;var s=Y.p;try{Y.p=8,Vs(t,e,a,r)}finally{Y.p=s,E.T=u}}function Vs(t,e,a,r){if(Do){var u=ks(r);if(u===null)Ms(t,e,r,Bo,a),s0(t,r);else if($b(u,t,e,a,r))r.stopPropagation();else if(s0(t,r),e&4&&-1<Hb.indexOf(t)){for(;u!==null;){var s=Ja(u);if(s!==null)switch(s.tag){case 3:if(s=s.stateNode,s.current.memoizedState.isDehydrated){var p=ga(s.pendingLanes);if(p!==0){var b=s;for(b.pendingLanes|=2,b.entangledLanes|=2;p;){var A=1<<31-je(p);b.entanglements[1]|=A,p&=~A}sn(s),(Nt&6)===0&&(go=Ht()+500,Ur(0))}}break;case 13:b=ol(s,2),b!==null&&Be(b,s,2),So(),Gs(s,2)}if(s=ks(r),s===null&&Ms(t,e,r,Bo,a),s===u)break;u=s}u!==null&&r.stopPropagation()}else Ms(t,e,r,null,a)}}function ks(t){return t=Wu(t),Zs(t)}var Bo=null;function Zs(t){if(Bo=null,t=Ka(t),t!==null){var e=f(t);if(e===null)t=null;else{var a=e.tag;if(a===13){if(t=d(e),t!==null)return t;t=null}else if(a===3){if(e.stateNode.current.memoizedState.isDehydrated)return e.tag===3?e.stateNode.containerInfo:null;t=null}else e!==t&&(t=null)}}return Bo=t,null}function c0(t){switch(t){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(Wl()){case Ei:return 2;case Hu:return 8;case Ci:case Tg:return 32;case Ed:return 268435456;default:return 32}default:return 32}}var Qs=!1,ca=null,sa=null,fa=null,Vr=new Map,kr=new Map,da=[],Hb="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function s0(t,e){switch(t){case"focusin":case"focusout":ca=null;break;case"dragenter":case"dragleave":sa=null;break;case"mouseover":case"mouseout":fa=null;break;case"pointerover":case"pointerout":Vr.delete(e.pointerId);break;case"gotpointercapture":case"lostpointercapture":kr.delete(e.pointerId)}}function Zr(t,e,a,r,u,s){return t===null||t.nativeEvent!==s?(t={blockedOn:e,domEventName:a,eventSystemFlags:r,nativeEvent:s,targetContainers:[u]},e!==null&&(e=Ja(e),e!==null&&u0(e)),t):(t.eventSystemFlags|=r,e=t.targetContainers,u!==null&&e.indexOf(u)===-1&&e.push(u),t)}function $b(t,e,a,r,u){switch(e){case"focusin":return ca=Zr(ca,t,e,a,r,u),!0;case"dragenter":return sa=Zr(sa,t,e,a,r,u),!0;case"mouseover":return fa=Zr(fa,t,e,a,r,u),!0;case"pointerover":var s=u.pointerId;return Vr.set(s,Zr(Vr.get(s)||null,t,e,a,r,u)),!0;case"gotpointercapture":return s=u.pointerId,kr.set(s,Zr(kr.get(s)||null,t,e,a,r,u)),!0}return!1}function f0(t){var e=Ka(t.target);if(e!==null){var a=f(e);if(a!==null){if(e=a.tag,e===13){if(e=d(a),e!==null){t.blockedOn=e,Mg(t.priority,function(){if(a.tag===13){var r=De();r=Lu(r);var u=ol(a,r);u!==null&&Be(u,a,r),Gs(a,r)}});return}}else if(e===3&&a.stateNode.current.memoizedState.isDehydrated){t.blockedOn=a.tag===3?a.stateNode.containerInfo:null;return}}}t.blockedOn=null}function Uo(t){if(t.blockedOn!==null)return!1;for(var e=t.targetContainers;0<e.length;){var a=ks(t.nativeEvent);if(a===null){a=t.nativeEvent;var r=new a.constructor(a.type,a);Ju=r,a.target.dispatchEvent(r),Ju=null}else return e=Ja(a),e!==null&&u0(e),t.blockedOn=a,!1;e.shift()}return!0}function d0(t,e,a){Uo(t)&&a.delete(e)}function Lb(){Qs=!1,ca!==null&&Uo(ca)&&(ca=null),sa!==null&&Uo(sa)&&(sa=null),fa!==null&&Uo(fa)&&(fa=null),Vr.forEach(d0),kr.forEach(d0)}function Ho(t,e){t.blockedOn===e&&(t.blockedOn=null,Qs||(Qs=!0,n.unstable_scheduleCallback(n.unstable_NormalPriority,Lb)))}var $o=null;function h0(t){$o!==t&&($o=t,n.unstable_scheduleCallback(n.unstable_NormalPriority,function(){$o===t&&($o=null);for(var e=0;e<t.length;e+=3){var a=t[e],r=t[e+1],u=t[e+2];if(typeof r!="function"){if(Zs(r||a)===null)continue;break}var s=Ja(a);s!==null&&(t.splice(e,3),e-=3,Zc(s,{pending:!0,data:u,method:a.method,action:r},r,u))}}))}function Qr(t){function e(A){return Ho(A,t)}ca!==null&&Ho(ca,t),sa!==null&&Ho(sa,t),fa!==null&&Ho(fa,t),Vr.forEach(e),kr.forEach(e);for(var a=0;a<da.length;a++){var r=da[a];r.blockedOn===t&&(r.blockedOn=null)}for(;0<da.length&&(a=da[0],a.blockedOn===null);)f0(a),a.blockedOn===null&&da.shift();if(a=(t.ownerDocument||t).$$reactFormReplay,a!=null)for(r=0;r<a.length;r+=3){var u=a[r],s=a[r+1],p=u[ge]||null;if(typeof s=="function")p||h0(a);else if(p){var b=null;if(s&&s.hasAttribute("formAction")){if(u=s,p=s[ge]||null)b=p.formAction;else if(Zs(u)!==null)continue}else b=p.action;typeof b=="function"?a[r+1]=b:(a.splice(r,3),r-=3),h0(a)}}}function Ks(t){this._internalRoot=t}Lo.prototype.render=Ks.prototype.render=function(t){var e=this._internalRoot;if(e===null)throw Error(o(409));var a=e.current,r=De();i0(a,r,t,e,null,null)},Lo.prototype.unmount=Ks.prototype.unmount=function(){var t=this._internalRoot;if(t!==null){this._internalRoot=null;var e=t.containerInfo;i0(t.current,2,null,t,null,null),So(),e[Qa]=null}};function Lo(t){this._internalRoot=t}Lo.prototype.unstable_scheduleHydration=function(t){if(t){var e=Od();t={blockedOn:null,target:t,priority:e};for(var a=0;a<da.length&&e!==0&&e<da[a].priority;a++);da.splice(a,0,t),a===0&&f0(t)}};var m0=l.version;if(m0!=="19.1.0")throw Error(o(527,m0,"19.1.0"));Y.findDOMNode=function(t){var e=t._reactInternals;if(e===void 0)throw typeof t.render=="function"?Error(o(188)):(t=Object.keys(t).join(","),Error(o(268,t)));return t=y(e),t=t!==null?m(t):null,t=t===null?null:t.stateNode,t};var Yb={bundleType:0,version:"19.1.0",rendererPackageName:"react-dom",currentDispatcherRef:E,reconcilerVersion:"19.1.0"};if(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__<"u"){var Yo=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!Yo.isDisabled&&Yo.supportsFiber)try{Pl=Yo.inject(Yb),Ae=Yo}catch{}}return Jr.createRoot=function(t,e){if(!c(t))throw Error(o(299));var a=!1,r="",u=Om,s=Mm,p=Rm,b=null;return e!=null&&(e.unstable_strictMode===!0&&(a=!0),e.identifierPrefix!==void 0&&(r=e.identifierPrefix),e.onUncaughtError!==void 0&&(u=e.onUncaughtError),e.onCaughtError!==void 0&&(s=e.onCaughtError),e.onRecoverableError!==void 0&&(p=e.onRecoverableError),e.unstable_transitionCallbacks!==void 0&&(b=e.unstable_transitionCallbacks)),e=l0(t,1,!1,null,null,a,r,u,s,p,b,null),t[Qa]=e.current,Os(t),new Ks(e)},Jr.hydrateRoot=function(t,e,a){if(!c(t))throw Error(o(299));var r=!1,u="",s=Om,p=Mm,b=Rm,A=null,H=null;return a!=null&&(a.unstable_strictMode===!0&&(r=!0),a.identifierPrefix!==void 0&&(u=a.identifierPrefix),a.onUncaughtError!==void 0&&(s=a.onUncaughtError),a.onCaughtError!==void 0&&(p=a.onCaughtError),a.onRecoverableError!==void 0&&(b=a.onRecoverableError),a.unstable_transitionCallbacks!==void 0&&(A=a.unstable_transitionCallbacks),a.formState!==void 0&&(H=a.formState)),e=l0(t,1,!0,e,a??null,r,u,s,p,b,A,H),e.context=r0(null),a=e.current,r=De(),r=Lu(r),u=Jn(r),u.callback=null,Wn(a,u,r),a=r,e.current.lanes=a,Il(e,a),sn(e),t[Qa]=e.current,Os(t),new Lo(e)},Jr.version="19.1.0",Jr}var w0;function Jb(){if(w0)return Ps.exports;w0=1;function n(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(n)}catch(l){console.error(l)}}return n(),Ps.exports=Kb(),Ps.exports}var Wb=Jb(),dn=function(){return dn=Object.assign||function(l){for(var i,o=1,c=arguments.length;o<c;o++){i=arguments[o];for(var f in i)Object.prototype.hasOwnProperty.call(i,f)&&(l[f]=i[f])}return l},dn.apply(this,arguments)};function xv(n,l){var i={};for(var o in n)Object.prototype.hasOwnProperty.call(n,o)&&l.indexOf(o)<0&&(i[o]=n[o]);if(n!=null&&typeof Object.getOwnPropertySymbols=="function")for(var c=0,o=Object.getOwnPropertySymbols(n);c<o.length;c++)l.indexOf(o[c])<0&&Object.prototype.propertyIsEnumerable.call(n,o[c])&&(i[o[c]]=n[o[c]]);return i}function Pb(n,l,i){if(i||arguments.length===2)for(var o=0,c=l.length,f;o<c;o++)(f||!(o in l))&&(f||(f=Array.prototype.slice.call(l,0,o)),f[o]=l[o]);return n.concat(f||Array.prototype.slice.call(l))}var Po="right-scroll-bar-position",Fo="width-before-scroll-bar",Fb="with-scroll-bars-hidden",Ib="--removed-body-scroll-bar-size";function ef(n,l){return typeof n=="function"?n(l):n&&(n.current=l),n}function tS(n,l){var i=C.useState(function(){return{value:n,callback:l,facade:{get current(){return i.value},set current(o){var c=i.value;c!==o&&(i.value=o,i.callback(o,c))}}}})[0];return i.callback=l,i.facade}var eS=typeof window<"u"?C.useLayoutEffect:C.useEffect,E0=new WeakMap;function nS(n,l){var i=tS(null,function(o){return n.forEach(function(c){return ef(c,o)})});return eS(function(){var o=E0.get(i);if(o){var c=new Set(o),f=new Set(n),d=i.current;c.forEach(function(h){f.has(h)||ef(h,null)}),f.forEach(function(h){c.has(h)||ef(h,d)})}E0.set(i,n)},[n]),i}function aS(n){return n}function lS(n,l){l===void 0&&(l=aS);var i=[],o=!1,c={read:function(){if(o)throw new Error("Sidecar: could not `read` from an `assigned` medium. `read` could be used only with `useMedium`.");return i.length?i[i.length-1]:n},useMedium:function(f){var d=l(f,o);return i.push(d),function(){i=i.filter(function(h){return h!==d})}},assignSyncMedium:function(f){for(o=!0;i.length;){var d=i;i=[],d.forEach(f)}i={push:function(h){return f(h)},filter:function(){return i}}},assignMedium:function(f){o=!0;var d=[];if(i.length){var h=i;i=[],h.forEach(f),d=i}var y=function(){var v=d;d=[],v.forEach(f)},m=function(){return Promise.resolve().then(y)};m(),i={push:function(v){d.push(v),m()},filter:function(v){return d=d.filter(v),i}}}};return c}function rS(n){n===void 0&&(n={});var l=lS(null);return l.options=dn({async:!0,ssr:!1},n),l}var Tv=function(n){var l=n.sideCar,i=xv(n,["sideCar"]);if(!l)throw new Error("Sidecar: please provide `sideCar` property to import the right car");var o=l.read();if(!o)throw new Error("Sidecar medium not found");return C.createElement(o,dn({},i))};Tv.isSideCarExport=!0;function iS(n,l){return n.useMedium(l),Tv}var wv=rS(),nf=function(){},mu=C.forwardRef(function(n,l){var i=C.useRef(null),o=C.useState({onScrollCapture:nf,onWheelCapture:nf,onTouchMoveCapture:nf}),c=o[0],f=o[1],d=n.forwardProps,h=n.children,y=n.className,m=n.removeScrollBar,v=n.enabled,S=n.shards,_=n.sideCar,x=n.noIsolation,w=n.inert,N=n.allowPinchZoom,B=n.as,O=B===void 0?"div":B,j=n.gapMode,z=xv(n,["forwardProps","children","className","removeScrollBar","enabled","shards","sideCar","noIsolation","inert","allowPinchZoom","as","gapMode"]),L=_,$=nS([i,l]),V=dn(dn({},z),c);return C.createElement(C.Fragment,null,v&&C.createElement(L,{sideCar:wv,removeScrollBar:m,shards:S,noIsolation:x,inert:w,setCallbacks:f,allowPinchZoom:!!N,lockRef:i,gapMode:j}),d?C.cloneElement(C.Children.only(h),dn(dn({},V),{ref:$})):C.createElement(O,dn({},V,{className:y,ref:$}),h))});mu.defaultProps={enabled:!0,removeScrollBar:!0,inert:!1};mu.classNames={fullWidth:Fo,zeroRight:Po};var oS=function(){if(typeof __webpack_nonce__<"u")return __webpack_nonce__};function uS(){if(!document)return null;var n=document.createElement("style");n.type="text/css";var l=oS();return l&&n.setAttribute("nonce",l),n}function cS(n,l){n.styleSheet?n.styleSheet.cssText=l:n.appendChild(document.createTextNode(l))}function sS(n){var l=document.head||document.getElementsByTagName("head")[0];l.appendChild(n)}var fS=function(){var n=0,l=null;return{add:function(i){n==0&&(l=uS())&&(cS(l,i),sS(l)),n++},remove:function(){n--,!n&&l&&(l.parentNode&&l.parentNode.removeChild(l),l=null)}}},dS=function(){var n=fS();return function(l,i){C.useEffect(function(){return n.add(l),function(){n.remove()}},[l&&i])}},Ev=function(){var n=dS(),l=function(i){var o=i.styles,c=i.dynamic;return n(o,c),null};return l},hS={left:0,top:0,right:0,gap:0},af=function(n){return parseInt(n||"",10)||0},mS=function(n){var l=window.getComputedStyle(document.body),i=l[n==="padding"?"paddingLeft":"marginLeft"],o=l[n==="padding"?"paddingTop":"marginTop"],c=l[n==="padding"?"paddingRight":"marginRight"];return[af(i),af(o),af(c)]},yS=function(n){if(n===void 0&&(n="margin"),typeof window>"u")return hS;var l=mS(n),i=document.documentElement.clientWidth,o=window.innerWidth;return{left:l[0],top:l[1],right:l[2],gap:Math.max(0,o-i+l[2]-l[0])}},vS=Ev(),Hl="data-scroll-locked",pS=function(n,l,i,o){var c=n.left,f=n.top,d=n.right,h=n.gap;return i===void 0&&(i="margin"),`
  .`.concat(Fb,` {
   overflow: hidden `).concat(o,`;
   padding-right: `).concat(h,"px ").concat(o,`;
  }
  body[`).concat(Hl,`] {
    overflow: hidden `).concat(o,`;
    overscroll-behavior: contain;
    `).concat([l&&"position: relative ".concat(o,";"),i==="margin"&&`
    padding-left: `.concat(c,`px;
    padding-top: `).concat(f,`px;
    padding-right: `).concat(d,`px;
    margin-left:0;
    margin-top:0;
    margin-right: `).concat(h,"px ").concat(o,`;
    `),i==="padding"&&"padding-right: ".concat(h,"px ").concat(o,";")].filter(Boolean).join(""),`
  }
  
  .`).concat(Po,` {
    right: `).concat(h,"px ").concat(o,`;
  }
  
  .`).concat(Fo,` {
    margin-right: `).concat(h,"px ").concat(o,`;
  }
  
  .`).concat(Po," .").concat(Po,` {
    right: 0 `).concat(o,`;
  }
  
  .`).concat(Fo," .").concat(Fo,` {
    margin-right: 0 `).concat(o,`;
  }
  
  body[`).concat(Hl,`] {
    `).concat(Ib,": ").concat(h,`px;
  }
`)},C0=function(){var n=parseInt(document.body.getAttribute(Hl)||"0",10);return isFinite(n)?n:0},gS=function(){C.useEffect(function(){return document.body.setAttribute(Hl,(C0()+1).toString()),function(){var n=C0()-1;n<=0?document.body.removeAttribute(Hl):document.body.setAttribute(Hl,n.toString())}},[])},bS=function(n){var l=n.noRelative,i=n.noImportant,o=n.gapMode,c=o===void 0?"margin":o;gS();var f=C.useMemo(function(){return yS(c)},[c]);return C.createElement(vS,{styles:pS(f,!l,c,i?"":"!important")})},bf=!1;if(typeof window<"u")try{var qo=Object.defineProperty({},"passive",{get:function(){return bf=!0,!0}});window.addEventListener("test",qo,qo),window.removeEventListener("test",qo,qo)}catch{bf=!1}var Rl=bf?{passive:!1}:!1,SS=function(n){return n.tagName==="TEXTAREA"},Cv=function(n,l){if(!(n instanceof Element))return!1;var i=window.getComputedStyle(n);return i[l]!=="hidden"&&!(i.overflowY===i.overflowX&&!SS(n)&&i[l]==="visible")},_S=function(n){return Cv(n,"overflowY")},xS=function(n){return Cv(n,"overflowX")},A0=function(n,l){var i=l.ownerDocument,o=l;do{typeof ShadowRoot<"u"&&o instanceof ShadowRoot&&(o=o.host);var c=Av(n,o);if(c){var f=jv(n,o),d=f[1],h=f[2];if(d>h)return!0}o=o.parentNode}while(o&&o!==i.body);return!1},TS=function(n){var l=n.scrollTop,i=n.scrollHeight,o=n.clientHeight;return[l,i,o]},wS=function(n){var l=n.scrollLeft,i=n.scrollWidth,o=n.clientWidth;return[l,i,o]},Av=function(n,l){return n==="v"?_S(l):xS(l)},jv=function(n,l){return n==="v"?TS(l):wS(l)},ES=function(n,l){return n==="h"&&l==="rtl"?-1:1},CS=function(n,l,i,o,c){var f=ES(n,window.getComputedStyle(l).direction),d=f*o,h=i.target,y=l.contains(h),m=!1,v=d>0,S=0,_=0;do{var x=jv(n,h),w=x[0],N=x[1],B=x[2],O=N-B-f*w;(w||O)&&Av(n,h)&&(S+=O,_+=w),h instanceof ShadowRoot?h=h.host:h=h.parentNode}while(!y&&h!==document.body||y&&(l.contains(h)||l===h));return(v&&Math.abs(S)<1||!v&&Math.abs(_)<1)&&(m=!0),m},Xo=function(n){return"changedTouches"in n?[n.changedTouches[0].clientX,n.changedTouches[0].clientY]:[0,0]},j0=function(n){return[n.deltaX,n.deltaY]},N0=function(n){return n&&"current"in n?n.current:n},AS=function(n,l){return n[0]===l[0]&&n[1]===l[1]},jS=function(n){return`
  .block-interactivity-`.concat(n,` {pointer-events: none;}
  .allow-interactivity-`).concat(n,` {pointer-events: all;}
`)},NS=0,zl=[];function OS(n){var l=C.useRef([]),i=C.useRef([0,0]),o=C.useRef(),c=C.useState(NS++)[0],f=C.useState(Ev)[0],d=C.useRef(n);C.useEffect(function(){d.current=n},[n]),C.useEffect(function(){if(n.inert){document.body.classList.add("block-interactivity-".concat(c));var N=Pb([n.lockRef.current],(n.shards||[]).map(N0),!0).filter(Boolean);return N.forEach(function(B){return B.classList.add("allow-interactivity-".concat(c))}),function(){document.body.classList.remove("block-interactivity-".concat(c)),N.forEach(function(B){return B.classList.remove("allow-interactivity-".concat(c))})}}},[n.inert,n.lockRef.current,n.shards]);var h=C.useCallback(function(N,B){if("touches"in N&&N.touches.length===2||N.type==="wheel"&&N.ctrlKey)return!d.current.allowPinchZoom;var O=Xo(N),j=i.current,z="deltaX"in N?N.deltaX:j[0]-O[0],L="deltaY"in N?N.deltaY:j[1]-O[1],$,V=N.target,Z=Math.abs(z)>Math.abs(L)?"h":"v";if("touches"in N&&Z==="h"&&V.type==="range")return!1;var W=A0(Z,V);if(!W)return!0;if(W?$=Z:($=Z==="v"?"h":"v",W=A0(Z,V)),!W)return!1;if(!o.current&&"changedTouches"in N&&(z||L)&&(o.current=$),!$)return!0;var it=o.current||$;return CS(it,B,N,it==="h"?z:L)},[]),y=C.useCallback(function(N){var B=N;if(!(!zl.length||zl[zl.length-1]!==f)){var O="deltaY"in B?j0(B):Xo(B),j=l.current.filter(function($){return $.name===B.type&&($.target===B.target||B.target===$.shadowParent)&&AS($.delta,O)})[0];if(j&&j.should){B.cancelable&&B.preventDefault();return}if(!j){var z=(d.current.shards||[]).map(N0).filter(Boolean).filter(function($){return $.contains(B.target)}),L=z.length>0?h(B,z[0]):!d.current.noIsolation;L&&B.cancelable&&B.preventDefault()}}},[]),m=C.useCallback(function(N,B,O,j){var z={name:N,delta:B,target:O,should:j,shadowParent:MS(O)};l.current.push(z),setTimeout(function(){l.current=l.current.filter(function(L){return L!==z})},1)},[]),v=C.useCallback(function(N){i.current=Xo(N),o.current=void 0},[]),S=C.useCallback(function(N){m(N.type,j0(N),N.target,h(N,n.lockRef.current))},[]),_=C.useCallback(function(N){m(N.type,Xo(N),N.target,h(N,n.lockRef.current))},[]);C.useEffect(function(){return zl.push(f),n.setCallbacks({onScrollCapture:S,onWheelCapture:S,onTouchMoveCapture:_}),document.addEventListener("wheel",y,Rl),document.addEventListener("touchmove",y,Rl),document.addEventListener("touchstart",v,Rl),function(){zl=zl.filter(function(N){return N!==f}),document.removeEventListener("wheel",y,Rl),document.removeEventListener("touchmove",y,Rl),document.removeEventListener("touchstart",v,Rl)}},[]);var x=n.removeScrollBar,w=n.inert;return C.createElement(C.Fragment,null,w?C.createElement(f,{styles:jS(c)}):null,x?C.createElement(bS,{gapMode:n.gapMode}):null)}function MS(n){for(var l=null;n!==null;)n instanceof ShadowRoot&&(l=n.host,n=n.host),n=n.parentNode;return l}const RS=iS(wv,OS);var Nv=C.forwardRef(function(n,l){return C.createElement(mu,dn({},n,{ref:l,sideCar:RS}))});Nv.classNames=mu.classNames;function hn(n){return Object.keys(n)}function lf(n){return n&&typeof n=="object"&&!Array.isArray(n)}function qf(n,l){const i={...n},o=l;return lf(n)&&lf(l)&&Object.keys(l).forEach(c=>{lf(o[c])&&c in n?i[c]=qf(i[c],o[c]):i[c]=o[c]}),i}function zS(n){return n.replace(/[A-Z]/g,l=>`-${l.toLowerCase()}`)}function DS(n){var l;return typeof n!="string"||!n.includes("var(--mantine-scale)")?n:(l=n.match(/^calc\((.*?)\)$/))==null?void 0:l[1].split("*")[0].trim()}function BS(n){const l=DS(n);return typeof l=="number"?l:typeof l=="string"?l.includes("calc")||l.includes("var")?l:l.includes("px")?Number(l.replace("px","")):l.includes("rem")?Number(l.replace("rem",""))*16:l.includes("em")?Number(l.replace("em",""))*16:Number(l):NaN}function rf(n){return n==="0rem"?"0rem":`calc(${n} * var(--mantine-scale))`}function Ov(n,{shouldScale:l=!1}={}){function i(o){if(o===0||o==="0")return`0${n}`;if(typeof o=="number"){const c=`${o/16}${n}`;return l?rf(c):c}if(typeof o=="string"){if(o===""||o.startsWith("calc(")||o.startsWith("clamp(")||o.includes("rgba("))return o;if(o.includes(","))return o.split(",").map(f=>i(f)).join(",");if(o.includes(" "))return o.split(" ").map(f=>i(f)).join(" ");if(o.includes(n))return l?rf(o):o;const c=o.replace("px","");if(!Number.isNaN(Number(c))){const f=`${Number(c)/16}${n}`;return l?rf(f):f}}return o}return i}const tt=Ov("rem",{shouldScale:!0}),O0=Ov("em");function yu(n){return Object.keys(n).reduce((l,i)=>(n[i]!==void 0&&(l[i]=n[i]),l),{})}function Mv(n){if(typeof n=="number")return!0;if(typeof n=="string"){if(n.startsWith("calc(")||n.startsWith("var(")||n.includes(" ")&&n.trim()!=="")return!0;const l=/^[+-]?[0-9]+(\.[0-9]+)?(px|em|rem|ex|ch|lh|rlh|vw|vh|vmin|vmax|vb|vi|svw|svh|lvw|lvh|dvw|dvh|cm|mm|in|pt|pc|q|cqw|cqh|cqi|cqb|cqmin|cqmax|%)?$/;return n.trim().split(/\s+/).every(o=>l.test(o))}return!1}function US(n){return Array.isArray(n)||n===null?!1:typeof n=="object"?n.type!==C.Fragment:!1}function Xa(n){const l=C.createContext(null);return[({children:c,value:f})=>g.jsx(l.Provider,{value:f,children:c}),()=>{const c=C.useContext(l);if(c===null)throw new Error(n);return c}]}function Xf(n=null){const l=C.createContext(n);return[({children:c,value:f})=>g.jsx(l.Provider,{value:f,children:c}),()=>C.useContext(l)]}function M0(n,l){return i=>{if(typeof i!="string"||i.trim().length===0)throw new Error(l);return`${n}-${i}`}}function Sf(n,l){let i=n;for(;(i=i.parentElement)&&!i.matches(l););return i}function HS(n,l,i){for(let o=n-1;o>=0;o-=1)if(!l[o].disabled)return o;if(i){for(let o=l.length-1;o>-1;o-=1)if(!l[o].disabled)return o}return n}function $S(n,l,i){for(let o=n+1;o<l.length;o+=1)if(!l[o].disabled)return o;if(i){for(let o=0;o<l.length;o+=1)if(!l[o].disabled)return o}return n}function LS(n,l,i){return Sf(n,i)===Sf(l,i)}function YS({parentSelector:n,siblingSelector:l,onKeyDown:i,loop:o=!0,activateOnFocus:c=!1,dir:f="rtl",orientation:d}){return h=>{var w;i==null||i(h);const y=Array.from(((w=Sf(h.currentTarget,n))==null?void 0:w.querySelectorAll(l))||[]).filter(N=>LS(h.currentTarget,N,n)),m=y.findIndex(N=>h.currentTarget===N),v=$S(m,y,o),S=HS(m,y,o),_=f==="rtl"?S:v,x=f==="rtl"?v:S;switch(h.key){case"ArrowRight":{d==="horizontal"&&(h.stopPropagation(),h.preventDefault(),y[_].focus(),c&&y[_].click());break}case"ArrowLeft":{d==="horizontal"&&(h.stopPropagation(),h.preventDefault(),y[x].focus(),c&&y[x].click());break}case"ArrowUp":{d==="vertical"&&(h.stopPropagation(),h.preventDefault(),y[S].focus(),c&&y[S].click());break}case"ArrowDown":{d==="vertical"&&(h.stopPropagation(),h.preventDefault(),y[v].focus(),c&&y[v].click());break}case"Home":{h.stopPropagation(),h.preventDefault(),!y[0].disabled&&y[0].focus();break}case"End":{h.stopPropagation(),h.preventDefault();const N=y.length-1;!y[N].disabled&&y[N].focus();break}}}}const qS={app:100,modal:200,popover:300,overlay:400,max:9999};function kl(n){return qS[n]}function Vt(n,l="size",i=!0){if(n!==void 0)return Mv(n)?i?tt(n):n:`var(--${l}-${n})`}function Ll(n){return Vt(n,"mantine-spacing")}function vn(n){return n===void 0?"var(--mantine-radius-default)":Vt(n,"mantine-radius")}function we(n){return Vt(n,"mantine-font-size")}function Rv(n){return Vt(n,"mantine-line-height",!1)}function zv(n){if(n)return Vt(n,"mantine-shadow",!1)}function XS(n="mantine-"){return`${n}${Math.random().toString(36).slice(2,11)}`}function Ha(n){const l=C.useRef(n);return C.useEffect(()=>{l.current=n}),C.useMemo(()=>(...i)=>{var o;return(o=l.current)==null?void 0:o.call(l,...i)},[])}const GS=()=>{};function vu(n,l){const i=typeof l=="number"?l:l.delay,o=typeof l=="number"?!1:l.flushOnUnmount,c=Ha(n),f=C.useRef(0),d=Object.assign(C.useCallback((...h)=>{window.clearTimeout(f.current);const y=()=>{f.current!==0&&(f.current=0,c(...h))};d.flush=y,f.current=window.setTimeout(y,i)},[c,i]),{flush:GS});return C.useEffect(()=>()=>{window.clearTimeout(f.current),o&&d.flush()},[d,o]),d}function VS(n,l){try{return n.addEventListener("change",l),()=>n.removeEventListener("change",l)}catch{return n.addListener(l),()=>n.removeListener(l)}}function kS(n,l){return typeof window<"u"&&"matchMedia"in window?window.matchMedia(n).matches:!1}function ZS(n,l,{getInitialValueInEffect:i}={getInitialValueInEffect:!0}){const[o,c]=C.useState(i?l:kS(n)),f=C.useRef(null);return C.useEffect(()=>{if("matchMedia"in window)return f.current=window.matchMedia(n),c(f.current.matches),VS(f.current,d=>c(d.matches))},[n]),o}const hi=typeof document<"u"?C.useLayoutEffect:C.useEffect;function Dv(n,l){const i=C.useRef(!1);C.useEffect(()=>()=>{i.current=!1},[]),C.useEffect(()=>{if(i.current)return n();i.current=!0},l)}function QS({opened:n,shouldReturnFocus:l=!0}){const i=C.useRef(null),o=()=>{var c;i.current&&"focus"in i.current&&typeof i.current.focus=="function"&&((c=i.current)==null||c.focus({preventScroll:!0}))};return Dv(()=>{let c=-1;const f=d=>{d.key==="Tab"&&window.clearTimeout(c)};return document.addEventListener("keydown",f),n?i.current=document.activeElement:l&&(c=window.setTimeout(o,10)),()=>{window.clearTimeout(c),document.removeEventListener("keydown",f)}},[n,l]),o}const KS=/input|select|textarea|button|object/,Bv="a, input, select, textarea, button, object, [tabindex]";function JS(n){return n.style.display==="none"}function WS(n){if(n.getAttribute("aria-hidden")||n.getAttribute("hidden")||n.getAttribute("type")==="hidden")return!1;let i=n;for(;i&&!(i===document.body||i.nodeType===11);){if(JS(i))return!1;i=i.parentNode}return!0}function Uv(n){let l=n.getAttribute("tabindex");return l===null&&(l=void 0),parseInt(l,10)}function _f(n){const l=n.nodeName.toLowerCase(),i=!Number.isNaN(Uv(n));return(KS.test(l)&&!n.disabled||n instanceof HTMLAnchorElement&&n.href||i)&&WS(n)}function Hv(n){const l=Uv(n);return(Number.isNaN(l)||l>=0)&&_f(n)}function PS(n){return Array.from(n.querySelectorAll(Bv)).filter(Hv)}function FS(n,l){const i=PS(n);if(!i.length){l.preventDefault();return}const o=i[l.shiftKey?0:i.length-1],c=n.getRootNode();let f=o===c.activeElement||n===c.activeElement;const d=c.activeElement;if(d.tagName==="INPUT"&&d.getAttribute("type")==="radio"&&(f=i.filter(v=>v.getAttribute("type")==="radio"&&v.getAttribute("name")===d.getAttribute("name")).includes(o)),!f)return;l.preventDefault();const y=i[l.shiftKey?i.length-1:0];y&&y.focus()}function IS(n=!0){const l=C.useRef(null),i=c=>{let f=c.querySelector("[data-autofocus]");if(!f){const d=Array.from(c.querySelectorAll(Bv));f=d.find(Hv)||d.find(_f)||null,!f&&_f(c)&&(f=c)}f&&f.focus({preventScroll:!0})},o=C.useCallback(c=>{n&&c!==null&&l.current!==c&&(c?(setTimeout(()=>{c.getRootNode()&&i(c)}),l.current=c):l.current=null)},[n]);return C.useEffect(()=>{if(!n)return;l.current&&setTimeout(()=>i(l.current));const c=f=>{f.key==="Tab"&&l.current&&FS(l.current,f)};return document.addEventListener("keydown",c),()=>document.removeEventListener("keydown",c)},[n]),o}const t2=Yf.useId||(()=>{});function e2(){const n=t2();return n?`mantine-${n.replace(/:/g,"")}`:""}function pu(n){const l=e2(),[i,o]=C.useState(l);return hi(()=>{o(XS())},[]),typeof n=="string"?n:typeof window>"u"?l:i}function n2(n,l,i){C.useEffect(()=>(window.addEventListener(n,l,i),()=>window.removeEventListener(n,l,i)),[n,l])}function xf(n,l){if(typeof n=="function")return n(l);typeof n=="object"&&n!==null&&"current"in n&&(n.current=l)}function a2(...n){const l=new Map;return i=>{if(n.forEach(o=>{const c=xf(o,i);c&&l.set(o,c)}),l.size>0)return()=>{n.forEach(o=>{const c=l.get(o);c?c():xf(o,null)}),l.clear()}}}function Ga(...n){return C.useCallback(a2(...n),n)}function $v({value:n,defaultValue:l,finalValue:i,onChange:o=()=>{}}){const[c,f]=C.useState(l!==void 0?l:i),d=(h,...y)=>{f(h),o==null||o(h,...y)};return n!==void 0?[n,o,!0]:[c,d,!1]}function Lv(n,l){return ZS("(prefers-reduced-motion: reduce)",n,l)}function l2(n=!1,l){const{onOpen:i,onClose:o}={},[c,f]=C.useState(n),d=C.useCallback(()=>{f(m=>m||(i==null||i(),!0))},[i]),h=C.useCallback(()=>{f(m=>m&&(o==null||o(),!1))},[o]),y=C.useCallback(()=>{c?h():d()},[h,d,c]);return[c,{open:d,close:h,toggle:y}]}function Yv(n){var l,i,o="";if(typeof n=="string"||typeof n=="number")o+=n;else if(typeof n=="object")if(Array.isArray(n)){var c=n.length;for(l=0;l<c;l++)n[l]&&(i=Yv(n[l]))&&(o&&(o+=" "),o+=i)}else for(i in n)n[i]&&(o&&(o+=" "),o+=i);return o}function Ce(){for(var n,l,i=0,o="",c=arguments.length;i<c;i++)(n=arguments[i])&&(l=Yv(n))&&(o&&(o+=" "),o+=l);return o}const r2={};function i2(n){const l={};return n.forEach(i=>{Object.entries(i).forEach(([o,c])=>{l[o]?l[o]=Ce(l[o],c):l[o]=c})}),l}function gu({theme:n,classNames:l,props:i,stylesCtx:o}){const f=(Array.isArray(l)?l:[l]).map(d=>typeof d=="function"?d(n,i,o):d||r2);return i2(f)}function ru({theme:n,styles:l,props:i,stylesCtx:o}){return(Array.isArray(l)?l:[l]).reduce((f,d)=>typeof d=="function"?{...f,...d(n,i,o)}:{...f,...d},{})}const qv=C.createContext(null);function va(){const n=C.useContext(qv);if(!n)throw new Error("[@mantine/core] MantineProvider was not found in tree");return n}function o2(){return va().cssVariablesResolver}function u2(){return va().classNamesPrefix}function Gf(){return va().getStyleNonce}function c2(){return va().withStaticClasses}function s2(){return va().headless}function f2(){var n;return(n=va().stylesTransform)==null?void 0:n.sx}function d2(){var n;return(n=va().stylesTransform)==null?void 0:n.styles}function Xv(){return va().env||"default"}function h2(n){return/^#?([0-9A-F]{3}){1,2}([0-9A-F]{2})?$/i.test(n)}function m2(n){let l=n.replace("#","");if(l.length===3){const d=l.split("");l=[d[0],d[0],d[1],d[1],d[2],d[2]].join("")}if(l.length===8){const d=parseInt(l.slice(6,8),16)/255;return{r:parseInt(l.slice(0,2),16),g:parseInt(l.slice(2,4),16),b:parseInt(l.slice(4,6),16),a:d}}const i=parseInt(l,16),o=i>>16&255,c=i>>8&255,f=i&255;return{r:o,g:c,b:f,a:1}}function y2(n){const[l,i,o,c]=n.replace(/[^0-9,./]/g,"").split(/[/,]/).map(Number);return{r:l,g:i,b:o,a:c||1}}function v2(n){const l=/^hsla?\(\s*(\d+)\s*,\s*(\d+%)\s*,\s*(\d+%)\s*(,\s*(0?\.\d+|\d+(\.\d+)?))?\s*\)$/i,i=n.match(l);if(!i)return{r:0,g:0,b:0,a:1};const o=parseInt(i[1],10),c=parseInt(i[2],10)/100,f=parseInt(i[3],10)/100,d=i[5]?parseFloat(i[5]):void 0,h=(1-Math.abs(2*f-1))*c,y=o/60,m=h*(1-Math.abs(y%2-1)),v=f-h/2;let S,_,x;return y>=0&&y<1?(S=h,_=m,x=0):y>=1&&y<2?(S=m,_=h,x=0):y>=2&&y<3?(S=0,_=h,x=m):y>=3&&y<4?(S=0,_=m,x=h):y>=4&&y<5?(S=m,_=0,x=h):(S=h,_=0,x=m),{r:Math.round((S+v)*255),g:Math.round((_+v)*255),b:Math.round((x+v)*255),a:d||1}}function Vf(n){return h2(n)?m2(n):n.startsWith("rgb")?y2(n):n.startsWith("hsl")?v2(n):{r:0,g:0,b:0,a:1}}function Go(n,l){if(n.startsWith("var("))return`color-mix(in srgb, ${n}, black ${l*100}%)`;const{r:i,g:o,b:c,a:f}=Vf(n),d=1-l,h=y=>Math.round(y*d);return`rgba(${h(i)}, ${h(o)}, ${h(c)}, ${f})`}function oi(n,l){return typeof n.primaryShade=="number"?n.primaryShade:l==="dark"?n.primaryShade.dark:n.primaryShade.light}function of(n){return n<=.03928?n/12.92:((n+.055)/1.055)**2.4}function p2(n){const l=n.match(/oklch\((.*?)%\s/);return l?parseFloat(l[1]):null}function g2(n){if(n.startsWith("oklch("))return(p2(n)||0)/100;const{r:l,g:i,b:o}=Vf(n),c=l/255,f=i/255,d=o/255,h=of(c),y=of(f),m=of(d);return .2126*h+.7152*y+.0722*m}function Wr(n,l=.179){return n.startsWith("var(")?!1:g2(n)>l}function mi({color:n,theme:l,colorScheme:i}){if(typeof n!="string")throw new Error(`[@mantine/core] Failed to parse color. Expected color to be a string, instead got ${typeof n}`);if(n==="bright")return{color:n,value:i==="dark"?l.white:l.black,shade:void 0,isThemeColor:!1,isLight:Wr(i==="dark"?l.white:l.black,l.luminanceThreshold),variable:"--mantine-color-bright"};if(n==="dimmed")return{color:n,value:i==="dark"?l.colors.dark[2]:l.colors.gray[7],shade:void 0,isThemeColor:!1,isLight:Wr(i==="dark"?l.colors.dark[2]:l.colors.gray[6],l.luminanceThreshold),variable:"--mantine-color-dimmed"};if(n==="white"||n==="black")return{color:n,value:n==="white"?l.white:l.black,shade:void 0,isThemeColor:!1,isLight:Wr(n==="white"?l.white:l.black,l.luminanceThreshold),variable:`--mantine-color-${n}`};const[o,c]=n.split("."),f=c?Number(c):void 0,d=o in l.colors;if(d){const h=f!==void 0?l.colors[o][f]:l.colors[o][oi(l,i||"light")];return{color:o,value:h,shade:f,isThemeColor:d,isLight:Wr(h,l.luminanceThreshold),variable:c?`--mantine-color-${o}-${f}`:`--mantine-color-${o}-filled`}}return{color:n,value:n,isThemeColor:d,isLight:Wr(n,l.luminanceThreshold),shade:f,variable:void 0}}function mn(n,l){const i=mi({color:n||l.primaryColor,theme:l});return i.variable?`var(${i.variable})`:n}function Tf(n,l){const i={from:(n==null?void 0:n.from)||l.defaultGradient.from,to:(n==null?void 0:n.to)||l.defaultGradient.to,deg:(n==null?void 0:n.deg)??l.defaultGradient.deg??0},o=mn(i.from,l),c=mn(i.to,l);return`linear-gradient(${i.deg}deg, ${o} 0%, ${c} 100%)`}function fn(n,l){if(typeof n!="string"||l>1||l<0)return"rgba(0, 0, 0, 1)";if(n.startsWith("var(")){const f=(1-l)*100;return`color-mix(in srgb, ${n}, transparent ${f}%)`}if(n.startsWith("oklch"))return n.includes("/")?n.replace(/\/\s*[\d.]+\s*\)/,`/ ${l})`):n.replace(")",` / ${l})`);const{r:i,g:o,b:c}=Vf(n);return`rgba(${i}, ${o}, ${c}, ${l})`}const Dl=fn,b2=({color:n,theme:l,variant:i,gradient:o,autoContrast:c})=>{const f=mi({color:n,theme:l}),d=typeof c=="boolean"?c:l.autoContrast;if(i==="filled"){const h=d&&f.isLight?"var(--mantine-color-black)":"var(--mantine-color-white)";return f.isThemeColor?f.shade===void 0?{background:`var(--mantine-color-${n}-filled)`,hover:`var(--mantine-color-${n}-filled-hover)`,color:h,border:`${tt(1)} solid transparent`}:{background:`var(--mantine-color-${f.color}-${f.shade})`,hover:`var(--mantine-color-${f.color}-${f.shade===9?8:f.shade+1})`,color:h,border:`${tt(1)} solid transparent`}:{background:n,hover:Go(n,.1),color:h,border:`${tt(1)} solid transparent`}}if(i==="light"){if(f.isThemeColor){if(f.shade===void 0)return{background:`var(--mantine-color-${n}-light)`,hover:`var(--mantine-color-${n}-light-hover)`,color:`var(--mantine-color-${n}-light-color)`,border:`${tt(1)} solid transparent`};const h=l.colors[f.color][f.shade];return{background:fn(h,.1),hover:fn(h,.12),color:`var(--mantine-color-${f.color}-${Math.min(f.shade,6)})`,border:`${tt(1)} solid transparent`}}return{background:fn(n,.1),hover:fn(n,.12),color:n,border:`${tt(1)} solid transparent`}}if(i==="outline")return f.isThemeColor?f.shade===void 0?{background:"transparent",hover:`var(--mantine-color-${n}-outline-hover)`,color:`var(--mantine-color-${n}-outline)`,border:`${tt(1)} solid var(--mantine-color-${n}-outline)`}:{background:"transparent",hover:fn(l.colors[f.color][f.shade],.05),color:`var(--mantine-color-${f.color}-${f.shade})`,border:`${tt(1)} solid var(--mantine-color-${f.color}-${f.shade})`}:{background:"transparent",hover:fn(n,.05),color:n,border:`${tt(1)} solid ${n}`};if(i==="subtle"){if(f.isThemeColor){if(f.shade===void 0)return{background:"transparent",hover:`var(--mantine-color-${n}-light-hover)`,color:`var(--mantine-color-${n}-light-color)`,border:`${tt(1)} solid transparent`};const h=l.colors[f.color][f.shade];return{background:"transparent",hover:fn(h,.12),color:`var(--mantine-color-${f.color}-${Math.min(f.shade,6)})`,border:`${tt(1)} solid transparent`}}return{background:"transparent",hover:fn(n,.12),color:n,border:`${tt(1)} solid transparent`}}return i==="transparent"?f.isThemeColor?f.shade===void 0?{background:"transparent",hover:"transparent",color:`var(--mantine-color-${n}-light-color)`,border:`${tt(1)} solid transparent`}:{background:"transparent",hover:"transparent",color:`var(--mantine-color-${f.color}-${Math.min(f.shade,6)})`,border:`${tt(1)} solid transparent`}:{background:"transparent",hover:"transparent",color:n,border:`${tt(1)} solid transparent`}:i==="white"?f.isThemeColor?f.shade===void 0?{background:"var(--mantine-color-white)",hover:Go(l.white,.01),color:`var(--mantine-color-${n}-filled)`,border:`${tt(1)} solid transparent`}:{background:"var(--mantine-color-white)",hover:Go(l.white,.01),color:`var(--mantine-color-${f.color}-${f.shade})`,border:`${tt(1)} solid transparent`}:{background:"var(--mantine-color-white)",hover:Go(l.white,.01),color:n,border:`${tt(1)} solid transparent`}:i==="gradient"?{background:Tf(o,l),hover:Tf(o,l),color:"var(--mantine-color-white)",border:"none"}:i==="default"?{background:"var(--mantine-color-default)",hover:"var(--mantine-color-default-hover)",color:"var(--mantine-color-default-color)",border:`${tt(1)} solid var(--mantine-color-default-border)`}:{}},S2={dark:["#C9C9C9","#b8b8b8","#828282","#696969","#424242","#3b3b3b","#2e2e2e","#242424","#1f1f1f","#141414"],gray:["#f8f9fa","#f1f3f5","#e9ecef","#dee2e6","#ced4da","#adb5bd","#868e96","#495057","#343a40","#212529"],red:["#fff5f5","#ffe3e3","#ffc9c9","#ffa8a8","#ff8787","#ff6b6b","#fa5252","#f03e3e","#e03131","#c92a2a"],pink:["#fff0f6","#ffdeeb","#fcc2d7","#faa2c1","#f783ac","#f06595","#e64980","#d6336c","#c2255c","#a61e4d"],grape:["#f8f0fc","#f3d9fa","#eebefa","#e599f7","#da77f2","#cc5de8","#be4bdb","#ae3ec9","#9c36b5","#862e9c"],violet:["#f3f0ff","#e5dbff","#d0bfff","#b197fc","#9775fa","#845ef7","#7950f2","#7048e8","#6741d9","#5f3dc4"],indigo:["#edf2ff","#dbe4ff","#bac8ff","#91a7ff","#748ffc","#5c7cfa","#4c6ef5","#4263eb","#3b5bdb","#364fc7"],blue:["#e7f5ff","#d0ebff","#a5d8ff","#74c0fc","#4dabf7","#339af0","#228be6","#1c7ed6","#1971c2","#1864ab"],cyan:["#e3fafc","#c5f6fa","#99e9f2","#66d9e8","#3bc9db","#22b8cf","#15aabf","#1098ad","#0c8599","#0b7285"],teal:["#e6fcf5","#c3fae8","#96f2d7","#63e6be","#38d9a9","#20c997","#12b886","#0ca678","#099268","#087f5b"],green:["#ebfbee","#d3f9d8","#b2f2bb","#8ce99a","#69db7c","#51cf66","#40c057","#37b24d","#2f9e44","#2b8a3e"],lime:["#f4fce3","#e9fac8","#d8f5a2","#c0eb75","#a9e34b","#94d82d","#82c91e","#74b816","#66a80f","#5c940d"],yellow:["#fff9db","#fff3bf","#ffec99","#ffe066","#ffd43b","#fcc419","#fab005","#f59f00","#f08c00","#e67700"],orange:["#fff4e6","#ffe8cc","#ffd8a8","#ffc078","#ffa94d","#ff922b","#fd7e14","#f76707","#e8590c","#d9480f"]},R0="-apple-system, BlinkMacSystemFont, Segoe UI, Roboto, Helvetica, Arial, sans-serif, Apple Color Emoji, Segoe UI Emoji",kf={scale:1,fontSmoothing:!0,focusRing:"auto",white:"#fff",black:"#000",colors:S2,primaryShade:{light:6,dark:8},primaryColor:"blue",variantColorResolver:b2,autoContrast:!1,luminanceThreshold:.3,fontFamily:R0,fontFamilyMonospace:"ui-monospace, SFMono-Regular, Menlo, Monaco, Consolas, Liberation Mono, Courier New, monospace",respectReducedMotion:!1,cursorType:"default",defaultGradient:{from:"blue",to:"cyan",deg:45},defaultRadius:"sm",activeClassName:"mantine-active",focusClassName:"",headings:{fontFamily:R0,fontWeight:"700",textWrap:"wrap",sizes:{h1:{fontSize:tt(34),lineHeight:"1.3"},h2:{fontSize:tt(26),lineHeight:"1.35"},h3:{fontSize:tt(22),lineHeight:"1.4"},h4:{fontSize:tt(18),lineHeight:"1.45"},h5:{fontSize:tt(16),lineHeight:"1.5"},h6:{fontSize:tt(14),lineHeight:"1.5"}}},fontSizes:{xs:tt(12),sm:tt(14),md:tt(16),lg:tt(18),xl:tt(20)},lineHeights:{xs:"1.4",sm:"1.45",md:"1.55",lg:"1.6",xl:"1.65"},radius:{xs:tt(2),sm:tt(4),md:tt(8),lg:tt(16),xl:tt(32)},spacing:{xs:tt(10),sm:tt(12),md:tt(16),lg:tt(20),xl:tt(32)},breakpoints:{xs:"36em",sm:"48em",md:"62em",lg:"75em",xl:"88em"},shadows:{xs:`0 ${tt(1)} ${tt(3)} rgba(0, 0, 0, 0.05), 0 ${tt(1)} ${tt(2)} rgba(0, 0, 0, 0.1)`,sm:`0 ${tt(1)} ${tt(3)} rgba(0, 0, 0, 0.05), rgba(0, 0, 0, 0.05) 0 ${tt(10)} ${tt(15)} ${tt(-5)}, rgba(0, 0, 0, 0.04) 0 ${tt(7)} ${tt(7)} ${tt(-5)}`,md:`0 ${tt(1)} ${tt(3)} rgba(0, 0, 0, 0.05), rgba(0, 0, 0, 0.05) 0 ${tt(20)} ${tt(25)} ${tt(-5)}, rgba(0, 0, 0, 0.04) 0 ${tt(10)} ${tt(10)} ${tt(-5)}`,lg:`0 ${tt(1)} ${tt(3)} rgba(0, 0, 0, 0.05), rgba(0, 0, 0, 0.05) 0 ${tt(28)} ${tt(23)} ${tt(-7)}, rgba(0, 0, 0, 0.04) 0 ${tt(12)} ${tt(12)} ${tt(-7)}`,xl:`0 ${tt(1)} ${tt(3)} rgba(0, 0, 0, 0.05), rgba(0, 0, 0, 0.05) 0 ${tt(36)} ${tt(28)} ${tt(-7)}, rgba(0, 0, 0, 0.04) 0 ${tt(17)} ${tt(17)} ${tt(-7)}`},other:{},components:{}};function z0(n){return n==="auto"||n==="dark"||n==="light"}function _2({key:n="mantine-color-scheme-value"}={}){let l;return{get:i=>{if(typeof window>"u")return i;try{const o=window.localStorage.getItem(n);return z0(o)?o:i}catch{return i}},set:i=>{try{window.localStorage.setItem(n,i)}catch(o){console.warn("[@mantine/core] Local storage color scheme manager was unable to save color scheme.",o)}},subscribe:i=>{l=o=>{o.storageArea===window.localStorage&&o.key===n&&z0(o.newValue)&&i(o.newValue)},window.addEventListener("storage",l)},unsubscribe:()=>{window.removeEventListener("storage",l)},clear:()=>{window.localStorage.removeItem(n)}}}const x2="[@mantine/core] MantineProvider: Invalid theme.primaryColor, it accepts only key of theme.colors, learn more – https://mantine.dev/theming/colors/#primary-color",D0="[@mantine/core] MantineProvider: Invalid theme.primaryShade, it accepts only 0-9 integers or an object { light: 0-9, dark: 0-9 }";function uf(n){return n<0||n>9?!1:parseInt(n.toString(),10)===n}function B0(n){if(!(n.primaryColor in n.colors))throw new Error(x2);if(typeof n.primaryShade=="object"&&(!uf(n.primaryShade.dark)||!uf(n.primaryShade.light)))throw new Error(D0);if(typeof n.primaryShade=="number"&&!uf(n.primaryShade))throw new Error(D0)}function T2(n,l){var o;if(!l)return B0(n),n;const i=qf(n,l);return l.fontFamily&&!((o=l.headings)!=null&&o.fontFamily)&&(i.headings.fontFamily=l.fontFamily),B0(i),i}const Zf=C.createContext(null),w2=()=>C.useContext(Zf)||kf;function pn(){const n=C.useContext(Zf);if(!n)throw new Error("@mantine/core: MantineProvider was not found in component tree, make sure you have it in your app");return n}function Gv({theme:n,children:l,inherit:i=!0}){const o=w2(),c=C.useMemo(()=>T2(i?o:kf,n),[n,o,i]);return g.jsx(Zf.Provider,{value:c,children:l})}Gv.displayName="@mantine/core/MantineThemeProvider";function E2(){const n=pn(),l=Gf(),i=hn(n.breakpoints).reduce((o,c)=>{const f=n.breakpoints[c].includes("px"),d=BS(n.breakpoints[c]),h=f?`${d-.1}px`:O0(d-.1),y=f?`${d}px`:O0(d);return`${o}@media (max-width: ${h}) {.mantine-visible-from-${c} {display: none !important;}}@media (min-width: ${y}) {.mantine-hidden-from-${c} {display: none !important;}}`},"");return g.jsx("style",{"data-mantine-styles":"classes",nonce:l==null?void 0:l(),dangerouslySetInnerHTML:{__html:i}})}function cf(n){return Object.entries(n).map(([l,i])=>`${l}: ${i};`).join("")}function Pr(n,l){return(Array.isArray(n)?n:[n]).reduce((o,c)=>`${c}{${o}}`,l)}function C2(n,l){const i=cf(n.variables),o=i?Pr(l,i):"",c=cf(n.dark),f=cf(n.light),d=c?Pr(l===":host"?`${l}([data-mantine-color-scheme="dark"])`:`${l}[data-mantine-color-scheme="dark"]`,c):"",h=f?Pr(l===":host"?`${l}([data-mantine-color-scheme="light"])`:`${l}[data-mantine-color-scheme="light"]`,f):"";return`${o}${d}${h}`}function Vv({color:n,theme:l,autoContrast:i}){return(typeof i=="boolean"?i:l.autoContrast)&&mi({color:n||l.primaryColor,theme:l}).isLight?"var(--mantine-color-black)":"var(--mantine-color-white)"}function U0(n,l){return Vv({color:n.colors[n.primaryColor][oi(n,l)],theme:n,autoContrast:null})}function Vo({theme:n,color:l,colorScheme:i,name:o=l,withColorValues:c=!0}){if(!n.colors[l])return{};if(i==="light"){const h=oi(n,"light"),y={[`--mantine-color-${o}-text`]:`var(--mantine-color-${o}-filled)`,[`--mantine-color-${o}-filled`]:`var(--mantine-color-${o}-${h})`,[`--mantine-color-${o}-filled-hover`]:`var(--mantine-color-${o}-${h===9?8:h+1})`,[`--mantine-color-${o}-light`]:Dl(n.colors[l][h],.1),[`--mantine-color-${o}-light-hover`]:Dl(n.colors[l][h],.12),[`--mantine-color-${o}-light-color`]:`var(--mantine-color-${o}-${h})`,[`--mantine-color-${o}-outline`]:`var(--mantine-color-${o}-${h})`,[`--mantine-color-${o}-outline-hover`]:Dl(n.colors[l][h],.05)};return c?{[`--mantine-color-${o}-0`]:n.colors[l][0],[`--mantine-color-${o}-1`]:n.colors[l][1],[`--mantine-color-${o}-2`]:n.colors[l][2],[`--mantine-color-${o}-3`]:n.colors[l][3],[`--mantine-color-${o}-4`]:n.colors[l][4],[`--mantine-color-${o}-5`]:n.colors[l][5],[`--mantine-color-${o}-6`]:n.colors[l][6],[`--mantine-color-${o}-7`]:n.colors[l][7],[`--mantine-color-${o}-8`]:n.colors[l][8],[`--mantine-color-${o}-9`]:n.colors[l][9],...y}:y}const f=oi(n,"dark"),d={[`--mantine-color-${o}-text`]:`var(--mantine-color-${o}-4)`,[`--mantine-color-${o}-filled`]:`var(--mantine-color-${o}-${f})`,[`--mantine-color-${o}-filled-hover`]:`var(--mantine-color-${o}-${f===9?8:f+1})`,[`--mantine-color-${o}-light`]:Dl(n.colors[l][Math.max(0,f-2)],.15),[`--mantine-color-${o}-light-hover`]:Dl(n.colors[l][Math.max(0,f-2)],.2),[`--mantine-color-${o}-light-color`]:`var(--mantine-color-${o}-${Math.max(f-5,0)})`,[`--mantine-color-${o}-outline`]:`var(--mantine-color-${o}-${Math.max(f-4,0)})`,[`--mantine-color-${o}-outline-hover`]:Dl(n.colors[l][Math.max(f-4,0)],.05)};return c?{[`--mantine-color-${o}-0`]:n.colors[l][0],[`--mantine-color-${o}-1`]:n.colors[l][1],[`--mantine-color-${o}-2`]:n.colors[l][2],[`--mantine-color-${o}-3`]:n.colors[l][3],[`--mantine-color-${o}-4`]:n.colors[l][4],[`--mantine-color-${o}-5`]:n.colors[l][5],[`--mantine-color-${o}-6`]:n.colors[l][6],[`--mantine-color-${o}-7`]:n.colors[l][7],[`--mantine-color-${o}-8`]:n.colors[l][8],[`--mantine-color-${o}-9`]:n.colors[l][9],...d}:d}function A2(n){return!!n&&typeof n=="object"&&"mantine-virtual-color"in n}function Bl(n,l,i){hn(l).forEach(o=>Object.assign(n,{[`--mantine-${i}-${o}`]:l[o]}))}const kv=n=>{const l=oi(n,"light"),i=n.defaultRadius in n.radius?n.radius[n.defaultRadius]:tt(n.defaultRadius),o={variables:{"--mantine-scale":n.scale.toString(),"--mantine-cursor-type":n.cursorType,"--mantine-color-scheme":"light dark","--mantine-webkit-font-smoothing":n.fontSmoothing?"antialiased":"unset","--mantine-moz-font-smoothing":n.fontSmoothing?"grayscale":"unset","--mantine-color-white":n.white,"--mantine-color-black":n.black,"--mantine-line-height":n.lineHeights.md,"--mantine-font-family":n.fontFamily,"--mantine-font-family-monospace":n.fontFamilyMonospace,"--mantine-font-family-headings":n.headings.fontFamily,"--mantine-heading-font-weight":n.headings.fontWeight,"--mantine-heading-text-wrap":n.headings.textWrap,"--mantine-radius-default":i,"--mantine-primary-color-filled":`var(--mantine-color-${n.primaryColor}-filled)`,"--mantine-primary-color-filled-hover":`var(--mantine-color-${n.primaryColor}-filled-hover)`,"--mantine-primary-color-light":`var(--mantine-color-${n.primaryColor}-light)`,"--mantine-primary-color-light-hover":`var(--mantine-color-${n.primaryColor}-light-hover)`,"--mantine-primary-color-light-color":`var(--mantine-color-${n.primaryColor}-light-color)`},light:{"--mantine-primary-color-contrast":U0(n,"light"),"--mantine-color-bright":"var(--mantine-color-black)","--mantine-color-text":n.black,"--mantine-color-body":n.white,"--mantine-color-error":"var(--mantine-color-red-6)","--mantine-color-placeholder":"var(--mantine-color-gray-5)","--mantine-color-anchor":`var(--mantine-color-${n.primaryColor}-${l})`,"--mantine-color-default":"var(--mantine-color-white)","--mantine-color-default-hover":"var(--mantine-color-gray-0)","--mantine-color-default-color":"var(--mantine-color-black)","--mantine-color-default-border":"var(--mantine-color-gray-4)","--mantine-color-dimmed":"var(--mantine-color-gray-6)"},dark:{"--mantine-primary-color-contrast":U0(n,"dark"),"--mantine-color-bright":"var(--mantine-color-white)","--mantine-color-text":"var(--mantine-color-dark-0)","--mantine-color-body":"var(--mantine-color-dark-7)","--mantine-color-error":"var(--mantine-color-red-8)","--mantine-color-placeholder":"var(--mantine-color-dark-3)","--mantine-color-anchor":`var(--mantine-color-${n.primaryColor}-4)`,"--mantine-color-default":"var(--mantine-color-dark-6)","--mantine-color-default-hover":"var(--mantine-color-dark-5)","--mantine-color-default-color":"var(--mantine-color-white)","--mantine-color-default-border":"var(--mantine-color-dark-4)","--mantine-color-dimmed":"var(--mantine-color-dark-2)"}};Bl(o.variables,n.breakpoints,"breakpoint"),Bl(o.variables,n.spacing,"spacing"),Bl(o.variables,n.fontSizes,"font-size"),Bl(o.variables,n.lineHeights,"line-height"),Bl(o.variables,n.shadows,"shadow"),Bl(o.variables,n.radius,"radius"),n.colors[n.primaryColor].forEach((f,d)=>{o.variables[`--mantine-primary-color-${d}`]=`var(--mantine-color-${n.primaryColor}-${d})`}),hn(n.colors).forEach(f=>{const d=n.colors[f];if(A2(d)){Object.assign(o.light,Vo({theme:n,name:d.name,color:d.light,colorScheme:"light",withColorValues:!0})),Object.assign(o.dark,Vo({theme:n,name:d.name,color:d.dark,colorScheme:"dark",withColorValues:!0}));return}d.forEach((h,y)=>{o.variables[`--mantine-color-${f}-${y}`]=h}),Object.assign(o.light,Vo({theme:n,color:f,colorScheme:"light",withColorValues:!1})),Object.assign(o.dark,Vo({theme:n,color:f,colorScheme:"dark",withColorValues:!1}))});const c=n.headings.sizes;return hn(c).forEach(f=>{o.variables[`--mantine-${f}-font-size`]=c[f].fontSize,o.variables[`--mantine-${f}-line-height`]=c[f].lineHeight,o.variables[`--mantine-${f}-font-weight`]=c[f].fontWeight||n.headings.fontWeight}),o};function j2({theme:n,generator:l}){const i=kv(n),o=l==null?void 0:l(n);return o?qf(i,o):i}const sf=kv(kf);function N2(n){const l={variables:{},light:{},dark:{}};return hn(n.variables).forEach(i=>{sf.variables[i]!==n.variables[i]&&(l.variables[i]=n.variables[i])}),hn(n.light).forEach(i=>{sf.light[i]!==n.light[i]&&(l.light[i]=n.light[i])}),hn(n.dark).forEach(i=>{sf.dark[i]!==n.dark[i]&&(l.dark[i]=n.dark[i])}),l}function O2(n){return`
  ${n}[data-mantine-color-scheme="dark"] { --mantine-color-scheme: dark; }
  ${n}[data-mantine-color-scheme="light"] { --mantine-color-scheme: light; }
`}function Zv({cssVariablesSelector:n,deduplicateCssVariables:l}){const i=pn(),o=Gf(),c=o2(),f=j2({theme:i,generator:c}),d=n===":root"&&l,h=d?N2(f):f,y=C2(h,n);return y?g.jsx("style",{"data-mantine-styles":!0,nonce:o==null?void 0:o(),dangerouslySetInnerHTML:{__html:`${y}${d?"":O2(n)}`}}):null}Zv.displayName="@mantine/CssVariables";function M2(){const n=console.error;console.error=(...l)=>{l.length>1&&typeof l[0]=="string"&&l[0].toLowerCase().includes("extra attributes from the server")&&typeof l[1]=="string"&&l[1].toLowerCase().includes("data-mantine-color-scheme")||n(...l)}}function Ul(n,l){var c,f;const i=typeof window<"u"&&"matchMedia"in window&&((c=window.matchMedia("(prefers-color-scheme: dark)"))==null?void 0:c.matches),o=n!=="auto"?n:i?"dark":"light";(f=l())==null||f.setAttribute("data-mantine-color-scheme",o)}function R2({manager:n,defaultColorScheme:l,getRootElement:i,forceColorScheme:o}){const c=C.useRef(null),[f,d]=C.useState(()=>n.get(l)),h=o||f,y=C.useCallback(v=>{o||(Ul(v,i),d(v),n.set(v))},[n.set,h,o]),m=C.useCallback(()=>{d(l),Ul(l,i),n.clear()},[n.clear,l]);return C.useEffect(()=>(n.subscribe(y),n.unsubscribe),[n.subscribe,n.unsubscribe]),hi(()=>{Ul(n.get(l),i)},[]),C.useEffect(()=>{var S;if(o)return Ul(o,i),()=>{};o===void 0&&Ul(f,i),typeof window<"u"&&"matchMedia"in window&&(c.current=window.matchMedia("(prefers-color-scheme: dark)"));const v=_=>{f==="auto"&&Ul(_.matches?"dark":"light",i)};return(S=c.current)==null||S.addEventListener("change",v),()=>{var _;return(_=c.current)==null?void 0:_.removeEventListener("change",v)}},[f,o]),{colorScheme:h,setColorScheme:y,clearColorScheme:m}}function z2({respectReducedMotion:n,getRootElement:l}){hi(()=>{var i;n&&((i=l())==null||i.setAttribute("data-respect-reduced-motion","true"))},[n])}M2();function Qv({theme:n,children:l,getStyleNonce:i,withStaticClasses:o=!0,withGlobalClasses:c=!0,deduplicateCssVariables:f=!0,withCssVariables:d=!0,cssVariablesSelector:h=":root",classNamesPrefix:y="mantine",colorSchemeManager:m=_2(),defaultColorScheme:v="light",getRootElement:S=()=>document.documentElement,cssVariablesResolver:_,forceColorScheme:x,stylesTransform:w,env:N}){const{colorScheme:B,setColorScheme:O,clearColorScheme:j}=R2({defaultColorScheme:v,forceColorScheme:x,manager:m,getRootElement:S});return z2({respectReducedMotion:(n==null?void 0:n.respectReducedMotion)||!1,getRootElement:S}),g.jsx(qv.Provider,{value:{colorScheme:B,setColorScheme:O,clearColorScheme:j,getRootElement:S,classNamesPrefix:y,getStyleNonce:i,cssVariablesResolver:_,cssVariablesSelector:h,withStaticClasses:o,stylesTransform:w,env:N},children:g.jsxs(Gv,{theme:n,children:[d&&g.jsx(Zv,{cssVariablesSelector:h,deduplicateCssVariables:f}),c&&g.jsx(E2,{}),l]})})}Qv.displayName="@mantine/core/MantineProvider";function Kv({classNames:n,styles:l,props:i,stylesCtx:o}){const c=pn();return{resolvedClassNames:gu({theme:c,classNames:n,props:i,stylesCtx:o||void 0}),resolvedStyles:ru({theme:c,styles:l,props:i,stylesCtx:o||void 0})}}const D2={always:"mantine-focus-always",auto:"mantine-focus-auto",never:"mantine-focus-never"};function B2({theme:n,options:l,unstyled:i}){return Ce((l==null?void 0:l.focusable)&&!i&&(n.focusClassName||D2[n.focusRing]),(l==null?void 0:l.active)&&!i&&n.activeClassName)}function U2({selector:n,stylesCtx:l,options:i,props:o,theme:c}){return gu({theme:c,classNames:i==null?void 0:i.classNames,props:(i==null?void 0:i.props)||o,stylesCtx:l})[n]}function H0({selector:n,stylesCtx:l,theme:i,classNames:o,props:c}){return gu({theme:i,classNames:o,props:c,stylesCtx:l})[n]}function H2({rootSelector:n,selector:l,className:i}){return n===l?i:void 0}function $2({selector:n,classes:l,unstyled:i}){return i?void 0:l[n]}function L2({themeName:n,classNamesPrefix:l,selector:i,withStaticClass:o}){return o===!1?[]:n.map(c=>`${l}-${c}-${i}`)}function Y2({themeName:n,theme:l,selector:i,props:o,stylesCtx:c}){return n.map(f=>{var d,h;return(h=gu({theme:l,classNames:(d=l.components[f])==null?void 0:d.classNames,props:o,stylesCtx:c}))==null?void 0:h[i]})}function q2({options:n,classes:l,selector:i,unstyled:o}){return n!=null&&n.variant&&!o?l[`${i}--${n.variant}`]:void 0}function X2({theme:n,options:l,themeName:i,selector:o,classNamesPrefix:c,classNames:f,classes:d,unstyled:h,className:y,rootSelector:m,props:v,stylesCtx:S,withStaticClasses:_,headless:x,transformedStyles:w}){return Ce(B2({theme:n,options:l,unstyled:h||x}),Y2({theme:n,themeName:i,selector:o,props:v,stylesCtx:S}),q2({options:l,classes:d,selector:o,unstyled:h}),H0({selector:o,stylesCtx:S,theme:n,classNames:f,props:v}),H0({selector:o,stylesCtx:S,theme:n,classNames:w,props:v}),U2({selector:o,stylesCtx:S,options:l,props:v,theme:n}),H2({rootSelector:m,selector:o,className:y}),$2({selector:o,classes:d,unstyled:h||x}),_&&!x&&L2({themeName:i,classNamesPrefix:c,selector:o,withStaticClass:l==null?void 0:l.withStaticClass}),l==null?void 0:l.className)}function G2({theme:n,themeName:l,props:i,stylesCtx:o,selector:c}){return l.map(f=>{var d;return ru({theme:n,styles:(d=n.components[f])==null?void 0:d.styles,props:i,stylesCtx:o})[c]}).reduce((f,d)=>({...f,...d}),{})}function wf({style:n,theme:l}){return Array.isArray(n)?[...n].reduce((i,o)=>({...i,...wf({style:o,theme:l})}),{}):typeof n=="function"?n(l):n??{}}function V2(n){return n.reduce((l,i)=>(i&&Object.keys(i).forEach(o=>{l[o]={...l[o],...yu(i[o])}}),l),{})}function k2({vars:n,varsResolver:l,theme:i,props:o,stylesCtx:c,selector:f,themeName:d,headless:h}){var y;return(y=V2([h?{}:l==null?void 0:l(i,o,c),...d.map(m=>{var v,S,_;return(_=(S=(v=i.components)==null?void 0:v[m])==null?void 0:S.vars)==null?void 0:_.call(S,i,o,c)}),n==null?void 0:n(i,o,c)]))==null?void 0:y[f]}function Z2({theme:n,themeName:l,selector:i,options:o,props:c,stylesCtx:f,rootSelector:d,styles:h,style:y,vars:m,varsResolver:v,headless:S,withStylesTransform:_}){return{...!_&&G2({theme:n,themeName:l,props:c,stylesCtx:f,selector:i}),...!_&&ru({theme:n,styles:h,props:c,stylesCtx:f})[i],...!_&&ru({theme:n,styles:o==null?void 0:o.styles,props:(o==null?void 0:o.props)||c,stylesCtx:f})[i],...k2({theme:n,props:c,stylesCtx:f,vars:m,varsResolver:v,selector:i,themeName:l,headless:S}),...d===i?wf({style:y,theme:n}):null,...wf({style:o==null?void 0:o.style,theme:n})}}function Q2({props:n,stylesCtx:l,themeName:i}){var d;const o=pn(),c=(d=d2())==null?void 0:d();return{getTransformedStyles:h=>c?[...h.map(m=>c(m,{props:n,theme:o,ctx:l})),...i.map(m=>{var v;return c((v=o.components[m])==null?void 0:v.styles,{props:n,theme:o,ctx:l})})].filter(Boolean):[],withStylesTransform:!!c}}function jt({name:n,classes:l,props:i,stylesCtx:o,className:c,style:f,rootSelector:d="root",unstyled:h,classNames:y,styles:m,vars:v,varsResolver:S}){const _=pn(),x=u2(),w=c2(),N=s2(),B=(Array.isArray(n)?n:[n]).filter(z=>z),{withStylesTransform:O,getTransformedStyles:j}=Q2({props:i,stylesCtx:o,themeName:B});return(z,L)=>({className:X2({theme:_,options:L,themeName:B,selector:z,classNamesPrefix:x,classNames:y,classes:l,unstyled:h,className:c,rootSelector:d,props:i,stylesCtx:o,withStaticClasses:w,headless:N,transformedStyles:j([L==null?void 0:L.styles,m])}),style:Z2({theme:_,themeName:B,selector:z,options:L,props:i,stylesCtx:o,rootSelector:d,styles:m,style:f,vars:v,varsResolver:S,headless:N,withStylesTransform:O})})}function K2(n,l){return typeof n=="boolean"?n:l.autoContrast}function st(n,l,i){var d;const o=pn(),c=(d=o.components[n])==null?void 0:d.defaultProps,f=typeof c=="function"?c(o):c;return{...l,...f,...yu(i)}}function ff(n){return hn(n).reduce((l,i)=>n[i]!==void 0?`${l}${zS(i)}:${n[i]};`:l,"").trim()}function J2({selector:n,styles:l,media:i,container:o}){const c=l?ff(l):"",f=Array.isArray(i)?i.map(h=>`@media${h.query}{${n}{${ff(h.styles)}}}`):[],d=Array.isArray(o)?o.map(h=>`@container ${h.query}{${n}{${ff(h.styles)}}}`):[];return`${c?`${n}{${c}}`:""}${f.join("")}${d.join("")}`.trim()}function Jv(n){const l=Gf();return g.jsx("style",{"data-mantine-styles":"inline",nonce:l==null?void 0:l(),dangerouslySetInnerHTML:{__html:J2(n)}})}function bu(n){const{m:l,mx:i,my:o,mt:c,mb:f,ml:d,mr:h,me:y,ms:m,p:v,px:S,py:_,pt:x,pb:w,pl:N,pr:B,pe:O,ps:j,bd:z,bg:L,c:$,opacity:V,ff:Z,fz:W,fw:it,lts:ht,ta:pt,lh:at,fs:lt,tt:R,td:Q,w:E,miw:Y,maw:k,h:et,mih:T,mah:X,bgsz:I,bgp:P,bgr:nt,bga:ut,pos:rt,top:bt,left:vt,bottom:Ut,right:ae,inset:pe,display:$e,flex:ln,hiddenFrom:Sn,visibleFrom:pa,lightHidden:ce,darkHidden:Ht,sx:Wl,...Ei}=n;return{styleProps:yu({m:l,mx:i,my:o,mt:c,mb:f,ml:d,mr:h,me:y,ms:m,p:v,px:S,py:_,pt:x,pb:w,pl:N,pr:B,pe:O,ps:j,bd:z,bg:L,c:$,opacity:V,ff:Z,fz:W,fw:it,lts:ht,ta:pt,lh:at,fs:lt,tt:R,td:Q,w:E,miw:Y,maw:k,h:et,mih:T,mah:X,bgsz:I,bgp:P,bgr:nt,bga:ut,pos:rt,top:bt,left:vt,bottom:Ut,right:ae,inset:pe,display:$e,flex:ln,hiddenFrom:Sn,visibleFrom:pa,lightHidden:ce,darkHidden:Ht,sx:Wl}),rest:Ei}}const W2={m:{type:"spacing",property:"margin"},mt:{type:"spacing",property:"marginTop"},mb:{type:"spacing",property:"marginBottom"},ml:{type:"spacing",property:"marginLeft"},mr:{type:"spacing",property:"marginRight"},ms:{type:"spacing",property:"marginInlineStart"},me:{type:"spacing",property:"marginInlineEnd"},mx:{type:"spacing",property:"marginInline"},my:{type:"spacing",property:"marginBlock"},p:{type:"spacing",property:"padding"},pt:{type:"spacing",property:"paddingTop"},pb:{type:"spacing",property:"paddingBottom"},pl:{type:"spacing",property:"paddingLeft"},pr:{type:"spacing",property:"paddingRight"},ps:{type:"spacing",property:"paddingInlineStart"},pe:{type:"spacing",property:"paddingInlineEnd"},px:{type:"spacing",property:"paddingInline"},py:{type:"spacing",property:"paddingBlock"},bd:{type:"border",property:"border"},bg:{type:"color",property:"background"},c:{type:"textColor",property:"color"},opacity:{type:"identity",property:"opacity"},ff:{type:"fontFamily",property:"fontFamily"},fz:{type:"fontSize",property:"fontSize"},fw:{type:"identity",property:"fontWeight"},lts:{type:"size",property:"letterSpacing"},ta:{type:"identity",property:"textAlign"},lh:{type:"lineHeight",property:"lineHeight"},fs:{type:"identity",property:"fontStyle"},tt:{type:"identity",property:"textTransform"},td:{type:"identity",property:"textDecoration"},w:{type:"spacing",property:"width"},miw:{type:"spacing",property:"minWidth"},maw:{type:"spacing",property:"maxWidth"},h:{type:"spacing",property:"height"},mih:{type:"spacing",property:"minHeight"},mah:{type:"spacing",property:"maxHeight"},bgsz:{type:"size",property:"backgroundSize"},bgp:{type:"identity",property:"backgroundPosition"},bgr:{type:"identity",property:"backgroundRepeat"},bga:{type:"identity",property:"backgroundAttachment"},pos:{type:"identity",property:"position"},top:{type:"size",property:"top"},left:{type:"size",property:"left"},bottom:{type:"size",property:"bottom"},right:{type:"size",property:"right"},inset:{type:"size",property:"inset"},display:{type:"identity",property:"display"},flex:{type:"identity",property:"flex"}};function Qf(n,l){const i=mi({color:n,theme:l});return i.color==="dimmed"?"var(--mantine-color-dimmed)":i.color==="bright"?"var(--mantine-color-bright)":i.variable?`var(${i.variable})`:i.color}function P2(n,l){const i=mi({color:n,theme:l});return i.isThemeColor&&i.shade===void 0?`var(--mantine-color-${i.color}-text)`:Qf(n,l)}function F2(n,l){if(typeof n=="number")return tt(n);if(typeof n=="string"){const[i,o,...c]=n.split(" ").filter(d=>d.trim()!=="");let f=`${tt(i)}`;return o&&(f+=` ${o}`),c.length>0&&(f+=` ${Qf(c.join(" "),l)}`),f.trim()}return n}const $0={text:"var(--mantine-font-family)",mono:"var(--mantine-font-family-monospace)",monospace:"var(--mantine-font-family-monospace)",heading:"var(--mantine-font-family-headings)",headings:"var(--mantine-font-family-headings)"};function I2(n){return typeof n=="string"&&n in $0?$0[n]:n}const t_=["h1","h2","h3","h4","h5","h6"];function e_(n,l){return typeof n=="string"&&n in l.fontSizes?`var(--mantine-font-size-${n})`:typeof n=="string"&&t_.includes(n)?`var(--mantine-${n}-font-size)`:typeof n=="number"||typeof n=="string"?tt(n):n}function n_(n){return n}const a_=["h1","h2","h3","h4","h5","h6"];function l_(n,l){return typeof n=="string"&&n in l.lineHeights?`var(--mantine-line-height-${n})`:typeof n=="string"&&a_.includes(n)?`var(--mantine-${n}-line-height)`:n}function r_(n){return typeof n=="number"?tt(n):n}function i_(n,l){if(typeof n=="number")return tt(n);if(typeof n=="string"){const i=n.replace("-","");if(!(i in l.spacing))return tt(n);const o=`--mantine-spacing-${i}`;return n.startsWith("-")?`calc(var(${o}) * -1)`:`var(${o})`}return n}const df={color:Qf,textColor:P2,fontSize:e_,spacing:i_,identity:n_,size:r_,lineHeight:l_,fontFamily:I2,border:F2};function L0(n){return n.replace("(min-width: ","").replace("em)","")}function o_({media:n,...l}){const o=Object.keys(n).sort((c,f)=>Number(L0(c))-Number(L0(f))).map(c=>({query:c,styles:n[c]}));return{...l,media:o}}function u_(n){if(typeof n!="object"||n===null)return!1;const l=Object.keys(n);return!(l.length===1&&l[0]==="base")}function c_(n){return typeof n=="object"&&n!==null?"base"in n?n.base:void 0:n}function s_(n){return typeof n=="object"&&n!==null?hn(n).filter(l=>l!=="base"):[]}function f_(n,l){return typeof n=="object"&&n!==null&&l in n?n[l]:n}function Wv({styleProps:n,data:l,theme:i}){return o_(hn(n).reduce((o,c)=>{if(c==="hiddenFrom"||c==="visibleFrom"||c==="sx")return o;const f=l[c],d=Array.isArray(f.property)?f.property:[f.property],h=c_(n[c]);if(!u_(n[c]))return d.forEach(m=>{o.inlineStyles[m]=df[f.type](h,i)}),o;o.hasResponsiveStyles=!0;const y=s_(n[c]);return d.forEach(m=>{h&&(o.styles[m]=df[f.type](h,i)),y.forEach(v=>{const S=`(min-width: ${i.breakpoints[v]})`;o.media[S]={...o.media[S],[m]:df[f.type](f_(n[c],v),i)}})}),o},{hasResponsiveStyles:!1,styles:{},inlineStyles:{},media:{}}))}function Pv(){return`__m__-${C.useId().replace(/:/g,"")}`}function Fv(n){return n.startsWith("data-")?n:`data-${n}`}function d_(n){return Object.keys(n).reduce((l,i)=>{const o=n[i];return o===void 0||o===""||o===!1||o===null||(l[Fv(i)]=n[i]),l},{})}function Iv(n){return n?typeof n=="string"?{[Fv(n)]:!0}:Array.isArray(n)?[...n].reduce((l,i)=>({...l,...Iv(i)}),{}):d_(n):null}function Ef(n,l){return Array.isArray(n)?[...n].reduce((i,o)=>({...i,...Ef(o,l)}),{}):typeof n=="function"?n(l):n??{}}function h_({theme:n,style:l,vars:i,styleProps:o}){const c=Ef(l,n),f=Ef(i,n);return{...c,...f,...o}}const tp=C.forwardRef(({component:n,style:l,__vars:i,className:o,variant:c,mod:f,size:d,hiddenFrom:h,visibleFrom:y,lightHidden:m,darkHidden:v,renderRoot:S,__size:_,...x},w)=>{var W;const N=pn(),B=n||"div",{styleProps:O,rest:j}=bu(x),z=f2(),L=(W=z==null?void 0:z())==null?void 0:W(O.sx),$=Pv(),V=Wv({styleProps:O,theme:N,data:W2}),Z={ref:w,style:h_({theme:N,style:l,vars:i,styleProps:V.inlineStyles}),className:Ce(o,L,{[$]:V.hasResponsiveStyles,"mantine-light-hidden":m,"mantine-dark-hidden":v,[`mantine-hidden-from-${h}`]:h,[`mantine-visible-from-${y}`]:y}),"data-variant":c,"data-size":Mv(d)?void 0:d||void 0,size:_,...Iv(f),...j};return g.jsxs(g.Fragment,{children:[V.hasResponsiveStyles&&g.jsx(Jv,{selector:`.${$}`,styles:V.styles,media:V.media}),typeof S=="function"?S(Z):g.jsx(B,{...Z})]})});tp.displayName="@mantine/core/Box";const ft=tp;function ep(n){return n}function _t(n){const l=C.forwardRef(n);return l.extend=ep,l.withProps=i=>{const o=C.forwardRef((c,f)=>g.jsx(l,{...i,...c,ref:f}));return o.extend=l.extend,o.displayName=`WithProps(${l.displayName})`,o},l}function gn(n){const l=C.forwardRef(n);return l.withProps=i=>{const o=C.forwardRef((c,f)=>g.jsx(l,{...i,...c,ref:f}));return o.extend=l.extend,o.displayName=`WithProps(${l.displayName})`,o},l.extend=ep,l}const m_=C.createContext({dir:"ltr",toggleDirection:()=>{},setDirection:()=>{}});function np(){return C.useContext(m_)}var ap=_v();const y_=Sv(ap);function v_(n){return C.useMemo(()=>n.every(l=>l==null)?null:l=>{n.forEach(i=>{typeof i=="function"?i(l):i!=null&&(i.current=l)})},n)}const[p_,Je]=Xa("ScrollArea.Root component was not found in tree");function Yl(n,l){const i=Ha(l);hi(()=>{let o=0;if(n){const c=new ResizeObserver(()=>{cancelAnimationFrame(o),o=window.requestAnimationFrame(i)});return c.observe(n),()=>{window.cancelAnimationFrame(o),c.unobserve(n)}}},[n,i])}const g_=C.forwardRef((n,l)=>{const{style:i,...o}=n,c=Je(),[f,d]=C.useState(0),[h,y]=C.useState(0),m=!!(f&&h);return Yl(c.scrollbarX,()=>{var S;const v=((S=c.scrollbarX)==null?void 0:S.offsetHeight)||0;c.onCornerHeightChange(v),y(v)}),Yl(c.scrollbarY,()=>{var S;const v=((S=c.scrollbarY)==null?void 0:S.offsetWidth)||0;c.onCornerWidthChange(v),d(v)}),m?g.jsx("div",{...o,ref:l,style:{...i,width:f,height:h}}):null}),b_=C.forwardRef((n,l)=>{const i=Je(),o=!!(i.scrollbarX&&i.scrollbarY);return i.type!=="scroll"&&o?g.jsx(g_,{...n,ref:l}):null}),S_={scrollHideDelay:1e3,type:"hover"},lp=C.forwardRef((n,l)=>{const i=st("ScrollAreaRoot",S_,n),{type:o,scrollHideDelay:c,scrollbars:f,...d}=i,[h,y]=C.useState(null),[m,v]=C.useState(null),[S,_]=C.useState(null),[x,w]=C.useState(null),[N,B]=C.useState(null),[O,j]=C.useState(0),[z,L]=C.useState(0),[$,V]=C.useState(!1),[Z,W]=C.useState(!1),it=Ga(l,ht=>y(ht));return g.jsx(p_,{value:{type:o,scrollHideDelay:c,scrollArea:h,viewport:m,onViewportChange:v,content:S,onContentChange:_,scrollbarX:x,onScrollbarXChange:w,scrollbarXEnabled:$,onScrollbarXEnabledChange:V,scrollbarY:N,onScrollbarYChange:B,scrollbarYEnabled:Z,onScrollbarYEnabledChange:W,onCornerWidthChange:j,onCornerHeightChange:L},children:g.jsx(ft,{...d,ref:it,__vars:{"--sa-corner-width":f!=="xy"?"0px":`${O}px`,"--sa-corner-height":f!=="xy"?"0px":`${z}px`}})})});lp.displayName="@mantine/core/ScrollAreaRoot";function rp(n,l){const i=n/l;return Number.isNaN(i)?0:i}function Su(n){const l=rp(n.viewport,n.content),i=n.scrollbar.paddingStart+n.scrollbar.paddingEnd,o=(n.scrollbar.size-i)*l;return Math.max(o,18)}function ip(n,l){return i=>{if(n[0]===n[1]||l[0]===l[1])return l[0];const o=(l[1]-l[0])/(n[1]-n[0]);return l[0]+o*(i-n[0])}}function __(n,[l,i]){return Math.min(i,Math.max(l,n))}function Y0(n,l,i="ltr"){const o=Su(l),c=l.scrollbar.paddingStart+l.scrollbar.paddingEnd,f=l.scrollbar.size-c,d=l.content-l.viewport,h=f-o,y=i==="ltr"?[0,d]:[d*-1,0],m=__(n,y);return ip([0,d],[0,h])(m)}function x_(n,l,i,o="ltr"){const c=Su(i),f=c/2,d=l||f,h=c-d,y=i.scrollbar.paddingStart+d,m=i.scrollbar.size-i.scrollbar.paddingEnd-h,v=i.content-i.viewport,S=o==="ltr"?[0,v]:[v*-1,0];return ip([y,m],S)(n)}function op(n,l){return n>0&&n<l}function iu(n){return n?parseInt(n,10):0}function La(n,l,{checkForDefaultPrevented:i=!0}={}){return o=>{n==null||n(o),(i===!1||!o.defaultPrevented)&&(l==null||l(o))}}const[T_,up]=Xa("ScrollAreaScrollbar was not found in tree"),cp=C.forwardRef((n,l)=>{const{sizes:i,hasThumb:o,onThumbChange:c,onThumbPointerUp:f,onThumbPointerDown:d,onThumbPositionChange:h,onDragScroll:y,onWheelScroll:m,onResize:v,...S}=n,_=Je(),[x,w]=C.useState(null),N=Ga(l,W=>w(W)),B=C.useRef(null),O=C.useRef(""),{viewport:j}=_,z=i.content-i.viewport,L=Ha(m),$=Ha(h),V=vu(v,10),Z=W=>{if(B.current){const it=W.clientX-B.current.left,ht=W.clientY-B.current.top;y({x:it,y:ht})}};return C.useEffect(()=>{const W=it=>{const ht=it.target;(x==null?void 0:x.contains(ht))&&L(it,z)};return document.addEventListener("wheel",W,{passive:!1}),()=>document.removeEventListener("wheel",W,{passive:!1})},[j,x,z,L]),C.useEffect($,[i,$]),Yl(x,V),Yl(_.content,V),g.jsx(T_,{value:{scrollbar:x,hasThumb:o,onThumbChange:Ha(c),onThumbPointerUp:Ha(f),onThumbPositionChange:$,onThumbPointerDown:Ha(d)},children:g.jsx("div",{...S,ref:N,"data-mantine-scrollbar":!0,style:{position:"absolute",...S.style},onPointerDown:La(n.onPointerDown,W=>{W.preventDefault(),W.button===0&&(W.target.setPointerCapture(W.pointerId),B.current=x.getBoundingClientRect(),O.current=document.body.style.webkitUserSelect,document.body.style.webkitUserSelect="none",Z(W))}),onPointerMove:La(n.onPointerMove,Z),onPointerUp:La(n.onPointerUp,W=>{const it=W.target;it.hasPointerCapture(W.pointerId)&&(W.preventDefault(),it.releasePointerCapture(W.pointerId))}),onLostPointerCapture:()=>{document.body.style.webkitUserSelect=O.current,B.current=null}})})}),sp=C.forwardRef((n,l)=>{const{sizes:i,onSizesChange:o,style:c,...f}=n,d=Je(),[h,y]=C.useState(),m=C.useRef(null),v=Ga(l,m,d.onScrollbarXChange);return C.useEffect(()=>{m.current&&y(getComputedStyle(m.current))},[m]),g.jsx(cp,{"data-orientation":"horizontal",...f,ref:v,sizes:i,style:{...c,"--sa-thumb-width":`${Su(i)}px`},onThumbPointerDown:S=>n.onThumbPointerDown(S.x),onDragScroll:S=>n.onDragScroll(S.x),onWheelScroll:(S,_)=>{if(d.viewport){const x=d.viewport.scrollLeft+S.deltaX;n.onWheelScroll(x),op(x,_)&&S.preventDefault()}},onResize:()=>{m.current&&d.viewport&&h&&o({content:d.viewport.scrollWidth,viewport:d.viewport.offsetWidth,scrollbar:{size:m.current.clientWidth,paddingStart:iu(h.paddingLeft),paddingEnd:iu(h.paddingRight)}})}})});sp.displayName="@mantine/core/ScrollAreaScrollbarX";const fp=C.forwardRef((n,l)=>{const{sizes:i,onSizesChange:o,style:c,...f}=n,d=Je(),[h,y]=C.useState(),m=C.useRef(null),v=Ga(l,m,d.onScrollbarYChange);return C.useEffect(()=>{m.current&&y(window.getComputedStyle(m.current))},[]),g.jsx(cp,{...f,"data-orientation":"vertical",ref:v,sizes:i,style:{"--sa-thumb-height":`${Su(i)}px`,...c},onThumbPointerDown:S=>n.onThumbPointerDown(S.y),onDragScroll:S=>n.onDragScroll(S.y),onWheelScroll:(S,_)=>{if(d.viewport){const x=d.viewport.scrollTop+S.deltaY;n.onWheelScroll(x),op(x,_)&&S.preventDefault()}},onResize:()=>{m.current&&d.viewport&&h&&o({content:d.viewport.scrollHeight,viewport:d.viewport.offsetHeight,scrollbar:{size:m.current.clientHeight,paddingStart:iu(h.paddingTop),paddingEnd:iu(h.paddingBottom)}})}})});fp.displayName="@mantine/core/ScrollAreaScrollbarY";const _u=C.forwardRef((n,l)=>{const{orientation:i="vertical",...o}=n,{dir:c}=np(),f=Je(),d=C.useRef(null),h=C.useRef(0),[y,m]=C.useState({content:0,viewport:0,scrollbar:{size:0,paddingStart:0,paddingEnd:0}}),v=rp(y.viewport,y.content),S={...o,sizes:y,onSizesChange:m,hasThumb:v>0&&v<1,onThumbChange:x=>{d.current=x},onThumbPointerUp:()=>{h.current=0},onThumbPointerDown:x=>{h.current=x}},_=(x,w)=>x_(x,h.current,y,w);return i==="horizontal"?g.jsx(sp,{...S,ref:l,onThumbPositionChange:()=>{if(f.viewport&&d.current){const x=f.viewport.scrollLeft,w=Y0(x,y,c);d.current.style.transform=`translate3d(${w}px, 0, 0)`}},onWheelScroll:x=>{f.viewport&&(f.viewport.scrollLeft=x)},onDragScroll:x=>{f.viewport&&(f.viewport.scrollLeft=_(x,c))}}):i==="vertical"?g.jsx(fp,{...S,ref:l,onThumbPositionChange:()=>{if(f.viewport&&d.current){const x=f.viewport.scrollTop,w=Y0(x,y);y.scrollbar.size===0?d.current.style.setProperty("--thumb-opacity","0"):d.current.style.setProperty("--thumb-opacity","1"),d.current.style.transform=`translate3d(0, ${w}px, 0)`}},onWheelScroll:x=>{f.viewport&&(f.viewport.scrollTop=x)},onDragScroll:x=>{f.viewport&&(f.viewport.scrollTop=_(x))}}):null});_u.displayName="@mantine/core/ScrollAreaScrollbarVisible";const Kf=C.forwardRef((n,l)=>{const i=Je(),{forceMount:o,...c}=n,[f,d]=C.useState(!1),h=n.orientation==="horizontal",y=vu(()=>{if(i.viewport){const m=i.viewport.offsetWidth<i.viewport.scrollWidth,v=i.viewport.offsetHeight<i.viewport.scrollHeight;d(h?m:v)}},10);return Yl(i.viewport,y),Yl(i.content,y),o||f?g.jsx(_u,{"data-state":f?"visible":"hidden",...c,ref:l}):null});Kf.displayName="@mantine/core/ScrollAreaScrollbarAuto";const dp=C.forwardRef((n,l)=>{const{forceMount:i,...o}=n,c=Je(),[f,d]=C.useState(!1);return C.useEffect(()=>{const{scrollArea:h}=c;let y=0;if(h){const m=()=>{window.clearTimeout(y),d(!0)},v=()=>{y=window.setTimeout(()=>d(!1),c.scrollHideDelay)};return h.addEventListener("pointerenter",m),h.addEventListener("pointerleave",v),()=>{window.clearTimeout(y),h.removeEventListener("pointerenter",m),h.removeEventListener("pointerleave",v)}}},[c.scrollArea,c.scrollHideDelay]),i||f?g.jsx(Kf,{"data-state":f?"visible":"hidden",...o,ref:l}):null});dp.displayName="@mantine/core/ScrollAreaScrollbarHover";const w_=C.forwardRef((n,l)=>{const{forceMount:i,...o}=n,c=Je(),f=n.orientation==="horizontal",[d,h]=C.useState("hidden"),y=vu(()=>h("idle"),100);return C.useEffect(()=>{if(d==="idle"){const m=window.setTimeout(()=>h("hidden"),c.scrollHideDelay);return()=>window.clearTimeout(m)}},[d,c.scrollHideDelay]),C.useEffect(()=>{const{viewport:m}=c,v=f?"scrollLeft":"scrollTop";if(m){let S=m[v];const _=()=>{const x=m[v];S!==x&&(h("scrolling"),y()),S=x};return m.addEventListener("scroll",_),()=>m.removeEventListener("scroll",_)}},[c.viewport,f,y]),i||d!=="hidden"?g.jsx(_u,{"data-state":d==="hidden"?"hidden":"visible",...o,ref:l,onPointerEnter:La(n.onPointerEnter,()=>h("interacting")),onPointerLeave:La(n.onPointerLeave,()=>h("idle"))}):null}),Cf=C.forwardRef((n,l)=>{const{forceMount:i,...o}=n,c=Je(),{onScrollbarXEnabledChange:f,onScrollbarYEnabledChange:d}=c,h=n.orientation==="horizontal";return C.useEffect(()=>(h?f(!0):d(!0),()=>{h?f(!1):d(!1)}),[h,f,d]),c.type==="hover"?g.jsx(dp,{...o,ref:l,forceMount:i}):c.type==="scroll"?g.jsx(w_,{...o,ref:l,forceMount:i}):c.type==="auto"?g.jsx(Kf,{...o,ref:l,forceMount:i}):c.type==="always"?g.jsx(_u,{...o,ref:l}):null});Cf.displayName="@mantine/core/ScrollAreaScrollbar";function E_(n,l=()=>{}){let i={left:n.scrollLeft,top:n.scrollTop},o=0;return function c(){const f={left:n.scrollLeft,top:n.scrollTop},d=i.left!==f.left,h=i.top!==f.top;(d||h)&&l(),i=f,o=window.requestAnimationFrame(c)}(),()=>window.cancelAnimationFrame(o)}const hp=C.forwardRef((n,l)=>{const{style:i,...o}=n,c=Je(),f=up(),{onThumbPositionChange:d}=f,h=Ga(l,v=>f.onThumbChange(v)),y=C.useRef(void 0),m=vu(()=>{y.current&&(y.current(),y.current=void 0)},100);return C.useEffect(()=>{const{viewport:v}=c;if(v){const S=()=>{if(m(),!y.current){const _=E_(v,d);y.current=_,d()}};return d(),v.addEventListener("scroll",S),()=>v.removeEventListener("scroll",S)}},[c.viewport,m,d]),g.jsx("div",{"data-state":f.hasThumb?"visible":"hidden",...o,ref:h,style:{width:"var(--sa-thumb-width)",height:"var(--sa-thumb-height)",...i},onPointerDownCapture:La(n.onPointerDownCapture,v=>{const _=v.target.getBoundingClientRect(),x=v.clientX-_.left,w=v.clientY-_.top;f.onThumbPointerDown({x,y:w})}),onPointerUp:La(n.onPointerUp,f.onThumbPointerUp)})});hp.displayName="@mantine/core/ScrollAreaThumb";const Af=C.forwardRef((n,l)=>{const{forceMount:i,...o}=n,c=up();return i||c.hasThumb?g.jsx(hp,{ref:l,...o}):null});Af.displayName="@mantine/core/ScrollAreaThumb";const mp=C.forwardRef(({children:n,style:l,...i},o)=>{const c=Je(),f=Ga(o,c.onViewportChange);return g.jsx(ft,{...i,ref:f,style:{overflowX:c.scrollbarXEnabled?"scroll":"hidden",overflowY:c.scrollbarYEnabled?"scroll":"hidden",...l},children:g.jsx("div",{style:{minWidth:"100%",display:"table"},ref:c.onContentChange,children:n})})});mp.displayName="@mantine/core/ScrollAreaViewport";var Jf={root:"m_d57069b5",viewport:"m_c0783ff9",viewportInner:"m_f8f631dd",scrollbar:"m_c44ba933",thumb:"m_d8b5e363",corner:"m_21657268"};const yp={scrollHideDelay:1e3,type:"hover",scrollbars:"xy"},C_=(n,{scrollbarSize:l,overscrollBehavior:i})=>({root:{"--scrollarea-scrollbar-size":tt(l),"--scrollarea-over-scroll-behavior":i}}),yi=_t((n,l)=>{const i=st("ScrollArea",yp,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,scrollbarSize:y,vars:m,type:v,scrollHideDelay:S,viewportProps:_,viewportRef:x,onScrollPositionChange:w,children:N,offsetScrollbars:B,scrollbars:O,onBottomReached:j,onTopReached:z,overscrollBehavior:L,...$}=i,[V,Z]=C.useState(!1),[W,it]=C.useState(!1),[ht,pt]=C.useState(!1),at=jt({name:"ScrollArea",props:i,classes:Jf,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:m,varsResolver:C_}),lt=C.useRef(null),R=v_([x,lt]);return C.useEffect(()=>{if(!lt.current||B!=="present")return;const Q=lt.current,E=new ResizeObserver(()=>{const{scrollHeight:Y,clientHeight:k,scrollWidth:et,clientWidth:T}=Q;it(Y>k),pt(et>T)});return E.observe(Q),()=>E.disconnect()},[lt,B]),g.jsxs(lp,{type:v==="never"?"always":v,scrollHideDelay:S,ref:l,scrollbars:O,...at("root"),...$,children:[g.jsx(mp,{..._,...at("viewport",{style:_==null?void 0:_.style}),ref:R,"data-offset-scrollbars":B===!0?"xy":B||void 0,"data-scrollbars":O||void 0,"data-horizontal-hidden":B==="present"&&!ht?"true":void 0,"data-vertical-hidden":B==="present"&&!W?"true":void 0,onScroll:Q=>{var et;(et=_==null?void 0:_.onScroll)==null||et.call(_,Q),w==null||w({x:Q.currentTarget.scrollLeft,y:Q.currentTarget.scrollTop});const{scrollTop:E,scrollHeight:Y,clientHeight:k}=Q.currentTarget;E-(Y-k)>=0&&(j==null||j()),E===0&&(z==null||z())},children:N}),(O==="xy"||O==="x")&&g.jsx(Cf,{...at("scrollbar"),orientation:"horizontal","data-hidden":v==="never"||!ht||void 0,forceMount:!0,onMouseEnter:()=>Z(!0),onMouseLeave:()=>Z(!1),children:g.jsx(Af,{...at("thumb")})}),(O==="xy"||O==="y")&&g.jsx(Cf,{...at("scrollbar"),orientation:"vertical","data-hidden":v==="never"||!W||void 0,forceMount:!0,onMouseEnter:()=>Z(!0),onMouseLeave:()=>Z(!1),children:g.jsx(Af,{...at("thumb")})}),g.jsx(b_,{...at("corner"),"data-hovered":V||void 0,"data-hidden":v==="never"||void 0})]})});yi.displayName="@mantine/core/ScrollArea";const Wf=_t((n,l)=>{const{children:i,classNames:o,styles:c,scrollbarSize:f,scrollHideDelay:d,type:h,dir:y,offsetScrollbars:m,viewportRef:v,onScrollPositionChange:S,unstyled:_,variant:x,viewportProps:w,scrollbars:N,style:B,vars:O,onBottomReached:j,onTopReached:z,...L}=st("ScrollAreaAutosize",yp,n);return g.jsx(ft,{...L,ref:l,style:[{display:"flex",overflow:"auto"},B],children:g.jsx(ft,{style:{display:"flex",flexDirection:"column",flex:1},children:g.jsx(yi,{classNames:o,styles:c,scrollHideDelay:d,scrollbarSize:f,type:h,dir:y,offsetScrollbars:m,viewportRef:v,onScrollPositionChange:S,unstyled:_,variant:x,viewportProps:w,vars:O,scrollbars:N,onBottomReached:j,onTopReached:z,children:i})})})});yi.classes=Jf;Wf.displayName="@mantine/core/ScrollAreaAutosize";Wf.classes=Jf;yi.Autosize=Wf;var vp={root:"m_87cf2631"};const A_={__staticSelector:"UnstyledButton"},Zl=gn((n,l)=>{const i=st("UnstyledButton",A_,n),{className:o,component:c="button",__staticSelector:f,unstyled:d,classNames:h,styles:y,style:m,...v}=i,S=jt({name:f,props:i,classes:vp,className:o,style:m,classNames:h,styles:y,unstyled:d});return g.jsx(ft,{...S("root",{focusable:!0}),component:c,ref:l,type:c==="button"?"button":void 0,...v})});Zl.classes=vp;Zl.displayName="@mantine/core/UnstyledButton";var pp={root:"m_515a97f8"};const j_={},Pf=_t((n,l)=>{const i=st("VisuallyHidden",j_,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,...m}=i,v=jt({name:"VisuallyHidden",classes:pp,props:i,className:c,style:f,classNames:o,styles:d,unstyled:h});return g.jsx(ft,{component:"span",ref:l,...v("root"),...m})});Pf.classes=pp;Pf.displayName="@mantine/core/VisuallyHidden";var gp={root:"m_1b7284a3"};const N_={},O_=(n,{radius:l,shadow:i})=>({root:{"--paper-radius":l===void 0?void 0:vn(l),"--paper-shadow":zv(i)}}),Ue=gn((n,l)=>{const i=st("Paper",N_,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,withBorder:y,vars:m,radius:v,shadow:S,variant:_,mod:x,...w}=i,N=jt({name:"Paper",props:i,classes:gp,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:m,varsResolver:O_});return g.jsx(ft,{ref:l,mod:[{"data-with-border":y},x],...N("root"),variant:_,...w})});Ue.classes=gp;Ue.displayName="@mantine/core/Paper";var bp={root:"m_9814e45f"};const M_={zIndex:kl("modal")},R_=(n,{gradient:l,color:i,backgroundOpacity:o,blur:c,radius:f,zIndex:d})=>({root:{"--overlay-bg":l||(i!==void 0||o!==void 0)&&fn(i||"#000",o??.6)||void 0,"--overlay-filter":c?`blur(${tt(c)})`:void 0,"--overlay-radius":f===void 0?void 0:vn(f),"--overlay-z-index":d==null?void 0:d.toString()}}),Ff=gn((n,l)=>{const i=st("Overlay",M_,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,fixed:m,center:v,children:S,radius:_,zIndex:x,gradient:w,blur:N,color:B,backgroundOpacity:O,mod:j,...z}=i,L=jt({name:"Overlay",props:i,classes:bp,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:y,varsResolver:R_});return g.jsx(ft,{ref:l,...L("root"),mod:[{center:v,fixed:m},j],...z,children:S})});Ff.classes=bp;Ff.displayName="@mantine/core/Overlay";function hf(n){const l=document.createElement("div");return l.setAttribute("data-portal","true"),typeof n.className=="string"&&l.classList.add(...n.className.split(" ").filter(Boolean)),typeof n.style=="object"&&Object.assign(l.style,n.style),typeof n.id=="string"&&l.setAttribute("id",n.id),l}function z_({target:n,reuseTargetNode:l,...i}){if(n)return typeof n=="string"?document.querySelector(n)||hf(i):n;if(l){const o=document.querySelector("[data-mantine-shared-portal-node]");if(o)return o;const c=hf(i);return c.setAttribute("data-mantine-shared-portal-node","true"),document.body.appendChild(c),c}return hf(i)}const D_={},Sp=_t((n,l)=>{const{children:i,target:o,reuseTargetNode:c,...f}=st("Portal",D_,n),[d,h]=C.useState(!1),y=C.useRef(null);return hi(()=>(h(!0),y.current=z_({target:o,reuseTargetNode:c,...f}),xf(l,y.current),!o&&!c&&y.current&&document.body.appendChild(y.current),()=>{!o&&!c&&y.current&&document.body.removeChild(y.current)}),[o]),!d||!y.current?null:ap.createPortal(g.jsx(g.Fragment,{children:i}),y.current)});Sp.displayName="@mantine/core/Portal";const _p=_t(({withinPortal:n=!0,children:l,...i},o)=>Xv()==="test"||!n?g.jsx(g.Fragment,{children:l}):g.jsx(Sp,{ref:o,...i,children:l}));_p.displayName="@mantine/core/OptionalPortal";const Fr=n=>({in:{opacity:1,transform:"scale(1)"},out:{opacity:0,transform:`scale(.9) translateY(${n==="bottom"?10:-10}px)`},transitionProperty:"transform, opacity"}),ko={fade:{in:{opacity:1},out:{opacity:0},transitionProperty:"opacity"},"fade-up":{in:{opacity:1,transform:"translateY(0)"},out:{opacity:0,transform:"translateY(30px)"},transitionProperty:"opacity, transform"},"fade-down":{in:{opacity:1,transform:"translateY(0)"},out:{opacity:0,transform:"translateY(-30px)"},transitionProperty:"opacity, transform"},"fade-left":{in:{opacity:1,transform:"translateX(0)"},out:{opacity:0,transform:"translateX(30px)"},transitionProperty:"opacity, transform"},"fade-right":{in:{opacity:1,transform:"translateX(0)"},out:{opacity:0,transform:"translateX(-30px)"},transitionProperty:"opacity, transform"},scale:{in:{opacity:1,transform:"scale(1)"},out:{opacity:0,transform:"scale(0)"},common:{transformOrigin:"top"},transitionProperty:"transform, opacity"},"scale-y":{in:{opacity:1,transform:"scaleY(1)"},out:{opacity:0,transform:"scaleY(0)"},common:{transformOrigin:"top"},transitionProperty:"transform, opacity"},"scale-x":{in:{opacity:1,transform:"scaleX(1)"},out:{opacity:0,transform:"scaleX(0)"},common:{transformOrigin:"left"},transitionProperty:"transform, opacity"},"skew-up":{in:{opacity:1,transform:"translateY(0) skew(0deg, 0deg)"},out:{opacity:0,transform:"translateY(-20px) skew(-10deg, -5deg)"},common:{transformOrigin:"top"},transitionProperty:"transform, opacity"},"skew-down":{in:{opacity:1,transform:"translateY(0) skew(0deg, 0deg)"},out:{opacity:0,transform:"translateY(20px) skew(-10deg, -5deg)"},common:{transformOrigin:"bottom"},transitionProperty:"transform, opacity"},"rotate-left":{in:{opacity:1,transform:"translateY(0) rotate(0deg)"},out:{opacity:0,transform:"translateY(20px) rotate(-5deg)"},common:{transformOrigin:"bottom"},transitionProperty:"transform, opacity"},"rotate-right":{in:{opacity:1,transform:"translateY(0) rotate(0deg)"},out:{opacity:0,transform:"translateY(20px) rotate(5deg)"},common:{transformOrigin:"top"},transitionProperty:"transform, opacity"},"slide-down":{in:{opacity:1,transform:"translateY(0)"},out:{opacity:0,transform:"translateY(-100%)"},common:{transformOrigin:"top"},transitionProperty:"transform, opacity"},"slide-up":{in:{opacity:1,transform:"translateY(0)"},out:{opacity:0,transform:"translateY(100%)"},common:{transformOrigin:"bottom"},transitionProperty:"transform, opacity"},"slide-left":{in:{opacity:1,transform:"translateX(0)"},out:{opacity:0,transform:"translateX(100%)"},common:{transformOrigin:"left"},transitionProperty:"transform, opacity"},"slide-right":{in:{opacity:1,transform:"translateX(0)"},out:{opacity:0,transform:"translateX(-100%)"},common:{transformOrigin:"right"},transitionProperty:"transform, opacity"},pop:{...Fr("bottom"),common:{transformOrigin:"center center"}},"pop-bottom-left":{...Fr("bottom"),common:{transformOrigin:"bottom left"}},"pop-bottom-right":{...Fr("bottom"),common:{transformOrigin:"bottom right"}},"pop-top-left":{...Fr("top"),common:{transformOrigin:"top left"}},"pop-top-right":{...Fr("top"),common:{transformOrigin:"top right"}}},q0={entering:"in",entered:"in",exiting:"out",exited:"out","pre-exiting":"out","pre-entering":"out"};function B_({transition:n,state:l,duration:i,timingFunction:o}){const c={WebkitBackfaceVisibility:"hidden",willChange:"transform, opacity",transitionDuration:`${i}ms`,transitionTimingFunction:o};return typeof n=="string"?n in ko?{transitionProperty:ko[n].transitionProperty,...c,...ko[n].common,...ko[n][q0[l]]}:{}:{transitionProperty:n.transitionProperty,...c,...n.common,...n[q0[l]]}}function U_({duration:n,exitDuration:l,timingFunction:i,mounted:o,onEnter:c,onExit:f,onEntered:d,onExited:h,enterDelay:y,exitDelay:m}){const v=pn(),S=Lv(),_=v.respectReducedMotion?S:!1,[x,w]=C.useState(_?0:n),[N,B]=C.useState(o?"entered":"exited"),O=C.useRef(-1),j=C.useRef(-1),z=C.useRef(-1),L=V=>{const Z=V?c:f,W=V?d:h;window.clearTimeout(O.current);const it=_?0:V?n:l;w(it),it===0?(typeof Z=="function"&&Z(),typeof W=="function"&&W(),B(V?"entered":"exited")):z.current=requestAnimationFrame(()=>{y_.flushSync(()=>{B(V?"pre-entering":"pre-exiting")}),z.current=requestAnimationFrame(()=>{typeof Z=="function"&&Z(),B(V?"entering":"exiting"),O.current=window.setTimeout(()=>{typeof W=="function"&&W(),B(V?"entered":"exited")},it)})})},$=V=>{if(window.clearTimeout(j.current),typeof(V?y:m)!="number"){L(V);return}j.current=window.setTimeout(()=>{L(V)},V?y:m)};return Dv(()=>{$(o)},[o]),C.useEffect(()=>()=>{window.clearTimeout(O.current),cancelAnimationFrame(z.current)},[]),{transitionDuration:x,transitionStatus:N,transitionTimingFunction:i||"ease"}}function vi({keepMounted:n,transition:l="fade",duration:i=250,exitDuration:o=i,mounted:c,children:f,timingFunction:d="ease",onExit:h,onEntered:y,onEnter:m,onExited:v,enterDelay:S,exitDelay:_}){const x=Xv(),{transitionDuration:w,transitionStatus:N,transitionTimingFunction:B}=U_({mounted:c,exitDuration:o,duration:i,timingFunction:d,onExit:h,onEntered:y,onEnter:m,onExited:v,enterDelay:S,exitDelay:_});return w===0||x==="test"?c?g.jsx(g.Fragment,{children:f({})}):n?f({display:"none"}):null:N==="exited"?n?f({display:"none"}):null:g.jsx(g.Fragment,{children:f(B_({transition:l,duration:w,state:N,timingFunction:B}))})}vi.displayName="@mantine/core/Transition";function If({children:n,active:l=!0,refProp:i="ref",innerRef:o}){const c=IS(l),f=Ga(c,o);return US(n)?C.cloneElement(n,{[i]:f}):n}function xp(n){return g.jsx(Pf,{tabIndex:-1,"data-autofocus":!0,...n})}If.displayName="@mantine/core/FocusTrap";xp.displayName="@mantine/core/FocusTrapInitialFocus";If.InitialFocus=xp;var nn={root:"m_5ae2e3c",barsLoader:"m_7a2bd4cd",bar:"m_870bb79","bars-loader-animation":"m_5d2b3b9d",dotsLoader:"m_4e3f22d7",dot:"m_870c4af","loader-dots-animation":"m_aac34a1",ovalLoader:"m_b34414df","oval-loader-animation":"m_f8e89c4b"};const Tp=C.forwardRef(({className:n,...l},i)=>g.jsxs(ft,{component:"span",className:Ce(nn.barsLoader,n),...l,ref:i,children:[g.jsx("span",{className:nn.bar}),g.jsx("span",{className:nn.bar}),g.jsx("span",{className:nn.bar})]}));Tp.displayName="@mantine/core/Bars";const wp=C.forwardRef(({className:n,...l},i)=>g.jsxs(ft,{component:"span",className:Ce(nn.dotsLoader,n),...l,ref:i,children:[g.jsx("span",{className:nn.dot}),g.jsx("span",{className:nn.dot}),g.jsx("span",{className:nn.dot})]}));wp.displayName="@mantine/core/Dots";const Ep=C.forwardRef(({className:n,...l},i)=>g.jsx(ft,{component:"span",className:Ce(nn.ovalLoader,n),...l,ref:i}));Ep.displayName="@mantine/core/Oval";const Cp={bars:Tp,oval:Ep,dots:wp},H_={loaders:Cp,type:"oval"},$_=(n,{size:l,color:i})=>({root:{"--loader-size":Vt(l,"loader-size"),"--loader-color":i?mn(i,n):void 0}}),pi=_t((n,l)=>{const i=st("Loader",H_,n),{size:o,color:c,type:f,vars:d,className:h,style:y,classNames:m,styles:v,unstyled:S,loaders:_,variant:x,children:w,...N}=i,B=jt({name:"Loader",props:i,classes:nn,className:h,style:y,classNames:m,styles:v,unstyled:S,vars:d,varsResolver:$_});return w?g.jsx(ft,{...B("root"),ref:l,...N,children:w}):g.jsx(ft,{...B("root"),ref:l,component:_[f],variant:x,size:o,...N})});pi.defaultLoaders=Cp;pi.classes=nn;pi.displayName="@mantine/core/Loader";var Ql={root:"m_8d3f4000",icon:"m_8d3afb97",loader:"m_302b9fb1",group:"m_1a0f1b21",groupSection:"m_437b6484"};const X0={orientation:"horizontal"},L_=(n,{borderWidth:l})=>({group:{"--ai-border-width":tt(l)}}),td=_t((n,l)=>{const i=st("ActionIconGroup",X0,n),{className:o,style:c,classNames:f,styles:d,unstyled:h,orientation:y,vars:m,borderWidth:v,variant:S,mod:_,...x}=st("ActionIconGroup",X0,n),w=jt({name:"ActionIconGroup",props:i,classes:Ql,className:o,style:c,classNames:f,styles:d,unstyled:h,vars:m,varsResolver:L_,rootSelector:"group"});return g.jsx(ft,{...w("group"),ref:l,variant:S,mod:[{"data-orientation":y},_],role:"group",...x})});td.classes=Ql;td.displayName="@mantine/core/ActionIconGroup";const G0={},Y_=(n,{radius:l,color:i,gradient:o,variant:c,autoContrast:f,size:d})=>{const h=n.variantColorResolver({color:i||n.primaryColor,theme:n,gradient:o,variant:c||"filled",autoContrast:f});return{groupSection:{"--section-height":Vt(d,"section-height"),"--section-padding-x":Vt(d,"section-padding-x"),"--section-fz":we(d),"--section-radius":l===void 0?void 0:vn(l),"--section-bg":i||c?h.background:void 0,"--section-color":h.color,"--section-bd":i||c?h.border:void 0}}},ed=_t((n,l)=>{const i=st("ActionIconGroupSection",G0,n),{className:o,style:c,classNames:f,styles:d,unstyled:h,vars:y,variant:m,gradient:v,radius:S,autoContrast:_,...x}=st("ActionIconGroupSection",G0,n),w=jt({name:"ActionIconGroupSection",props:i,classes:Ql,className:o,style:c,classNames:f,styles:d,unstyled:h,vars:y,varsResolver:Y_,rootSelector:"groupSection"});return g.jsx(ft,{...w("groupSection"),ref:l,variant:m,...x})});ed.classes=Ql;ed.displayName="@mantine/core/ActionIconGroupSection";const q_={},X_=(n,{size:l,radius:i,variant:o,gradient:c,color:f,autoContrast:d})=>{const h=n.variantColorResolver({color:f||n.primaryColor,theme:n,gradient:c,variant:o||"filled",autoContrast:d});return{root:{"--ai-size":Vt(l,"ai-size"),"--ai-radius":i===void 0?void 0:vn(i),"--ai-bg":f||o?h.background:void 0,"--ai-hover":f||o?h.hover:void 0,"--ai-hover-color":f||o?h.hoverColor:void 0,"--ai-color":h.color,"--ai-bd":f||o?h.border:void 0}}},gi=gn((n,l)=>{const i=st("ActionIcon",q_,n),{className:o,unstyled:c,variant:f,classNames:d,styles:h,style:y,loading:m,loaderProps:v,size:S,color:_,radius:x,__staticSelector:w,gradient:N,vars:B,children:O,disabled:j,"data-disabled":z,autoContrast:L,mod:$,...V}=i,Z=jt({name:["ActionIcon",w],props:i,className:o,style:y,classes:Ql,classNames:d,styles:h,unstyled:c,vars:B,varsResolver:X_});return g.jsxs(Zl,{...Z("root",{active:!j&&!m&&!z}),...V,unstyled:c,variant:f,size:S,disabled:j||m,ref:l,mod:[{loading:m,disabled:j||z},$],children:[g.jsx(vi,{mounted:!!m,transition:"slide-down",duration:150,children:W=>g.jsx(ft,{component:"span",...Z("loader",{style:W}),"aria-hidden":!0,children:g.jsx(pi,{color:"var(--ai-color)",size:"calc(var(--ai-size) * 0.55)",...v})})}),g.jsx(ft,{component:"span",mod:{loading:m},...Z("icon"),children:O})]})});gi.classes=Ql;gi.displayName="@mantine/core/ActionIcon";gi.Group=td;gi.GroupSection=ed;const Ap=C.forwardRef(({size:n="var(--cb-icon-size, 70%)",style:l,...i},o)=>g.jsx("svg",{viewBox:"0 0 15 15",fill:"none",xmlns:"http://www.w3.org/2000/svg",style:{...l,width:n,height:n},ref:o,...i,children:g.jsx("path",{d:"M11.7816 4.03157C12.0062 3.80702 12.0062 3.44295 11.7816 3.2184C11.5571 2.99385 11.193 2.99385 10.9685 3.2184L7.50005 6.68682L4.03164 3.2184C3.80708 2.99385 3.44301 2.99385 3.21846 3.2184C2.99391 3.44295 2.99391 3.80702 3.21846 4.03157L6.68688 7.49999L3.21846 10.9684C2.99391 11.193 2.99391 11.557 3.21846 11.7816C3.44301 12.0061 3.80708 12.0061 4.03164 11.7816L7.50005 8.31316L10.9685 11.7816C11.193 12.0061 11.5571 12.0061 11.7816 11.7816C12.0062 11.557 12.0062 11.193 11.7816 10.9684L8.31322 7.49999L11.7816 4.03157Z",fill:"currentColor",fillRule:"evenodd",clipRule:"evenodd"})}));Ap.displayName="@mantine/core/CloseIcon";var jp={root:"m_86a44da5","root--subtle":"m_220c80f2"};const G_={variant:"subtle"},V_=(n,{size:l,radius:i,iconSize:o})=>({root:{"--cb-size":Vt(l,"cb-size"),"--cb-radius":i===void 0?void 0:vn(i),"--cb-icon-size":tt(o)}}),Va=gn((n,l)=>{const i=st("CloseButton",G_,n),{iconSize:o,children:c,vars:f,radius:d,className:h,classNames:y,style:m,styles:v,unstyled:S,"data-disabled":_,disabled:x,variant:w,icon:N,mod:B,__staticSelector:O,...j}=i,z=jt({name:O||"CloseButton",props:i,className:h,style:m,classes:jp,classNames:y,styles:v,unstyled:S,vars:f,varsResolver:V_});return g.jsxs(Zl,{ref:l,...j,unstyled:S,variant:w,disabled:x,mod:[{disabled:x||_},B],...z("root",{variant:w,active:!x&&!_}),children:[N||g.jsx(Ap,{}),c]})});Va.classes=jp;Va.displayName="@mantine/core/CloseButton";function k_(n){return C.Children.toArray(n).filter(Boolean)}var Np={root:"m_4081bf90"};const Z_={preventGrowOverflow:!0,gap:"md",align:"center",justify:"flex-start",wrap:"wrap"},Q_=(n,{grow:l,preventGrowOverflow:i,gap:o,align:c,justify:f,wrap:d},{childWidth:h})=>({root:{"--group-child-width":l&&i?h:void 0,"--group-gap":Ll(o),"--group-align":c,"--group-justify":f,"--group-wrap":d}}),Ee=_t((n,l)=>{const i=st("Group",Z_,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,children:y,gap:m,align:v,justify:S,wrap:_,grow:x,preventGrowOverflow:w,vars:N,variant:B,__size:O,mod:j,...z}=i,L=k_(y),$=L.length,V=Ll(m??"md"),W={childWidth:`calc(${100/$}% - (${V} - ${V} / ${$}))`},it=jt({name:"Group",props:i,stylesCtx:W,className:c,style:f,classes:Np,classNames:o,styles:d,unstyled:h,vars:N,varsResolver:Q_});return g.jsx(ft,{...it("root"),ref:l,variant:B,mod:[{grow:x},j],size:O,...z,children:L})});Ee.classes=Np;Ee.displayName="@mantine/core/Group";const[K_,Xn]=Xa("ModalBase component was not found in tree");function J_({opened:n,transitionDuration:l}){const[i,o]=C.useState(n),c=C.useRef(-1),d=Lv()?0:l;return C.useEffect(()=>(n?(o(!0),window.clearTimeout(c.current)):d===0?o(!1):c.current=window.setTimeout(()=>o(!1),d),()=>window.clearTimeout(c.current)),[n,d]),i}function W_({id:n,transitionProps:l,opened:i,trapFocus:o,closeOnEscape:c,onClose:f,returnFocus:d}){const h=pu(n),[y,m]=C.useState(!1),[v,S]=C.useState(!1),_=typeof(l==null?void 0:l.duration)=="number"?l==null?void 0:l.duration:200,x=J_({opened:i,transitionDuration:_});return n2("keydown",w=>{var N;w.key==="Escape"&&c&&i&&((N=w.target)==null?void 0:N.getAttribute("data-mantine-stop-propagation"))!=="true"&&f()},{capture:!0}),QS({opened:i,shouldReturnFocus:o&&d}),{_id:h,titleMounted:y,bodyMounted:v,shouldLockScroll:x,setTitleMounted:m,setBodyMounted:S}}const Op=C.forwardRef(({keepMounted:n,opened:l,onClose:i,id:o,transitionProps:c,onExitTransitionEnd:f,onEnterTransitionEnd:d,trapFocus:h,closeOnEscape:y,returnFocus:m,closeOnClickOutside:v,withinPortal:S,portalProps:_,lockScroll:x,children:w,zIndex:N,shadow:B,padding:O,__vars:j,unstyled:z,removeScrollProps:L,...$},V)=>{const{_id:Z,titleMounted:W,bodyMounted:it,shouldLockScroll:ht,setTitleMounted:pt,setBodyMounted:at}=W_({id:o,transitionProps:c,opened:l,trapFocus:h,closeOnEscape:y,onClose:i,returnFocus:m}),{key:lt,...R}=L||{};return g.jsx(_p,{..._,withinPortal:S,children:g.jsx(K_,{value:{opened:l,onClose:i,closeOnClickOutside:v,onExitTransitionEnd:f,onEnterTransitionEnd:d,transitionProps:{...c,keepMounted:n},getTitleId:()=>`${Z}-title`,getBodyId:()=>`${Z}-body`,titleMounted:W,bodyMounted:it,setTitleMounted:pt,setBodyMounted:at,trapFocus:h,closeOnEscape:y,zIndex:N,unstyled:z},children:g.jsx(Nv,{enabled:ht&&x,...R,children:g.jsx(ft,{ref:V,...$,__vars:{...j,"--mb-z-index":(N||kl("modal")).toString(),"--mb-shadow":zv(B),"--mb-padding":Ll(O)},children:w})},lt)})})});Op.displayName="@mantine/core/ModalBase";function P_(){const n=Xn();return C.useEffect(()=>(n.setBodyMounted(!0),()=>n.setBodyMounted(!1)),[]),n.getBodyId()}var ql={title:"m_615af6c9",header:"m_b5489c3c",inner:"m_60c222c7",content:"m_fd1ab0aa",close:"m_606cb269",body:"m_5df29311"};const Mp=C.forwardRef(({className:n,...l},i)=>{const o=P_(),c=Xn();return g.jsx(ft,{ref:i,...l,id:o,className:Ce({[ql.body]:!c.unstyled},n)})});Mp.displayName="@mantine/core/ModalBaseBody";const Rp=C.forwardRef(({className:n,onClick:l,...i},o)=>{const c=Xn();return g.jsx(Va,{ref:o,...i,onClick:f=>{c.onClose(),l==null||l(f)},className:Ce({[ql.close]:!c.unstyled},n),unstyled:c.unstyled})});Rp.displayName="@mantine/core/ModalBaseCloseButton";const zp=C.forwardRef(({transitionProps:n,className:l,innerProps:i,onKeyDown:o,style:c,...f},d)=>{const h=Xn();return g.jsx(vi,{mounted:h.opened,transition:"pop",...h.transitionProps,onExited:()=>{var y,m,v;(y=h.onExitTransitionEnd)==null||y.call(h),(v=(m=h.transitionProps)==null?void 0:m.onExited)==null||v.call(m)},onEntered:()=>{var y,m,v;(y=h.onEnterTransitionEnd)==null||y.call(h),(v=(m=h.transitionProps)==null?void 0:m.onEntered)==null||v.call(m)},...n,children:y=>g.jsx("div",{...i,className:Ce({[ql.inner]:!h.unstyled},i.className),children:g.jsx(If,{active:h.opened&&h.trapFocus,innerRef:d,children:g.jsx(Ue,{...f,component:"section",role:"dialog",tabIndex:-1,"aria-modal":!0,"aria-describedby":h.bodyMounted?h.getBodyId():void 0,"aria-labelledby":h.titleMounted?h.getTitleId():void 0,style:[c,y],className:Ce({[ql.content]:!h.unstyled},l),unstyled:h.unstyled,children:f.children})})})})});zp.displayName="@mantine/core/ModalBaseContent";const Dp=C.forwardRef(({className:n,...l},i)=>{const o=Xn();return g.jsx(ft,{component:"header",ref:i,className:Ce({[ql.header]:!o.unstyled},n),...l})});Dp.displayName="@mantine/core/ModalBaseHeader";const F_={duration:200,timingFunction:"ease",transition:"fade"};function I_(n){const l=Xn();return{...F_,...l.transitionProps,...n}}const Bp=C.forwardRef(({onClick:n,transitionProps:l,style:i,visible:o,...c},f)=>{const d=Xn(),h=I_(l);return g.jsx(vi,{mounted:o!==void 0?o:d.opened,...h,transition:"fade",children:y=>g.jsx(Ff,{ref:f,fixed:!0,style:[i,y],zIndex:d.zIndex,unstyled:d.unstyled,onClick:m=>{n==null||n(m),d.closeOnClickOutside&&d.onClose()},...c})})});Bp.displayName="@mantine/core/ModalBaseOverlay";function tx(){const n=Xn();return C.useEffect(()=>(n.setTitleMounted(!0),()=>n.setTitleMounted(!1)),[]),n.getTitleId()}const Up=C.forwardRef(({className:n,...l},i)=>{const o=tx(),c=Xn();return g.jsx(ft,{component:"h2",ref:i,className:Ce({[ql.title]:!c.unstyled},n),...l,id:o})});Up.displayName="@mantine/core/ModalBaseTitle";function ex({children:n}){return g.jsx(g.Fragment,{children:n})}const[nx,ax]=Xf({size:"sm"}),lx={},Hp=_t((n,l)=>{const i=st("InputClearButton",lx,n),{size:o,variant:c,vars:f,classNames:d,styles:h,...y}=i,m=ax(),{resolvedClassNames:v,resolvedStyles:S}=Kv({classNames:d,styles:h,props:i});return g.jsx(Va,{variant:c||"transparent",ref:l,size:o||(m==null?void 0:m.size)||"sm",classNames:v,styles:S,__staticSelector:"InputClearButton",...y})});Hp.displayName="@mantine/core/InputClearButton";const[rx,xu]=Xf({offsetBottom:!1,offsetTop:!1,describedBy:void 0,getStyles:null,inputId:void 0,labelId:void 0});var We={wrapper:"m_6c018570",input:"m_8fb7ebe7",section:"m_82577fc2",placeholder:"m_88bacfd0",root:"m_46b77525",label:"m_8fdc1311",required:"m_78a94662",error:"m_8f816625",description:"m_fe47ce59"};const V0={},ix=(n,{size:l})=>({description:{"--input-description-size":l===void 0?void 0:`calc(${we(l)} - ${tt(2)})`}}),Tu=_t((n,l)=>{const i=st("InputDescription",V0,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,size:m,__staticSelector:v,__inheritStyles:S=!0,variant:_,...x}=st("InputDescription",V0,i),w=xu(),N=jt({name:["InputWrapper",v],props:i,classes:We,className:c,style:f,classNames:o,styles:d,unstyled:h,rootSelector:"description",vars:y,varsResolver:ix}),B=S&&(w==null?void 0:w.getStyles)||N;return g.jsx(ft,{component:"p",ref:l,variant:_,size:m,...B("description",w!=null&&w.getStyles?{className:c,style:f}:void 0),...x})});Tu.classes=We;Tu.displayName="@mantine/core/InputDescription";const ox={},ux=(n,{size:l})=>({error:{"--input-error-size":l===void 0?void 0:`calc(${we(l)} - ${tt(2)})`}}),wu=_t((n,l)=>{const i=st("InputError",ox,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,size:m,__staticSelector:v,__inheritStyles:S=!0,variant:_,...x}=i,w=jt({name:["InputWrapper",v],props:i,classes:We,className:c,style:f,classNames:o,styles:d,unstyled:h,rootSelector:"error",vars:y,varsResolver:ux}),N=xu(),B=S&&(N==null?void 0:N.getStyles)||w;return g.jsx(ft,{component:"p",ref:l,variant:_,size:m,...B("error",N!=null&&N.getStyles?{className:c,style:f}:void 0),...x})});wu.classes=We;wu.displayName="@mantine/core/InputError";const k0={labelElement:"label"},cx=(n,{size:l})=>({label:{"--input-label-size":we(l),"--input-asterisk-color":void 0}}),Eu=_t((n,l)=>{const i=st("InputLabel",k0,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,labelElement:m,size:v,required:S,htmlFor:_,onMouseDown:x,children:w,__staticSelector:N,variant:B,mod:O,...j}=st("InputLabel",k0,i),z=jt({name:["InputWrapper",N],props:i,classes:We,className:c,style:f,classNames:o,styles:d,unstyled:h,rootSelector:"label",vars:y,varsResolver:cx}),L=xu(),$=(L==null?void 0:L.getStyles)||z;return g.jsxs(ft,{...$("label",L!=null&&L.getStyles?{className:c,style:f}:void 0),component:m,variant:B,size:v,ref:l,htmlFor:m==="label"?_:void 0,mod:[{required:S},O],onMouseDown:V=>{x==null||x(V),!V.defaultPrevented&&V.detail>1&&V.preventDefault()},...j,children:[w,S&&g.jsx("span",{...$("required"),"aria-hidden":!0,children:" *"})]})});Eu.classes=We;Eu.displayName="@mantine/core/InputLabel";const Z0={},nd=_t((n,l)=>{const i=st("InputPlaceholder",Z0,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,__staticSelector:m,variant:v,error:S,mod:_,...x}=st("InputPlaceholder",Z0,i),w=jt({name:["InputPlaceholder",m],props:i,classes:We,className:c,style:f,classNames:o,styles:d,unstyled:h,rootSelector:"placeholder"});return g.jsx(ft,{...w("placeholder"),mod:[{error:!!S},_],component:"span",variant:v,ref:l,...x})});nd.classes=We;nd.displayName="@mantine/core/InputPlaceholder";function sx(n,{hasDescription:l,hasError:i}){const o=n.findIndex(y=>y==="input"),c=n.slice(0,o),f=n.slice(o+1),d=l&&c.includes("description")||i&&c.includes("error");return{offsetBottom:l&&f.includes("description")||i&&f.includes("error"),offsetTop:d}}const fx={labelElement:"label",inputContainer:n=>n,inputWrapperOrder:["label","description","input","error"]},dx=(n,{size:l})=>({label:{"--input-label-size":we(l),"--input-asterisk-color":void 0},error:{"--input-error-size":l===void 0?void 0:`calc(${we(l)} - ${tt(2)})`},description:{"--input-description-size":l===void 0?void 0:`calc(${we(l)} - ${tt(2)})`}}),ad=_t((n,l)=>{const i=st("InputWrapper",fx,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,size:m,variant:v,__staticSelector:S,inputContainer:_,inputWrapperOrder:x,label:w,error:N,description:B,labelProps:O,descriptionProps:j,errorProps:z,labelElement:L,children:$,withAsterisk:V,id:Z,required:W,__stylesApiProps:it,mod:ht,...pt}=i,at=jt({name:["InputWrapper",S],props:it||i,classes:We,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:y,varsResolver:dx}),lt={size:m,variant:v,__staticSelector:S},R=pu(Z),Q=typeof V=="boolean"?V:W,E=(z==null?void 0:z.id)||`${R}-error`,Y=(j==null?void 0:j.id)||`${R}-description`,k=R,et=!!N&&typeof N!="boolean",T=!!B,X=`${et?E:""} ${T?Y:""}`,I=X.trim().length>0?X.trim():void 0,P=(O==null?void 0:O.id)||`${R}-label`,nt=w&&g.jsx(Eu,{labelElement:L,id:P,htmlFor:k,required:Q,...lt,...O,children:w},"label"),ut=T&&g.jsx(Tu,{...j,...lt,size:(j==null?void 0:j.size)||lt.size,id:(j==null?void 0:j.id)||Y,children:B},"description"),rt=g.jsx(C.Fragment,{children:_($)},"input"),bt=et&&C.createElement(wu,{...z,...lt,size:(z==null?void 0:z.size)||lt.size,key:"error",id:(z==null?void 0:z.id)||E},N),vt=x.map(Ut=>{switch(Ut){case"label":return nt;case"input":return rt;case"description":return ut;case"error":return bt;default:return null}});return g.jsx(rx,{value:{getStyles:at,describedBy:I,inputId:k,labelId:P,...sx(x,{hasDescription:T,hasError:et})},children:g.jsx(ft,{ref:l,variant:v,size:m,mod:[{error:!!N},ht],...at("root"),...pt,children:vt})})});ad.classes=We;ad.displayName="@mantine/core/InputWrapper";const hx={variant:"default",leftSectionPointerEvents:"none",rightSectionPointerEvents:"none",withAria:!0,withErrorStyles:!0},mx=(n,l,i)=>({wrapper:{"--input-margin-top":i.offsetTop?"calc(var(--mantine-spacing-xs) / 2)":void 0,"--input-margin-bottom":i.offsetBottom?"calc(var(--mantine-spacing-xs) / 2)":void 0,"--input-height":Vt(l.size,"input-height"),"--input-fz":we(l.size),"--input-radius":l.radius===void 0?void 0:vn(l.radius),"--input-left-section-width":l.leftSectionWidth!==void 0?tt(l.leftSectionWidth):void 0,"--input-right-section-width":l.rightSectionWidth!==void 0?tt(l.rightSectionWidth):void 0,"--input-padding-y":l.multiline?Vt(l.size,"input-padding-y"):void 0,"--input-left-section-pointer-events":l.leftSectionPointerEvents,"--input-right-section-pointer-events":l.rightSectionPointerEvents}}),Qt=gn((n,l)=>{const i=st("Input",hx,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,required:y,__staticSelector:m,__stylesApiProps:v,size:S,wrapperProps:_,error:x,disabled:w,leftSection:N,leftSectionProps:B,leftSectionWidth:O,rightSection:j,rightSectionProps:z,rightSectionWidth:L,rightSectionPointerEvents:$,leftSectionPointerEvents:V,variant:Z,vars:W,pointer:it,multiline:ht,radius:pt,id:at,withAria:lt,withErrorStyles:R,mod:Q,inputSize:E,__clearSection:Y,__clearable:k,__defaultRightSection:et,...T}=i,{styleProps:X,rest:I}=bu(T),P=xu(),nt={offsetBottom:P==null?void 0:P.offsetBottom,offsetTop:P==null?void 0:P.offsetTop},ut=jt({name:["Input",m],props:v||i,classes:We,className:c,style:f,classNames:o,styles:d,unstyled:h,stylesCtx:nt,rootSelector:"wrapper",vars:W,varsResolver:mx}),rt=lt?{required:y,disabled:w,"aria-invalid":!!x,"aria-describedby":P==null?void 0:P.describedBy,id:(P==null?void 0:P.inputId)||at}:{},bt=j||k&&Y||et;return g.jsx(nx,{value:{size:S||"sm"},children:g.jsxs(ft,{...ut("wrapper"),...X,..._,mod:[{error:!!x&&R,pointer:it,disabled:w,multiline:ht,"data-with-right-section":!!bt,"data-with-left-section":!!N},Q],variant:Z,size:S,children:[N&&g.jsx("div",{...B,"data-position":"left",...ut("section",{className:B==null?void 0:B.className,style:B==null?void 0:B.style}),children:N}),g.jsx(ft,{component:"input",...I,...rt,ref:l,required:y,mod:{disabled:w,error:!!x&&R},variant:Z,__size:E,...ut("input")}),bt&&g.jsx("div",{...z,"data-position":"right",...ut("section",{className:z==null?void 0:z.className,style:z==null?void 0:z.style}),children:bt})]})})});Qt.classes=We;Qt.Wrapper=ad;Qt.Label=Eu;Qt.Error=wu;Qt.Description=Tu;Qt.Placeholder=nd;Qt.ClearButton=Hp;Qt.displayName="@mantine/core/Input";function yx(n,l,i){const o=st(n,l,i),{label:c,description:f,error:d,required:h,classNames:y,styles:m,className:v,unstyled:S,__staticSelector:_,__stylesApiProps:x,errorProps:w,labelProps:N,descriptionProps:B,wrapperProps:O,id:j,size:z,style:L,inputContainer:$,inputWrapperOrder:V,withAsterisk:Z,variant:W,vars:it,mod:ht,...pt}=o,{styleProps:at,rest:lt}=bu(pt),R={label:c,description:f,error:d,required:h,classNames:y,className:v,__staticSelector:_,__stylesApiProps:x||o,errorProps:w,labelProps:N,descriptionProps:B,unstyled:S,styles:m,size:z,style:L,inputContainer:$,inputWrapperOrder:V,withAsterisk:Z,variant:W,id:j,mod:ht,...O};return{...lt,classNames:y,styles:m,unstyled:S,wrapperProps:{...R,...at},inputProps:{required:h,classNames:y,styles:m,unstyled:S,size:z,__staticSelector:_,__stylesApiProps:x||o,error:d,variant:W,id:j}}}const vx={__staticSelector:"InputBase",withAria:!0},bi=gn((n,l)=>{const{inputProps:i,wrapperProps:o,...c}=yx("InputBase",vx,n);return g.jsx(Qt.Wrapper,{...o,children:g.jsx(Qt,{...i,...c,ref:l})})});bi.classes={...Qt.classes,...Qt.Wrapper.classes};bi.displayName="@mantine/core/InputBase";const px={gap:{type:"spacing",property:"gap"},rowGap:{type:"spacing",property:"rowGap"},columnGap:{type:"spacing",property:"columnGap"},align:{type:"identity",property:"alignItems"},justify:{type:"identity",property:"justifyContent"},wrap:{type:"identity",property:"flexWrap"},direction:{type:"identity",property:"flexDirection"}};var $p={root:"m_8bffd616"};const gx={},ld=gn((n,l)=>{const i=st("Flex",gx,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,gap:m,rowGap:v,columnGap:S,align:_,justify:x,wrap:w,direction:N,...B}=i,O=jt({name:"Flex",classes:$p,props:i,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:y}),j=pn(),z=Pv(),L=Wv({styleProps:{gap:m,rowGap:v,columnGap:S,align:_,justify:x,wrap:w,direction:N},theme:j,data:px});return g.jsxs(g.Fragment,{children:[L.hasResponsiveStyles&&g.jsx(Jv,{selector:`.${z}`,styles:L.styles,media:L.media}),g.jsx(ft,{ref:l,...O("root",{className:z,style:yu(L.inlineStyles)}),...B})]})});ld.classes=$p;ld.displayName="@mantine/core/Flex";var Lp={root:"m_b6d8b162"};function bx(n){if(n==="start")return"start";if(n==="end"||n)return"end"}const Sx={inherit:!1},_x=(n,{variant:l,lineClamp:i,gradient:o,size:c,color:f})=>({root:{"--text-fz":we(c),"--text-lh":Rv(c),"--text-gradient":l==="gradient"?Tf(o,n):void 0,"--text-line-clamp":typeof i=="number"?i.toString():void 0,"--text-color":f?mn(f,n):void 0}}),Wt=gn((n,l)=>{const i=st("Text",Sx,n),{lineClamp:o,truncate:c,inline:f,inherit:d,gradient:h,span:y,__staticSelector:m,vars:v,className:S,style:_,classNames:x,styles:w,unstyled:N,variant:B,mod:O,size:j,...z}=i,L=jt({name:["Text",m],props:i,classes:Lp,className:S,style:_,classNames:x,styles:w,unstyled:N,vars:v,varsResolver:_x});return g.jsx(ft,{...L("root",{focusable:!0}),ref:l,component:y?"span":"p",variant:B,mod:[{"data-truncate":bx(c),"data-line-clamp":typeof o=="number","data-inline":f,"data-inherit":d},O],size:j,...z})});Wt.classes=Lp;Wt.displayName="@mantine/core/Text";var Kl={root:"m_77c9d27d",inner:"m_80f1301b",label:"m_811560b9",section:"m_a74036a",loader:"m_a25b86ee",group:"m_80d6d844",groupSection:"m_70be2a01"};const Q0={orientation:"horizontal"},xx=(n,{borderWidth:l})=>({group:{"--button-border-width":tt(l)}}),rd=_t((n,l)=>{const i=st("ButtonGroup",Q0,n),{className:o,style:c,classNames:f,styles:d,unstyled:h,orientation:y,vars:m,borderWidth:v,variant:S,mod:_,...x}=st("ButtonGroup",Q0,n),w=jt({name:"ButtonGroup",props:i,classes:Kl,className:o,style:c,classNames:f,styles:d,unstyled:h,vars:m,varsResolver:xx,rootSelector:"group"});return g.jsx(ft,{...w("group"),ref:l,variant:S,mod:[{"data-orientation":y},_],role:"group",...x})});rd.classes=Kl;rd.displayName="@mantine/core/ButtonGroup";const K0={},Tx=(n,{radius:l,color:i,gradient:o,variant:c,autoContrast:f,size:d})=>{const h=n.variantColorResolver({color:i||n.primaryColor,theme:n,gradient:o,variant:c||"filled",autoContrast:f});return{groupSection:{"--section-height":Vt(d,"section-height"),"--section-padding-x":Vt(d,"section-padding-x"),"--section-fz":d!=null&&d.includes("compact")?we(d.replace("compact-","")):we(d),"--section-radius":l===void 0?void 0:vn(l),"--section-bg":i||c?h.background:void 0,"--section-color":h.color,"--section-bd":i||c?h.border:void 0}}},id=_t((n,l)=>{const i=st("ButtonGroupSection",K0,n),{className:o,style:c,classNames:f,styles:d,unstyled:h,vars:y,variant:m,gradient:v,radius:S,autoContrast:_,...x}=st("ButtonGroupSection",K0,n),w=jt({name:"ButtonGroupSection",props:i,classes:Kl,className:o,style:c,classNames:f,styles:d,unstyled:h,vars:y,varsResolver:Tx,rootSelector:"groupSection"});return g.jsx(ft,{...w("groupSection"),ref:l,variant:m,...x})});id.classes=Kl;id.displayName="@mantine/core/ButtonGroupSection";const wx={in:{opacity:1,transform:`translate(-50%, calc(-50% + ${tt(1)}))`},out:{opacity:0,transform:"translate(-50%, -200%)"},common:{transformOrigin:"center"},transitionProperty:"transform, opacity"},Ex={},Cx=(n,{radius:l,color:i,gradient:o,variant:c,size:f,justify:d,autoContrast:h})=>{const y=n.variantColorResolver({color:i||n.primaryColor,theme:n,gradient:o,variant:c||"filled",autoContrast:h});return{root:{"--button-justify":d,"--button-height":Vt(f,"button-height"),"--button-padding-x":Vt(f,"button-padding-x"),"--button-fz":f!=null&&f.includes("compact")?we(f.replace("compact-","")):we(f),"--button-radius":l===void 0?void 0:vn(l),"--button-bg":i||c?y.background:void 0,"--button-hover":i||c?y.hover:void 0,"--button-color":y.color,"--button-bd":i||c?y.border:void 0,"--button-hover-color":i||c?y.hoverColor:void 0}}},oe=gn((n,l)=>{const i=st("Button",Ex,n),{style:o,vars:c,className:f,color:d,disabled:h,children:y,leftSection:m,rightSection:v,fullWidth:S,variant:_,radius:x,loading:w,loaderProps:N,gradient:B,classNames:O,styles:j,unstyled:z,"data-disabled":L,autoContrast:$,mod:V,...Z}=i,W=jt({name:"Button",props:i,classes:Kl,className:f,style:o,classNames:O,styles:j,unstyled:z,vars:c,varsResolver:Cx}),it=!!m,ht=!!v;return g.jsxs(Zl,{ref:l,...W("root",{active:!h&&!w&&!L}),unstyled:z,variant:_,disabled:h||w,mod:[{disabled:h||L,loading:w,block:S,"with-left-section":it,"with-right-section":ht},V],...Z,children:[g.jsx(vi,{mounted:!!w,transition:wx,duration:150,children:pt=>g.jsx(ft,{component:"span",...W("loader",{style:pt}),"aria-hidden":!0,children:g.jsx(pi,{color:"var(--button-color)",size:"calc(var(--button-height) / 1.8)",...N})})}),g.jsxs("span",{...W("inner"),children:[m&&g.jsx(ft,{component:"span",...W("section"),mod:{position:"left"},children:m}),g.jsx(ft,{component:"span",mod:{loading:w},...W("label"),children:y}),v&&g.jsx(ft,{component:"span",...W("section"),mod:{position:"right"},children:v})]})]})});oe.classes=Kl;oe.displayName="@mantine/core/Button";oe.Group=rd;oe.GroupSection=id;var Yp={root:"m_7485cace"};const Ax={},jx=(n,{size:l,fluid:i})=>({root:{"--container-size":i?void 0:Vt(l,"container-size")}}),ue=_t((n,l)=>{const i=st("Container",Ax,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,fluid:m,mod:v,...S}=i,_=jt({name:"Container",classes:Yp,props:i,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:y,varsResolver:jx});return g.jsx(ft,{ref:l,mod:[{fluid:m},v],..._("root"),...S})});ue.classes=Yp;ue.displayName="@mantine/core/Container";var qp={root:"m_3eebeb36",label:"m_9e365f20"};const Nx={orientation:"horizontal"},Ox=(n,{color:l,variant:i,size:o})=>({root:{"--divider-color":l?mn(l,n):void 0,"--divider-border-style":i,"--divider-size":Vt(o,"divider-size")}}),Cu=_t((n,l)=>{const i=st("Divider",Nx,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,color:m,orientation:v,label:S,labelPosition:_,mod:x,...w}=i,N=jt({name:"Divider",classes:qp,props:i,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:y,varsResolver:Ox});return g.jsx(ft,{ref:l,mod:[{orientation:v,"with-label":!!S},x],...N("root"),...w,role:"separator",children:S&&g.jsx(ft,{component:"span",mod:{position:_},...N("label"),children:S})})});Cu.classes=qp;Cu.displayName="@mantine/core/Divider";const[Mx,Rx]=Xa("List component was not found in tree");var od={root:"m_abbac491",item:"m_abb6bec2",itemWrapper:"m_75cd9f71",itemIcon:"m_60f83e5b"};const zx={},ud=_t((n,l)=>{const i=st("ListItem",zx,n),{classNames:o,className:c,style:f,styles:d,vars:h,icon:y,children:m,mod:v,...S}=i,_=Rx(),x=y||_.icon,w={classNames:o,styles:d};return g.jsx(ft,{..._.getStyles("item",{...w,className:c,style:f}),component:"li",mod:[{"with-icon":!!x,centered:_.center},v],ref:l,...S,children:g.jsxs("div",{..._.getStyles("itemWrapper",w),children:[x&&g.jsx("span",{..._.getStyles("itemIcon",w),children:x}),g.jsx("span",{..._.getStyles("itemLabel",w),children:m})]})})});ud.classes=od;ud.displayName="@mantine/core/ListItem";const Dx={type:"unordered"},Bx=(n,{size:l,spacing:i})=>({root:{"--list-fz":we(l),"--list-lh":Rv(l),"--list-spacing":Ll(i)}}),Pt=_t((n,l)=>{const i=st("List",Dx,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,children:m,type:v,withPadding:S,icon:_,spacing:x,center:w,listStyleType:N,mod:B,...O}=i,j=jt({name:"List",classes:od,props:i,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:y,varsResolver:Bx});return g.jsx(Mx,{value:{center:w,icon:_,getStyles:j},children:g.jsx(ft,{...j("root",{style:{listStyleType:N}}),component:v==="unordered"?"ul":"ol",mod:[{"with-padding":S},B],ref:l,...O,children:m})})});Pt.classes=od;Pt.displayName="@mantine/core/List";Pt.Item=ud;const[Ux,Jl]=Xa("Modal component was not found in tree");var Gn={root:"m_9df02822",content:"m_54c44539",inner:"m_1f958f16",header:"m_d0e2b9cd"};const Hx={},Au=_t((n,l)=>{const i=st("ModalBody",Hx,n),{classNames:o,className:c,style:f,styles:d,vars:h,...y}=i,m=Jl();return g.jsx(Mp,{ref:l,...m.getStyles("body",{classNames:o,style:f,styles:d,className:c}),...y})});Au.classes=Gn;Au.displayName="@mantine/core/ModalBody";const $x={},ju=_t((n,l)=>{const i=st("ModalCloseButton",$x,n),{classNames:o,className:c,style:f,styles:d,vars:h,...y}=i,m=Jl();return g.jsx(Rp,{ref:l,...m.getStyles("close",{classNames:o,style:f,styles:d,className:c}),...y})});ju.classes=Gn;ju.displayName="@mantine/core/ModalCloseButton";const Lx={},Nu=_t((n,l)=>{const i=st("ModalContent",Lx,n),{classNames:o,className:c,style:f,styles:d,vars:h,children:y,__hidden:m,...v}=i,S=Jl(),_=S.scrollAreaComponent||ex;return g.jsx(zp,{...S.getStyles("content",{className:c,style:f,styles:d,classNames:o}),innerProps:S.getStyles("inner",{className:c,style:f,styles:d,classNames:o}),"data-full-screen":S.fullScreen||void 0,"data-modal-content":!0,"data-hidden":m||void 0,ref:l,...v,children:g.jsx(_,{style:{maxHeight:S.fullScreen?"100dvh":`calc(100dvh - (${tt(S.yOffset)} * 2))`},children:y})})});Nu.classes=Gn;Nu.displayName="@mantine/core/ModalContent";const Yx={},Ou=_t((n,l)=>{const i=st("ModalHeader",Yx,n),{classNames:o,className:c,style:f,styles:d,vars:h,...y}=i,m=Jl();return g.jsx(Dp,{ref:l,...m.getStyles("header",{classNames:o,style:f,styles:d,className:c}),...y})});Ou.classes=Gn;Ou.displayName="@mantine/core/ModalHeader";const qx={},Mu=_t((n,l)=>{const i=st("ModalOverlay",qx,n),{classNames:o,className:c,style:f,styles:d,vars:h,...y}=i,m=Jl();return g.jsx(Bp,{ref:l,...m.getStyles("overlay",{classNames:o,style:f,styles:d,className:c}),...y})});Mu.classes=Gn;Mu.displayName="@mantine/core/ModalOverlay";const Xx={__staticSelector:"Modal",closeOnClickOutside:!0,withinPortal:!0,lockScroll:!0,trapFocus:!0,returnFocus:!0,closeOnEscape:!0,keepMounted:!1,zIndex:kl("modal"),transitionProps:{duration:200,transition:"fade-down"},yOffset:"5dvh"},Gx=(n,{radius:l,size:i,yOffset:o,xOffset:c})=>({root:{"--modal-radius":l===void 0?void 0:vn(l),"--modal-size":Vt(i,"modal-size"),"--modal-y-offset":tt(o),"--modal-x-offset":tt(c)}}),Ru=_t((n,l)=>{const i=st("ModalRoot",Xx,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,yOffset:m,scrollAreaComponent:v,radius:S,fullScreen:_,centered:x,xOffset:w,__staticSelector:N,...B}=i,O=jt({name:N,classes:Gn,props:i,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:y,varsResolver:Gx});return g.jsx(Ux,{value:{yOffset:m,scrollAreaComponent:v,getStyles:O,fullScreen:_},children:g.jsx(Op,{ref:l,...O("root"),"data-full-screen":_||void 0,"data-centered":x||void 0,unstyled:h,...B})})});Ru.classes=Gn;Ru.displayName="@mantine/core/ModalRoot";const[Vx,kx]=Xf();function Xp({children:n}){const[l,i]=C.useState([]),[o,c]=C.useState(kl("modal"));return g.jsx(Vx,{value:{stack:l,addModal:(f,d)=>{i(h=>[...new Set([...h,f])]),c(h=>typeof d=="number"&&typeof h=="number"?Math.max(h,d):h)},removeModal:f=>i(d=>d.filter(h=>h!==f)),getZIndex:f=>`calc(${o} + ${l.indexOf(f)} + 1)`,currentId:l[l.length-1],maxZIndex:o},children:n})}Xp.displayName="@mantine/core/ModalStack";const Zx={},zu=_t((n,l)=>{const i=st("ModalTitle",Zx,n),{classNames:o,className:c,style:f,styles:d,vars:h,...y}=i,m=Jl();return g.jsx(Up,{ref:l,...m.getStyles("title",{classNames:o,style:f,styles:d,className:c}),...y})});zu.classes=Gn;zu.displayName="@mantine/core/ModalTitle";const Qx={closeOnClickOutside:!0,withinPortal:!0,lockScroll:!0,trapFocus:!0,returnFocus:!0,closeOnEscape:!0,keepMounted:!1,zIndex:kl("modal"),transitionProps:{duration:200,transition:"fade-down"},withOverlay:!0,withCloseButton:!0},Pe=_t((n,l)=>{const{title:i,withOverlay:o,overlayProps:c,withCloseButton:f,closeButtonProps:d,children:h,radius:y,opened:m,stackId:v,zIndex:S,..._}=st("Modal",Qx,n),x=kx(),w=!!i||f,N=x&&v?{closeOnEscape:x.currentId===v,trapFocus:x.currentId===v,zIndex:x.getZIndex(v)}:{},B=o===!1?!1:v&&x?x.currentId===v:m;return C.useEffect(()=>{x&&v&&(m?x.addModal(v,S||kl("modal")):x.removeModal(v))},[m,v,S]),g.jsxs(Ru,{ref:l,radius:y,opened:m,zIndex:x&&v?x.getZIndex(v):S,..._,...N,children:[o&&g.jsx(Mu,{visible:B,transitionProps:x&&v?{duration:0}:void 0,...c}),g.jsxs(Nu,{radius:y,__hidden:x&&v&&m?v!==x.currentId:!1,children:[w&&g.jsxs(Ou,{children:[i&&g.jsx(zu,{children:i}),f&&g.jsx(ju,{...d})]}),g.jsx(Au,{children:h})]})]})});Pe.classes=Gn;Pe.displayName="@mantine/core/Modal";Pe.Root=Ru;Pe.Overlay=Mu;Pe.Content=Nu;Pe.Body=Au;Pe.Header=Ou;Pe.Title=zu;Pe.CloseButton=ju;Pe.Stack=Xp;const Kx=({reveal:n})=>g.jsx("svg",{viewBox:"0 0 15 15",fill:"none",xmlns:"http://www.w3.org/2000/svg",style:{width:"var(--psi-icon-size)",height:"var(--psi-icon-size)"},children:g.jsx("path",{d:n?"M13.3536 2.35355C13.5488 2.15829 13.5488 1.84171 13.3536 1.64645C13.1583 1.45118 12.8417 1.45118 12.6464 1.64645L10.6828 3.61012C9.70652 3.21671 8.63759 3 7.5 3C4.30786 3 1.65639 4.70638 0.0760002 7.23501C-0.0253338 7.39715 -0.0253334 7.60288 0.0760014 7.76501C0.902945 9.08812 2.02314 10.1861 3.36061 10.9323L1.64645 12.6464C1.45118 12.8417 1.45118 13.1583 1.64645 13.3536C1.84171 13.5488 2.15829 13.5488 2.35355 13.3536L4.31723 11.3899C5.29348 11.7833 6.36241 12 7.5 12C10.6921 12 13.3436 10.2936 14.924 7.76501C15.0253 7.60288 15.0253 7.39715 14.924 7.23501C14.0971 5.9119 12.9769 4.81391 11.6394 4.06771L13.3536 2.35355ZM9.90428 4.38861C9.15332 4.1361 8.34759 4 7.5 4C4.80285 4 2.52952 5.37816 1.09622 7.50001C1.87284 8.6497 2.89609 9.58106 4.09974 10.1931L9.90428 4.38861ZM5.09572 10.6114L10.9003 4.80685C12.1039 5.41894 13.1272 6.35031 13.9038 7.50001C12.4705 9.62183 10.1971 11 7.5 11C6.65241 11 5.84668 10.8639 5.09572 10.6114Z":"M7.5 11C4.80285 11 2.52952 9.62184 1.09622 7.50001C2.52952 5.37816 4.80285 4 7.5 4C10.1971 4 12.4705 5.37816 13.9038 7.50001C12.4705 9.62183 10.1971 11 7.5 11ZM7.5 3C4.30786 3 1.65639 4.70638 0.0760002 7.23501C-0.0253338 7.39715 -0.0253334 7.60288 0.0760014 7.76501C1.65639 10.2936 4.30786 12 7.5 12C10.6921 12 13.3436 10.2936 14.924 7.76501C15.0253 7.60288 15.0253 7.39715 14.924 7.23501C13.3436 4.70638 10.6921 3 7.5 3ZM7.5 9.5C8.60457 9.5 9.5 8.60457 9.5 7.5C9.5 6.39543 8.60457 5.5 7.5 5.5C6.39543 5.5 5.5 6.39543 5.5 7.5C5.5 8.60457 6.39543 9.5 7.5 9.5Z",fill:"currentColor",fillRule:"evenodd",clipRule:"evenodd"})});var jf={root:"m_f61ca620",input:"m_ccf8da4c",innerInput:"m_f2d85dd2",visibilityToggle:"m_b1072d44"};const Jx={visibilityToggleIcon:Kx},Wx=(n,{size:l})=>({root:{"--psi-icon-size":Vt(l,"psi-icon-size"),"--psi-button-size":Vt(l,"psi-button-size")}}),cd=_t((n,l)=>{const i=st("PasswordInput",Jx,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,required:m,error:v,leftSection:S,disabled:_,id:x,variant:w,inputContainer:N,description:B,label:O,size:j,errorProps:z,descriptionProps:L,labelProps:$,withAsterisk:V,inputWrapperOrder:Z,wrapperProps:W,radius:it,rightSection:ht,rightSectionWidth:pt,rightSectionPointerEvents:at,leftSectionWidth:lt,visible:R,defaultVisible:Q,onVisibilityChange:E,visibilityToggleIcon:Y,visibilityToggleButtonProps:k,rightSectionProps:et,leftSectionProps:T,leftSectionPointerEvents:X,withErrorStyles:I,mod:P,...nt}=i,ut=pu(x),[rt,bt]=$v({value:R,defaultValue:Q,finalValue:!1,onChange:E}),vt=()=>bt(!rt),Ut=jt({name:"PasswordInput",classes:jf,props:i,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:y,varsResolver:Wx}),{resolvedClassNames:ae,resolvedStyles:pe}=Kv({classNames:o,styles:d,props:i}),{styleProps:$e,rest:ln}=bu(nt),Sn=Y,pa=g.jsx(gi,{...Ut("visibilityToggle"),disabled:_,radius:it,"aria-hidden":!k,tabIndex:-1,...k,variant:(k==null?void 0:k.variant)??"subtle",color:"gray",unstyled:h,onTouchEnd:ce=>{var Ht;ce.preventDefault(),(Ht=k==null?void 0:k.onTouchEnd)==null||Ht.call(k,ce),vt()},onMouseDown:ce=>{var Ht;ce.preventDefault(),(Ht=k==null?void 0:k.onMouseDown)==null||Ht.call(k,ce),vt()},onKeyDown:ce=>{var Ht;(Ht=k==null?void 0:k.onKeyDown)==null||Ht.call(k,ce),ce.key===" "&&(ce.preventDefault(),vt())},children:g.jsx(Sn,{reveal:rt})});return g.jsx(Qt.Wrapper,{required:m,id:ut,label:O,error:v,description:B,size:j,classNames:ae,styles:pe,__staticSelector:"PasswordInput",errorProps:z,descriptionProps:L,unstyled:h,withAsterisk:V,inputWrapperOrder:Z,inputContainer:N,variant:w,labelProps:{...$,htmlFor:ut},mod:P,...Ut("root"),...$e,...W,children:g.jsx(Qt,{component:"div",error:v,leftSection:S,size:j,classNames:{...ae,input:Ce(jf.input,ae.input)},styles:pe,radius:it,disabled:_,__staticSelector:"PasswordInput",rightSectionWidth:pt,rightSection:ht??pa,variant:w,unstyled:h,leftSectionWidth:lt,rightSectionPointerEvents:at||"all",rightSectionProps:et,leftSectionProps:T,leftSectionPointerEvents:X,withAria:!1,withErrorStyles:I,children:g.jsx("input",{required:m,"data-invalid":!!v||void 0,"data-with-left-section":!!S||void 0,...Ut("innerInput"),disabled:_,id:ut,ref:l,...ln,autoComplete:ln.autoComplete||"off",type:rt?"text":"password"})})})});cd.classes={...bi.classes,...jf};cd.displayName="@mantine/core/PasswordInput";const[Px,Fx]=Xa("Table component was not found in the tree");var Si={table:"m_b23fa0ef",th:"m_4e7aa4f3",tr:"m_4e7aa4fd",td:"m_4e7aa4ef",tbody:"m_b2404537",thead:"m_b242d975",caption:"m_9e5a3ac7",scrollContainer:"m_a100c15",scrollContainerInner:"m_62259741"};function Ix(n,l){if(!l)return;const i={};return l.columnBorder&&n.withColumnBorders&&(i["data-with-column-border"]=!0),l.rowBorder&&n.withRowBorders&&(i["data-with-row-border"]=!0),l.striped&&n.striped&&(i["data-striped"]=n.striped),l.highlightOnHover&&n.highlightOnHover&&(i["data-hover"]=!0),l.captionSide&&n.captionSide&&(i["data-side"]=n.captionSide),l.stickyHeader&&n.stickyHeader&&(i["data-sticky"]=!0),i}function ka(n,l){const i=`Table${n.charAt(0).toUpperCase()}${n.slice(1)}`,o=_t((c,f)=>{const d=st(i,{},c),{classNames:h,className:y,style:m,styles:v,...S}=d,_=Fx();return g.jsx(ft,{component:n,ref:f,...Ix(_,l),..._.getStyles(n,{className:y,classNames:h,style:m,styles:v,props:d}),...S})});return o.displayName=`@mantine/core/${i}`,o.classes=Si,o}const Nf=ka("th",{columnBorder:!0}),Gp=ka("td",{columnBorder:!0}),Io=ka("tr",{rowBorder:!0,striped:!0,highlightOnHover:!0}),Vp=ka("thead",{stickyHeader:!0}),kp=ka("tbody"),Zp=ka("tfoot"),Qp=ka("caption",{captionSide:!0});function sd({data:n}){return g.jsxs(g.Fragment,{children:[n.caption&&g.jsx(Qp,{children:n.caption}),n.head&&g.jsx(Vp,{children:g.jsx(Io,{children:n.head.map((l,i)=>g.jsx(Nf,{children:l},i))})}),n.body&&g.jsx(kp,{children:n.body.map((l,i)=>g.jsx(Io,{children:l.map((o,c)=>g.jsx(Gp,{children:o},c))},i))}),n.foot&&g.jsx(Zp,{children:g.jsx(Io,{children:n.foot.map((l,i)=>g.jsx(Nf,{children:l},i))})})]})}sd.displayName="@mantine/core/TableDataRenderer";const t3={type:"scrollarea"},e3=(n,{minWidth:l,type:i})=>({scrollContainer:{"--table-min-width":tt(l),"--table-overflow":i==="native"?"auto":void 0}}),fd=_t((n,l)=>{const i=st("TableScrollContainer",t3,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,children:m,minWidth:v,type:S,..._}=i,x=jt({name:"TableScrollContainer",classes:Si,props:i,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:y,varsResolver:e3,rootSelector:"scrollContainer"});return g.jsx(ft,{component:S==="scrollarea"?yi:"div",...S==="scrollarea"?{offsetScrollbars:"x"}:{},ref:l,...x("scrollContainer"),..._,children:g.jsx("div",{...x("scrollContainerInner"),children:m})})});fd.classes=Si;fd.displayName="@mantine/core/TableScrollContainer";const n3={withRowBorders:!0,verticalSpacing:7},a3=(n,{layout:l,captionSide:i,horizontalSpacing:o,verticalSpacing:c,borderColor:f,stripedColor:d,highlightOnHoverColor:h,striped:y,highlightOnHover:m,stickyHeaderOffset:v,stickyHeader:S})=>({table:{"--table-layout":l,"--table-caption-side":i,"--table-horizontal-spacing":Ll(o),"--table-vertical-spacing":Ll(c),"--table-border-color":f?mn(f,n):void 0,"--table-striped-color":y&&d?mn(d,n):void 0,"--table-highlight-on-hover-color":m&&h?mn(h,n):void 0,"--table-sticky-header-offset":S?tt(v):void 0}}),ne=_t((n,l)=>{const i=st("Table",n3,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,vars:y,horizontalSpacing:m,verticalSpacing:v,captionSide:S,stripedColor:_,highlightOnHoverColor:x,striped:w,highlightOnHover:N,withColumnBorders:B,withRowBorders:O,withTableBorder:j,borderColor:z,layout:L,variant:$,data:V,children:Z,stickyHeader:W,stickyHeaderOffset:it,mod:ht,tabularNums:pt,...at}=i,lt=jt({name:"Table",props:i,className:c,style:f,classes:Si,classNames:o,styles:d,unstyled:h,rootSelector:"table",vars:y,varsResolver:a3});return g.jsx(Px,{value:{getStyles:lt,stickyHeader:W,striped:w===!0?"odd":w||void 0,highlightOnHover:N,withColumnBorders:B,withRowBorders:O,captionSide:S||"bottom"},children:g.jsx(ft,{component:"table",variant:$,ref:l,mod:[{"data-with-table-border":j,"data-tabular-nums":pt},ht],...lt("table"),...at,children:Z||!!V&&g.jsx(sd,{data:V})})})});ne.classes=Si;ne.displayName="@mantine/core/Table";ne.Td=Gp;ne.Th=Nf;ne.Tr=Io;ne.Thead=Vp;ne.Tbody=kp;ne.Tfoot=Zp;ne.Caption=Qp;ne.ScrollContainer=fd;ne.DataRenderer=sd;const[l3,dd]=Xa("Tabs component was not found in the tree");var _i={root:"m_89d60db1","list--default":"m_576c9d4",list:"m_89d33d6d",panel:"m_b0c91715",tab:"m_4ec4dce6",tabSection:"m_fc420b1f","tab--default":"m_539e827b","list--outline":"m_6772fbd5","tab--outline":"m_b59ab47c","tab--pills":"m_c3381914"};const r3={},hd=_t((n,l)=>{const i=st("TabsList",r3,n),{children:o,className:c,grow:f,justify:d,classNames:h,styles:y,style:m,mod:v,...S}=i,_=dd();return g.jsx(ft,{...S,..._.getStyles("list",{className:c,style:m,classNames:h,styles:y,props:i,variant:_.variant}),ref:l,role:"tablist",variant:_.variant,mod:[{grow:f,orientation:_.orientation,placement:_.orientation==="vertical"&&_.placement,inverted:_.inverted},v],"aria-orientation":_.orientation,__vars:{"--tabs-justify":d},children:o})});hd.classes=_i;hd.displayName="@mantine/core/TabsList";const i3={},md=_t((n,l)=>{const i=st("TabsPanel",i3,n),{children:o,className:c,value:f,classNames:d,styles:h,style:y,mod:m,keepMounted:v,...S}=i,_=dd(),x=_.value===f,w=_.keepMounted||v||x?o:null;return g.jsx(ft,{...S,..._.getStyles("panel",{className:c,classNames:d,styles:h,style:[y,x?void 0:{display:"none"}],props:i}),ref:l,mod:[{orientation:_.orientation},m],role:"tabpanel",id:_.getPanelId(f),"aria-labelledby":_.getTabId(f),children:w})});md.classes=_i;md.displayName="@mantine/core/TabsPanel";const o3={},yd=_t((n,l)=>{const i=st("TabsTab",o3,n),{className:o,children:c,rightSection:f,leftSection:d,value:h,onClick:y,onKeyDown:m,disabled:v,color:S,style:_,classNames:x,styles:w,vars:N,mod:B,tabIndex:O,...j}=i,z=pn(),{dir:L}=np(),$=dd(),V=h===$.value,Z=it=>{$.onChange($.allowTabDeactivation&&h===$.value?null:h),y==null||y(it)},W={classNames:x,styles:w,props:i};return g.jsxs(Zl,{...j,...$.getStyles("tab",{className:o,style:_,variant:$.variant,...W}),disabled:v,unstyled:$.unstyled,variant:$.variant,mod:[{active:V,disabled:v,orientation:$.orientation,inverted:$.inverted,placement:$.orientation==="vertical"&&$.placement},B],ref:l,role:"tab",id:$.getTabId(h),"aria-selected":V,tabIndex:O!==void 0?O:V||$.value===null?0:-1,"aria-controls":$.getPanelId(h),onClick:Z,__vars:{"--tabs-color":S?mn(S,z):void 0},onKeyDown:YS({siblingSelector:'[role="tab"]',parentSelector:'[role="tablist"]',activateOnFocus:$.activateTabWithKeyboard,loop:$.loop,orientation:$.orientation||"horizontal",dir:L,onKeyDown:m}),children:[d&&g.jsx("span",{...$.getStyles("tabSection",W),"data-position":"left",children:d}),c&&g.jsx("span",{...$.getStyles("tabLabel",W),children:c}),f&&g.jsx("span",{...$.getStyles("tabSection",W),"data-position":"right",children:f})]})});yd.classes=_i;yd.displayName="@mantine/core/TabsTab";const J0="Tabs.Tab or Tabs.Panel component was rendered with invalid value or without value",u3={keepMounted:!0,orientation:"horizontal",loop:!0,activateTabWithKeyboard:!0,allowTabDeactivation:!1,unstyled:!1,inverted:!1,variant:"default",placement:"left"},c3=(n,{radius:l,color:i,autoContrast:o})=>({root:{"--tabs-radius":vn(l),"--tabs-color":mn(i,n),"--tabs-text-color":K2(o,n)?Vv({color:i,theme:n,autoContrast:o}):void 0}}),ve=_t((n,l)=>{const i=st("Tabs",u3,n),{defaultValue:o,value:c,onChange:f,orientation:d,children:h,loop:y,id:m,activateTabWithKeyboard:v,allowTabDeactivation:S,variant:_,color:x,radius:w,inverted:N,placement:B,keepMounted:O,classNames:j,styles:z,unstyled:L,className:$,style:V,vars:Z,autoContrast:W,mod:it,...ht}=i,pt=pu(m),[at,lt]=$v({value:c,defaultValue:o,finalValue:null,onChange:f}),R=jt({name:"Tabs",props:i,classes:_i,className:$,style:V,classNames:j,styles:z,unstyled:L,vars:Z,varsResolver:c3});return g.jsx(l3,{value:{placement:B,value:at,orientation:d,id:pt,loop:y,activateTabWithKeyboard:v,getTabId:M0(`${pt}-tab`,J0),getPanelId:M0(`${pt}-panel`,J0),onChange:lt,allowTabDeactivation:S,variant:_,color:x,radius:w,inverted:N,keepMounted:O,unstyled:L,getStyles:R},children:g.jsx(ft,{ref:l,id:pt,variant:_,mod:[{orientation:d,inverted:d==="horizontal"&&N,placement:d==="vertical"&&B},it],...R("root"),...ht,children:h})})});ve.classes=_i;ve.displayName="@mantine/core/Tabs";ve.Tab=yd;ve.Panel=md;ve.List=hd;const s3={},vd=_t((n,l)=>{const i=st("TextInput",s3,n);return g.jsx(bi,{component:"input",ref:l,...i,__staticSelector:"TextInput"})});vd.classes=bi.classes;vd.displayName="@mantine/core/TextInput";const f3=["h1","h2","h3","h4","h5","h6"],d3=["xs","sm","md","lg","xl"];function h3(n,l){const i=l!==void 0?l:`h${n}`;return f3.includes(i)?{fontSize:`var(--mantine-${i}-font-size)`,fontWeight:`var(--mantine-${i}-font-weight)`,lineHeight:`var(--mantine-${i}-line-height)`}:d3.includes(i)?{fontSize:`var(--mantine-font-size-${i})`,fontWeight:`var(--mantine-h${n}-font-weight)`,lineHeight:`var(--mantine-h${n}-line-height)`}:{fontSize:tt(i),fontWeight:`var(--mantine-h${n}-font-weight)`,lineHeight:`var(--mantine-h${n}-line-height)`}}var Kp={root:"m_8a5d1357"};const m3={order:1},y3=(n,{order:l,size:i,lineClamp:o,textWrap:c})=>{const f=h3(l,i);return{root:{"--title-fw":f.fontWeight,"--title-lh":f.lineHeight,"--title-fz":f.fontSize,"--title-line-clamp":typeof o=="number"?o.toString():void 0,"--title-text-wrap":c}}},Za=_t((n,l)=>{const i=st("Title",m3,n),{classNames:o,className:c,style:f,styles:d,unstyled:h,order:y,vars:m,size:v,variant:S,lineClamp:_,textWrap:x,mod:w,...N}=i,B=jt({name:"Title",props:i,classes:Kp,className:c,style:f,classNames:o,styles:d,unstyled:h,vars:m,varsResolver:y3});return[1,2,3,4,5,6].includes(y)?g.jsx(ft,{...B("root"),component:`h${y}`,variant:S,ref:l,mod:[{order:y,"data-line-clamp":typeof _=="number"},w],size:v,...N}):null});Za.classes=Kp;Za.displayName="@mantine/core/Title";var Ir={},W0;function v3(){if(W0)return Ir;W0=1,Object.defineProperty(Ir,"__esModule",{value:!0}),Ir.parse=d,Ir.serialize=m;const n=/^[\u0021-\u003A\u003C\u003E-\u007E]+$/,l=/^[\u0021-\u003A\u003C-\u007E]*$/,i=/^([.]?[a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?)([.][a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?)*$/i,o=/^[\u0020-\u003A\u003D-\u007E]*$/,c=Object.prototype.toString,f=(()=>{const _=function(){};return _.prototype=Object.create(null),_})();function d(_,x){const w=new f,N=_.length;if(N<2)return w;const B=(x==null?void 0:x.decode)||v;let O=0;do{const j=_.indexOf("=",O);if(j===-1)break;const z=_.indexOf(";",O),L=z===-1?N:z;if(j>L){O=_.lastIndexOf(";",j-1)+1;continue}const $=h(_,O,j),V=y(_,j,$),Z=_.slice($,V);if(w[Z]===void 0){let W=h(_,j+1,L),it=y(_,L,W);const ht=B(_.slice(W,it));w[Z]=ht}O=L+1}while(O<N);return w}function h(_,x,w){do{const N=_.charCodeAt(x);if(N!==32&&N!==9)return x}while(++x<w);return w}function y(_,x,w){for(;x>w;){const N=_.charCodeAt(--x);if(N!==32&&N!==9)return x+1}return w}function m(_,x,w){const N=(w==null?void 0:w.encode)||encodeURIComponent;if(!n.test(_))throw new TypeError(`argument name is invalid: ${_}`);const B=N(x);if(!l.test(B))throw new TypeError(`argument val is invalid: ${x}`);let O=_+"="+B;if(!w)return O;if(w.maxAge!==void 0){if(!Number.isInteger(w.maxAge))throw new TypeError(`option maxAge is invalid: ${w.maxAge}`);O+="; Max-Age="+w.maxAge}if(w.domain){if(!i.test(w.domain))throw new TypeError(`option domain is invalid: ${w.domain}`);O+="; Domain="+w.domain}if(w.path){if(!o.test(w.path))throw new TypeError(`option path is invalid: ${w.path}`);O+="; Path="+w.path}if(w.expires){if(!S(w.expires)||!Number.isFinite(w.expires.valueOf()))throw new TypeError(`option expires is invalid: ${w.expires}`);O+="; Expires="+w.expires.toUTCString()}if(w.httpOnly&&(O+="; HttpOnly"),w.secure&&(O+="; Secure"),w.partitioned&&(O+="; Partitioned"),w.priority)switch(typeof w.priority=="string"?w.priority.toLowerCase():void 0){case"low":O+="; Priority=Low";break;case"medium":O+="; Priority=Medium";break;case"high":O+="; Priority=High";break;default:throw new TypeError(`option priority is invalid: ${w.priority}`)}if(w.sameSite)switch(typeof w.sameSite=="string"?w.sameSite.toLowerCase():w.sameSite){case!0:case"strict":O+="; SameSite=Strict";break;case"lax":O+="; SameSite=Lax";break;case"none":O+="; SameSite=None";break;default:throw new TypeError(`option sameSite is invalid: ${w.sameSite}`)}return O}function v(_){if(_.indexOf("%")===-1)return _;try{return decodeURIComponent(_)}catch{return _}}function S(_){return c.call(_)==="[object Date]"}return Ir}var tu=v3();function p3(){const n=typeof global>"u"?void 0:global.TEST_HAS_DOCUMENT_COOKIE;return typeof n=="boolean"?n:typeof document=="object"&&typeof document.cookie=="string"}function g3(n){return typeof n=="string"?tu.parse(n):typeof n=="object"&&n!==null?n:{}}function mf(n,l={}){const i=b3(n);if(!l.doNotParse)try{return JSON.parse(i)}catch{}return n}function b3(n){return n&&n[0]==="j"&&n[1]===":"?n.substr(2):n}class S3{constructor(l,i={}){this.changeListeners=[],this.HAS_DOCUMENT_COOKIE=!1,this.update=()=>{if(!this.HAS_DOCUMENT_COOKIE)return;const c=this.cookies;this.cookies=tu.parse(document.cookie),this._checkChanges(c)};const o=typeof document>"u"?"":document.cookie;this.cookies=g3(l||o),this.defaultSetOptions=i,this.HAS_DOCUMENT_COOKIE=p3()}_emitChange(l){for(let i=0;i<this.changeListeners.length;++i)this.changeListeners[i](l)}_checkChanges(l){new Set(Object.keys(l).concat(Object.keys(this.cookies))).forEach(o=>{l[o]!==this.cookies[o]&&this._emitChange({name:o,value:mf(this.cookies[o])})})}_startPolling(){this.pollingInterval=setInterval(this.update,300)}_stopPolling(){this.pollingInterval&&clearInterval(this.pollingInterval)}get(l,i={}){return i.doNotUpdate||this.update(),mf(this.cookies[l],i)}getAll(l={}){l.doNotUpdate||this.update();const i={};for(let o in this.cookies)i[o]=mf(this.cookies[o],l);return i}set(l,i,o){o?o=Object.assign(Object.assign({},this.defaultSetOptions),o):o=this.defaultSetOptions;const c=typeof i=="string"?i:JSON.stringify(i);this.cookies=Object.assign(Object.assign({},this.cookies),{[l]:c}),this.HAS_DOCUMENT_COOKIE&&(document.cookie=tu.serialize(l,c,o)),this._emitChange({name:l,value:i,options:o})}remove(l,i){const o=i=Object.assign(Object.assign(Object.assign({},this.defaultSetOptions),i),{expires:new Date(1970,1,1,0,0,1),maxAge:0});this.cookies=Object.assign({},this.cookies),delete this.cookies[l],this.HAS_DOCUMENT_COOKIE&&(document.cookie=tu.serialize(l,"",o)),this._emitChange({name:l,value:void 0,options:i})}addChangeListener(l){this.changeListeners.push(l),this.HAS_DOCUMENT_COOKIE&&this.changeListeners.length===1&&(typeof window=="object"&&"cookieStore"in window?window.cookieStore.addEventListener("change",this.update):this._startPolling())}removeChangeListener(l){const i=this.changeListeners.indexOf(l);i>=0&&this.changeListeners.splice(i,1),this.HAS_DOCUMENT_COOKIE&&this.changeListeners.length===0&&(typeof window=="object"&&"cookieStore"in window?window.cookieStore.removeEventListener("change",this.update):this._stopPolling())}removeAllChangeListeners(){for(;this.changeListeners.length>0;)this.removeChangeListener(this.changeListeners[0])}}const Zo=window.location.origin,Hn=new S3(null,{path:"/"}),$n={get:async(n,l={})=>{const i={"Content-Type":"application/json",...l};Hn.get("token")&&(i.Authorization=`Bearer ${Hn.get("token")}`);const o=await fetch(`${Zo}${n}`,{method:"GET",headers:i});if(!o.ok){const c=await o.json().catch(()=>null),f=new Error(`HTTP error ${o.status}: ${o.statusText}`);throw f.status=o.status,f.headers=Object.fromEntries(o.headers.entries()),f.data=c,f}return await o.json()},post:async(n,l,i={})=>{let o=l;l instanceof URLSearchParams||(o=JSON.stringify(l));const c={"Content-Type":"application/json",...i};Hn.get("token")&&(c.Authorization=`Bearer ${Hn.get("token")}`);const f=await fetch(`${Zo}${n}`,{method:"POST",headers:c,body:o});if(!f.ok){const d=await f.json().catch(()=>null),h=new Error(`HTTP error ${f.status}: ${f.statusText}`);throw h.status=f.status,h.headers=Object.fromEntries(f.headers.entries()),h.data=d,h}return await f.json()},put:async(n,l,i={})=>{let o=l;l instanceof URLSearchParams||(o=JSON.stringify(l));const c={"Content-Type":"application/json",...i};Hn.get("token")&&(c.Authorization=`Bearer ${Hn.get("token")}`);const f=await fetch(`${Zo}${n}`,{method:"PUT",headers:c,body:o});if(!f.ok){const d=await f.json().catch(()=>null),h=new Error(`HTTP error ${f.status}: ${f.statusText}`);throw h.status=f.status,h.headers=Object.fromEntries(f.headers.entries()),h.data=d,h}return await f.json()},delete:async(n,l={})=>{const i={...l};Hn.get("token")&&(i.Authorization=`Bearer ${Hn.get("token")}`);const o=await fetch(`${Zo}${n}`,{method:"DELETE",headers:i});if(!o.ok){const c=await o.json().catch(()=>null),f=new Error(`HTTP error ${o.status}: ${o.statusText}`);throw f.status=o.status,f.headers=Object.fromEntries(o.headers.entries()),f.data=c,f}return await o.json()}},Ln="/autodla-admin/",xi=n=>({get_all:async(l=10,i=0)=>await $n.get(`${Ln}${n}/list?limit=${l}&skip=${i}`),get_by_id:async l=>await $n.get(`${Ln}${n}/get/${l}`),get_history:async l=>await $n.get(`${Ln}${n}/get_history/${l}`),get_tables:async()=>await $n.get(`${Ln}${n}/table`),create:async l=>await $n.post(`${Ln}${n}/new`,l),edit:async(l,i)=>await $n.put(`${Ln}${n}/edit/${l}`,i),delete:async l=>await $n.delete(`${Ln}${n}/delete/${l}`)}),_3=async()=>await $n.get(`${Ln}admin/get_json_schema`),x3=async(n,l)=>await $n.post(`${Ln}admin/token`,new URLSearchParams({username:n,password:l,grant_type:"password"}),{"Content-Type":"application/x-www-form-urlencoded"});function ui(n){const l=d=>[!0,!1].includes(d)?{true:"TRUE",false:"FALSE"}[d]:d,i=n.data;if(i.length==0)return g.jsx("div",{});const o=Object.keys(i[0]),c=g.jsx(ne.Tr,{children:o.map(d=>g.jsx(ne.Th,{children:d},crypto.randomUUID()))}),f=i.map(d=>g.jsx(ne.Tr,{onClick:()=>n.onRowSelect(d),children:Object.keys(d).map(h=>g.jsx(ne.Td,{children:l(d[h])},crypto.randomUUID()))},crypto.randomUUID()));return g.jsxs(ne,{highlightOnHover:!0,withColumnBorders:!0,mt:"sm",children:[g.jsx(ne.Thead,{bg:"neutral-gray.4",children:c}),g.jsx(ne.Tbody,{children:f})]})}/**
 * @license @tabler/icons-react v3.31.0 - MIT
 *
 * This source code is licensed under the MIT license.
 * See the LICENSE file in the root directory of this source tree.
 */var T3={outline:{xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:2,strokeLinecap:"round",strokeLinejoin:"round"},filled:{xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"currentColor",stroke:"none"}};/**
 * @license @tabler/icons-react v3.31.0 - MIT
 *
 * This source code is licensed under the MIT license.
 * See the LICENSE file in the root directory of this source tree.
 */const pd=(n,l,i,o)=>{const c=C.forwardRef(({color:f="currentColor",size:d=24,stroke:h=2,title:y,className:m,children:v,...S},_)=>C.createElement("svg",{ref:_,...T3[n],width:d,height:d,className:["tabler-icon",`tabler-icon-${l}`,m].join(" "),strokeWidth:h,stroke:f,...S},[y&&C.createElement("title",{key:"svg-title"},y),...o.map(([x,w])=>C.createElement(x,w)),...Array.isArray(v)?v:[v]]));return c.displayName=`${i}`,c};/**
 * @license @tabler/icons-react v3.31.0 - MIT
 *
 * This source code is licensed under the MIT license.
 * See the LICENSE file in the root directory of this source tree.
 */var w3=pd("outline","edit","IconEdit",[["path",{d:"M7 7h-1a2 2 0 0 0 -2 2v9a2 2 0 0 0 2 2h9a2 2 0 0 0 2 -2v-1",key:"svg-0"}],["path",{d:"M20.385 6.585a2.1 2.1 0 0 0 -2.97 -2.97l-8.415 8.385v3h3l8.385 -8.415z",key:"svg-1"}],["path",{d:"M16 5l3 3",key:"svg-2"}]]);/**
 * @license @tabler/icons-react v3.31.0 - MIT
 *
 * This source code is licensed under the MIT license.
 * See the LICENSE file in the root directory of this source tree.
 */var E3=pd("outline","reload","IconReload",[["path",{d:"M19.933 13.041a8 8 0 1 1 -9.925 -8.788c3.899 -1 7.935 1.007 9.425 4.747",key:"svg-0"}],["path",{d:"M20 4v5h-5",key:"svg-1"}]]);/**
 * @license @tabler/icons-react v3.31.0 - MIT
 *
 * This source code is licensed under the MIT license.
 * See the LICENSE file in the root directory of this source tree.
 */var P0=pd("outline","trash","IconTrash",[["path",{d:"M4 7l16 0",key:"svg-0"}],["path",{d:"M10 11l0 6",key:"svg-1"}],["path",{d:"M14 11l0 6",key:"svg-2"}],["path",{d:"M5 7l1 12a2 2 0 0 0 2 2h8a2 2 0 0 0 2 -2l1 -12",key:"svg-3"}],["path",{d:"M9 7v-3a1 1 0 0 1 1 -1h4a1 1 0 0 1 1 1v3",key:"svg-4"}]]);const Xl=n=>g.jsx(ue,{fluid:!0,...n,children:g.jsxs(Ee,{gap:"0",children:[n.onReload&&g.jsx(oe,{size:"sm",radius:"xl",px:"xs",variant:"light",onClick:n.onReload,children:g.jsx(E3,{})}),g.jsx(ue,{fluid:!0,children:g.jsx(Za,{order:4,children:n.children})}),n.onClose&&g.jsx(Va,{onClick:n.onClose})]})}),Jp=n=>{const l={};Object.keys(n.schema[n.obj]).map(m=>l[m]="---");const[i,o]=C.useState([l]),[c,f]=C.useState(0),d=xi(n.obj),h=async(m=0)=>{const v=await d.get_all(10,m*10);let S=[];v.map(_=>{const x={};Object.keys(_).map(w=>{if(n.schema[n.obj][w].type.slice(0,4)=="list"||n.schema[n.obj][w].depends){x[w]=g.jsx("strong",{children:n.schema[n.obj][w].type});return}x[w]=_[w]}),S.push(x)}),S.length===0&&(S=[l]),o(S)},y=m=>{m!=null&&n.onSelect(m),n.removeChildFunc()};return C.useEffect(()=>{C.useEffect(()=>{isNaN(c)?f(0):h(c)},[c])},[]),g.jsxs(ue,{fluid:!0,px:"0",children:[g.jsxs(Xl,{onReload:()=>h(c),onClose:()=>y(),mt:"lg",children:["Select ",n.obj]}),g.jsx(ui,{data:i,onRowSelect:y}),g.jsxs(Ee,{children:[g.jsx(oe,{onClick:()=>f(c-1),children:"<"}),g.jsx(Wt,{children:c}),g.jsx(oe,{onClick:()=>f(c+1),children:">"})]})]})},C3=n=>{const l=n.schema[n.obj],i=Object.keys(l).filter(j=>l[j].type=="primary_key")[0];let o={};Object.keys(l).filter(j=>l[j].nullable).map(j=>o[j]=null),o={...o,...n.data};const[c,f]=C.useState({}),[d,h]=C.useState({}),[y,m]=C.useState(o),[v,S]=C.useState(!1),_=j=>{f(L=>(L&&n.removeChildFunc(L),j))},x=(j,z,L)=>{L(V=>({...V,[j]:z(V[j])}))},w=xi(n.obj);let N=!1;const B=()=>{const j=async L=>{if(!N){N=!0,await S(!0);try{const $={};Object.keys(L).map(V=>{const Z=l[V];if(["int","float"].includes(Z.type))$[V]=JSON.parse(L[V]);else if(Z.is_list)try{$[V]=JSON.parse(L[V])}catch{$[V]=L[V]}else $[V]=L[V]}),console.log($),await w.edit(n.data[i],$),n.onUpdate(),n.removeChildFunc()}catch($){$.data&&$.data.detail?(h({}),$.data.detail.map(V=>{x(V.loc[1],Z=>V.msg,h)})):console.log($)}finally{N=!1,S(!1)}}};m(L=>(j(L),L))},O=(j,z)=>{const L=crypto.randomUUID();n.addChildFunc({ele:Jp,props:{obj:j,schema:n.schema,onSelect:z}},L),_(L)};return g.jsxs(ue,{fluid:!0,px:"0",children:[g.jsxs(Xl,{onClose:()=>n.removeChildFunc(),mt:"lg",children:["Edit ",n.obj]}),g.jsxs(ue,{fluid:!0,children:[Object.keys(l).map(j=>g.jsx(ft,{mt:"md",children:g.jsx(Qt.Wrapper,{label:g.jsxs(Ee,{children:[g.jsx(Wt,{fw:"bold",children:j}),g.jsxs(Wt,{c:"yellow",children:["< ",l[j].is_list?`list[${l[j].type}]`:l[j].type," >"]}),l[j].nullable&&g.jsx(Wt,{c:"red",children:"< nullable >"})]}),error:d[j],children:l[j].primary_key?g.jsx(Qt,{placeholder:"auto generated",disabled:!0,value:n.data[i]}):l[j].depends?l[j].is_list?g.jsx(Ue,{withBorder:!0,p:"sm",style:d[j]?{borderColor:"red"}:{},onClick:()=>O(l[j].depends.slice(5),z=>x(j,L=>[...L||[],z],m)),children:y[j]?g.jsx(Pt,{spacing:"sm",children:y[j].map(z=>g.jsx(Pt.Item,{children:g.jsx(Ue,{withBorder:!0,p:"sm",children:g.jsxs(Ee,{children:[g.jsx(Pt,{listStyleType:"disc",children:Object.keys(z).map(L=>g.jsxs(Pt.Item,{children:[g.jsx("strong",{children:L}),": ",JSON.stringify(z[L])]}))}),g.jsx(Va,{})]})})}))}):g.jsx("div",{})}):g.jsx(Ue,{withBorder:!0,p:"sm",style:d[j]?{borderColor:"red"}:{},onClick:()=>O(l[j].depends.slice(5),z=>x(j,L=>z,m)),children:y[j]?g.jsx(Ue,{withBorder:!0,p:"sm",children:g.jsx(Pt,{listStyleType:"disc",children:Object.keys(y[j]).map(z=>g.jsxs(Pt.Item,{children:[g.jsx("strong",{children:z}),": ",JSON.stringify(y[j][z])]}))})}):g.jsx("div",{})}):g.jsx(Qt,{placeholder:j,onChange:z=>x(j,L=>z.target.value,m),error:d[j],value:y[j]})})})),g.jsx(oe,{fullWidth:!0,mt:"lg",onClick:()=>{v||B()},disabled:v,children:"Update"})]})]})},A3=n=>{const l=n.schema[n.obj],o={[Object.keys(l).filter(j=>l[j].type=="primary_key")[0]]:"id"};Object.keys(l).filter(j=>l[j].nullable).map(j=>o[j]=null);const[c,f]=C.useState({}),[d,h]=C.useState({}),[y,m]=C.useState(o),[v,S]=C.useState(!1),_=j=>{f(L=>(L&&n.removeChildFunc(L),j))},x=(j,z,L)=>{L(V=>({...V,[j]:z(V[j])}))},w=xi(n.obj);let N=!1;const B=()=>{const j=async L=>{if(!N){N=!0,await S(!0);try{const $={};Object.keys(L).map(V=>{const Z=l[V];if(["int","float"].includes(Z.type))$[V]=JSON.parse(L[V]);else if(Z.is_list)try{$[V]=JSON.parse(L[V])}catch{$[V]=L[V]}else $[V]=L[V]}),await w.create($),n.onUpdate(),n.removeChildFunc()}catch($){$.data&&$.data.detail?(h({}),$.data.detail.map(V=>{x(V.loc[1],Z=>V.msg,h)})):console.log($)}finally{N=!1,S(!1)}}};m(L=>(j(L),L))},O=(j,z)=>{const L=crypto.randomUUID();n.addChildFunc({ele:Jp,props:{obj:j,schema:n.schema,onSelect:z}},L),_(L)};return g.jsxs(ue,{fluid:!0,px:"0",children:[g.jsxs(Xl,{onClose:()=>n.removeChildFunc(),mt:"lg",children:["Create ",n.obj]}),g.jsxs(ue,{fluid:!0,children:[Object.keys(l).map(j=>g.jsx(ft,{mt:"md",children:g.jsx(Qt.Wrapper,{label:g.jsxs(Ee,{children:[g.jsx(Wt,{fw:"bold",children:j}),g.jsxs(Wt,{c:"yellow",children:["< ",l[j].is_list?`list[${l[j].type}]`:l[j].type," >"]}),l[j].nullable&&g.jsx(Wt,{c:"red",children:"< nullable >"})]}),error:d[j],children:l[j].primary_key?g.jsx(Qt,{placeholder:"auto generated",disabled:!0}):l[j].depends?l[j].is_list?g.jsx(Ue,{withBorder:!0,p:"sm",style:d[j]?{borderColor:"red"}:{},onClick:()=>O(l[j].depends.slice(5),z=>x(j,L=>[...L||[],z],m)),children:y[j]?g.jsx(Pt,{spacing:"sm",children:y[j].map(z=>g.jsx(Pt.Item,{children:g.jsx(Ue,{withBorder:!0,p:"sm",children:g.jsxs(Ee,{children:[g.jsx(Pt,{listStyleType:"disc",children:Object.keys(z).map(L=>g.jsxs(Pt.Item,{children:[g.jsx("strong",{children:L}),": ",JSON.stringify(z[L])]}))}),g.jsx(Va,{})]})})}))}):g.jsx("div",{})}):g.jsx(Ue,{withBorder:!0,p:"sm",style:d[j]?{borderColor:"red"}:{},onClick:()=>O(l[j].depends.slice(5),z=>x(j,L=>z,m)),children:y[j]?g.jsx(Ue,{withBorder:!0,p:"sm",children:g.jsx(Pt,{listStyleType:"disc",children:Object.keys(y[j]).map(z=>g.jsxs(Pt.Item,{children:[g.jsx("strong",{children:z}),": ",JSON.stringify(y[j][z])]}))})}):g.jsx("div",{})}):g.jsx(Qt,{placeholder:j,onChange:z=>x(j,L=>z.target.value,m),error:d[j]})})})),g.jsx(oe,{fullWidth:!0,mt:"lg",onClick:()=>{v||B()},disabled:v,children:"Create"})]})]})},Wp=n=>{const[l,i]=C.useState(void 0),[o,c]=l2(!1),f=O=>{i(z=>(z&&n.removeChildFunc(z),O))},[d,h]=C.useState(void 0),[y,m]=C.useState(void 0),v=Object.keys(n.schema[n.obj]).filter(O=>n.schema[n.obj][O].type=="primary_key")[0],S=xi(n.obj),_=async()=>{const O=n.data[v];if(!["---",void 0].includes(O)){const j=await S.get_by_id(n.data[v]);j===void 0&&n.removeChildFunc(),n.onUpdate&&n.onUpdate(),h(j)}},x=async()=>{const O=n.data[v];if(!["---",void 0].includes(O)){const j=await S.get_history(n.data[v]);j!=null&&(console.log(j),m(j))}},w=()=>{const O=async z=>{await S.delete(z),n.removeChildFunc(),n.onUpdate&&n.onUpdate()},j=z=>(O(z[v]),j);h(j)},N=(O,j,z)=>{const L=crypto.randomUUID();n.addChildFunc({ele:z,props:{obj:O,data:j,schema:n.schema,onUpdate:_}},L),f(L)};C.useEffect(()=>{_()},[]),C.useEffect(()=>{console.log(o)},[o]);const B=(O,j)=>{const z=O[j];if([null,void 0].includes(z))return"----";const L=n.schema[n.obj][j].is_list===!0;let $;n.schema[n.obj][j].depends&&($=n.schema[n.obj][j].depends.slice(5));const V=Z=>$?g.jsx(Ue,{withBorder:!0,p:"sm",onClick:()=>N($,Z,Wp),children:g.jsx(Pt,{listStyleType:"disc",children:Object.keys(Z).map(W=>g.jsxs(Pt.Item,{children:[g.jsx("strong",{children:W}),": ",JSON.stringify(Z[W])]}))})}):JSON.stringify(Z);return L?z.length==0?"----":g.jsx(Pt,{spacing:"sm",listStyleType:"disc",children:z.map(Z=>g.jsx(Pt.Item,{children:V(Z)}))}):V(z)};return g.jsxs(ue,{fluid:!0,px:"0",children:[g.jsxs(Xl,{onClose:()=>n.removeChildFunc(),mt:"lg",onReload:_,children:[n.obj," details"]}),d&&g.jsxs(g.Fragment,{children:[g.jsx(ue,{fluid:!0,children:g.jsxs(Ee,{my:"lg",justify:"flex-end",children:[g.jsx(oe,{onClick:()=>N(n.obj,d,C3),children:g.jsx(w3,{})}),g.jsx(oe,{color:"red",onClick:c.open,children:g.jsx(P0,{})})]})}),g.jsx(ue,{fluid:!0,children:Object.keys(d).map(O=>g.jsxs(ft,{mt:"md",children:[g.jsxs(Ee,{children:[g.jsx(Wt,{fw:"bold",children:O}),g.jsxs(Wt,{c:"yellow",children:["< ",n.schema[n.obj][O].is_list?`list[${n.schema[n.obj][O].type}]`:n.schema[n.obj][O].type," >"]}),n.schema[n.obj][O].nullable&&g.jsx(Wt,{c:"red",children:"< nullable >"})]}),g.jsx(Ue,{shadow:"xs",withBorder:!0,p:"sm",bg:"neutral-gray.1",children:B(d,O)})]}))}),g.jsxs(Pe,{opened:o,onClose:c.close,title:"Warning",centered:!0,children:[g.jsx(Wt,{children:"You are about to delete:"}),g.jsx(Wt,{children:`${n.obj} <${d[v]}>`}),g.jsx(Wt,{mt:"lg",children:"Are you sure you want to continue?"}),g.jsxs(Ee,{mt:"lg",justify:"flex-end",children:[g.jsxs(oe,{bg:"red",onClick:w,children:[g.jsx(P0,{})," Yes"]}),g.jsx(oe,{bg:"neutral-gray.1",c:"black",onClick:c.close,children:"No"})]})]})]}),g.jsx(Xl,{mt:"lg",onReload:x,children:"History"}),y?g.jsxs(g.Fragment,{children:[g.jsx(ui,{data:y.self}),y.dependencies&&Object.keys(y.dependencies).map(O=>g.jsxs(g.Fragment,{children:[g.jsx(Wt,{mt:"lg",children:O}),g.jsx(ui,{data:y.dependencies[O]})]}))]}):g.jsx(Wt,{mt:"lg",align:"center",children:"..."})]})},j3=n=>{const[l,i]=C.useState(void 0),[o,c]=C.useState(0),f=_=>{i(w=>(w&&n.removeChildFunc(w),_))},d=(_,x)=>{const w=crypto.randomUUID();n.addChildFunc({ele:x,props:{obj:n.obj,data:_,schema:n.schema,onUpdate:S}},w),f(w)},h={};Object.keys(n.schema[n.obj]).map(_=>h[_]="---");const[y,m]=C.useState([h]),v=xi(n.obj),S=async(_=0)=>{const x=await v.get_all(10,_*10);let w=[];x.map(N=>{const B={};Object.keys(N).map(O=>{if(n.schema[n.obj][O].type.slice(0,4)=="list"||n.schema[n.obj][O].depends){B[O]=g.jsx("strong",{children:n.schema[n.obj][O].type});return}B[O]=N[O]}),w.push(B)}),w.length===0&&(w=[h]),m(w)};return C.useEffect(()=>{isNaN(o)?c(0):S(o)},[o]),g.jsxs(ue,{fluid:!0,px:"0",children:[g.jsxs(Ee,{my:"lg",children:[g.jsx(Za,{children:n.obj}),g.jsx(oe,{onClick:()=>d(void 0,A3),ml:"auto",children:"New"})]}),g.jsx(Xl,{onReload:()=>S(o),children:"Objects"}),g.jsx(ui,{data:y,onRowSelect:_=>d(_,Wp)}),g.jsxs(Ee,{children:[g.jsx(oe,{onClick:()=>c(o-1),children:"<"}),g.jsx(Wt,{children:o}),g.jsx(oe,{onClick:()=>c(o+1),children:">"})]})]})},N3=n=>g.jsxs(ue,{fluid:!0,style:{width:"100%",overflowX:"hidden",display:"flex",overflowY:"auto",height:"100%"},miw:"40vw",p:"0",...n,children:[g.jsx(Cu,{orientation:"vertical",h:"auto",maw:"sm",mr:"sm"}),g.jsx(ue,{fluid:!0,style:{width:"100%",overflowX:"auto"},pb:"100",children:n.children})]}),O3=n=>{const[l,i]=C.useState([]),o=()=>{const d=[];n.elementsList.map(h=>{const y=crypto.randomUUID(),m=g.jsx(h.ele,{...h.props,addChildFunc:(v,S=void 0)=>c(v,S),removeChildFunc:v=>f(v||y)},y);d.push(m)}),i(d)},c=(d,h=void 0)=>{i(m=>{const v=m.slice();let S=h;S||(S=crypto.randomUUID());const _=g.jsx(d.ele,{...d.props,addChildFunc:(x,w=void 0)=>c(x,w),removeChildFunc:x=>f(x||S)},S);return v.push(_),v})},f=d=>{i(y=>{const m=y.slice(),v=m.filter(_=>_.key==d);if(v.length<=0)return m;const S=m.indexOf(v[0]);return S>-1&&m.splice(S,1),m})};return C.useEffect(()=>{o()},[]),g.jsx(ld,{gap:"0",style:{height:"100%"},children:l.map(d=>g.jsx(N3,{children:d}))})};function M3(n){return g.jsx(ue,{fluid:!0,style:{height:"100%"},children:g.jsx(O3,{elementsList:[{ele:j3,props:{obj:n.obj,schema:n.schema}}]})})}function R3(n){const l=n.schema;return g.jsxs(ve,{orientation:"vertical",ml:"lg",style:{height:"100%"},children:[g.jsx(ve.List,{children:Object.keys(l).map(i=>g.jsx(ve.Tab,{value:i,children:i},crypto.randomUUID()))}),Object.keys(l).map(i=>g.jsx(ve.Panel,{value:i,style:{maxWidth:"90vw",overflow:"scroll",height:"100%"},p:"0",pb:"lg",children:g.jsx(M3,{obj:i,schema:l})},crypto.randomUUID()))]})}(function(){try{if(typeof document<"u"){var n=document.createElement("style");n.appendChild(document.createTextNode("._dotSvgContainer_64xbd_1{position:absolute;width:100%;height:100%;inset-block-start:0;inset-inline-start:0}._container_1xxyl_1{position:relative;overflow:hidden;width:100%;height:100%}._canvasWrapper_1xxyl_8{position:relative;width:100%;height:100%;overflow:hidden;background-color:var(--kf-color-bg-gray);cursor:grab;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none}._canvasWrapper_1xxyl_8._panning_1xxyl_21{cursor:grabbing}._canvas_1xxyl_8{position:relative;z-index:2;width:100%;height:100%}._canvas_1xxyl_8>div:first-of-type{position:absolute;inset-inline-start:0;inset-block-start:0;transition:transform .01s ease}._contentWrapper_1xxyl_39{position:absolute}._verticalScrollBar_17thb_1{position:absolute;inset-inline-end:2px;inset-block-start:0;inset-block-end:0;background:transparent}._horizontalScrollBar_17thb_9{position:absolute;inset-block-end:2px;inset-inline-start:0;inset-inline-end:0;background:transparent}._verticalScrollBar_17thb_1>div,._horizontalScrollBar_17thb_9>div{position:absolute;border-radius:50px;z-index:8;cursor:pointer}._verticalScrollBar_17thb_1>div{width:100%}._horizontalScrollBar_17thb_9>div{height:100%}")),document.head.appendChild(n)}}catch(l){console.error("vite-plugin-css-injected-by-js",l)}})();var Of="http://www.w3.org/1999/xhtml";const F0={svg:"http://www.w3.org/2000/svg",xhtml:Of,xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"};function Du(n){var l=n+="",i=l.indexOf(":");return i>=0&&(l=n.slice(0,i))!=="xmlns"&&(n=n.slice(i+1)),F0.hasOwnProperty(l)?{space:F0[l],local:n}:n}function z3(n){return function(){var l=this.ownerDocument,i=this.namespaceURI;return i===Of&&l.documentElement.namespaceURI===Of?l.createElement(n):l.createElementNS(i,n)}}function D3(n){return function(){return this.ownerDocument.createElementNS(n.space,n.local)}}function Pp(n){var l=Du(n);return(l.local?D3:z3)(l)}function B3(){}function gd(n){return n==null?B3:function(){return this.querySelector(n)}}function U3(n){typeof n!="function"&&(n=gd(n));for(var l=this._groups,i=l.length,o=new Array(i),c=0;c<i;++c)for(var f=l[c],d=f.length,h=o[c]=new Array(d),y,m,v=0;v<d;++v)(y=f[v])&&(m=n.call(y,y.__data__,v,f))&&("__data__"in y&&(m.__data__=y.__data__),h[v]=m);return new He(o,this._parents)}function H3(n){return n==null?[]:Array.isArray(n)?n:Array.from(n)}function $3(){return[]}function Fp(n){return n==null?$3:function(){return this.querySelectorAll(n)}}function L3(n){return function(){return H3(n.apply(this,arguments))}}function Y3(n){typeof n=="function"?n=L3(n):n=Fp(n);for(var l=this._groups,i=l.length,o=[],c=[],f=0;f<i;++f)for(var d=l[f],h=d.length,y,m=0;m<h;++m)(y=d[m])&&(o.push(n.call(y,y.__data__,m,d)),c.push(y));return new He(o,c)}function Ip(n){return function(){return this.matches(n)}}function tg(n){return function(l){return l.matches(n)}}var q3=Array.prototype.find;function X3(n){return function(){return q3.call(this.children,n)}}function G3(){return this.firstElementChild}function V3(n){return this.select(n==null?G3:X3(typeof n=="function"?n:tg(n)))}var k3=Array.prototype.filter;function Z3(){return Array.from(this.children)}function Q3(n){return function(){return k3.call(this.children,n)}}function K3(n){return this.selectAll(n==null?Z3:Q3(typeof n=="function"?n:tg(n)))}function J3(n){typeof n!="function"&&(n=Ip(n));for(var l=this._groups,i=l.length,o=new Array(i),c=0;c<i;++c)for(var f=l[c],d=f.length,h=o[c]=[],y,m=0;m<d;++m)(y=f[m])&&n.call(y,y.__data__,m,f)&&h.push(y);return new He(o,this._parents)}function eg(n){return new Array(n.length)}function W3(){return new He(this._enter||this._groups.map(eg),this._parents)}function ou(n,l){this.ownerDocument=n.ownerDocument,this.namespaceURI=n.namespaceURI,this._next=null,this._parent=n,this.__data__=l}ou.prototype={constructor:ou,appendChild:function(n){return this._parent.insertBefore(n,this._next)},insertBefore:function(n,l){return this._parent.insertBefore(n,l)},querySelector:function(n){return this._parent.querySelector(n)},querySelectorAll:function(n){return this._parent.querySelectorAll(n)}};function P3(n){return function(){return n}}function F3(n,l,i,o,c,f){for(var d=0,h,y=l.length,m=f.length;d<m;++d)(h=l[d])?(h.__data__=f[d],o[d]=h):i[d]=new ou(n,f[d]);for(;d<y;++d)(h=l[d])&&(c[d]=h)}function I3(n,l,i,o,c,f,d){var h,y,m=new Map,v=l.length,S=f.length,_=new Array(v),x;for(h=0;h<v;++h)(y=l[h])&&(_[h]=x=d.call(y,y.__data__,h,l)+"",m.has(x)?c[h]=y:m.set(x,y));for(h=0;h<S;++h)x=d.call(n,f[h],h,f)+"",(y=m.get(x))?(o[h]=y,y.__data__=f[h],m.delete(x)):i[h]=new ou(n,f[h]);for(h=0;h<v;++h)(y=l[h])&&m.get(_[h])===y&&(c[h]=y)}function tT(n){return n.__data__}function eT(n,l){if(!arguments.length)return Array.from(this,tT);var i=l?I3:F3,o=this._parents,c=this._groups;typeof n!="function"&&(n=P3(n));for(var f=c.length,d=new Array(f),h=new Array(f),y=new Array(f),m=0;m<f;++m){var v=o[m],S=c[m],_=S.length,x=nT(n.call(v,v&&v.__data__,m,o)),w=x.length,N=h[m]=new Array(w),B=d[m]=new Array(w),O=y[m]=new Array(_);i(v,S,N,B,O,x,l);for(var j=0,z=0,L,$;j<w;++j)if(L=N[j]){for(j>=z&&(z=j+1);!($=B[z])&&++z<w;);L._next=$||null}}return d=new He(d,o),d._enter=h,d._exit=y,d}function nT(n){return typeof n=="object"&&"length"in n?n:Array.from(n)}function aT(){return new He(this._exit||this._groups.map(eg),this._parents)}function lT(n,l,i){var o=this.enter(),c=this,f=this.exit();return typeof n=="function"?(o=n(o),o&&(o=o.selection())):o=o.append(n+""),l!=null&&(c=l(c),c&&(c=c.selection())),i==null?f.remove():i(f),o&&c?o.merge(c).order():c}function rT(n){for(var l=n.selection?n.selection():n,i=this._groups,o=l._groups,c=i.length,f=o.length,d=Math.min(c,f),h=new Array(c),y=0;y<d;++y)for(var m=i[y],v=o[y],S=m.length,_=h[y]=new Array(S),x,w=0;w<S;++w)(x=m[w]||v[w])&&(_[w]=x);for(;y<c;++y)h[y]=i[y];return new He(h,this._parents)}function iT(){for(var n=this._groups,l=-1,i=n.length;++l<i;)for(var o=n[l],c=o.length-1,f=o[c],d;--c>=0;)(d=o[c])&&(f&&d.compareDocumentPosition(f)^4&&f.parentNode.insertBefore(d,f),f=d);return this}function oT(n){n||(n=uT);function l(S,_){return S&&_?n(S.__data__,_.__data__):!S-!_}for(var i=this._groups,o=i.length,c=new Array(o),f=0;f<o;++f){for(var d=i[f],h=d.length,y=c[f]=new Array(h),m,v=0;v<h;++v)(m=d[v])&&(y[v]=m);y.sort(l)}return new He(c,this._parents).order()}function uT(n,l){return n<l?-1:n>l?1:n>=l?0:NaN}function cT(){var n=arguments[0];return arguments[0]=this,n.apply(null,arguments),this}function sT(){return Array.from(this)}function fT(){for(var n=this._groups,l=0,i=n.length;l<i;++l)for(var o=n[l],c=0,f=o.length;c<f;++c){var d=o[c];if(d)return d}return null}function dT(){let n=0;for(const l of this)++n;return n}function hT(){return!this.node()}function mT(n){for(var l=this._groups,i=0,o=l.length;i<o;++i)for(var c=l[i],f=0,d=c.length,h;f<d;++f)(h=c[f])&&n.call(h,h.__data__,f,c);return this}function yT(n){return function(){this.removeAttribute(n)}}function vT(n){return function(){this.removeAttributeNS(n.space,n.local)}}function pT(n,l){return function(){this.setAttribute(n,l)}}function gT(n,l){return function(){this.setAttributeNS(n.space,n.local,l)}}function bT(n,l){return function(){var i=l.apply(this,arguments);i==null?this.removeAttribute(n):this.setAttribute(n,i)}}function ST(n,l){return function(){var i=l.apply(this,arguments);i==null?this.removeAttributeNS(n.space,n.local):this.setAttributeNS(n.space,n.local,i)}}function _T(n,l){var i=Du(n);if(arguments.length<2){var o=this.node();return i.local?o.getAttributeNS(i.space,i.local):o.getAttribute(i)}return this.each((l==null?i.local?vT:yT:typeof l=="function"?i.local?ST:bT:i.local?gT:pT)(i,l))}function ng(n){return n.ownerDocument&&n.ownerDocument.defaultView||n.document&&n||n.defaultView}function xT(n){return function(){this.style.removeProperty(n)}}function TT(n,l,i){return function(){this.style.setProperty(n,l,i)}}function wT(n,l,i){return function(){var o=l.apply(this,arguments);o==null?this.style.removeProperty(n):this.style.setProperty(n,o,i)}}function ET(n,l,i){return arguments.length>1?this.each((l==null?xT:typeof l=="function"?wT:TT)(n,l,i??"")):Gl(this.node(),n)}function Gl(n,l){return n.style.getPropertyValue(l)||ng(n).getComputedStyle(n,null).getPropertyValue(l)}function CT(n){return function(){delete this[n]}}function AT(n,l){return function(){this[n]=l}}function jT(n,l){return function(){var i=l.apply(this,arguments);i==null?delete this[n]:this[n]=i}}function NT(n,l){return arguments.length>1?this.each((l==null?CT:typeof l=="function"?jT:AT)(n,l)):this.node()[n]}function ag(n){return n.trim().split(/^|\s+/)}function bd(n){return n.classList||new lg(n)}function lg(n){this._node=n,this._names=ag(n.getAttribute("class")||"")}lg.prototype={add:function(n){var l=this._names.indexOf(n);l<0&&(this._names.push(n),this._node.setAttribute("class",this._names.join(" ")))},remove:function(n){var l=this._names.indexOf(n);l>=0&&(this._names.splice(l,1),this._node.setAttribute("class",this._names.join(" ")))},contains:function(n){return this._names.indexOf(n)>=0}};function rg(n,l){for(var i=bd(n),o=-1,c=l.length;++o<c;)i.add(l[o])}function ig(n,l){for(var i=bd(n),o=-1,c=l.length;++o<c;)i.remove(l[o])}function OT(n){return function(){rg(this,n)}}function MT(n){return function(){ig(this,n)}}function RT(n,l){return function(){(l.apply(this,arguments)?rg:ig)(this,n)}}function zT(n,l){var i=ag(n+"");if(arguments.length<2){for(var o=bd(this.node()),c=-1,f=i.length;++c<f;)if(!o.contains(i[c]))return!1;return!0}return this.each((typeof l=="function"?RT:l?OT:MT)(i,l))}function DT(){this.textContent=""}function BT(n){return function(){this.textContent=n}}function UT(n){return function(){var l=n.apply(this,arguments);this.textContent=l??""}}function HT(n){return arguments.length?this.each(n==null?DT:(typeof n=="function"?UT:BT)(n)):this.node().textContent}function $T(){this.innerHTML=""}function LT(n){return function(){this.innerHTML=n}}function YT(n){return function(){var l=n.apply(this,arguments);this.innerHTML=l??""}}function qT(n){return arguments.length?this.each(n==null?$T:(typeof n=="function"?YT:LT)(n)):this.node().innerHTML}function XT(){this.nextSibling&&this.parentNode.appendChild(this)}function GT(){return this.each(XT)}function VT(){this.previousSibling&&this.parentNode.insertBefore(this,this.parentNode.firstChild)}function kT(){return this.each(VT)}function ZT(n){var l=typeof n=="function"?n:Pp(n);return this.select(function(){return this.appendChild(l.apply(this,arguments))})}function QT(){return null}function KT(n,l){var i=typeof n=="function"?n:Pp(n),o=l==null?QT:typeof l=="function"?l:gd(l);return this.select(function(){return this.insertBefore(i.apply(this,arguments),o.apply(this,arguments)||null)})}function JT(){var n=this.parentNode;n&&n.removeChild(this)}function WT(){return this.each(JT)}function PT(){var n=this.cloneNode(!1),l=this.parentNode;return l?l.insertBefore(n,this.nextSibling):n}function FT(){var n=this.cloneNode(!0),l=this.parentNode;return l?l.insertBefore(n,this.nextSibling):n}function IT(n){return this.select(n?FT:PT)}function tw(n){return arguments.length?this.property("__data__",n):this.node().__data__}function ew(n){return function(l){n.call(this,l,this.__data__)}}function nw(n){return n.trim().split(/^|\s+/).map(function(l){var i="",o=l.indexOf(".");return o>=0&&(i=l.slice(o+1),l=l.slice(0,o)),{type:l,name:i}})}function aw(n){return function(){var l=this.__on;if(l){for(var i=0,o=-1,c=l.length,f;i<c;++i)f=l[i],(!n.type||f.type===n.type)&&f.name===n.name?this.removeEventListener(f.type,f.listener,f.options):l[++o]=f;++o?l.length=o:delete this.__on}}}function lw(n,l,i){return function(){var o=this.__on,c,f=ew(l);if(o){for(var d=0,h=o.length;d<h;++d)if((c=o[d]).type===n.type&&c.name===n.name){this.removeEventListener(c.type,c.listener,c.options),this.addEventListener(c.type,c.listener=f,c.options=i),c.value=l;return}}this.addEventListener(n.type,f,i),c={type:n.type,name:n.name,value:l,listener:f,options:i},o?o.push(c):this.__on=[c]}}function rw(n,l,i){var o=nw(n+""),c,f=o.length,d;if(arguments.length<2){var h=this.node().__on;if(h){for(var y=0,m=h.length,v;y<m;++y)for(c=0,v=h[y];c<f;++c)if((d=o[c]).type===v.type&&d.name===v.name)return v.value}return}for(h=l?lw:aw,c=0;c<f;++c)this.each(h(o[c],l,i));return this}function og(n,l,i){var o=ng(n),c=o.CustomEvent;typeof c=="function"?c=new c(l,i):(c=o.document.createEvent("Event"),i?(c.initEvent(l,i.bubbles,i.cancelable),c.detail=i.detail):c.initEvent(l,!1,!1)),n.dispatchEvent(c)}function iw(n,l){return function(){return og(this,n,l)}}function ow(n,l){return function(){return og(this,n,l.apply(this,arguments))}}function uw(n,l){return this.each((typeof l=="function"?ow:iw)(n,l))}function*cw(){for(var n=this._groups,l=0,i=n.length;l<i;++l)for(var o=n[l],c=0,f=o.length,d;c<f;++c)(d=o[c])&&(yield d)}var ug=[null];function He(n,l){this._groups=n,this._parents=l}function Ti(){return new He([[document.documentElement]],ug)}function sw(){return this}He.prototype=Ti.prototype={constructor:He,select:U3,selectAll:Y3,selectChild:V3,selectChildren:K3,filter:J3,data:eT,enter:W3,exit:aT,join:lT,merge:rT,selection:sw,order:iT,sort:oT,call:cT,nodes:sT,node:fT,size:dT,empty:hT,each:mT,attr:_T,style:ET,property:NT,classed:zT,text:HT,html:qT,raise:GT,lower:kT,append:ZT,insert:KT,remove:WT,clone:IT,datum:tw,on:rw,dispatch:uw,[Symbol.iterator]:cw};function ye(n){return typeof n=="string"?new He([[document.querySelector(n)]],[document.documentElement]):new He([[n]],ug)}function fw(n){let l;for(;l=n.sourceEvent;)n=l;return n}function ma(n,l){if(n=fw(n),l===void 0&&(l=n.currentTarget),l){var i=l.ownerSVGElement||l;if(i.createSVGPoint){var o=i.createSVGPoint();return o.x=n.clientX,o.y=n.clientY,o=o.matrixTransform(l.getScreenCTM().inverse()),[o.x,o.y]}if(l.getBoundingClientRect){var c=l.getBoundingClientRect();return[n.clientX-c.left-l.clientLeft,n.clientY-c.top-l.clientTop]}}return[n.pageX,n.pageY]}var dw={value:()=>{}};function Sd(){for(var n=0,l=arguments.length,i={},o;n<l;++n){if(!(o=arguments[n]+"")||o in i||/[\s.]/.test(o))throw new Error("illegal type: "+o);i[o]=[]}return new eu(i)}function eu(n){this._=n}function hw(n,l){return n.trim().split(/^|\s+/).map(function(i){var o="",c=i.indexOf(".");if(c>=0&&(o=i.slice(c+1),i=i.slice(0,c)),i&&!l.hasOwnProperty(i))throw new Error("unknown type: "+i);return{type:i,name:o}})}eu.prototype=Sd.prototype={constructor:eu,on:function(n,l){var i=this._,o=hw(n+"",i),c,f=-1,d=o.length;if(arguments.length<2){for(;++f<d;)if((c=(n=o[f]).type)&&(c=mw(i[c],n.name)))return c;return}if(l!=null&&typeof l!="function")throw new Error("invalid callback: "+l);for(;++f<d;)if(c=(n=o[f]).type)i[c]=I0(i[c],n.name,l);else if(l==null)for(c in i)i[c]=I0(i[c],n.name,null);return this},copy:function(){var n={},l=this._;for(var i in l)n[i]=l[i].slice();return new eu(n)},call:function(n,l){if((c=arguments.length-2)>0)for(var i=new Array(c),o=0,c,f;o<c;++o)i[o]=arguments[o+2];if(!this._.hasOwnProperty(n))throw new Error("unknown type: "+n);for(f=this._[n],o=0,c=f.length;o<c;++o)f[o].value.apply(l,i)},apply:function(n,l,i){if(!this._.hasOwnProperty(n))throw new Error("unknown type: "+n);for(var o=this._[n],c=0,f=o.length;c<f;++c)o[c].value.apply(l,i)}};function mw(n,l){for(var i=0,o=n.length,c;i<o;++i)if((c=n[i]).name===l)return c.value}function I0(n,l,i){for(var o=0,c=n.length;o<c;++o)if(n[o].name===l){n[o]=dw,n=n.slice(0,o).concat(n.slice(o+1));break}return i!=null&&n.push({name:l,value:i}),n}const Mf={capture:!0,passive:!1};function Rf(n){n.preventDefault(),n.stopImmediatePropagation()}function yw(n){var l=n.document.documentElement,i=ye(n).on("dragstart.drag",Rf,Mf);"onselectstart"in l?i.on("selectstart.drag",Rf,Mf):(l.__noselect=l.style.MozUserSelect,l.style.MozUserSelect="none")}function vw(n,l){var i=n.document.documentElement,o=ye(n).on("dragstart.drag",null);l&&(o.on("click.drag",Rf,Mf),setTimeout(function(){o.on("click.drag",null)},0)),"onselectstart"in i?o.on("selectstart.drag",null):(i.style.MozUserSelect=i.__noselect,delete i.__noselect)}function _d(n,l,i){n.prototype=l.prototype=i,i.constructor=n}function cg(n,l){var i=Object.create(n.prototype);for(var o in l)i[o]=l[o];return i}function wi(){}var ci=.7,uu=1/ci,$l="\\s*([+-]?\\d+)\\s*",si="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)\\s*",yn="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)%\\s*",pw=/^#([0-9a-f]{3,8})$/,gw=new RegExp(`^rgb\\(${$l},${$l},${$l}\\)$`),bw=new RegExp(`^rgb\\(${yn},${yn},${yn}\\)$`),Sw=new RegExp(`^rgba\\(${$l},${$l},${$l},${si}\\)$`),_w=new RegExp(`^rgba\\(${yn},${yn},${yn},${si}\\)$`),xw=new RegExp(`^hsl\\(${si},${yn},${yn}\\)$`),Tw=new RegExp(`^hsla\\(${si},${yn},${yn},${si}\\)$`),tv={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};_d(wi,fi,{copy(n){return Object.assign(new this.constructor,this,n)},displayable(){return this.rgb().displayable()},hex:ev,formatHex:ev,formatHex8:ww,formatHsl:Ew,formatRgb:nv,toString:nv});function ev(){return this.rgb().formatHex()}function ww(){return this.rgb().formatHex8()}function Ew(){return sg(this).formatHsl()}function nv(){return this.rgb().formatRgb()}function fi(n){var l,i;return n=(n+"").trim().toLowerCase(),(l=pw.exec(n))?(i=l[1].length,l=parseInt(l[1],16),i===6?av(l):i===3?new Te(l>>8&15|l>>4&240,l>>4&15|l&240,(l&15)<<4|l&15,1):i===8?Qo(l>>24&255,l>>16&255,l>>8&255,(l&255)/255):i===4?Qo(l>>12&15|l>>8&240,l>>8&15|l>>4&240,l>>4&15|l&240,((l&15)<<4|l&15)/255):null):(l=gw.exec(n))?new Te(l[1],l[2],l[3],1):(l=bw.exec(n))?new Te(l[1]*255/100,l[2]*255/100,l[3]*255/100,1):(l=Sw.exec(n))?Qo(l[1],l[2],l[3],l[4]):(l=_w.exec(n))?Qo(l[1]*255/100,l[2]*255/100,l[3]*255/100,l[4]):(l=xw.exec(n))?iv(l[1],l[2]/100,l[3]/100,1):(l=Tw.exec(n))?iv(l[1],l[2]/100,l[3]/100,l[4]):tv.hasOwnProperty(n)?av(tv[n]):n==="transparent"?new Te(NaN,NaN,NaN,0):null}function av(n){return new Te(n>>16&255,n>>8&255,n&255,1)}function Qo(n,l,i,o){return o<=0&&(n=l=i=NaN),new Te(n,l,i,o)}function Cw(n){return n instanceof wi||(n=fi(n)),n?(n=n.rgb(),new Te(n.r,n.g,n.b,n.opacity)):new Te}function zf(n,l,i,o){return arguments.length===1?Cw(n):new Te(n,l,i,o??1)}function Te(n,l,i,o){this.r=+n,this.g=+l,this.b=+i,this.opacity=+o}_d(Te,zf,cg(wi,{brighter(n){return n=n==null?uu:Math.pow(uu,n),new Te(this.r*n,this.g*n,this.b*n,this.opacity)},darker(n){return n=n==null?ci:Math.pow(ci,n),new Te(this.r*n,this.g*n,this.b*n,this.opacity)},rgb(){return this},clamp(){return new Te(Ya(this.r),Ya(this.g),Ya(this.b),cu(this.opacity))},displayable(){return-.5<=this.r&&this.r<255.5&&-.5<=this.g&&this.g<255.5&&-.5<=this.b&&this.b<255.5&&0<=this.opacity&&this.opacity<=1},hex:lv,formatHex:lv,formatHex8:Aw,formatRgb:rv,toString:rv}));function lv(){return`#${$a(this.r)}${$a(this.g)}${$a(this.b)}`}function Aw(){return`#${$a(this.r)}${$a(this.g)}${$a(this.b)}${$a((isNaN(this.opacity)?1:this.opacity)*255)}`}function rv(){const n=cu(this.opacity);return`${n===1?"rgb(":"rgba("}${Ya(this.r)}, ${Ya(this.g)}, ${Ya(this.b)}${n===1?")":`, ${n})`}`}function cu(n){return isNaN(n)?1:Math.max(0,Math.min(1,n))}function Ya(n){return Math.max(0,Math.min(255,Math.round(n)||0))}function $a(n){return n=Ya(n),(n<16?"0":"")+n.toString(16)}function iv(n,l,i,o){return o<=0?n=l=i=NaN:i<=0||i>=1?n=l=NaN:l<=0&&(n=NaN),new en(n,l,i,o)}function sg(n){if(n instanceof en)return new en(n.h,n.s,n.l,n.opacity);if(n instanceof wi||(n=fi(n)),!n)return new en;if(n instanceof en)return n;n=n.rgb();var l=n.r/255,i=n.g/255,o=n.b/255,c=Math.min(l,i,o),f=Math.max(l,i,o),d=NaN,h=f-c,y=(f+c)/2;return h?(l===f?d=(i-o)/h+(i<o)*6:i===f?d=(o-l)/h+2:d=(l-i)/h+4,h/=y<.5?f+c:2-f-c,d*=60):h=y>0&&y<1?0:d,new en(d,h,y,n.opacity)}function jw(n,l,i,o){return arguments.length===1?sg(n):new en(n,l,i,o??1)}function en(n,l,i,o){this.h=+n,this.s=+l,this.l=+i,this.opacity=+o}_d(en,jw,cg(wi,{brighter(n){return n=n==null?uu:Math.pow(uu,n),new en(this.h,this.s,this.l*n,this.opacity)},darker(n){return n=n==null?ci:Math.pow(ci,n),new en(this.h,this.s,this.l*n,this.opacity)},rgb(){var n=this.h%360+(this.h<0)*360,l=isNaN(n)||isNaN(this.s)?0:this.s,i=this.l,o=i+(i<.5?i:1-i)*l,c=2*i-o;return new Te(yf(n>=240?n-240:n+120,c,o),yf(n,c,o),yf(n<120?n+240:n-120,c,o),this.opacity)},clamp(){return new en(ov(this.h),Ko(this.s),Ko(this.l),cu(this.opacity))},displayable(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1},formatHsl(){const n=cu(this.opacity);return`${n===1?"hsl(":"hsla("}${ov(this.h)}, ${Ko(this.s)*100}%, ${Ko(this.l)*100}%${n===1?")":`, ${n})`}`}}));function ov(n){return n=(n||0)%360,n<0?n+360:n}function Ko(n){return Math.max(0,Math.min(1,n||0))}function yf(n,l,i){return(n<60?l+(i-l)*n/60:n<180?i:n<240?l+(i-l)*(240-n)/60:l)*255}const fg=n=>()=>n;function Nw(n,l){return function(i){return n+i*l}}function Ow(n,l,i){return n=Math.pow(n,i),l=Math.pow(l,i)-n,i=1/i,function(o){return Math.pow(n+o*l,i)}}function Mw(n){return(n=+n)==1?dg:function(l,i){return i-l?Ow(l,i,n):fg(isNaN(l)?i:l)}}function dg(n,l){var i=l-n;return i?Nw(n,i):fg(isNaN(n)?l:n)}const uv=function n(l){var i=Mw(l);function o(c,f){var d=i((c=zf(c)).r,(f=zf(f)).r),h=i(c.g,f.g),y=i(c.b,f.b),m=dg(c.opacity,f.opacity);return function(v){return c.r=d(v),c.g=h(v),c.b=y(v),c.opacity=m(v),c+""}}return o.gamma=n,o}(1);function ya(n,l){return n=+n,l=+l,function(i){return n*(1-i)+l*i}}var Df=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,vf=new RegExp(Df.source,"g");function Rw(n){return function(){return n}}function zw(n){return function(l){return n(l)+""}}function Dw(n,l){var i=Df.lastIndex=vf.lastIndex=0,o,c,f,d=-1,h=[],y=[];for(n=n+"",l=l+"";(o=Df.exec(n))&&(c=vf.exec(l));)(f=c.index)>i&&(f=l.slice(i,f),h[d]?h[d]+=f:h[++d]=f),(o=o[0])===(c=c[0])?h[d]?h[d]+=c:h[++d]=c:(h[++d]=null,y.push({i:d,x:ya(o,c)})),i=vf.lastIndex;return i<l.length&&(f=l.slice(i),h[d]?h[d]+=f:h[++d]=f),h.length<2?y[0]?zw(y[0].x):Rw(l):(l=y.length,function(m){for(var v=0,S;v<l;++v)h[(S=y[v]).i]=S.x(m);return h.join("")})}var cv=180/Math.PI,hg={translateX:0,translateY:0,rotate:0,skewX:0,scaleX:1,scaleY:1};function mg(n,l,i,o,c,f){var d,h,y;return(d=Math.sqrt(n*n+l*l))&&(n/=d,l/=d),(y=n*i+l*o)&&(i-=n*y,o-=l*y),(h=Math.sqrt(i*i+o*o))&&(i/=h,o/=h,y/=h),n*o<l*i&&(n=-n,l=-l,y=-y,d=-d),{translateX:c,translateY:f,rotate:Math.atan2(l,n)*cv,skewX:Math.atan(y)*cv,scaleX:d,scaleY:h}}var Jo;function Bw(n){const l=new(typeof DOMMatrix=="function"?DOMMatrix:WebKitCSSMatrix)(n+"");return l.isIdentity?hg:mg(l.a,l.b,l.c,l.d,l.e,l.f)}function Uw(n){return n==null||(Jo||(Jo=document.createElementNS("http://www.w3.org/2000/svg","g")),Jo.setAttribute("transform",n),!(n=Jo.transform.baseVal.consolidate()))?hg:(n=n.matrix,mg(n.a,n.b,n.c,n.d,n.e,n.f))}function yg(n,l,i,o){function c(m){return m.length?m.pop()+" ":""}function f(m,v,S,_,x,w){if(m!==S||v!==_){var N=x.push("translate(",null,l,null,i);w.push({i:N-4,x:ya(m,S)},{i:N-2,x:ya(v,_)})}else(S||_)&&x.push("translate("+S+l+_+i)}function d(m,v,S,_){m!==v?(m-v>180?v+=360:v-m>180&&(m+=360),_.push({i:S.push(c(S)+"rotate(",null,o)-2,x:ya(m,v)})):v&&S.push(c(S)+"rotate("+v+o)}function h(m,v,S,_){m!==v?_.push({i:S.push(c(S)+"skewX(",null,o)-2,x:ya(m,v)}):v&&S.push(c(S)+"skewX("+v+o)}function y(m,v,S,_,x,w){if(m!==S||v!==_){var N=x.push(c(x)+"scale(",null,",",null,")");w.push({i:N-4,x:ya(m,S)},{i:N-2,x:ya(v,_)})}else(S!==1||_!==1)&&x.push(c(x)+"scale("+S+","+_+")")}return function(m,v){var S=[],_=[];return m=n(m),v=n(v),f(m.translateX,m.translateY,v.translateX,v.translateY,S,_),d(m.rotate,v.rotate,S,_),h(m.skewX,v.skewX,S,_),y(m.scaleX,m.scaleY,v.scaleX,v.scaleY,S,_),m=v=null,function(x){for(var w=-1,N=_.length,B;++w<N;)S[(B=_[w]).i]=B.x(x);return S.join("")}}}var Hw=yg(Bw,"px, ","px)","deg)"),$w=yg(Uw,", ",")",")"),Lw=1e-12;function sv(n){return((n=Math.exp(n))+1/n)/2}function Yw(n){return((n=Math.exp(n))-1/n)/2}function qw(n){return((n=Math.exp(2*n))-1)/(n+1)}const Xw=function n(l,i,o){function c(f,d){var h=f[0],y=f[1],m=f[2],v=d[0],S=d[1],_=d[2],x=v-h,w=S-y,N=x*x+w*w,B,O;if(N<Lw)O=Math.log(_/m)/l,B=function(Z){return[h+Z*x,y+Z*w,m*Math.exp(l*Z*O)]};else{var j=Math.sqrt(N),z=(_*_-m*m+o*N)/(2*m*i*j),L=(_*_-m*m-o*N)/(2*_*i*j),$=Math.log(Math.sqrt(z*z+1)-z),V=Math.log(Math.sqrt(L*L+1)-L);O=(V-$)/l,B=function(Z){var W=Z*O,it=sv($),ht=m/(i*j)*(it*qw(l*W+$)-Yw($));return[h+ht*x,y+ht*w,m*it/sv(l*W+$)]}}return B.duration=O*1e3*l/Math.SQRT2,B}return c.rho=function(f){var d=Math.max(.001,+f),h=d*d,y=h*h;return n(d,h,y)},c}(Math.SQRT2,2,4);var Vl=0,ni=0,ti=0,vg=1e3,su,ai,fu=0,qa=0,Bu=0,di=typeof performance=="object"&&performance.now?performance:Date,pg=typeof window=="object"&&window.requestAnimationFrame?window.requestAnimationFrame.bind(window):function(n){setTimeout(n,17)};function xd(){return qa||(pg(Gw),qa=di.now()+Bu)}function Gw(){qa=0}function du(){this._call=this._time=this._next=null}du.prototype=gg.prototype={constructor:du,restart:function(n,l,i){if(typeof n!="function")throw new TypeError("callback is not a function");i=(i==null?xd():+i)+(l==null?0:+l),!this._next&&ai!==this&&(ai?ai._next=this:su=this,ai=this),this._call=n,this._time=i,Bf()},stop:function(){this._call&&(this._call=null,this._time=1/0,Bf())}};function gg(n,l,i){var o=new du;return o.restart(n,l,i),o}function Vw(){xd(),++Vl;for(var n=su,l;n;)(l=qa-n._time)>=0&&n._call.call(void 0,l),n=n._next;--Vl}function fv(){qa=(fu=di.now())+Bu,Vl=ni=0;try{Vw()}finally{Vl=0,Zw(),qa=0}}function kw(){var n=di.now(),l=n-fu;l>vg&&(Bu-=l,fu=n)}function Zw(){for(var n,l=su,i,o=1/0;l;)l._call?(o>l._time&&(o=l._time),n=l,l=l._next):(i=l._next,l._next=null,l=n?n._next=i:su=i);ai=n,Bf(o)}function Bf(n){if(!Vl){ni&&(ni=clearTimeout(ni));var l=n-qa;l>24?(n<1/0&&(ni=setTimeout(fv,n-di.now()-Bu)),ti&&(ti=clearInterval(ti))):(ti||(fu=di.now(),ti=setInterval(kw,vg)),Vl=1,pg(fv))}}function dv(n,l,i){var o=new du;return l=l==null?0:+l,o.restart(c=>{o.stop(),n(c+l)},l,i),o}var Qw=Sd("start","end","cancel","interrupt"),Kw=[],bg=0,hv=1,Uf=2,nu=3,mv=4,Hf=5,au=6;function Uu(n,l,i,o,c,f){var d=n.__transition;if(!d)n.__transition={};else if(i in d)return;Jw(n,i,{name:l,index:o,group:c,on:Qw,tween:Kw,time:f.time,delay:f.delay,duration:f.duration,ease:f.ease,timer:null,state:bg})}function Td(n,l){var i=an(n,l);if(i.state>bg)throw new Error("too late; already scheduled");return i}function bn(n,l){var i=an(n,l);if(i.state>nu)throw new Error("too late; already running");return i}function an(n,l){var i=n.__transition;if(!i||!(i=i[l]))throw new Error("transition not found");return i}function Jw(n,l,i){var o=n.__transition,c;o[l]=i,i.timer=gg(f,0,i.time);function f(m){i.state=hv,i.timer.restart(d,i.delay,i.time),i.delay<=m&&d(m-i.delay)}function d(m){var v,S,_,x;if(i.state!==hv)return y();for(v in o)if(x=o[v],x.name===i.name){if(x.state===nu)return dv(d);x.state===mv?(x.state=au,x.timer.stop(),x.on.call("interrupt",n,n.__data__,x.index,x.group),delete o[v]):+v<l&&(x.state=au,x.timer.stop(),x.on.call("cancel",n,n.__data__,x.index,x.group),delete o[v])}if(dv(function(){i.state===nu&&(i.state=mv,i.timer.restart(h,i.delay,i.time),h(m))}),i.state=Uf,i.on.call("start",n,n.__data__,i.index,i.group),i.state===Uf){for(i.state=nu,c=new Array(_=i.tween.length),v=0,S=-1;v<_;++v)(x=i.tween[v].value.call(n,n.__data__,i.index,i.group))&&(c[++S]=x);c.length=S+1}}function h(m){for(var v=m<i.duration?i.ease.call(null,m/i.duration):(i.timer.restart(y),i.state=Hf,1),S=-1,_=c.length;++S<_;)c[S].call(n,v);i.state===Hf&&(i.on.call("end",n,n.__data__,i.index,i.group),y())}function y(){i.state=au,i.timer.stop(),delete o[l];for(var m in o)return;delete n.__transition}}function lu(n,l){var i=n.__transition,o,c,f=!0,d;if(i){l=l==null?null:l+"";for(d in i){if((o=i[d]).name!==l){f=!1;continue}c=o.state>Uf&&o.state<Hf,o.state=au,o.timer.stop(),o.on.call(c?"interrupt":"cancel",n,n.__data__,o.index,o.group),delete i[d]}f&&delete n.__transition}}function Ww(n){return this.each(function(){lu(this,n)})}function Pw(n,l){var i,o;return function(){var c=bn(this,n),f=c.tween;if(f!==i){o=i=f;for(var d=0,h=o.length;d<h;++d)if(o[d].name===l){o=o.slice(),o.splice(d,1);break}}c.tween=o}}function Fw(n,l,i){var o,c;if(typeof i!="function")throw new Error;return function(){var f=bn(this,n),d=f.tween;if(d!==o){c=(o=d).slice();for(var h={name:l,value:i},y=0,m=c.length;y<m;++y)if(c[y].name===l){c[y]=h;break}y===m&&c.push(h)}f.tween=c}}function Iw(n,l){var i=this._id;if(n+="",arguments.length<2){for(var o=an(this.node(),i).tween,c=0,f=o.length,d;c<f;++c)if((d=o[c]).name===n)return d.value;return null}return this.each((l==null?Pw:Fw)(i,n,l))}function wd(n,l,i){var o=n._id;return n.each(function(){var c=bn(this,o);(c.value||(c.value={}))[l]=i.apply(this,arguments)}),function(c){return an(c,o).value[l]}}function Sg(n,l){var i;return(typeof l=="number"?ya:l instanceof fi?uv:(i=fi(l))?(l=i,uv):Dw)(n,l)}function tE(n){return function(){this.removeAttribute(n)}}function eE(n){return function(){this.removeAttributeNS(n.space,n.local)}}function nE(n,l,i){var o,c=i+"",f;return function(){var d=this.getAttribute(n);return d===c?null:d===o?f:f=l(o=d,i)}}function aE(n,l,i){var o,c=i+"",f;return function(){var d=this.getAttributeNS(n.space,n.local);return d===c?null:d===o?f:f=l(o=d,i)}}function lE(n,l,i){var o,c,f;return function(){var d,h=i(this),y;return h==null?void this.removeAttribute(n):(d=this.getAttribute(n),y=h+"",d===y?null:d===o&&y===c?f:(c=y,f=l(o=d,h)))}}function rE(n,l,i){var o,c,f;return function(){var d,h=i(this),y;return h==null?void this.removeAttributeNS(n.space,n.local):(d=this.getAttributeNS(n.space,n.local),y=h+"",d===y?null:d===o&&y===c?f:(c=y,f=l(o=d,h)))}}function iE(n,l){var i=Du(n),o=i==="transform"?$w:Sg;return this.attrTween(n,typeof l=="function"?(i.local?rE:lE)(i,o,wd(this,"attr."+n,l)):l==null?(i.local?eE:tE)(i):(i.local?aE:nE)(i,o,l))}function oE(n,l){return function(i){this.setAttribute(n,l.call(this,i))}}function uE(n,l){return function(i){this.setAttributeNS(n.space,n.local,l.call(this,i))}}function cE(n,l){var i,o;function c(){var f=l.apply(this,arguments);return f!==o&&(i=(o=f)&&uE(n,f)),i}return c._value=l,c}function sE(n,l){var i,o;function c(){var f=l.apply(this,arguments);return f!==o&&(i=(o=f)&&oE(n,f)),i}return c._value=l,c}function fE(n,l){var i="attr."+n;if(arguments.length<2)return(i=this.tween(i))&&i._value;if(l==null)return this.tween(i,null);if(typeof l!="function")throw new Error;var o=Du(n);return this.tween(i,(o.local?cE:sE)(o,l))}function dE(n,l){return function(){Td(this,n).delay=+l.apply(this,arguments)}}function hE(n,l){return l=+l,function(){Td(this,n).delay=l}}function mE(n){var l=this._id;return arguments.length?this.each((typeof n=="function"?dE:hE)(l,n)):an(this.node(),l).delay}function yE(n,l){return function(){bn(this,n).duration=+l.apply(this,arguments)}}function vE(n,l){return l=+l,function(){bn(this,n).duration=l}}function pE(n){var l=this._id;return arguments.length?this.each((typeof n=="function"?yE:vE)(l,n)):an(this.node(),l).duration}function gE(n,l){if(typeof l!="function")throw new Error;return function(){bn(this,n).ease=l}}function bE(n){var l=this._id;return arguments.length?this.each(gE(l,n)):an(this.node(),l).ease}function SE(n,l){return function(){var i=l.apply(this,arguments);if(typeof i!="function")throw new Error;bn(this,n).ease=i}}function _E(n){if(typeof n!="function")throw new Error;return this.each(SE(this._id,n))}function xE(n){typeof n!="function"&&(n=Ip(n));for(var l=this._groups,i=l.length,o=new Array(i),c=0;c<i;++c)for(var f=l[c],d=f.length,h=o[c]=[],y,m=0;m<d;++m)(y=f[m])&&n.call(y,y.__data__,m,f)&&h.push(y);return new qn(o,this._parents,this._name,this._id)}function TE(n){if(n._id!==this._id)throw new Error;for(var l=this._groups,i=n._groups,o=l.length,c=i.length,f=Math.min(o,c),d=new Array(o),h=0;h<f;++h)for(var y=l[h],m=i[h],v=y.length,S=d[h]=new Array(v),_,x=0;x<v;++x)(_=y[x]||m[x])&&(S[x]=_);for(;h<o;++h)d[h]=l[h];return new qn(d,this._parents,this._name,this._id)}function wE(n){return(n+"").trim().split(/^|\s+/).every(function(l){var i=l.indexOf(".");return i>=0&&(l=l.slice(0,i)),!l||l==="start"})}function EE(n,l,i){var o,c,f=wE(l)?Td:bn;return function(){var d=f(this,n),h=d.on;h!==o&&(c=(o=h).copy()).on(l,i),d.on=c}}function CE(n,l){var i=this._id;return arguments.length<2?an(this.node(),i).on.on(n):this.each(EE(i,n,l))}function AE(n){return function(){var l=this.parentNode;for(var i in this.__transition)if(+i!==n)return;l&&l.removeChild(this)}}function jE(){return this.on("end.remove",AE(this._id))}function NE(n){var l=this._name,i=this._id;typeof n!="function"&&(n=gd(n));for(var o=this._groups,c=o.length,f=new Array(c),d=0;d<c;++d)for(var h=o[d],y=h.length,m=f[d]=new Array(y),v,S,_=0;_<y;++_)(v=h[_])&&(S=n.call(v,v.__data__,_,h))&&("__data__"in v&&(S.__data__=v.__data__),m[_]=S,Uu(m[_],l,i,_,m,an(v,i)));return new qn(f,this._parents,l,i)}function OE(n){var l=this._name,i=this._id;typeof n!="function"&&(n=Fp(n));for(var o=this._groups,c=o.length,f=[],d=[],h=0;h<c;++h)for(var y=o[h],m=y.length,v,S=0;S<m;++S)if(v=y[S]){for(var _=n.call(v,v.__data__,S,y),x,w=an(v,i),N=0,B=_.length;N<B;++N)(x=_[N])&&Uu(x,l,i,N,_,w);f.push(_),d.push(v)}return new qn(f,d,l,i)}var ME=Ti.prototype.constructor;function RE(){return new ME(this._groups,this._parents)}function zE(n,l){var i,o,c;return function(){var f=Gl(this,n),d=(this.style.removeProperty(n),Gl(this,n));return f===d?null:f===i&&d===o?c:c=l(i=f,o=d)}}function _g(n){return function(){this.style.removeProperty(n)}}function DE(n,l,i){var o,c=i+"",f;return function(){var d=Gl(this,n);return d===c?null:d===o?f:f=l(o=d,i)}}function BE(n,l,i){var o,c,f;return function(){var d=Gl(this,n),h=i(this),y=h+"";return h==null&&(y=h=(this.style.removeProperty(n),Gl(this,n))),d===y?null:d===o&&y===c?f:(c=y,f=l(o=d,h))}}function UE(n,l){var i,o,c,f="style."+l,d="end."+f,h;return function(){var y=bn(this,n),m=y.on,v=y.value[f]==null?h||(h=_g(l)):void 0;(m!==i||c!==v)&&(o=(i=m).copy()).on(d,c=v),y.on=o}}function HE(n,l,i){var o=(n+="")=="transform"?Hw:Sg;return l==null?this.styleTween(n,zE(n,o)).on("end.style."+n,_g(n)):typeof l=="function"?this.styleTween(n,BE(n,o,wd(this,"style."+n,l))).each(UE(this._id,n)):this.styleTween(n,DE(n,o,l),i).on("end.style."+n,null)}function $E(n,l,i){return function(o){this.style.setProperty(n,l.call(this,o),i)}}function LE(n,l,i){var o,c;function f(){var d=l.apply(this,arguments);return d!==c&&(o=(c=d)&&$E(n,d,i)),o}return f._value=l,f}function YE(n,l,i){var o="style."+(n+="");if(arguments.length<2)return(o=this.tween(o))&&o._value;if(l==null)return this.tween(o,null);if(typeof l!="function")throw new Error;return this.tween(o,LE(n,l,i??""))}function qE(n){return function(){this.textContent=n}}function XE(n){return function(){var l=n(this);this.textContent=l??""}}function GE(n){return this.tween("text",typeof n=="function"?XE(wd(this,"text",n)):qE(n==null?"":n+""))}function VE(n){return function(l){this.textContent=n.call(this,l)}}function kE(n){var l,i;function o(){var c=n.apply(this,arguments);return c!==i&&(l=(i=c)&&VE(c)),l}return o._value=n,o}function ZE(n){var l="text";if(arguments.length<1)return(l=this.tween(l))&&l._value;if(n==null)return this.tween(l,null);if(typeof n!="function")throw new Error;return this.tween(l,kE(n))}function QE(){for(var n=this._name,l=this._id,i=xg(),o=this._groups,c=o.length,f=0;f<c;++f)for(var d=o[f],h=d.length,y,m=0;m<h;++m)if(y=d[m]){var v=an(y,l);Uu(y,n,i,m,d,{time:v.time+v.delay+v.duration,delay:0,duration:v.duration,ease:v.ease})}return new qn(o,this._parents,n,i)}function KE(){var n,l,i=this,o=i._id,c=i.size();return new Promise(function(f,d){var h={value:d},y={value:function(){--c===0&&f()}};i.each(function(){var m=bn(this,o),v=m.on;v!==n&&(l=(n=v).copy(),l._.cancel.push(h),l._.interrupt.push(h),l._.end.push(y)),m.on=l}),c===0&&f()})}var JE=0;function qn(n,l,i,o){this._groups=n,this._parents=l,this._name=i,this._id=o}function xg(){return++JE}var Un=Ti.prototype;qn.prototype={constructor:qn,select:NE,selectAll:OE,selectChild:Un.selectChild,selectChildren:Un.selectChildren,filter:xE,merge:TE,selection:RE,transition:QE,call:Un.call,nodes:Un.nodes,node:Un.node,size:Un.size,empty:Un.empty,each:Un.each,on:CE,attr:iE,attrTween:fE,style:HE,styleTween:YE,text:GE,textTween:ZE,remove:jE,tween:Iw,delay:mE,duration:pE,ease:bE,easeVarying:_E,end:KE,[Symbol.iterator]:Un[Symbol.iterator]};function WE(n){return((n*=2)<=1?n*n*n:(n-=2)*n*n+2)/2}var PE={time:null,delay:0,duration:250,ease:WE};function FE(n,l){for(var i;!(i=n.__transition)||!(i=i[l]);)if(!(n=n.parentNode))throw new Error(`transition ${l} not found`);return i}function IE(n){var l,i;n instanceof qn?(l=n._id,n=n._name):(l=xg(),(i=PE).time=xd(),n=n==null?null:n+"");for(var o=this._groups,c=o.length,f=0;f<c;++f)for(var d=o[f],h=d.length,y,m=0;m<h;++m)(y=d[m])&&Uu(y,n,l,m,d,i||FE(y,l));return new qn(o,this._parents,n,l)}Ti.prototype.interrupt=Ww;Ti.prototype.transition=IE;const Wo=n=>()=>n;function t4(n,{sourceEvent:l,target:i,transform:o,dispatch:c}){Object.defineProperties(this,{type:{value:n,enumerable:!0,configurable:!0},sourceEvent:{value:l,enumerable:!0,configurable:!0},target:{value:i,enumerable:!0,configurable:!0},transform:{value:o,enumerable:!0,configurable:!0},_:{value:c}})}function Yn(n,l,i){this.k=n,this.x=l,this.y=i}Yn.prototype={constructor:Yn,scale:function(n){return n===1?this:new Yn(this.k*n,this.x,this.y)},translate:function(n,l){return n===0&l===0?this:new Yn(this.k,this.x+this.k*n,this.y+this.k*l)},apply:function(n){return[n[0]*this.k+this.x,n[1]*this.k+this.y]},applyX:function(n){return n*this.k+this.x},applyY:function(n){return n*this.k+this.y},invert:function(n){return[(n[0]-this.x)/this.k,(n[1]-this.y)/this.k]},invertX:function(n){return(n-this.x)/this.k},invertY:function(n){return(n-this.y)/this.k},rescaleX:function(n){return n.copy().domain(n.range().map(this.invertX,this).map(n.invert,n))},rescaleY:function(n){return n.copy().domain(n.range().map(this.invertY,this).map(n.invert,n))},toString:function(){return"translate("+this.x+","+this.y+") scale("+this.k+")"}};var li=new Yn(1,0,0);Yn.prototype;function pf(n){n.stopImmediatePropagation()}function ei(n){n.preventDefault(),n.stopImmediatePropagation()}function e4(n){return(!n.ctrlKey||n.type==="wheel")&&!n.button}function n4(){var n=this;return n instanceof SVGElement?(n=n.ownerSVGElement||n,n.hasAttribute("viewBox")?(n=n.viewBox.baseVal,[[n.x,n.y],[n.x+n.width,n.y+n.height]]):[[0,0],[n.width.baseVal.value,n.height.baseVal.value]]):[[0,0],[n.clientWidth,n.clientHeight]]}function yv(){return this.__zoom||li}function a4(n){return-n.deltaY*(n.deltaMode===1?.05:n.deltaMode?1:.002)*(n.ctrlKey?10:1)}function l4(){return navigator.maxTouchPoints||"ontouchstart"in this}function r4(n,l,i){var o=n.invertX(l[0][0])-i[0][0],c=n.invertX(l[1][0])-i[1][0],f=n.invertY(l[0][1])-i[0][1],d=n.invertY(l[1][1])-i[1][1];return n.translate(c>o?(o+c)/2:Math.min(0,o)||Math.max(0,c),d>f?(f+d)/2:Math.min(0,f)||Math.max(0,d))}function vv(){var n=e4,l=n4,i=r4,o=a4,c=l4,f=[0,1/0],d=[[-1/0,-1/0],[1/0,1/0]],h=250,y=Xw,m=Sd("start","zoom","end"),v,S,_,x=500,w=150,N=0,B=10;function O(R){R.property("__zoom",yv).on("wheel.zoom",W,{passive:!1}).on("mousedown.zoom",it).on("dblclick.zoom",ht).filter(c).on("touchstart.zoom",pt).on("touchmove.zoom",at).on("touchend.zoom touchcancel.zoom",lt).style("-webkit-tap-highlight-color","rgba(0,0,0,0)")}O.transform=function(R,Q,E,Y){var k=R.selection?R.selection():R;k.property("__zoom",yv),R!==k?$(R,Q,E,Y):k.interrupt().each(function(){V(this,arguments).event(Y).start().zoom(null,typeof Q=="function"?Q.apply(this,arguments):Q).end()})},O.scaleBy=function(R,Q,E,Y){O.scaleTo(R,function(){var k=this.__zoom.k,et=typeof Q=="function"?Q.apply(this,arguments):Q;return k*et},E,Y)},O.scaleTo=function(R,Q,E,Y){O.transform(R,function(){var k=l.apply(this,arguments),et=this.__zoom,T=E==null?L(k):typeof E=="function"?E.apply(this,arguments):E,X=et.invert(T),I=typeof Q=="function"?Q.apply(this,arguments):Q;return i(z(j(et,I),T,X),k,d)},E,Y)},O.translateBy=function(R,Q,E,Y){O.transform(R,function(){return i(this.__zoom.translate(typeof Q=="function"?Q.apply(this,arguments):Q,typeof E=="function"?E.apply(this,arguments):E),l.apply(this,arguments),d)},null,Y)},O.translateTo=function(R,Q,E,Y,k){O.transform(R,function(){var et=l.apply(this,arguments),T=this.__zoom,X=Y==null?L(et):typeof Y=="function"?Y.apply(this,arguments):Y;return i(li.translate(X[0],X[1]).scale(T.k).translate(typeof Q=="function"?-Q.apply(this,arguments):-Q,typeof E=="function"?-E.apply(this,arguments):-E),et,d)},Y,k)};function j(R,Q){return Q=Math.max(f[0],Math.min(f[1],Q)),Q===R.k?R:new Yn(Q,R.x,R.y)}function z(R,Q,E){var Y=Q[0]-E[0]*R.k,k=Q[1]-E[1]*R.k;return Y===R.x&&k===R.y?R:new Yn(R.k,Y,k)}function L(R){return[(+R[0][0]+ +R[1][0])/2,(+R[0][1]+ +R[1][1])/2]}function $(R,Q,E,Y){R.on("start.zoom",function(){V(this,arguments).event(Y).start()}).on("interrupt.zoom end.zoom",function(){V(this,arguments).event(Y).end()}).tween("zoom",function(){var k=this,et=arguments,T=V(k,et).event(Y),X=l.apply(k,et),I=E==null?L(X):typeof E=="function"?E.apply(k,et):E,P=Math.max(X[1][0]-X[0][0],X[1][1]-X[0][1]),nt=k.__zoom,ut=typeof Q=="function"?Q.apply(k,et):Q,rt=y(nt.invert(I).concat(P/nt.k),ut.invert(I).concat(P/ut.k));return function(bt){if(bt===1)bt=ut;else{var vt=rt(bt),Ut=P/vt[2];bt=new Yn(Ut,I[0]-vt[0]*Ut,I[1]-vt[1]*Ut)}T.zoom(null,bt)}})}function V(R,Q,E){return!E&&R.__zooming||new Z(R,Q)}function Z(R,Q){this.that=R,this.args=Q,this.active=0,this.sourceEvent=null,this.extent=l.apply(R,Q),this.taps=0}Z.prototype={event:function(R){return R&&(this.sourceEvent=R),this},start:function(){return++this.active===1&&(this.that.__zooming=this,this.emit("start")),this},zoom:function(R,Q){return this.mouse&&R!=="mouse"&&(this.mouse[1]=Q.invert(this.mouse[0])),this.touch0&&R!=="touch"&&(this.touch0[1]=Q.invert(this.touch0[0])),this.touch1&&R!=="touch"&&(this.touch1[1]=Q.invert(this.touch1[0])),this.that.__zoom=Q,this.emit("zoom"),this},end:function(){return--this.active===0&&(delete this.that.__zooming,this.emit("end")),this},emit:function(R){var Q=ye(this.that).datum();m.call(R,this.that,new t4(R,{sourceEvent:this.sourceEvent,target:O,transform:this.that.__zoom,dispatch:m}),Q)}};function W(R,...Q){if(!n.apply(this,arguments))return;var E=V(this,Q).event(R),Y=this.__zoom,k=Math.max(f[0],Math.min(f[1],Y.k*Math.pow(2,o.apply(this,arguments)))),et=ma(R);if(E.wheel)(E.mouse[0][0]!==et[0]||E.mouse[0][1]!==et[1])&&(E.mouse[1]=Y.invert(E.mouse[0]=et)),clearTimeout(E.wheel);else{if(Y.k===k)return;E.mouse=[et,Y.invert(et)],lu(this),E.start()}ei(R),E.wheel=setTimeout(T,w),E.zoom("mouse",i(z(j(Y,k),E.mouse[0],E.mouse[1]),E.extent,d));function T(){E.wheel=null,E.end()}}function it(R,...Q){if(_||!n.apply(this,arguments))return;var E=R.currentTarget,Y=V(this,Q,!0).event(R),k=ye(R.view).on("mousemove.zoom",I,!0).on("mouseup.zoom",P,!0),et=ma(R,E),T=R.clientX,X=R.clientY;yw(R.view),pf(R),Y.mouse=[et,this.__zoom.invert(et)],lu(this),Y.start();function I(nt){if(ei(nt),!Y.moved){var ut=nt.clientX-T,rt=nt.clientY-X;Y.moved=ut*ut+rt*rt>N}Y.event(nt).zoom("mouse",i(z(Y.that.__zoom,Y.mouse[0]=ma(nt,E),Y.mouse[1]),Y.extent,d))}function P(nt){k.on("mousemove.zoom mouseup.zoom",null),vw(nt.view,Y.moved),ei(nt),Y.event(nt).end()}}function ht(R,...Q){if(n.apply(this,arguments)){var E=this.__zoom,Y=ma(R.changedTouches?R.changedTouches[0]:R,this),k=E.invert(Y),et=E.k*(R.shiftKey?.5:2),T=i(z(j(E,et),Y,k),l.apply(this,Q),d);ei(R),h>0?ye(this).transition().duration(h).call($,T,Y,R):ye(this).call(O.transform,T,Y,R)}}function pt(R,...Q){if(n.apply(this,arguments)){var E=R.touches,Y=E.length,k=V(this,Q,R.changedTouches.length===Y).event(R),et,T,X,I;for(pf(R),T=0;T<Y;++T)X=E[T],I=ma(X,this),I=[I,this.__zoom.invert(I),X.identifier],k.touch0?!k.touch1&&k.touch0[2]!==I[2]&&(k.touch1=I,k.taps=0):(k.touch0=I,et=!0,k.taps=1+!!v);v&&(v=clearTimeout(v)),et&&(k.taps<2&&(S=I[0],v=setTimeout(function(){v=null},x)),lu(this),k.start())}}function at(R,...Q){if(this.__zooming){var E=V(this,Q).event(R),Y=R.changedTouches,k=Y.length,et,T,X,I;for(ei(R),et=0;et<k;++et)T=Y[et],X=ma(T,this),E.touch0&&E.touch0[2]===T.identifier?E.touch0[0]=X:E.touch1&&E.touch1[2]===T.identifier&&(E.touch1[0]=X);if(T=E.that.__zoom,E.touch1){var P=E.touch0[0],nt=E.touch0[1],ut=E.touch1[0],rt=E.touch1[1],bt=(bt=ut[0]-P[0])*bt+(bt=ut[1]-P[1])*bt,vt=(vt=rt[0]-nt[0])*vt+(vt=rt[1]-nt[1])*vt;T=j(T,Math.sqrt(bt/vt)),X=[(P[0]+ut[0])/2,(P[1]+ut[1])/2],I=[(nt[0]+rt[0])/2,(nt[1]+rt[1])/2]}else if(E.touch0)X=E.touch0[0],I=E.touch0[1];else return;E.zoom("touch",i(z(T,X,I),E.extent,d))}}function lt(R,...Q){if(this.__zooming){var E=V(this,Q).event(R),Y=R.changedTouches,k=Y.length,et,T;for(pf(R),_&&clearTimeout(_),_=setTimeout(function(){_=null},x),et=0;et<k;++et)T=Y[et],E.touch0&&E.touch0[2]===T.identifier?delete E.touch0:E.touch1&&E.touch1[2]===T.identifier&&delete E.touch1;if(E.touch1&&!E.touch0&&(E.touch0=E.touch1,delete E.touch1),E.touch0)E.touch0[1]=this.__zoom.invert(E.touch0[0]);else if(E.end(),E.taps===2&&(T=ma(T,this),Math.hypot(S[0]-T[0],S[1]-T[1])<B)){var X=ye(this).on("dblclick.zoom");X&&X.apply(this,arguments)}}}return O.wheelDelta=function(R){return arguments.length?(o=typeof R=="function"?R:Wo(+R),O):o},O.filter=function(R){return arguments.length?(n=typeof R=="function"?R:Wo(!!R),O):n},O.touchable=function(R){return arguments.length?(c=typeof R=="function"?R:Wo(!!R),O):c},O.extent=function(R){return arguments.length?(l=typeof R=="function"?R:Wo([[+R[0][0],+R[0][1]],[+R[1][0],+R[1][1]]]),O):l},O.scaleExtent=function(R){return arguments.length?(f[0]=+R[0],f[1]=+R[1],O):[f[0],f[1]]},O.translateExtent=function(R){return arguments.length?(d[0][0]=+R[0][0],d[1][0]=+R[1][0],d[0][1]=+R[0][1],d[1][1]=+R[1][1],O):[[d[0][0],d[0][1]],[d[1][0],d[1][1]]]},O.constrain=function(R){return arguments.length?(i=R,O):i},O.duration=function(R){return arguments.length?(h=+R,O):h},O.interpolate=function(R){return arguments.length?(y=R,O):y},O.on=function(){var R=m.on.apply(m,arguments);return R===m?O:R},O.clickDistance=function(R){return arguments.length?(N=(R=+R)*R,O):Math.sqrt(N)},O.tapDistance=function(R){return arguments.length?(B=+R,O):B},O}const he={DEFAULT_MIN_ZOOM:.1,DEFAULT_MAX_ZOOM:4,FIT_TO_VIEW_MAX_ZOOM:1,INITIAL_POSITION_X:0,INITIAL_POSITION_Y:0,SCROLL_POS_RATIO:3,DEFAULT_LAYOUT:1e6,INITIAL_SCROLLBAR_SIZE:50,MIN_SCROLLBAR_SIZE:15,SCROLL_DELTA_DIFF:1.5},hu={TOP_LEFT:"top-left",TOP_RIGHT:"top-right",BOTTOM_LEFT:"bottom-left",BOTTOM_CENTER:"bottom-center"},Ke={...hu,TOP_CENTER:"top-center",BOTTOM_RIGHT:"bottom-right",CENTER_LEFT:"center-left",CENTER_RIGHT:"center-right",CENTER_CENTER:"center-center"},ri={BLOCK_EVENTS:"react-infinite-canvas-block-events",BLOCK_SCROLL_CLASS:"react-infinite-canvas-block-scroll",BLOCK_ZOOM_CLASS:"react-infinite-canvas-block-zoom",BLOCK_PAN_CLASS:"react-infinite-canvas-block-pan"},ii=({value:n,min:l=0,max:i})=>Math.min(Math.max(n,l),i),i4=({position:n,svgBounds:l,nodeBounds:i,currentTranslateX:o,currentTranslateY:c,currentScale:f,updatedScale:d,customOffset:h})=>{let[y,m,v]=[0,0,1];switch(d!==f&&(v=d/f,o*=v,c*=v),n){case Ke.TOP_LEFT:return{updatedX:o+((l.x-i.x)*v+h.x),updatedY:c+((l.y-i.y)*v+h.y)};case Ke.TOP_CENTER:return y=l.width/2-(i.x*v+i.width*v/2)+l.x,{updatedX:o+y+h.x,updatedY:c+((l.y-i.y)*v+h.y)};case Ke.TOP_RIGHT:return{updatedX:o+(l.width-(Math.abs(l.x-i.x*v)+i.width*v))+h.x,updatedY:c+((l.y-i.y)*v+h.y)};case Ke.CENTER_LEFT:return m=l.height/2-(Math.abs(l.y-i.y)*v+i.height*v/2),{updatedX:o+((l.x-i.x)*v+h.x),updatedY:c+m+h.y};case Ke.CENTER_CENTER:return y=l.width/2-(Math.abs(l.x-i.x)*v+i.width*v/2),m=l.height/2-(Math.abs(l.y-i.y)*v+i.height*v/2),{updatedX:o+y+h.x,updatedY:c+m+h.y};case Ke.CENTER_RIGHT:return m=l.height/2-(Math.abs(l.y-i.y)*v+i.height*v/2),{updatedX:o+(l.width-(Math.abs(l.x-i.x)*v+i.width*v))+h.x,updatedY:c+m+h.y};case Ke.BOTTOM_LEFT:return{updatedX:o+((l.x-i.x)*v+h.x),updatedY:c+(l.height-(Math.abs(l.y-i.y)*v+i.height*v))+h.y};case Ke.BOTTOM_CENTER:return y=l.width/2-(Math.abs(l.x-i.x)*v+i.width*v/2),{updatedX:o+y+h.x,updatedY:c+(l.height-(Math.abs(l.y-i.y)*v+i.height*v))+h.y};case Ke.BOTTOM_RIGHT:return{updatedX:o+(l.width-(Math.abs(l.x-i.x)*v+i.width*v))+h.x,updatedY:c+(l.height-(Math.abs(l.y-i.y)*v+i.height))+h.y};default:return{updatedX:o,updatedY:c}}},pv=n=>{const[l,i]=[Math.max(n-.2,.2),n];return[he.INITIAL_SCROLLBAR_SIZE/l,he.INITIAL_SCROLLBAR_SIZE/i]},o4=({isVertical:n,state:l,scrollDelta:i})=>{const o=n?l.verticalPos:l.horizontalPos,c=n?l.verticalSize:l.horizontalSize,f=n?l.verticalSizeDecrease:l.horizontalSizeDecrease,d=n?i.deltaY:i.deltaX;return{scrollPos:o,scrollSize:c,scrollSizeDecrease:f,deltaValue:d}},gv=({isVertical:n=!0,state:l,scrollDelta:i,scrollLength:o})=>{const{scrollPos:c,scrollSize:f,scrollSizeDecrease:d,deltaValue:h}=o4({isVertical:n,state:l,scrollDelta:i}),y=h/10||0,m=h/100||0;let[v,S]=[c+y,d];if(v<1)v=0,S-=Math.abs(m);else if(v+f>o){const _=o-Math.max(he.MIN_SCROLLBAR_SIZE,f-Math.abs(S));v=Math.min(o-(f+S),_),S-=m}else if(S!==0){const _=v>o/2;v=_?o-f:0,_?(S-=m,S<=0&&(S=0)):(S+=Math.abs(m),S>=0&&(S=0))}return[v,S]},u4=n=>{const l=n.target;return!!(l.closest(`.${ri.BLOCK_PAN_CLASS}`)||l.closest(`.${ri.BLOCK_EVENTS}`))},c4=n=>{const l=n.target;if(l.closest(`.${ri.BLOCK_EVENTS}`))return!0;const i=n.ctrlKey||n.metaKey;return!!(!i&&l.closest(`.${ri.BLOCK_SCROLL_CLASS}`)||i&&l.closest(`.${ri.BLOCK_ZOOM_CLASS}`))},s4="_dotSvgContainer_64xbd_1",f4={dotSvgContainer:s4},d4=({id:n="",size:l=1,minSize:i=.3,maxZoom:o=4,gap:c=20,zoomTransform:f={scale:1,translateX:0,translateY:0},className:d="",minOpacity:h=.8,maxOpacity:y=1,elementColor:m="#afb7c7",backgroundColor:v})=>{const{scale:S,translateX:_,translateY:x}=f,w=ii({value:S*10/(o*10),min:h,max:y}),N=c*S,B=l*S,O=Math.max(i,B),j=`patternId-${n}`;return g.jsxs("svg",{className:`${d} ${f4.dotSvgContainer}`,style:v?{backgroundColor:v}:{},role:"img","aria-label":"Background pattern",children:[g.jsx("pattern",{id:j,x:_%N,y:x%N,width:N,height:N,patternUnits:"userSpaceOnUse",patternTransform:`translate(-${B},-${B})`,children:g.jsx("circle",{cx:O,cy:O,r:O,fill:m,opacity:w})}),g.jsx("rect",{x:"0",y:"0",width:"100%",height:"100%",fill:`url(#${j})`})]})},h4="_container_1xxyl_1",m4="_canvasWrapper_1xxyl_8",y4="_panning_1xxyl_21",v4="_canvas_1xxyl_8",p4="_contentWrapper_1xxyl_39",Ua={container:h4,canvasWrapper:m4,panning:y4,canvas:v4,contentWrapper:p4},g4="_verticalScrollBar_17thb_1",b4="_horizontalScrollBar_17thb_9",bv={verticalScrollBar:g4,horizontalScrollBar:b4},S4=C.forwardRef(({scale:n,startingPosition:l,offset:i={x:0,y:0},color:o="gray",thickness:c="8px",minSize:f="15px",verticalOffsetHeight:d,horizontalOffsetWidth:h,onScrollDeltaHandler:y,getContainerOffset:m},v)=>{const[S,_]=pv(n),x=C.useRef(null),w=C.useRef(null),N=C.useRef({isDragging:!1,vertical:!0,initialOffset:0}),[B,O]=C.useState({scale:n,horizontalSize:_,horizontalPos:l?l.x:(h??0)/he.SCROLL_POS_RATIO,horizontalSizeDecrease:0,verticalSize:S,verticalPos:l?l.y:(d??0)/he.SCROLL_POS_RATIO,verticalSizeDecrease:0});C.useImperativeHandle(v,()=>({resetScrollPos:()=>{O(Z=>({...Z,horizontalPos:(h??0)/he.SCROLL_POS_RATIO,verticalPos:(d??0)/he.SCROLL_POS_RATIO,verticalSizeDecrease:0,horizontalSizeDecrease:0}))},onScrollDeltaChangeHandler:j,onMouseUp:L})),C.useEffect(function(){O(Z=>{let W;const[it,ht]=pv(n),pt=n/Z.scale||0;return n<Z.scale?(W={horizontalPos:Z.horizontalPos-pt,verticalPos:Z.verticalPos-pt},{...Z,scale:n,horizontalSize:ht,verticalSize:it,...W}):(W={horizontalPos:Z.horizontalPos+pt,verticalPos:Z.verticalPos+pt},{...Z,scale:n,horizontalSize:ht,verticalSize:it,...W})})},[n]);function j(Z){N.current.isDragging||O(W=>{const[it,ht]=gv({state:W,scrollDelta:Z,scrollLength:d??0}),[pt,at]=gv({isVertical:!1,state:W,scrollDelta:Z,scrollLength:h??0});return{...W,horizontalPos:pt,horizontalSizeDecrease:at,verticalPos:it,verticalSizeDecrease:ht}})}const z=C.useCallback(function(Z,W){if(!N.current.isDragging)return;Z.stopPropagation();const{movementX:it,movementY:ht,clientX:pt,clientY:at,offsetX:lt,offsetY:R}=Z;N.current.initialOffset<1&&(N.current.initialOffset=W?R:lt);const Q=W?ht:it,E=(W?at:pt)-N.current.initialOffset-m(W),Y=T=>W?["verticalPos",T.verticalSize,d??0,i.y]:["horizontalPos",T.horizontalSize,h??0,i.x];O(T=>{const[X,I,P,nt]=Y(T);return{...T,[X]:ii({value:E,min:nt,max:P-I})}});const k=()=>{const T=Q>0?Q+he.SCROLL_DELTA_DIFF:Q-he.SCROLL_DELTA_DIFF;return W?{deltaX:0,deltaY:T}:{deltaX:T,deltaY:0}},et=(W?d??0:h??0)-(W?B.verticalSize:B.horizontalSize);!(E<=0||E>=et)&&Q!==0&&y(k())},[d,h,i.y,i.x]),L=C.useCallback(function(){N.current.isDragging=!1,N.current.initialOffset=0,window.removeEventListener("mousemove",Z=>z(Z,N.current.vertical)),window.removeEventListener("mouseup",L)},[z]),$=C.useCallback(function(){window.addEventListener("mousemove",Z=>z(Z,N.current.vertical)),window.addEventListener("mouseup",L)},[z,L]),V=Z=>{Z.stopPropagation()};return g.jsxs(C.Fragment,{children:[g.jsx("div",{className:bv.verticalScrollBar,style:{width:c},ref:x,children:g.jsx("div",{"data-id":"vertical-scrollbar",style:{insetBlockStart:`${ii({value:B.verticalPos,min:i.y,max:d})}px`,height:`${B.verticalSize+B.verticalSizeDecrease}px`,background:o,minHeight:f},onMouseDownCapture:Z=>{Z.stopPropagation(),N.current={...N.current,isDragging:!0,vertical:!0},$()},onDragEnterCapture:V,onDragCapture:V,onDragEndCapture:V})}),g.jsx("div",{className:bv.horizontalScrollBar,ref:w,style:{height:c},children:g.jsx("div",{"data-id":"horizontal-scrollbar",style:{insetInlineStart:`${ii({value:B.horizontalPos,min:i.x,max:h})}px`,width:`${B.horizontalSize+B.horizontalSizeDecrease}px`,background:o,minWidth:f},onMouseDownCapture:Z=>{Z.stopPropagation(),N.current={...N.current,isDragging:!0,vertical:!1},$()},onDragEnterCapture:V,onDragCapture:V,onDragEndCapture:V})})]})}),$f=/^((?!chrome|android).)*safari/i.test(navigator.userAgent),gf=$f?600:300,_4=C.forwardRef(({children:n,...l},i)=>{const o=Yf.useRef(null);return g.jsx(x4,{innerRef:i,...l,children:g.jsx("div",{ref:o,style:{width:"max-content",height:"max-content"},children:n})})}),x4=C.memo(({children:n,className:l="",innerRef:i,minZoom:o=he.DEFAULT_MIN_ZOOM,maxZoom:c=he.DEFAULT_MAX_ZOOM,panOnScroll:f=!0,customComponents:d=[],scrollBarConfig:h={},backgroundConfig:y={},onCanvasMount:m=()=>{}})=>{const v=C.useRef(null),S=C.useRef(null),_=C.useRef(null),x=C.useRef(null),w=C.useRef(null),N=n.ref,B=C.useRef(null),O=C.useMemo(()=>vv().scaleExtent([o,c]),[c,o]),j=C.useRef(ye(_.current).call(O)),[z,L]=C.useState({translateX:0,translateY:0,scale:1});C.useImperativeHandle(i,()=>({scrollNodeToCenter:({nodeElement:at,offset:lt,scale:R,shouldUpdateMaxScale:Q,maxScale:E,transitionDuration:Y})=>Z({nodeElement:at,offset:lt,scale:R,shouldUpdateMaxScale:Q,maxScale:E,transitionDuration:Y,position:Ke.CENTER_CENTER}),scrollNodeHandler:Z,scrollContentHorizontallyCenter:W,fitContentToView:V,getCanvasState:it})),C.useEffect(function(){j.current=ye(_.current).call(O);const at=ye(x.current);S.current=v.current?v.current.getBoundingClientRect():null,O.filter(lt=>(lt.type==="mousedown"&&!B.current&&(B.current=!0,ht()),lt.ctrlKey||lt.type!=="wheel")).on("zoom",lt=>{var R,Q,E;const Y=(R=lt.sourceEvent)==null?void 0:R.target;if(Y&&u4({target:Y}))return;((Q=lt.sourceEvent)==null?void 0:Q.ctrlKey)===!1&&lt.type==="zoom"&&((E=v.current)==null||E.classList.add(Ua.panning));const k=lt.transform,{x:et,y:T,k:X}=k,I=x.current;L({translateX:et,translateY:T,scale:X}),$f&&I?I.style.transform=`translate(${et}px, ${T}px) scale(${X})`:at.attr("transform",`translate(${et},${T}) scale(${X})`)}),O.on("end",()=>{var lt;B.current=!1,(lt=v.current)==null||lt.classList.remove(Ua.panning)}),m({scrollNodeToCenter:({nodeElement:lt,offset:R,scale:Q,shouldUpdateMaxScale:E,maxScale:Y,transitionDuration:k})=>Z({nodeElement:lt,offset:R,scale:Q,shouldUpdateMaxScale:E,maxScale:Y,transitionDuration:k,position:Ke.CENTER_CENTER}),scrollNodeHandler:Z,scrollContentHorizontallyCenter:W,fitContentToView:V,getCanvasState:it})},[]),j.current.call(vv).on("wheel.zoom",at=>{var lt;if(c4({...at,target:at.target}))return;at.preventDefault();const R=j.current.property("__zoom").k||1;if(f&&!at.ctrlKey){const Q={deltaX:at.deltaX,deltaY:at.deltaY};(lt=w.current)==null||lt.onScrollDeltaChangeHandler(Q),$(Q)}else{const Q=R*2**(-at.deltaY*.01),E=j.current;E&&O.scaleTo(E,Q,ma(at))}},{passive:!1,capture:!0});const $=at=>{const lt=j.current.property("__zoom").k||1,R=j.current;R&&O.translateBy(R,-(at.deltaX/lt),-(at.deltaY/lt))},V=C.useCallback(function({duration:at=500,offset:lt={x:0,y:0},scale:R,maxZoomLimit:Q=he.FIT_TO_VIEW_MAX_ZOOM,disableVerticalCenter:E=!1}){requestIdleCallback(()=>{var Y,k;if(!N.current)return;const et=ye(_.current),T=N.current.getBoundingClientRect(),X=j.current.property("__zoom").k||1,I=(Y=_.current)==null?void 0:Y.getBoundingClientRect(),{width:P=0,height:nt=0}=I||{},ut=1/X,rt=T.width*ut,bt=T.height*ut,vt=nt/bt,Ut=P/rt,ae=R??ii({value:Math.min(Q,Math.min(vt,Ut)),min:o,max:c}),pe=P-rt*ae,$e=nt-bt*ae,ln=!E&&vt>Ut,Sn=pe/2,pa=ln?$e/2:0,ce=Sn+lt.x,Ht=pa+lt.y,Wl=li.translate(ce,Ht).scale(ae);L({translateX:ce,translateY:Ht,scale:ae}),(k=w.current)==null||k.resetScrollPos(),et.transition().duration(at).call(O.transform,Wl)},{timeout:gf})},[c,o]),Z=({nodeElement:at,offset:lt={x:0,y:0},scale:R,shouldUpdateMaxScale:Q=!0,maxScale:E,transitionDuration:Y=300,position:k=Ke.TOP_CENTER})=>{requestIdleCallback(()=>{if(!at)return;const et=j.current.property("__zoom"),{k:T,x:X,y:I}=et,P=ye(_.current),nt=(()=>{const ae=$e=>E?Math.min(E,$e):$e;if(!R)return ae(T);let pe=R;return Q&&(pe=Math.max(R,T)),ae(pe)})(),ut=_.current.getBoundingClientRect(),rt=at.getBoundingClientRect(),{updatedX:bt,updatedY:vt}=i4({position:k,svgBounds:ut,nodeBounds:rt,currentTranslateX:X,currentTranslateY:I,currentScale:T,updatedScale:nt,customOffset:{x:lt.x,y:lt.y}}),Ut=li.translate(bt,vt).scale(nt);P.transition().duration(Y).call(O.transform,Ut)},{timeout:gf})},W=({offset:at=0,transitionDuration:lt=300})=>{N.current&&requestIdleCallback(()=>{const R=j.current.property("__zoom"),{k:Q,y:E}=R,Y=ye(_.current),k=_.current.getBoundingClientRect(),et=N.current.getBoundingClientRect(),T=1/Q,X=et.width*T,I=(k.width-X*Q)/2+at;L({...z,translateX:I});const P=li.translate(I,E).scale(Q);Y.transition().duration(lt).call(O.transform,P)},{timeout:gf})},it=()=>({canvasNode:ye(_.current),zoomNode:ye(x.current),currentPosition:j.current.property("__zoom"),d3Zoom:O}),ht=()=>{const at=document.body;if(at){const lt=new MouseEvent("mousedown",{bubbles:!0,cancelable:!0,view:window});at.dispatchEvent(lt)}},pt=C.useCallback(function(at=!0){var lt,R;const Q=S.current;return at?(lt=Q==null?void 0:Q.top)!=null?lt:0:(R=Q==null?void 0:Q.left)!=null?R:0},[]);return g.jsxs("div",{className:Ua.container,children:[g.jsx("div",{ref:v,className:`${Ua.canvasWrapper} ${l}`,children:$f?g.jsx("div",{ref:_,className:Ua.canvas,children:g.jsx("div",{ref:x,children:g.jsx("div",{className:Ua.contentWrapper,children:n})})}):g.jsx("svg",{ref:_,className:Ua.canvas,"aria-label":"Infinite canvas",role:"application",children:g.jsx("g",{ref:x,children:g.jsx("foreignObject",{x:he.INITIAL_POSITION_X,y:he.INITIAL_POSITION_Y,width:he.DEFAULT_LAYOUT,height:he.DEFAULT_LAYOUT,children:n})})})}),y.disable?null:g.jsx(d4,{maxZoom:c,zoomTransform:z,...y}),h.renderScrollBar&&v.current&&g.jsx(S4,{ref:w,scale:z.scale,...h,verticalOffsetHeight:v.current.offsetHeight,horizontalOffsetWidth:v.current.offsetWidth,getContainerOffset:pt,onScrollDeltaHandler:$}),d.map(at=>{const{component:lt,position:R=hu.BOTTOM_LEFT,offset:Q={x:0,y:0},overlap:E=!0,className:Y=""}=at,k=`${R}-${Q.x}-${Q.y}-${E}`;return g.jsx(T4,{component:lt,position:R,offset:Q,overlap:E,zoomState:{...z,minZoom:o,maxZoom:c},className:Y},k)})]})}),T4=({component:n,position:l,offset:i,overlap:o,zoomState:c,className:f})=>{const d=C.useMemo(()=>{const y=Object.values(hu).includes(l)?l:hu.BOTTOM_LEFT,[m,v]=y.split("-");return{[v]:i.x,[m]:i.y}},[l,i]),h=Yf.cloneElement(n,{zoomState:c});return g.jsx("div",{style:{position:"absolute",...d,zIndex:o?20:1},className:f,children:h})};function w4(n){const l=C.useRef();return g.jsx(ft,{style:{height:"100%"},children:g.jsx(_4,{ref:l,onCanvasMount:i=>{i.fitContentToView({scale:1})},children:n.children})})}function E4(n){return g.jsx("div",{style:{position:"absolute",left:n.x||0,top:n.y||0,backgroundColor:"white",border:"solid black 0.2rem",padding:"1.5rem",borderRadius:"1rem"},children:n.children})}function C4({x1:n,y1:l,x2:i,y2:o,color:c="black",width:f=2}){const d=Math.min(n,i),h=Math.min(l,o),y=Math.abs(i-n),m=Math.abs(o-l);return g.jsxs("svg",{style:{position:"absolute",left:d,top:h,pointerEvents:"none",overflow:"visible",zIndex:0},width:y||1,height:m||1,children:[g.jsx("line",{x1:n-d,y1:l-h,x2:i-d,y2:o-h,stroke:c,strokeWidth:f}),g.jsx("circle",{r:"3",cx:n-d,cy:l-h,fill:c}),g.jsx("circle",{r:"3",cx:i-d,cy:o-h,fill:c})]})}function A4(n){const[l,i]=C.useState(void 0),o=C.useRef(0);return C.useEffect(()=>{if(o.current){let c=o.current;const f=c.getBoundingClientRect();for(;c.parentElement.parentElement.tagName.toLowerCase()!="foreignobject";)c.parentElement&&(c=c.parentElement);const d=c.getBoundingClientRect();i({x:f.left-d.left+f.width,y:f.top-d.top+f.height/2})}},[o]),g.jsxs(g.Fragment,{children:[g.jsx(Wt,{ref:o,children:n.children}),l&&g.jsx(C4,{x1:l.x,y1:l.y,x2:n.x,y2:n.y})]})}function j4(n){const l=n.schema;let i={},o=[],c=0;Object.keys(l).map(d=>{const h={position:{x:c*300,y:0},table_data:[]};Object.keys(l[d]).map(y=>{const m=l[d][y],v={field_name:y,field_type:m.is_list?`list[${m.type}]`:m.type,refears_to:m.depends?m.depends.slice(5):void 0};if(v.refears_to){const S=i[v.refears_to];S&&(c--,h.position={x:S.position.x-400,y:S.position.y-300}),v.field_type={refears_to_name:v.field_type,refears_to_pos:S.position}}h.table_data.push({Field:v.field_name,Type:v.field_type})}),i[d]=h,c++,console.log(h)});let f={x:0,y:0};return Object.values(i).map(d=>{d.position.y<f.y&&(f.y=d.position.y),d.position.x<f.x&&(f.x=d.position.x)}),Object.keys(i).map(d=>{const h=i[d];h.position={x:h.position.x-f.x,y:h.position.y-f.y},h.table_data.map(y=>{if(y.Type.refears_to_name){const m=y.Type.refears_to_pos,v={x:m.x-f.x,y:m.y-f.y};y.Type=g.jsx(A4,{x:v.x+20,y:v.y+20,children:y.Type.refears_to_name})}}),i[d]=h}),Object.keys(i).map(d=>{const h=i[d];o.push(g.jsxs(E4,{x:h.position.x,y:h.position.y,children:[g.jsx(Za,{children:d}),g.jsx(ui,{data:h.table_data})]})),console.log(h)}),g.jsx(w4,{children:o})}const N4=n=>{const[l,i]=C.useState({username:"",password:""}),[o,c]=C.useState(!1),[f,d]=C.useState(!1);let h=!1;const y=()=>{const m=async S=>{if(!h){h=!0,c(!0);try{const _=await x3(S.username,S.password);Hn.set("token",_.access_token),n.onSuccess(),d(!1)}catch(_){_&&(d(!0),console.error(_))}h=!1,c(!1)}};i(S=>(m(S),S))};return g.jsxs(Pe,{opened:!0,centered:!0,withCloseButton:!1,onClose:()=>{},children:[g.jsx(Ee,{justify:"center",children:g.jsxs(Za,{color:"primary",fw:"normal",children:[g.jsx("strong",{children:"AUTODLA"})," - WEB"]})}),g.jsxs(ue,{fluid:!0,mt:"lg",children:[g.jsx(vd,{label:"Username",onChange:m=>i(v=>({...v,username:m.target.value})),mt:"lg",error:f}),g.jsx(cd,{label:"Password",onChange:m=>i(v=>({...v,password:m.target.value})),mt:"sm",error:f}),g.jsx(oe,{fullWidth:!0,onClick:()=>y(),my:"30",loading:o,children:"Login"})]})]})},O4={primaryColor:"autodla-blue",primaryShade:5,colors:{"autodla-blue":["#E0E6EB","#C2CED9","#A3B6C8","#859EB6","#6686A5","#4F9DDE","#3E6B87","#2F556C","#1F3E51","#1E2A38"],"audit-green":["#D3F7EA","#A8F0D5","#7DE9C0","#52E2AB","#3ECF8E","#2BB975","#1DA55F","#118F4A","#097A38","#036529"],"warning-coral":["#FFE5E5","#FFC9C9","#FFAFAF","#FF9595","#FF7B7B","#FF6B6B","#E55D5D","#CC5050","#B34343","#993737"],"neutral-gray":["#F9FAFB","#F5F7FA","#E4E7EB","#D1D6DB","#B0B8C1","#8B95A1","#6B7785","#4C5561","#333B46","#1E1E1E"]}};function M4(){const[n,l]=C.useState({}),[i,o]=C.useState(!1),c=async()=>{const d=await _3();l(d),o(!0)},f=async()=>{try{c()}catch(d){d&&o(!1)}};return C.useEffect(()=>{f()},[]),g.jsx(Qv,{theme:O4,children:i?g.jsxs(g.Fragment,{children:[g.jsx(ft,{p:"lg",children:g.jsxs(Za,{color:"primary",fw:"normal",children:[g.jsx("strong",{children:"AUTODLA"})," - WEB"]})}),g.jsx(Cu,{}),g.jsxs(ve,{defaultValue:"schema",orientation:"vertical",style:{height:"100%"},children:[g.jsxs(ve.List,{children:[g.jsx(ve.Tab,{value:"schema",children:"Schema"}),g.jsx(ve.Tab,{value:"objects",children:"Objects"})]}),g.jsx(ve.Panel,{value:"schema",children:g.jsx(j4,{schema:n})}),g.jsx(ve.Panel,{value:"objects",children:g.jsx(R3,{schema:n})})]})]}):g.jsx(N4,{onSuccess:()=>f()})})}Wb.createRoot(document.getElementById("root")).render(g.jsx(C.StrictMode,{children:g.jsx(M4,{})}));
